Package: classmap
Type: Package
Title: Visualizing Classification Results
Date: 2026-04-28
Version: 1.2.7
Authors@R: c(
            person(given="Jakob",family="Raymaekers", role = c("aut","cre"),
            email="jakob.raymaekers@kuleuven.be"),
            person(given="Peter",family="Rousseeuw", role = c("aut")) 
             )
Author: Jakob Raymaekers [aut, cre],
  Peter Rousseeuw [aut]
Depends: R (>= 4.1.0)
Suggests: knitr, reshape2, svd, rpart.plot, nnet, robCompositions,
        rmarkdown, fairml, robustHD, vioplot
Imports: stats, graphics, ggplot2, robustbase, e1071, cellWise,
        cluster, kernlab, gridExtra, rpart, randomForest, scales,
        lpSolve, diptest
Maintainer: Jakob Raymaekers <jakob.raymaekers@kuleuven.be>
Description: Tools to visualize the results of a classification or a regression.
    The graphical displays include stacked plots, silhouette plots, quasi residual plots, class maps, predictions plots, and predictions correlation plots.
    Implements the techniques described and illustrated in Raymaekers J., Rousseeuw P.J., Hubert M. (2022). Class maps for visualizing classification results. \emph{Technometrics}, 64(2), 151–165. <doi:10.1080/00401706.2021.1927849>
 (open access), Raymaekers J., Rousseeuw P.J.(2022). Silhouettes and quasi residual plots for neural nets and tree-based classifiers. \emph{Journal of Computational and Graphical Statistics}, 31(4), 1332–1343. <doi:10.1080/10618600.2022.2050249>, and Rousseeuw, P.J. (2026). Explainable Linear and Generalized Linear Models by the Predictions Plot. The American Statistician, 80, 157-163, <doi:10.1080/00031305.2025.2539235> (open access), and Montalcini, C., Rousseeuw, P.J. (2025). The bixplot: A variation on the boxplot suited for bimodal data, <doi:10.48550/arXiv.2510.09276> (open access).
 Examples can be found in the vignettes:
    "Discriminant_analysis_examples","K_nearest_neighbors_examples",
    "Support_vector_machine_examples", "Rpart_examples", "Random_forest_examples",
    "Neural_net_examples", "predsplot_examples", and "bixplot_examples".
URL: <doi:10.1080/00401706.2021.1927849>,
        <doi:10.1080/10618600.2022.2050249>,
        <doi:10.1080/00031305.2025.2539235>,
        <doi:10.48550/arXiv.2510.09276>
License: GPL (>= 2)
Encoding: UTF-8
LazyData: true
VignetteBuilder: knitr, rmarkdown
RoxygenNote: 7.1.2
NeedsCompilation: no
Packaged: 2026-04-29 11:09:36 UTC; Jakob Raymaekers
Repository: CRAN
Date/Publication: 2026-04-29 11:40:08 UTC
Built: R 4.6.0; ; 2026-05-01 21:19:34 UTC; unix

import("ggplot2")
import("gridExtra")
import("rpart")
importFrom("e1071", "svm")
importFrom("robustbase", "covMcd")
importFrom("randomForest", "randomForest", "importance")
importFrom("cellWise", "transfo", "truncPC")
importFrom("cluster", "daisy", "pam", "silhouette")
importFrom("grDevices", "adjustcolor", "rainbow", "colorRampPalette")
importFrom("graphics", "abline", "axis", "par", "rect", "title", "lines",
           "plot", "layout", "image", "mtext", "hist","strwidth", "curve",
           "points", "polygon", "box", "segments")
importFrom("stats", "coef", "cov", "dist", "mad", "mahalanobis",
               "median", "pnorm", "prcomp", "qnorm", "sd", "predict",
           "as.dist", "loess", "quantile", "na.omit", "terms.formula",
           "na.pass", "cor", "as.dendrogram", "hclust", "reorder", "order.dendrogram",
            "model.matrix", "approx", "density", "fivenum")
importFrom("kernlab", "stringdot", "kernelMatrix")
importFrom("scales", "breaks_extended")
importFrom("diptest", "dip.test")
importFrom("lpSolve", "lp.transport")


export(makeKernel)
export(makeFV)

export(vcr.da.train)
export(vcr.da.newdata)

export(vcr.knn.train)
export(vcr.knn.newdata)

export(vcr.svm.train)
export(vcr.svm.newdata)

export(confmat.vcr)
export(stackedplot)
export(classmap)

export(vcr.rpart.train)
export(vcr.rpart.newdata)

export(vcr.forest.train)
export(vcr.forest.newdata)

export(vcr.neural.train)
export(vcr.neural.newdata)

export(silplot)
export(qresplot)

export(predsplot)
export(predscor)


export(bixplot)
export(pamc1d)

## ----echo = FALSE-------------------------------------------------------------
knitr::opts_chunk$set(
  fig.width  = 7,
  fig.height = 4,
  fig.align  = 'center'
)
oldpar <- list(mar = par()$mar, mfrow = par()$mfrow)

## -----------------------------------------------------------------------------
library(vioplot)
library(robustHD)
library(classmap)

## -----------------------------------------------------------------------------
set.seed(1)
dat1 <- rnorm(120, 0, 2.5)
dat2 <- c(rnorm(80, -3, 1), rnorm(40, 3, 1))
dat3 <- c(rnorm(25, -4, 0.8), rnorm(50, 0, 0.8), rnorm(40, 4, 0.8))
xlist <- list(Unimodal = dat1, Bimodal = dat2, Multimodal = dat3)

## -----------------------------------------------------------------------------
bixout <- bixplot(xlist, main = "bixplot")

## -----------------------------------------------------------------------------
bixout

## ----fig.width = 8, fig.height = 4--------------------------------------------
ylim <- c(-8.5, 9)
par(las = 1, mfrow = c(1, 2))
par(mar = c(2.1, 2.2, 1.7, 2))
viocol <- adjustcolor("chocolate3", alpha.f = 0.5)
vioplot::vioplot(xlist, ylim = ylim, main = "", col = viocol)
title(main = "violin plot", line = 0.5, cex.main = 1)
par(mar = c(2.1, 2.2, 1.7, 0.2))
bixplot(xlist, ylim = ylim, main = "")
title(main = "bixplot", line = 0.5, cex.main = 1)
par(oldpar)

## ----fig.width = 10, fig.height = 7-------------------------------------------
data("data_latenc")
dim(data_latenc) # 40 rows, 3 columns

par(las = 1, mfrow = c(2, 2))
mar1 <- c(2.1, 2.4, 2.3, 2)
mar2 <- c(2.1, 2.2, 2.3, 0.2)
viocol <- adjustcolor("chocolate3", alpha.f = 0.5)

par(mar = mar1)
vioplot::vioplot(data_latenc, main = "", col = viocol)
title(main = "violin plot", cex.main = 1.2, line = 0.6)

par(mar = mar2)
mymodeCol <- c("cadetblue3", "hotpink2")
bixplot(data_latenc, main = "", cutmin = 0, cutmax = 300,
        ylim = c(0, 300), modeCol = mymodeCol)
title(main = "bixplot", cex.main = 1.2, line = 0.6)

# Score the islands in a meaningful order
islandscore <- rep(NA, length(penguins$island))
islandscore[penguins$island == "Torgersen"] <- 1
islandscore[penguins$island == "Biscoe"]    <- 2
islandscore[penguins$island == "Dream"]     <- 3
ylim <- c(29, 62)

par(mar = mar1)
vioplot::vioplot(bill_len ~ reorder(island, islandscore, mean),
                 data = penguins, main = "",
                 col = viocol, xlab = "", ylab = "",
                 ylim = ylim, cex.axis = 1)
title(main = "violin plot", cex.main = 1.2, line = 0.6)

par(mar = mar2)
bixplot(bill_len ~ reorder(island, islandscore, mean),
        data = penguins, main = "", ylim = ylim,
        bodyCol = "gray40", bodyOpaque = 0.3)
title(main = "bixplot", cex.main = 1.2, line = 0.6)
par(oldpar)

## ----fig.width = 10, fig.height = 5-------------------------------------------
par(mfrow = c(1, 2))
par(mar = c(4, 2, 2, 0.1))

diris <- data.frame(scale(iris[, 1:4]))
colnames(diris) <- c("Sepal.L", "Sepal.W", "Petal.L", "Petal.W")

mymodeCol <- c("cadetblue3", "hotpink2",
               "cadetblue3", "hotpink2",
               "lawngreen")
bixplot(diris, main = "", cut = 3, col = "gray75",
        bodyOpaque = 0.6, rugW = 0.16,
        rugoutCol = "red", curveLwd = 0.5,
        modeCol = mymodeCol,
        ylim = c(-3.6, 4.2), yaxs = "i",
        xlab = "standardized variables")
title(main = "bixplot display of iris data", cex.main = 1, line = 0.7)

## ----fig.width = 5, fig.height = 5--------------------------------------------
par(mar = c(4, 4, 2, 0.2))
x <- diris$Petal.L
y <- diris$Petal.W
xlim <- c(-2.3, 2.4)
ylim <- c(-2.2, 2.4)
xyratio <- (xlim[2] - xlim[1]) / (ylim[2] - ylim[1])

plot(x, y, xlim = xlim, ylim = ylim, pch = 16,
     xlab = "", ylab = "", xaxs = "i", yaxs = "i")
title(xlab = "Petal.L", line = 2)
title(ylab = "Petal.W", line = 2)
title(main = "petal length versus petal width", cex.main = 1, line = 0.7)

bixplot(x, add = TRUE, horizontal = TRUE,
        at = ylim[1] + 0.015, cutmin = xlim[1],
        boxwex = 0.9, curveLwd = 0.5,
        border = "black", side = "second",
        bodyOpaque = 0.6)
bixplot(y, add = TRUE, horizontal = FALSE,
        at = xlim[1] + 0.015, cutmin = ylim[1],
        boxwex = xyratio * 0.9,
        modeCol = c("cadetblue3", "hotpink2", "lawngreen"),
        curveLwd = 0.5, side = "second",
        bodyOpaque = 0.6)
par(oldpar)

## ----fig.width = 6, fig.height = 2.4------------------------------------------
par(mfrow = c(1, 3))
par(mar = c(2.5, 2, 2, 1))

for (bs in c("width_is_constant", "area_is_constant", "area_from_count")) {
  bixplot(diris[, 3], main = "", cut = 3, col = "gray75",
          bodyOpaque = 0.5, bodysize = bs, curveLwd = 0.5,
          ylim = c(-2.2, 2.4), yaxs = "i", names = "Petal.L")
  title(main = switch(bs,
                      width_is_constant = "equal width",
                      area_is_constant  = "equal area",
                      area_from_count   = "area from count"),
        cex.main = 1, line = 0.7)
}
par(oldpar)

## ----fig.width = 12, fig.height = 6-------------------------------------------
par(las = 1, mfrow = c(1, 2))

islscore <- rep(NA, length(penguins$island))
islscore[penguins$island == "Torgersen"] <- 1
islscore[penguins$island == "Biscoe"]    <- 2
islscore[penguins$island == "Dream"]     <- 3
ylim    <- c(27, 64)
mynames <- c("Torger.F", "Torger.M", "Biscoe.F",
             "Biscoe.M", "Dream.F",  "Dream.M")
mycol      <- c("slateblue2", "orange")
mymodeCol  <- c("darkorchid1", "slateblue3",
                "goldenrod1",  "darkorange2")

par(mar = c(2.9, 2, 0.8, 1))
bixplot(bill_len ~ sex + reorder(island, islscore, mean),
        data = penguins, names = mynames, modeCol = mymodeCol,
        bodyOpaque = 0.6, ylim = ylim, bodyW = 0.9,
        col = mycol, main = "", rugCol = "black", las = 1)
legend("topleft", legend = c("female", "male"),
       fill = c("slateblue3", "orange"), cex = 1.5)

par(mar = c(2.9, 3, 0.8, 0.1))
bixplot(bill_len ~ sex + reorder(island, islscore, mean),
        data = penguins, main = "", rugCol = "black",
        bodyOpaque = 0.6, ylim = ylim, bodyW = 0.9,
        col = mycol, stickCol = "gray10", stickLwd = 1,
        side = "both", modeCol = mymodeCol, las = 1,
        names = c("Torgersen", "Biscoe", "Dream"))
legend("topleft", legend = c("female", "male"),
       fill = mycol, cex = 1.5)
par(oldpar)

## ----fig.width = 5, fig.height = 5--------------------------------------------
par(las = 1, mfrow = c(1, 1))
par(mar = c(3.2, 3.2, 2, 2.6))

scpenguins <- scale(penguins[, c(4, 3, 5)])
varnames   <- c("bill_depth", "bill_length", "flipper_length")

bixplot(scpenguins, rugNumeric = penguins$body_mass,
        main = "", boxW = 0.30, rugW = 0.24,
        names = varnames, boxLwd = 1.2, boxOpaque = 1,
        curveLwd = 2, colorbarW = 0.16, cex.colorbar = 0.8,
        bodysize = "width_is_constant",
        bodyCol = "grey90", modeCol = "grey90",
        xlab = "feature", ylab = "standardized value")
title(main = "penguins with rug color by body_mass       ",
      line = 0.6)
par(oldpar)

## ----fig.width = 5, fig.height = 5--------------------------------------------
par(mfrow = c(1, 1))
par(mar = c(3.2, 3.8, 2, 0.2))

rugFactorColors <- c("red", "blue", "forestgreen")

bixplot(scpenguins, rugFactor = penguins$species,
        main = "", boxW = 0.30, rugW = 0.24,
        ylim = c(-3, 4), names = varnames,
        boxLwd = 1.2, boxOpaque = 1, curveLwd = 2,
        horizontal = TRUE, bodysize = "width_is_constant",
        bodyCol = "grey90", modeCol = "grey90",
        xlab = "standardized value", ylab = "feature",
        rugFactorColors = rugFactorColors, las = 0)
title(main = "penguins with rug color by species", line = 0.6)
legend(x = 2.36, y = 2.8,
       legend = c("Adelie", "Chinstrap", "Gentoo"),
       fill = rugFactorColors, cex = 0.88)
par(oldpar)

## -----------------------------------------------------------------------------
data(TopGear, package = "robustHD")
scars        <- TopGear[, c(14, 12, 5, 7, 8)]
scars[, 3]   <- log(scars[, 3])
colnames(scars)[3] <- "log(Price)"
scars[, 1:4] <- scale(scars[, 1:4])

## -----------------------------------------------------------------------------
bixplot(scars[, 1:4])

## -----------------------------------------------------------------------------
which.min(TopGear$Weight) # row 199
TopGear[199, c(1, 2, 14, 12, 5, 7)]

## -----------------------------------------------------------------------------
scars[199, 1] <- NA
bixplot(scars[, 1:4])

## -----------------------------------------------------------------------------
which.min(TopGear$TopSpeed) # row 220
TopGear[220, c(1, 2, 14, 12, 5, 7)]

## ----fig.width = 11, fig.height = 5.5-----------------------------------------
par(las = 1, mfrow = c(1, 2))
par(mar = c(4, 2, 2, 0.1))

bixplot(scars[, 1:4], ylim = c(-3.8, 5.2), main = "",
        col = "darkgoldenrod2", yaxs = "i",
        modeCol = c("cadetblue3", "hotpink2"),
        bodyOpaque = 0.6, las = 1)
title(main = "standardized Top Gear variables",
      cex.main = 1, line = 0.7)

par(mar = c(4, 4, 2, 0.1))
x <- jitter(scars[, 4])
y <- jitter(scars[, 2])
mycol <- rep(NA, nrow(scars))
mycol[scars[, 5] == "Front"] <- "red"
mycol[scars[, 5] == "Rear"]  <- "forestgreen"
mycol[scars[, 5] == "4WD"]   <- "orange"

plot(x, y, xlim = c(-1.999, 3.999), ylim = c(-2.999, 4.5),
     xaxs = "i", yaxs = "i", xlab = "", ylab = "",
     pch = 16, cex = 1, col = mycol, las = 1)
title(xlab = "Displacement", line = 2)
title(ylab = "TopSpeed",     line = 2)
title(main = "TopSpeed versus Displacement",
      cex.main = 1, line = 0.7)

bixplot(scars[, 4], add = TRUE, at = -2.98, horizontal = TRUE,
        side = "second", boxwex = 2, bodyOpaque = 0.6)
bixplot(scars[, 2], add = TRUE, at = -1.985, horizontal = FALSE,
        side = "second", boxwex = 1.4, bodyOpaque = 0.6)

legend(x = -1.5, y = 4, title = "DriveWheel",
       legend = c("Front", "Rear", "4WD"),
       fill = c("red", "forestgreen", "orange"), cex = 1)
par(oldpar)

## ----fig.width = 9.2, fig.height = 4.6----------------------------------------
par(las = 0, mfrow = c(1, 2))
par(mar = c(3.5, 2, 2, 1))

bixplot(len ~ supp * dose,
        data = ToothGrowth, side = "both",
        col = c("orange", "slateblue3"),
        main = "", xlab = "Vitamin C dose (mg)",
        ylab = "tooth length", bodyOpaque = 0.7,
        ylim = c(-1, 44), yaxs = "i", las = 0,
        stickCol = "black", stickLwd = 1)
title(main = "guinea pigs' tooth growth by supplement type",
      cex.main = 1, line = 0.7)
legend("topleft", title = "supplement",
       legend = c("OJ", "AA"),
       fill = c("orange", "slateblue3"), cex = 1)

par(mar = c(3.5, 2.8, 2, 0.1))
diris <- data.frame(scale(iris[, 1:4]))
colnames(diris) <- c("Sepal.L", "Sepal.W", "Petal.L", "Petal.W")

bixplot(diris, side = "both", main = "",
        bodysize = "width_is_constant",
        col = c("orange", "slateblue3"),
        modeCol = c("cadetblue3", "hotpink2",
                    "lawngreen", "gray60", "cyan3"),
        stickCol = "red", stickLwd = 1,
        bodyOpaque = 0.7, las = 0,
        xlab = "standardized measurements")
title(main = "iris data by length and width",
      cex.main = 1, line = 0.7)
legend("top", legend = c("length", "width"),
       fill = c("orange", "slateblue3"), cex = 1)
par(oldpar)

## ----fig.width = 5.5, fig.height = 5------------------------------------------
par(las = 1, mfrow = c(1, 1))
par(mar = c(2.2, 2.2, 2, 2.6))

bixplot(diris, main = "",
        rugNumeric = iris$Sepal.Length, colorbarW = 0.15,
        bodyCol = "grey80", modeCol = "grey80", las = 1)
title(main = "iris data with rug color by sepal length      ",
      line = 0.6)
par(oldpar)

## ----fig.width = 5, fig.height = 5--------------------------------------------
par(las = 1, mfrow = c(1, 1))
par(mar = c(2.2, 2.2, 2, 0.2))

rugFactorColors <- c("red", "blue", "forestgreen")
bixplot(diris, main = "", rugFactor = iris$Species,
        bodyCol = "grey80", modeCol = "grey80",
        rugFactorColors = rugFactorColors, las = 1)
title(main = "iris data with rug color by species", line = 0.6)
legend("topright",
       legend = c("setosa", "versicolor", "virginica"),
       fill = rugFactorColors, cex = 1)
par(oldpar)

## -----------------------------------------------------------------------------
data(data_titanic, package = "classmap")
titanic         <- data_titanic[1:100, ]
titanic$logFare <- log(titanic$Fare + 1)
titanic$Pclass  <- as.factor(titanic$Pclass)
titanic[, c(5, 14)] <- scale(titanic[, c(5, 14)])

xt <- list(titanic[titanic$y == "casualty",  5],
           titanic[titanic$y == "survived",  5],
           titanic[titanic$y == "casualty",  14],
           titanic[titanic$y == "survived",  14])
names(xt) <- c("standardized Age.C",
               "standardized Age.S",
               "standardized log(Fare).C",
               "standardized log(Fare).S")

## ----fig.width = 9.2, fig.height = 4.6----------------------------------------
par(las = 1, mfrow = c(1, 2))
par(mar = c(2.4, 2, 2, 1))

mycol <- c("coral2", "cadetblue3")

bixplot(xt, side = "both", main = "", col = mycol,
        stickLwd = 1, stickCol = "purple",
        boxW = 0.18, rugW = 0.10, las = 1)
title(main = "Titanic data by survival", line = 0.6)
legend("top", legend = c("casualty", "survived"),
       fill = mycol, cex = 0.9)

par(mar = c(2.4, 3, 2, 0.1))
bixplot(titanic[, c(5, 14)], main = "",
        rugFactor = titanic$Pclass, boxW = c(0.22, 0.18),
        names = c("standardized Age",
                  "standardized log(Fare)"), las = 1)
title(main = "Titanic data by cabin class", line = 0.6)
legend("top", title = "cabin class",
       legend = c("1", "2", "3"),
       fill = c("red", "blue", "forestgreen"), cex = 0.9)
par(oldpar)

## ----echo = FALSE-------------------------------------------------------------
knitr::opts_chunk$set(
 fig.width  = 5 ,
 fig.height = 3.5,
 fig.align  = 'center'
)
oldpar <- list(mar = par()$mar, mfrow = par()$mfrow)

## -----------------------------------------------------------------------------
library("classmap")
library("ggplot2")
library("gridExtra")

## -----------------------------------------------------------------------------
data(iris)
X <- iris[, 1:4]
y <- iris[, 5]
is.factor(y)
table(y)

pairs(X, col = as.numeric(y) + 1, pch = 19)

## -----------------------------------------------------------------------------
vcr.train <- vcr.da.train(X, y, rule = "QDA")
names(vcr.train)

## -----------------------------------------------------------------------------
vcr.train$predint 
vcr.train$pred[c(1:10, 51:60, 101:110)]
vcr.train$altint  
vcr.train$altlab[c(1:10, 51:60, 101:110)]

## -----------------------------------------------------------------------------
vcr.train$PAC[1:3] 
summary(vcr.train$PAC)

## -----------------------------------------------------------------------------
vcr.train$fig[1:5, ]

## -----------------------------------------------------------------------------
vcr.train$farness[1:5]
summary(vcr.train$farness)

## -----------------------------------------------------------------------------
summary(vcr.train$ofarness)

## -----------------------------------------------------------------------------
confmat.vcr(vcr.train, cutoff = 0.98)

## -----------------------------------------------------------------------------
confmat.vcr(vcr.train)

## -----------------------------------------------------------------------------
confmat.vcr(vcr.train, showClassNumbers = TRUE)

## -----------------------------------------------------------------------------
cols <- c("red", "darkgreen", "blue")
stackedplot(vcr.train, classCols = cols, separSize = 1.5,
            minSize = 1, showLegend = TRUE)

stackedplot(vcr.train, classCols = cols, separSize = 1.5,
            minSize = 1, showLegend = TRUE, cutoff = 0.98)

## -----------------------------------------------------------------------------
stplot <- stackedplot(vcr.train, classCols = cols, 
                     separSize = 1.5, minSize = 1,
                     main = "QDA on iris data")
stplot

## -----------------------------------------------------------------------------
# pdf("Iris_QDA_silhouettes.pdf", width=5.0, height=4.6)
silplot(vcr.train, classCols = cols, 
        main = "Silhouette plot of QDA on iris data")      
# dev.off()

## -----------------------------------------------------------------------------
classmap(vcr.train, 1, classCols = cols)
classmap(vcr.train, 2, classCols = cols)
classmap(vcr.train, 3, classCols = cols) # With the default cutoff no farness values stand out:

## -----------------------------------------------------------------------------
# With a lower cutoff:
classmap(vcr.train, 3, classCols = cols, cutoff = 0.98)
# Now one point is to the right of the vertical line.
# It also has a black border, meaning that it is flagged
# as an outlier, in the sense that its farness to _all_
# classes is above 0.98.

## -----------------------------------------------------------------------------
Xnew <- X[c(1:50, 101:150), ]
ynew <- y[c(1:50, 101:150)]
ynew[c(1:10, 51:60)] <- NA
pairs(X, col = as.numeric(y) + 1, pch = 19) # 3 colors
pairs(Xnew, col = as.numeric(ynew) + 1, pch = 19) # only red and blue

## -----------------------------------------------------------------------------
vcr.test <- vcr.da.newdata(Xnew, ynew, vcr.train)

## -----------------------------------------------------------------------------
plot(vcr.test$predint, vcr.train$predint[c(1:50, 101:150)]); abline(0, 1)
plot(vcr.test$altint, vcr.train$altint[c(1:50, 101:150)]); abline(0, 1)
plot(vcr.test$PAC, vcr.train$PAC[c(1:50, 101:150)]); abline(0, 1)
vcr.test$farness 
plot(vcr.test$farness, vcr.train$farness[c(1:50, 101:150)]); abline(0, 1)
plot(vcr.test$fig, vcr.train$fig[c(1:50, 101:150), ]); abline(0, 1)
vcr.test$ofarness 
plot(vcr.test$ofarness, vcr.train$ofarness[c(1:50, 101:150)]); abline(0, 1)

## -----------------------------------------------------------------------------
confmat.vcr(vcr.test)
confmat.vcr(vcr.test, cutoff = 0.98)

## -----------------------------------------------------------------------------
stplot # to compare with:
stackedplot(vcr.test, classCols = cols, separSize = 1.5, minSize = 1)

## -----------------------------------------------------------------------------
#pdf("Iris_test_QDA_silhouettes.pdf", width=5.0, height=4.3)
silplot(vcr.test, classCols = cols, 
        main = "Silhouette plot of QDA on iris subset") 
#dev.off()

## -----------------------------------------------------------------------------
classmap(vcr.train, 1, classCols = cols)
classmap(vcr.test, 1, classCols = cols) 

## ----error = TRUE-------------------------------------------------------------
try({
classmap(vcr.train, 2, classCols = cols)
classmap(vcr.test, 2, classCols = cols)
})

## -----------------------------------------------------------------------------
classmap(vcr.train, 3, classCols = cols)
classmap(vcr.test, 3, classCols = cols) 

## -----------------------------------------------------------------------------
data(data_floralbuds)
X <- as.matrix(data_floralbuds[, 1:6])
y <- data_floralbuds$y
dim(X) # 550  6
length(y) # 550
table(y)
# branch     bud  scales support 
#     49     363      94      44 

# Pairs plot
cols <- c("saddlebrown", "orange", "olivedrab4", "royalblue3")
pairs(X, gap = 0, col = cols[as.numeric(y)]) # hard to separate visually

## -----------------------------------------------------------------------------
vcr.obj <- vcr.da.train(X, y)

## -----------------------------------------------------------------------------
confmat <- confmat.vcr(vcr.obj, showOutliers = FALSE)

confmat.vcr(vcr.obj) 

## -----------------------------------------------------------------------------
stackedplot(vcr.obj, classCols = cols, separSize = 0.6,
            minSize = 1.5,  main = "stacked plot of QDA on floral buds")

# Version in paper:
# pdf("Floralbuds_QDA_stackplot_without_outliers.pdf",
#     width=5, height=4.3)
# stackedplot(vcr.obj, classCols = cols, separSize = 0.6,
#             minSize = 1.5, showOutliers = FALSE,
#             htitle = "given class", vtitle = "predicted class")
# dev.off()

## -----------------------------------------------------------------------------
#pdf("Floralbuds_QDA_silhouettes.pdf", width=5.0, height=4.3)
silplot(vcr.obj, classCols = cols,
        main = "Silhouette plot of QDA on floral bud data")      
#dev.off()

## -----------------------------------------------------------------------------
PAC <- vcr.obj$PAC
feat <- rowSums(X); xlab = "rowSums(X)"
# pdf("Floralbuds_QDA_quasi_residual_plot.pdf", width=5, height=4.8)
qresplot(PAC, feat, xlab = xlab, plotErrorBars = TRUE, fac = 2, 
         main = "Floral buds: quasi residual plot")
# dev.off()

cor.test(feat, PAC, method = "spearman") 

## -----------------------------------------------------------------------------
labels <- c("branch", "bud", "scale", "support")

# classmap of class "bud"
#
# To identify the points that stand out:
# classmap(vcr.obj, 2, classCols = cols, identify = TRUE)
# Press "Esc" to get out.
#
# pdf("Floralbuds_QDA_classmap_bud.pdf", width=7, height=7)
par(mar = c(3.6, 3.5, 2.4, 3.5))
coords <- classmap(vcr.obj, 2, classCols = cols,
         main = "predictions of buds",
         cex = 1.5, cex.lab = 1.5, cex.axis = 1.5,
         cex.main = 1.5) 
# For marking points:
indstomark <- c(294, 70, 69, 152, 204) # from identify = TRUE above
labs  <- letters[seq_len(5)]
xvals <- coords[indstomark, 1] +
  c(0, 0.10, 0.14, 0.10, 0.08) # visual finetuning
yvals <- coords[indstomark, 2] +
  c(0.04, 0.04, 0, -0.03, +0.04)
text(x = xvals, y = yvals, labels = labs, cex = 1.5)
legend("topleft", fill = cols[1:4], legend = labels, 
       cex = 1, ncol = 1, bg = "white")
# dev.off()
par(oldpar)

## ----fig.height = 8, fig.width = 8--------------------------------------------

#
# pdf(file = "Floralbuds_all_class_maps.pdf", width = 7, height = 7)
par(mfrow = c(2, 2))
par(mar = c(3.3, 3.2, 2.7, 1.0))
classmap(vcr.obj, 1, classCols = cols,
         main = "predictions of branches")
legend("topright", fill = cols, legend = labels,
       cex = 1, ncol = 1, bg = "white")
#
par(mar = c(3.3, 0.5, 2.7, 0.3))
classmap(vcr.obj, 2, classCols = cols,
         main = "predictions of buds")
labs  <- letters[seq_len(5)]
xvals <- coords[indstomark, 1] +
  c(0, 0.10, 0.14, 0.10, 0.08) # visual finetuning
yvals <- coords[indstomark, 2] +
  c(0.04, 0.04, 0, -0.03, 0.04)
# xvals <- c( 1.75, 1.68, 1.25, 3.25, 4.00)
# yvals <- c(0.045, 0.92, 0.54, 0.97, 0.045)
text(x = xvals, y = yvals, labels = labs, cex = 1.0)
legend("topleft", fill = cols, legend = labels,
       cex = 1, ncol = 1, bg = "white")
#
par(mar = c(3.3, 3.2, 2.7, 1.0))
classmap(vcr.obj, 3, classCols = cols,
         main = "predictions of scales")
legend("left", fill = cols, legend = labels,
       cex = 1, ncol = 1, bg = "white")
# 
par(mar = c(3.3, 0.5, 2.7, 0.3))
classmap(vcr.obj, 4, classCols = cols,
         main = "predictions of supports")
legend("topright", fill = cols, legend = labels,
       cex = 1, ncol = 1, bg = "white")
# dev.off()
par(oldpar)

## -----------------------------------------------------------------------------

mnist_url <- url("https://wis.kuleuven.be/statdatascience/robust/data/mnist-rdata")
load(mnist_url)

X_train <- mnist$train$x
y_train <- as.factor(mnist$train$y)

head(y_train)
# Levels: 0 1 2 3 4 5 6 7 8 9
dim(X_train) # 60000    28    28
length(y_train) # 60000

## ----fig.height = 1, fig.width = 1--------------------------------------------

plotImage = function(tempImage) {
  tdm = reshape2::melt(apply((tempImage), 2, rev))
  p = ggplot(tdm, aes(x = Var2, y = Var1, fill = (value))) +
    geom_raster() +
    guides(color = "none", size = "none", fill = "none") +
    theme(axis.title.x = element_blank(),
          axis.title.y = element_blank(),
          axis.text.x = element_blank(),
          axis.text.y = element_blank(),
          axis.ticks.x = element_blank(),
          axis.ticks.y = element_blank()) +
    scale_fill_gradient(low = "white", high = "black")
  p
}

plotImage(X_train[1, , ])
plotImage(X_train[2, , ])
plotImage(X_train[3, , ])

## ----fig.height = 2, fig.width = 5--------------------------------------------
# Change the dimensions of X for the sequel:
dim(X_train) <- c(60000, 28 * 28)
dim(X_train) # 60000    784

# Sampled digit images:
set.seed(123)
sampledigits <- list()
for (i in 0:9) {
  digit <- i
  idx <- sample(which(y_train == digit), size = 1)
  tempImage <- matrix(unlist(X_train[idx, ]), 28, 28)
  sampledigits[[i + 1]] <- plotImage(tempImage) 
}
psampledigits <- grid.arrange(grobs = sampledigits, ncol = 5)
# ggsave("MNIST_sampled_images.pdf", plot = psampledigits,
#        width = 10, height = 1)


# Averaged digit images:
meanPlots <- list()
for (j in 0:9) {
  m.out <- colMeans(X_train[which(y_train == j), ])
  dim(m.out) <- c(28, 28)
  meanPlots[[j + 1]] <- plotImage(m.out) 
}
meanplot <- grid.arrange(grobs = meanPlots, ncol = 5)
# ggsave("MNIST_averaged_images.pdf", plot = meanplot,
#        width = 10, height = 1)

## -----------------------------------------------------------------------------
library(svd)
ptm <- proc.time()
svd.out <- svd::propack.svd(X_train, neig = 50)
(proc.time() - ptm)[3]
loadings <- svd.out$v
rm(svd.out)
dataProj <- as.matrix(X_train %*% loadings)
dim(dataProj)

## -----------------------------------------------------------------------------
vcr.train <- vcr.da.train(X = dataProj, y_train)

## -----------------------------------------------------------------------------
confmat.vcr(vcr.train, showOutliers = FALSE)

cols <- c("red3", "darkorange", "gold2", "darkolivegreen3",
         "darkolivegreen4", "cadetblue3", "deepskyblue4", 
         "darkslateblue", "darkorchid3", "deeppink4")

# stacked plot in paper:
# pdf("MNIST_stackplot_with_outliers.pdf", width=5, height=4.3)
stackedplot(vcr.train, classCols = cols, separSize = 0.6,
            minSize = 1.5, htitle = "given class",
            main = "Stacked plot of QDA on MNIST training data", vtitle = "predicted class")
# dev.off()

## -----------------------------------------------------------------------------
# pdf("MNIST_QDA_silhouettes.pdf", width=5.0, height=4.6)
silplot(vcr.train, classCols = cols,
        main = "Silhouette plot of QDA on MNIST training data")      
# dev.off()

## -----------------------------------------------------------------------------
wnq <- function(string, qwrite=TRUE) { # auxiliary function
  # writes a line without quotes
  if (qwrite) write(noquote(string), file = "", ncolumns = 100)
}

showdigit <- function(digit=digit, i, plotIt = TRUE) {
  idx = which(y_train == digit)[i]
  # wnq(paste("Estimated digit: ", as.numeric(vcr.train$pred[idx]), sep=""))
  tempImage <- matrix(unlist(X_train[idx, ]), 28, 28)
  if (plotIt) {plot(plotImage(tempImage))}
  return(plotImage(tempImage))
}


## ----fig.height = 4, fig.width = 6--------------------------------------------
digit <- 0
#
# To identify outliers:
# classmap(vcr.train, digit+1, classCols = cols, identify = TRUE)
# Press "Esc" to get out.
#
# pdf(paste0("MNIST_classmap_digit", digit, ".pdf"), width = 7, height = 7)
par(mar = c(3.6, 3.5, 2.4, 3.5))
coords <- classmap(vcr.train, digit + 1, classCols = cols,
         main = paste0("predictions of digit ",digit),
         cex = 1.5, cex.lab = 1.5, cex.axis = 1.5, 
         cex.main = 1.5)
indstomark <- c(4000, 3964, 5891, 2485, 822, 
               2280, 2504, 3906, 5869, 1034) # from identify = TRUE
labs  <- letters[1:length(indstomark)]
xvals <- coords[indstomark, 1] +
  c(-0.04, -0.01, 0, -0.11, 0.06,
    0.07, 0.06, 0.10, 0.06, 0.09)
yvals <- coords[indstomark, 2] +
  c(-0.03, -0.03, -0.03, 0.022, -0.025, 
    -0.025, -0.035, -0.025, 0.03, 0.03)
text(x = xvals, y = yvals, labels = labs, cex = 1.5)
legend("left", fill = cols,
       legend = 0:9, cex = 1, ncol = 2, bg = "white")
# dev.off()
par(oldpar)

## ----fig.height = 2, fig.width = 5--------------------------------------------
pred <- vcr.train$pred # needed for discussion plots
tempPreds <- (pred[which(y_train == digit)])[indstomark]
discussionPlots <- list()
for (i in 1:length(indstomark)) {
  idx <- indstomark[i]
  tempplot <- showdigit(digit, idx, plotIt = FALSE)
  tempplot <- arrangeGrob(tempplot, 
    bottom = paste0("(", labs[i], ") \"", tempPreds[i], "\""))
  discussionPlots[[i]] = tempplot
}
discussionPlot <- grid.arrange(grobs = discussionPlots, ncol = 5)
# ggsave(paste0("MNIST_discussionplot_digit", digit, ".pdf"),
#        plot = discussionPlot, width = 5,
#        height = (length(indstomark) %/% 5 +
#                    (length(indstomark) %% 5 > 0)))

## ----fig.height = 4, fig.width = 6--------------------------------------------

digit <- 1
# pdf(paste0("MNIST_classmap_digit", digit, ".pdf"), width = 7, height = 7)
par(mar = c(3.6, 3.5, 2.4, 3.5))
classmap(vcr.train, digit + 1, classCols = cols,
         main = paste0("predictions of digit ", digit),
         cex = 1.5, cex.lab = 1.5, cex.axis = 1.5, 
         cex.main = 1.5)
legend("left", fill = cols,
       legend = 0:9, cex = 1, ncol = 2, bg = "white")
# dev.off()
par(oldpar)

## ----fig.height = 13, fig.width = 8-------------------------------------------
# indices of the 1s predicted as 2 (takes a while):
#
indstomark <- which(vcr.train$predint[which(y_train == digit)] == 3)
length(indstomark) # 104
labs  <- letters[1:length(indstomark)]
pred <- vcr.train$pred # needed for discussion plots
tempPreds    <- (pred[which(y_train == digit)])[indstomark]
discussionPlots <- list()
for (i in 1:length(indstomark)) {
  idx <- indstomark[i]
  tempplot <- showdigit(digit, idx, FALSE)
  tempplot <- arrangeGrob(tempplot, 
        bottom = paste0("\"", tempPreds[i], "\""))
  discussionPlots[[i]] <- tempplot
}
discussionPlot <- grid.arrange(grobs = discussionPlots, 
                                         ncol = 8)
# ggsave(paste0("MNIST_discussionplot_digit", digit, "predictedAs2b.pdf"),
#        plot = discussionPlot, width = 10,
#        height = (length(indstomark) %/% 10 +
#                    (length(indstomark) %% 10 > 0)))
                 
# The digits 1 predicted as a 2 are mostly ones written with
# a horizontal line at the bottom.

## ----fig.height = 4, fig.width = 6--------------------------------------------
digit <- 2
# To identify outliers:
# classmap(vcr.train, digit + 1, classCols = cols, identify = TRUE)
# Press "Esc" to get out.
#
# pdf(paste0("MNIST_classmap_digit", digit,".pdf"), width = 7, height = 7)
par(mar = c(3.6, 3.5, 2.4, 3.5))
coords <- classmap(vcr.train, digit + 1, classCols = cols,
                  main = paste0("predictions of digit", digit), cex = 1.5, cex.lab = 1.5, cex.axis = 1.5,
                  cex.main = 1.5)
indstomark <- c(3164, 5434, 2319 , 4224, 3682, 
               2642, 4920, 1233, 3741, 3993) # from identify = TRUE
labs  <- letters[1:length(indstomark)]
xvals <- coords[indstomark, 1] +
  c(0, 0.08, 0, 0, 0, 0, 0, 0, 0, 0)
yvals <- coords[indstomark, 2] +
  c(-0.03, -0.03, -0.03, -0.03, -0.03, 
    -0.03, -0.03, -0.03, 0.03, 0.03)  
text(x = xvals, y = yvals, labels = labs, cex = 1.5)
legend("right", fill = cols,
       legend = 0:9, cex = 1, ncol = 2, bg = "white")
# dev.off()
par(oldpar)

## ----fig.height = 2, fig.width = 5--------------------------------------------
pred <- vcr.train$pred # needed for discussion plots
tempPreds    <- (pred[which(y_train == digit)])[indstomark]
discussionPlots <- list()
for (i in 1:length(indstomark)) {
  idx <- indstomark[i]
  tempplot <- showdigit(digit, idx, FALSE)
  tempplot <- arrangeGrob(tempplot, 
        bottom = paste0("(", labs[i], ") \"", tempPreds[i], "\""))
  discussionPlots[[i]] <- tempplot
}
discussionPlot <- grid.arrange(grobs = discussionPlots, 
                                         ncol = 5)
# ggsave(paste0("MNIST_discussionplot_digit", digit, ".pdf"),
#        plot = discussionPlot, width = 5,
#        height = (length(indstomark) %/% 5 +
#                    (length(indstomark) %% 5 > 0)))

## ----fig.height = 1, fig.width = 1--------------------------------------------
X_test <- mnist$test$x
y_test <- as.factor(mnist$test$y)

head(y_test)
#
dim(X_test) # 10000    28    28
length(y_test) # 10000

plotImage(X_test[1, , ])
plotImage(X_test[2, , ])
plotImage(X_test[3, , ])

dim(X_test) <- c(10000, 28 * 28)
dim(X_test) # 10000  784

dataProj_test <- as.matrix(X_test %*% loadings)

## -----------------------------------------------------------------------------
vcr.test <- vcr.da.newdata(Xnew = dataProj_test,
                           ynew = y_test,
                           vcr.da.train.out = vcr.train)

## -----------------------------------------------------------------------------
confmat.vcr(vcr.test, showOutliers = FALSE, showClassNumbers = TRUE)

# In supplementary material:
# pdf("MNISTtest_stackplot_with_outliers.pdf", width = 5, height = 4.3)
stackedplot(vcr.test, classCols = cols, separSize = 0.6,
            main = "Stacked plot of QDA on MNIST test data",
            minSize = 1.5)
# dev.off()

## -----------------------------------------------------------------------------
#pdf("MNIST_test_QDA_silhouettes.pdf", width = 5.0, height = 4.6)
silplot(vcr.test, classCols = cols,
        main = "Silhouette plot of QDA on MNIST test data")      
#dev.off()

## ----fig.height = 4, fig.width = 6--------------------------------------------
showdigit_test <- function(digit = digit, i, plotIt = TRUE) {
  idx = which(y_test == digit)[i]
  # wnq(paste("Estimated digit: ", as.numeric(vcr.test$pred[idx]), sep = ""))
  tempImage <- matrix(unlist(X_test[idx, ]), 28, 28)
  if (plotIt) {plot(plotImage(tempImage))}
  return(plotImage(tempImage))
}

digit <- 0
# classmap(vcr.test, digit+1, classCols = cols, identify = TRUE)
# pdf(paste0("MNISTtest_classmap_digit", digit,".pdf"))
par(mar = c(3.6, 3.5, 2.4, 3.5))
coords <- classmap(vcr.test, digit + 1, classCols = cols,
         main = paste0("predictions of digit ", digit),
         cex = 1.5, cex.lab = 1.5, cex.axis = 1.5, 
         cex.main = 1.5)
indstomark <- c(140, 630, 241, 967, 189,
               377, 78, 943, 64, 354)
labs  <- letters[1:length(indstomark)]
xvals <- coords[indstomark, 1] +
  c(0.08, 0.07, -0.07, 0.06, 0,
    0.04, 0.05, 0.09, -0.04, 0.09)
yvals <- coords[indstomark, 2] +
  c(-0.025, -0.03, -0.024, -0.025, -0.03, 
    -0.03, -0.03, 0.022, 0.035, 0.03)
text(x = xvals, y = yvals, labels = labs, cex = 1.5)
legend("left", fill = cols,
       legend = 0:9, cex = 1, ncol = 2, bg = "white")
# dev.off()
par(oldpar)

## ----fig.height = 2, fig.width = 5--------------------------------------------
pred <- vcr.test$pred # needed for discussion plots
tempPreds <- (pred[which(y_test == digit)])[indstomark]
discussionPlots <- list()
for (i in 1:length(indstomark)) {
  idx <- indstomark[i]
  tempplot <- showdigit_test(digit, idx, FALSE)
  tempplot <- arrangeGrob(tempplot, 
      bottom = paste0("(", labs[i], ") \"", tempPreds[i], "\""))
  discussionPlots[[i]] <- tempplot
}
discussionPlot <- grid.arrange(grobs = discussionPlots, 
                                         ncol = 5)
# ggsave(paste0("MNISTtest_discussionplot_digit", digit, ".pdf"),
#        plot = discussionPlot, width = 5,
#        height = (length(indstomark) %/% 5 +
#                    (length(indstomark) %% 5 > 0)))

## ----fig.height = 4, fig.width = 6--------------------------------------------
digit <- 3
# classmap(vcr.test, digit + 1, classCols = cols, identify = TRUE)
# pdf(paste0("MNISTtest_classmap_digit", digit, ".pdf"))
par(mar = c(3.6, 3.5, 2.4, 3.5))
coords <- classmap(vcr.test, digit + 1, classCols = cols,
         main = paste0("predictions of digit ", digit),
         cex = 1.5, cex.lab = 1.5, cex.axis = 1.5, 
         cex.main = 1.5)
indstomark <- c(883, 659, 262, 60, 310,
               832, 223, 784, 835, 289)
labs  <- letters[1:length(indstomark)]
xvals <- coords[indstomark, 1] +
  c(-0.01, 0.08, -0.10, 0.06, 0.07, 
    0.06, 0.03, 0.11, 0.02, 0.06)
yvals <- coords[indstomark, 2] +
  c(0.035, 0.033, -0.017, -0.022, -0.025, 
    -0.025, -0.033, -0.022, 0.035, 0.038)
text(x = xvals, y = yvals, labels = labs, cex = 1.5)
legend("right", fill = cols,
       legend = 0:9, cex = 1, ncol = 2, bg = "white")
# dev.off()
par(oldpar)

## ----fig.height = 2, fig.width = 5--------------------------------------------
pred <- vcr.test$pred # needed for discussion plots
tempPreds    <- (pred[which(y_test == digit)])[indstomark]
discussionPlots <- list()
for (i in 1:length(indstomark)) {
  idx <- indstomark[i]
  tempplot <- showdigit_test(digit, idx, FALSE)
  tempplot <- arrangeGrob(tempplot, 
    bottom = paste0("(", labs[i], ") \"", tempPreds[i], "\""))
  discussionPlots[[i]] <- tempplot
}
discussionPlot <- grid.arrange(grobs = discussionPlots, 
                                         ncol = 5)
# ggsave(paste0("MNISTtest_discussionplot_digit", digit, ".pdf"),
#        plot = discussionPlot, width = 5,
#        height = (length(indstomark) %/% 5 +
#                    (length(indstomark) %% 5 > 0)))

## ----echo = FALSE-------------------------------------------------------------
knitr::opts_chunk$set(
 fig.width  = 5 ,
 fig.height = 3.5,
 fig.align  = 'center'
)
oldpar <- list(mar = par()$mar, mfrow = par()$mfrow)

## -----------------------------------------------------------------------------
library(classmap)

## -----------------------------------------------------------------------------
data(TopGear, package = "robustHD")
cars = TopGear; rm(TopGear)
dim(cars) # [1] 297  32
rownames(cars) = paste(cars[,1],cars[,2])
# Now the rownames are make and model of the cars.
colnames(cars)
colnames(cars)[4:17] = c("fuel","price","cyl","displ","drive",
                         "hp","torque","accel","topspeed",
                         "MPG","weight","length","width",
                         "height")
summary(cars$accel) # zero minimum is impossible
rownames(cars)[which(cars$accel == 0)]
# [1] "Citroen C5 Tourer" "Ford Mondeo"   "Lotus Elise"
# [4] "Renault Twizy"     "Ssangyong Rodius"
cars$accel[cars$accel == 0] = NA
summary(cars$accel) # OK

cars$weight[order(cars$weight)[1:4]]
rownames(cars)[order(cars$weight)[1:4]]
cars$weight[cars$weight == 210] = NA
summary(cars$weight) # OK
# save(cars, file = "Topgear.RData")
# load("Topgear.RData")

## -----------------------------------------------------------------------------
summary(cars$topspeed) # in mph
summary(cars$displ)    # in cc
summary(cars$length)   # in mm
cars$length = cars$length/1000 # in meters

fit = lm(hp ~ topspeed + length + displ, data = cars)
summary(fit)

predsplot(fit, main = "Top Gear data")

fact = 0.51
width = fact*10; height = fact*8
maxnchar = 6
main = "Top Gear data"
pdf(file = "topgear_numerical_predsplot.pdf",
    width = width, height = height)
predsplot(fit, main=main, maxchar.level = maxnchar)
dev.off()


predscor(fit, sort.by.stdev = FALSE) 
# fact = 0.6
# width = fact*10; height = fact*8
# pdf(file = "topgear_numerical_predscor.pdf", 
#     width = height, height = height)
# predscor(fit, sort.by.stdev = FALSE)
# dev.off()


## -----------------------------------------------------------------------------
cars$GPM = 1/cars$MPG

fit4 = lm(GPM ~ accel + drive + weight + fuel, data = cars)
summary(fit4)

## Figure 2:
## 
fact = 0.52
width = fact*10; height = fact*8
maxnchar = 6
main = "Top Gear data"
# pdf(file = "topgear_4_predsplot.pdf",
#     width = width, height = height)
predsplot(fit4, main=main, maxchar.level = maxnchar)
# dev.off()

## -----------------------------------------------------------------------------
## 
# fact = 0.6
# width = fact*10; height = fact*8
# pdf(file = "topgear_4_predscor.pdf", 
#     width = height, height = height)
predscor(fit4, sort.by.stdev = FALSE)
# dev.off()

car = "Bentley Continental"
# pdf(file = "topgear_4_predsplot_1_dens.pdf",
#     width = width, height = height)
predsplot(fit4, main = main, casetoshow=car, 
          displaytype = "density", 
          maxchar.level = maxnchar, 
          xlab = paste0("prediction for ",car))
# dev.off()

car = "Kia Rio"
# pdf(file = "topgear_4_predsplot_2_stairs.pdf",
#     width = width, height = height)
predsplot(fit4, main = main, casetoshow=car, 
          staircase = TRUE, maxchar.level = maxnchar, 
          xlab = paste0("prediction for ",car))
# dev.off()

## -----------------------------------------------------------------------------
summary(cars$AlarmSystem) 
cars$alarm = (cars$AlarmSystem == "standard")
summary(cars$alarm)

summary(cars$SatNav)
cars$navig = "y"
cars$navig[cars$SatNav == "no"] = "n"
summary(cars$navig)

table(cars$navig)
class(cars$navig) 
str(cars)

fitcombin = lm(1/MPG ~ accel + log(weight) + accel:torque + alarm + navig, data = cars)
summary(fitcombin)

fact = 0.45
width = fact*10; height = fact*8
main = "Example with transformation, interaction, logical, character"
# pdf(file = "topgear_logical+char_predsplot.pdf", width = width, height = height)
predsplot(fitcombin, main=main)
# dev.off()  

fact = 0.6
width = fact*10; height = fact*8
# pdf(file = "topgear_logical+char_predscor.pdf",
#     width = height, height = height)
predscor(fitcombin)
# dev.off()


## -----------------------------------------------------------------------------
data(data_titanic, package = "classmap")
titanic = data_titanic; rm(data_titanic)
dim(titanic) # 1309  13
# A data frame with 1309 observations on the following variables.
#   passengerId: a unique identifier for each passenger.
#   pclass: travel class of the passenger.
#   name: name of the passenger.
#   sex: sex of the passenger.
#   age: age of the passenger.
#   sibsp: number of siblings and spouses traveling with the passenger.
#   parch: number of parents and children traveling with the passenger.
#   ticket: ticket number of the passenger.
#   fare: fare paid for the ticket.
#   Cabin: cabin number of the passenger.
#   embarked: Port of embarkation. Takes the values
#             C (Cherbourg), Q (Queenstown) and
#             S (Southampton).
#   y: factor indicating casualty or survivor.
#   dataType: vector taking the values “train” or “test”
#             indicating whether the observation belongs
#             to training or test data.
#
colnames(titanic) = c("passengerId","pclass","name","sex",
                   "age", "sibsp", "parch", "ticket",
                   "fare", "cabin", "embarked", "survival",
                   "dataType")
str(titanic)

unique(titanic$pclass) # 3 1 2
titanic$pclass = factor(titanic$pclass, levels = unique(titanic$pclass))
#

unique(titanic$sex) 
titanic$sex = factor(titanic$sex, levels = unique(titanic$sex) )
titanic$sex = factor(titanic$sex, labels = c("M","F"))
head(titanic)
# save(titanic, file="Titanic.RData")
# load("Titanic.RData")

## -----------------------------------------------------------------------------
fit5 <- glm(survival ~ sex + age + sibsp + parch + pclass,
            family=binomial, data = titanic)
summary(fit5)

## -----------------------------------------------------------------------------
fact = 0.5
width = fact*10; height = fact*8
main = "Titanic data"
# pdf(file = "titanic_5_predsplot_dens.pdf",
#     width = width, height = height)
predsplot(fit5, main=main, displaytype = "density")

# The glm fit used the logit link function.
# dev.off()

# With other options for density estimation:
predsplot(fit5, main=main, displaytype = "density",
          bw = "SJ", adjust = 0.5)

## -----------------------------------------------------------------------------
# pdf(file = "titanic_5_predsplot_1.pdf",
#     width = width, height = height)
predsplot(fit5, main = main, casetoshow=1)
# dev.off()

# pdf(file = "titanic_5_predsplot_2_stairs.pdf",
#     width = width, height = height)
predsplot(fit5, main = main, casetoshow=2, staircase = TRUE)
# dev.off()

# fact = 0.6
# width = fact*10; height = fact*8
# pdf(file = "titanic_5_predscor.pdf", 
#     width = height, height = height)
predscor(fit5, sort.by.stdev = FALSE)
# dev.off()

## -----------------------------------------------------------------------------
data(german.credit, package = "fairml")
credit = german.credit; rm(german.credit)
dim(credit) # 1000 21
str(credit)
#
# Shorten variable names for plotting:
colnames(credit) <- c("checking","months","history","purpose","amount","savings","employment","rate","guarantors","residence","property","age","inst_plans","housing","nloans","job","nclients","phone","foreign","credit","sex")
#
# Give factors sex and purpose shorter labels for plotting:
credit$sex = factor(as.numeric(credit$sex), labels = c("F","M"))
levels(credit$purpose)
# [10] "retrainin"
levels(credit$purpose) = c("business","n.car","u.car","appliances","education","furniture","other","TV","repairs","retraining")
# save(credit, file = "German_credit.RData")
# load("German_credit.RData")

## -----------------------------------------------------------------------------
fit7 <- glm(credit ~ months + purpose + amount + rate + age +
            nclients + sex, family=binomial, data = credit)

## -----------------------------------------------------------------------------
main = "German credit data"
# fact = 0.48
# width = fact*10; height = fact*8
# pdf(file = "germancredit_7_predsplot.pdf",
#     width = width, height = height)
predsplot(fit7, main = main)       
# dev.off()

## -----------------------------------------------------------------------------
# fact = 0.48
# width = fact*10; height = fact*8
# pdf(file = "germancredit_7_predsplot_1.pdf",
#     width = width, height = height)
predsplot(fit7, main = main, casetoshow=1, 
          displaytype = "density")
# dev.off()

## -----------------------------------------------------------------------------
# pdf(file = "germancredit_7_predsplot_2_stairs.pdf",
#     width = width, height = height)
predsplot(fit7, main = main, casetoshow=2, staircase = TRUE) 
# dev.off()

## -----------------------------------------------------------------------------
# fact = 0.6
# width = fact*10; height = fact*8
# pdf(file = "germancredit_7_predscor_equalsizes.pdf", 
#     width = height, height = height)
predscor(fit7, sort.by.stdev = FALSE, cell.length = "equal")
# dev.off()
#
# pdf(file = "germancredit_7_predscor.pdf", 
#     width = height, height = height)
predscor(fit7, sort.by.stdev = FALSE)
# dev.off()

## -----------------------------------------------------------------------------
## Prediction for a new case:
newc = list("u.car", 36, 2, 6000, 55, "F", 1)
names(newc) = c("purpose","months","rate","amount",
                "age","sex","nclients")
# fact = 0.48
# width = fact*10; height = fact*8
# pdf(file = "germancredit_7_predsplot_new_stairs.pdf",
#     width = width, height = height)
predsplot(fit7, main = main, casetoshow = newc, 
          staircase = TRUE)
# dev.off()

## Figure with profile:
# pdf(file = "germancredit_7_predsplot_1_dens_profile.pdf",
#     width = width, height = height)
predsplot(fit7, main = main, casetoshow=1, 
               displaytype = "density", profile = TRUE)
# dev.off()


## -----------------------------------------------------------------------------
credit$x1 = credit$months + credit$nclients
credit$x2 = credit$months - credit$nclients
cor(credit$x1,credit$x2) # 0.998199
# But minus that for prediction terms!

fitart <- glm(credit ~ x1 + x2 + purpose + amount + rate + age +
           sex, family=binomial, data = credit)

## Figure 7:
## 
# fact = 0.48
# width = fact*10; height = fact*8
# main = "German credit data with artificial variables x1 and x2"
# pdf(file = "germancredit_7_artif_predsplot.pdf", 
#     width = width, height = height)
predsplot(fitart, main = main)
# dev.off()

## Figure 8:
## 
# fact = 0.6
# width = fact*10; height = fact*8
# pdf(file = "germancredit_7_artif_predscor.pdf", 
#     width = height, height = height)
predscor(fitart, sort.by.stdev = FALSE)
# dev.off()

## ----echo = FALSE-------------------------------------------------------------
knitr::opts_chunk$set(
 fig.width  = 5 ,
 fig.height = 3.5,
 fig.align  = 'center'
)
oldpar <- list(mar = par()$mar, mfrow = par()$mfrow)

## -----------------------------------------------------------------------------
library(rpart)
library(classmap)

## -----------------------------------------------------------------------------
data("data_titanic")

traindata <- data_titanic[which(data_titanic$dataType == "train"), -13]

dim(traindata) 
colnames(traindata)
# SibSp: number of siblings + spouses aboard
# Parch: number of parents + children aboard

str(traindata)
table(traindata$y)

## -----------------------------------------------------------------------------
set.seed(123) 
rpart.out <- rpart(y ~ Pclass + Sex + SibSp + 
                    Parch + Fare + Embarked,
                  data = traindata, method = 'class',
                  model = TRUE)

## plot the tree:
# pdf(file = "titanic_train_tree.pdf", width = 6, height = 6)
rpart.plot::rpart.plot(rpart.out, box.palette = "RdBu")
# dev.off()

## -----------------------------------------------------------------------------
rpart.out$variable.importance

## -----------------------------------------------------------------------------
mytype <- list(nominal = c("Name", "Sex", "Ticket", "Cabin", "Embarked"), ordratio = c("Pclass"))

## -----------------------------------------------------------------------------
x_train <- traindata[, -12]
y_train <- traindata[,  12]
vcrtrain <- vcr.rpart.train(x_train, y_train, rpart.out, mytype)
names(vcrtrain)

vcrtrain$predint[1:10] # prediction as integer
vcrtrain$pred[1:10]    # prediction as label
vcrtrain$altint[1:10]  # alternative label as integer
vcrtrain$altlab[1:10]  # alternative label

# Probability of Alternative Class (PAC) of each object:
vcrtrain$PAC[1:3] 
#
summary(vcrtrain$PAC)

# f(i, g) is the distance from case i to class g:
vcrtrain$fig[1:3, ] # for the first 3 objects:

# The farness of an object i is the f(i, g) to its own class: 
vcrtrain$farness[1:3]
#
summary(vcrtrain$farness)

# The "overall farness" of an object is defined as the 
# lowest f(i, g) it has to any class g (including its own):
summary(vcrtrain$ofarness)

sum(vcrtrain$ofarness > 0.99, na.rm = TRUE) 
# No farness is considered outlying in these data.

confmat.vcr(vcrtrain) 

cols <- c("firebrick", "blue")

## -----------------------------------------------------------------------------
stackedplot(vcrtrain, classCols = cols,
            main = "Stacked plot of rpart on Titanic training data")

## -----------------------------------------------------------------------------
silplot(vcrtrain, classCols = cols, 
        main = "Silhouettes of rpart on Titanic training data")
# silplot.out <- silplot(vcrtrain, classCols = cols)
# ggsave("titanic_train_silhouettes.pdf", silplot.out,
#        width = 5, height = 5)

## -----------------------------------------------------------------------------
hist(x_train$Age)

# Quasi residual plot versus age, for males only:

# pdf("titanic_qrp_versus_age_males.pdf", width = 4.8, height = 4.8)
PAC <- vcrtrain$PAC[which(x_train$Sex == "male")]
feat <- x_train$Age[which(x_train$Sex == "male")]
qresplot(PAC, feat, xlab = "Age (years)", opacity = 0.5,
         main = "quasi residual plot for male passengers",
         plotLoess = TRUE)
text(x = 14, y = 0.60, "loess curve", col = "red", cex = 1)
# dev.off()

## -----------------------------------------------------------------------------
classmap(vcrtrain, "casualty", classCols = cols)
# classmap(vcrtrain, "casualty", classCols = cols, identify = TRUE)

# blue points top right: cases "a" and "b" in the paper
x_train[which(y_train == "casualty")[119], ]
# Woman in 1st class, should have survived.
#
x_train[which(y_train == "casualty")[192], ]

# Similar, but child.

# red point most to the right: case "c" in the paper
x_train[which(y_train == "casualty")[268], ]


## -----------------------------------------------------------------------------
classmap(vcrtrain, "survived", classCols = cols)
# classmap(vcrtrain, "survived", classCols = cols, identify = TRUE)

# red point with highest farness among highest PAC: case "d" in the paper
x_train[which(y_train == "survived")[c(14)], ] 

# near-coinciding points with highest farness: cases "e" and "f" in the paper
#
x_train[which(y_train == "survived")[265], ]

x_train[which(y_train == "survived")[287], ] 

# man --> predicted as not survived. also paid the highest 
# fare in the data and has same ticket number as passenger 
# 265. Also embarked at the same place. It turns out that
# Gustave Lesueur was the man servant of the banker Thomas Cardeza:
  
# https://www.encyclopedia-titanica.org/titanic-survivor/thomas-cardeza.html
# https://www.encyclopedia-titanica.org/titanic-survivor/gustave-lesueur.html

# blue point bottom right: case "g" in the paper
x_train[which(y_train == "survived")[90], ]

# # woman  + first class so predicted in survivors.
# # paid highest fare in whole dataset 

## -----------------------------------------------------------------------------
testdata <- data_titanic[which(data_titanic$dataType == "test"), -13]

dim(testdata)
x_test <- testdata[, -12]
y_test <- testdata[, 12]
table(y_test)

## -----------------------------------------------------------------------------
vcrtest <- vcr.rpart.newdata(x_test, y_test, vcrtrain)

confmat.vcr(vcrtest)
cols <- c("firebrick", "blue")

## -----------------------------------------------------------------------------
stackedplot(vcrtest, classCols = cols,
            main = "Stacked plot of Titanic test data")

## -----------------------------------------------------------------------------
silplot(vcrtest, classCols = cols, 
        main = "Silhouettes of rpart on Titanic test data")

## -----------------------------------------------------------------------------
classmap(vcrtest, "casualty", classCols = cols) 

## -----------------------------------------------------------------------------
classmap(vcrtest, "survived", classCols = cols) 

## ----echo = FALSE-------------------------------------------------------------
knitr::opts_chunk$set(
 fig.width  = 5 ,
 fig.height = 3.5,
 fig.align  = 'center'
)
oldpar <- list(mar = par()$mar, mfrow = par()$mfrow)

## -----------------------------------------------------------------------------
library(nnet)
library(classmap)

## -----------------------------------------------------------------------------
set.seed(123)
nn.iris <- nnet(Species ~ ., data = iris, size = 3)
names(nn.iris)
head(nn.iris$fitted.values) # matrix of posterior probabilities.
nn.iris$wts # the "weights" (coefficients, some are negative).
summary(nn.iris) 

## -----------------------------------------------------------------------------
train_x <- as.matrix(iris[, -5]) # original input: 4 variables.
# extract "weights" (coefficients) from each layer:
bias1  <- nn.iris$wts[c(1, 6, 11)] # intercepts to hidden layer
betas1 <- cbind(nn.iris$wts[2:5], nn.iris$wts[7:10], 
                nn.iris$wts[12:15]) # slopes to hidden layer
bias2  <- nn.iris$wts[c(16, 20, 24)] # intercepts to output layer
betas2 <- cbind(nn.iris$wts[17:19], nn.iris$wts[21:23], 
                nn.iris$wts[25:27]) # slopes to output layer.

## -----------------------------------------------------------------------------
sigmoid <- function(x) 1 / (1 + exp(-x)) 

H <- t(t(train_x %*% betas1) + bias1) # = hidden layer
X <- t(t(sigmoid(H) %*% betas2) + bias2) # = layer before softmax.
pairs(X, col = unclass(iris$Species)) # G=3 classes ==> 3 dimensions.

## -----------------------------------------------------------------------------
softmax <- function(x) exp(x) / sum(exp(x)) 
outprobs <- t(apply(X, 1, softmax))
trainprobs <- nn.iris$fitted.values # this is after softmax
range(trainprobs - outprobs) # near 0, so we have a match.

## -----------------------------------------------------------------------------
y <- iris[, 5]
vcrtrain <- vcr.neural.train(X, y, trainprobs)

names(vcrtrain)
vcrtrain$predint[c(1:10, 51:60, 101:110)] # prediction as integer
vcrtrain$pred[c(1:10, 51:60, 101:110)]    # prediction as label
vcrtrain$altint[c(1:10, 51:60, 101:110)]  # alternative label as integer
vcrtrain$altlab[c(1:10, 51:60, 101:110)]  # alternative label

# Probability of Alternative Class (PAC) of each object:
vcrtrain$PAC[1:3] 
#
summary(vcrtrain$PAC)

# f(i, g) is the distance from case i to class g:
vcrtrain$fig[1:5, ] # for the first 5 objects:

# The farness of an object i is the f(i, g) to its own class: 
vcrtrain$farness[1:5]
#
summary(vcrtrain$farness)

# The "overall farness" of an object is defined as the 
# lowest f(i, g) it has to any class g (including its own):
summary(vcrtrain$ofarness)

confmat.vcr(vcrtrain)

## -----------------------------------------------------------------------------
cols <- c("red", "darkgreen", "blue")

# stacked mosaic plot:
stplot <- stackedplot(vcrtrain, classCols = cols,
                     main = "Stacked plot of nnet on iris data") 
stplot

# silhouette plot:
silplot(vcrtrain, classCols = cols, main =
          "Silhouette plot of nnet on iris data") 

# class maps:
classmap(vcrtrain, "setosa", classCols = cols)
# Very tight class (low PAC, no high farness).
classmap(vcrtrain, "versicolor", classCols = cols)
# Not so tight, one point is predicted as virginica.
classmap(vcrtrain, "virginica", classCols = cols)
# Also tight.

## -----------------------------------------------------------------------------
test_x <- train_x[1:100, ]
ynew <- y[1:100]
ynew[c(1:10, 51:60)] <- NA
pairs(train_x, col = as.numeric(y) + 1, pch = 19) # 3 colors
pairs(test_x, col = as.numeric(ynew) + 1, pch = 19) # only red and green

## -----------------------------------------------------------------------------
predprobs <- predict(nn.iris, test_x, type = "raw")
range(predprobs - trainprobs[1:100, ]) # perfect match

# Reconstruct this prediction:
Hnew <- t(t(test_x %*% betas1) + bias1) # hidden layer
Xnew <- t(t(sigmoid(Hnew) %*% betas2) + bias2) # layer before softmax
probsnew <- t(apply(Xnew, 1, softmax))
range(probsnew - predprobs) # ~0, so we have a match

## -----------------------------------------------------------------------------
vcrtest <- vcr.neural.newdata(Xnew, ynew, probsnew, vcrtrain)

plot(vcrtest$predint, vcrtrain$predint[1:100]); abline(0, 1) # identical, OK
plot(vcrtest$altint, vcrtrain$altint[1:100]); abline(0, 1) # identical when not NA, OK
plot(vcrtest$PAC, vcrtrain$PAC[1:100]); abline(0, 1) # OK
vcrtest$farness # length 100, with NA's where ynew is NA
plot(vcrtest$farness, vcrtrain$farness[1:100]); abline(0, 1) # identical where not NA
plot(vcrtest$fig, vcrtrain$fig[1:100, ]); abline(0, 1) # same, OK
vcrtest$ofarness # 100, withOUT NA's: ofarness even exists for cases with missing ynew, for which farness cannot exist.
plot(vcrtest$ofarness, vcrtrain$ofarness[1:100]); abline(0, 1) # same, OK

confmat.vcr(vcrtest) # as expected:

## -----------------------------------------------------------------------------
stplot # to compare with:
stackedplot(vcrtest, classCols = cols, separSize = 1.5,
            minSize = 1, main = "Stacked plot of nnet on subset of iris data")
silplot(vcrtest, classCols = cols, main = "Silhouette plot of nnet on subset of iris data")

classmap(vcrtrain, 1, classCols = cols)
classmap(vcrtest, 1, classCols = cols) # same, but fewer points
#
classmap(vcrtrain, 2, classCols = cols)
classmap(vcrtest, 2, classCols = cols) # same, but fewer points
#
# classmap(vcrtrain, 3, classCols = cols)
# classmap(vcrtest, 3, classCols = cols)
# # Class number 3 with label virginica has no objects to visualize.

## -----------------------------------------------------------------------------
data("data_floralbuds")

## -----------------------------------------------------------------------------
set.seed(123)
nn.buds <- nnet(y ~ ., data = data_floralbuds, size = 2)
names(nn.buds)
head(nn.buds$fitted.values) # matrix of posterior probabilities of each class
nn.buds$wts # the "weights" (coefficients, some negative)
summary(nn.buds) # A hidden layer with 2 neurons:

## -----------------------------------------------------------------------------
train_x <- as.matrix(data_floralbuds[, -7])
# extract weights for each layer:
bias1  <- nn.buds$wts[c(1, 8)]
betas1 <- cbind(nn.buds$wts[2:7], nn.buds$wts[9:14])
bias2  <- nn.buds$wts[c(15, 18, 21, 24)]
betas2 <- cbind(nn.buds$wts[16:17], nn.buds$wts[19:20], 
                nn.buds$wts[22:23], nn.buds$wts[25:26])

## -----------------------------------------------------------------------------
H <- t(t(train_x %*% betas1) + bias1) # hidden layer
X <- t(t(sigmoid(H) %*% betas2) + bias2) # layer before softmax
outprobs <- t(apply(X, 1, softmax))
trainprobs <- nn.buds$fitted.values # posterior probabilities
range(trainprobs - outprobs) # ~0
# OK, we have reconstructed the posterior probabilities
pairs(X, col = data_floralbuds$y)

## -----------------------------------------------------------------------------
y <- data_floralbuds$y
vcrtrain <- vcr.neural.train(X, y, trainprobs)
cols <- c("saddlebrown", "orange", "olivedrab4", "royalblue3")

## -----------------------------------------------------------------------------
stackedplot(vcrtrain, classCols = cols, main =
              "Stacked plot of nnet on floral buds data")

# Silhouette plot:
silplot(vcrtrain, classCols = cols, 
        main = "Silhouette plot of nnet on floral buds data")



# Quasi residual plot:

PAC <- vcrtrain$PAC
feat <- rowSums(train_x); xlab <- "rowSums(X)"
# pdf("Floralbuds_quasi_residual_plot.pdf", width = 5, height = 4.8)
qresplot(PAC, feat, xlab = xlab, plotErrorBars = TRUE, fac = 2, main = "Floral buds: quasi residual plot")
# images with higher sum are easier to classify
# dev.off()
cor.test(feat, PAC, method = "spearman") 
# rho = -0.238, p-value < 2e-8 
# ==> decreasing trend is significant



# Class maps:

classmap(vcrtrain, "branch", classCols = cols)

classmap(vcrtrain, "bud", classCols = cols)

classmap(vcrtrain, "scales", classCols = cols)

classmap(vcrtrain, "support", classCols = cols)

vcr.svm.train = function (X, y, svfit, ortho = FALSE){
  X <- as.matrix(X)
  if (nrow(X) == 1)
    X <- t(X)
  if (sum(is.na(as.vector(X))) > 0) {
    stop("The data matrix X has NA's.")
  }
  n <- nrow(X)
  d <- ncol(X)
  if (n < 2)
    stop(" X should have more than one row")
  checked <- checkLabels(y, n, training = TRUE)
  lab2int <- checked$lab2int
  levels <- checked$levels
  nlab <- length(levels)
  yint <- lab2int(y)
  indsv <- checked$indsv
  nvis <- length(indsv)
  yintv <- yint[indsv]
  if (ortho == TRUE & nlab != 2)
    stop(" ortho=TRUE is only possible for 2 classes")
  if (is.null(svfit$compprob))
    stop(" svfit$compprob is missing")
  if (svfit$compprob == FALSE) {
    stop(paste0("\n The object svfit needs to be obtained by ",
                "\n running e1071::svm with probability=TRUE"))
  }
  if (is.null(svfit$kernel)) {
    stop(" svfit$kernel is missing")
  }
  else {
    k <- svfit$kernel
  }
  if (!(k == 0 | k == "linear" | k == 1 | k == "polynomial" |
        k == 2 | k == "radial" | k == 3 | k == "sigmoid")) {
    stop(paste(" Unknown kernel type: ", k, sep = ""))
  }
  if (is.null(svfit$scaled)) {
    stop(" svfit$scaled is missing")
  }
  else {
    scaled <- svfit$scaled
  }
  if (is.null(svfit$degree)) {
    stop(" svfit$degree is missing")
  }
  else {
    degree <- svfit$degree
  }
  if (is.null(svfit$gamma)) {
    stop(" svfit$gamma is missing")
  }
  else {
    gamma <- svfit$gamma
  }
  if (is.null(svfit$coef0)) {
    stop(" svfit$coef0 is missing")
  }
  else {
    coef0 <- svfit$coef0
  }
  if (is.null(svfit$cost)) {
    stop(" svfit$cost is missing")
  }
  else {
    cost <- svfit$cost
  }
  if (is.null(svfit$fitted) | is.null(svfit$decision.values)) {
    if (k == 0) {
      ktype <- "linear"
    }
    if (k == 1) {
      ktype <- "polynomial"
    }
    if (k == 2) {
      ktype <- "radial"
    }
    if (k == 3) {
      ktype <- "sigmoid"
    }
    svfit <- svm(X, y, scale = scaled, kernel = ktype, degree = degree,
                 gamma = gamma, coef0 = coef0, cost = cost, probability = TRUE)
  }
  preds <- predict(object = svfit, newdata = data.frame(X = X),
                   probability = TRUE)
  predint <- lab2int(preds)
  altint <- PAC <- rep(NA, n)
  altintv <- PACv <- rep(NA, nvis)
  if (nlab == 2) {
    altintv <- 3 - yintv
    probs <- attr(preds, "probabilities")[indsv, ]
    probs <- probs[, order(lab2int(colnames(probs)))]
    for (i in seq_len(nvis)) PACv[i] <- probs[i, altintv[i]]
  }
  if (nlab > 2) {
    probs <- attr(preds, "probabilities")[indsv, ]
    probs <- probs[, order(lab2int(colnames(probs)))]
    for (i in seq_len(nvis)) {
      ptrue <- probs[i, yintv[i]]
      others <- (seq_len(nlab))[-yintv[i]]
      palt <- max(probs[i, others])
      altintv[i] <- others[which.max(probs[i, others])]
      PACv[i] <- palt/(ptrue + palt)
    }
  }
  altint[indsv] <- altintv
  PAC[indsv] <- PACv
  X <- X[indsv, ]
  Xloadings <- transfmat <- nbeta <- NULL
  if (k == 0) {
    Xfar <- X
    PCout <- cellWise::truncPC(Xfar, ncomp = min(20, ncol(Xfar)), center = FALSE,
                               scale = FALSE)
    eigs <- PCout$eigenvalues[seq_len(PCout$rank)]
    qkeep <- sum(eigs > 1e-08)
    Xloadings <- PCout$loadings[, seq_len(qkeep), drop = FALSE]
    Xfar <- Xfar %*% Xloadings
  }
  else { # k >= 1
    Xfar <- makeKernel(X, svfit = svfit)
    transfmat <- makeFV(Xfar, precS = 1e-12)$transfmat
    if(ncol(transfmat) > 20) transfmat = transfmat[, seq_len(20)]
    Xfar <- Xfar %*% transfmat
  }
  if (ortho == TRUE) {
    fitfs <- svm(Xfar, y, scale = FALSE, kernel = "linear",
                 cost = cost)
    beta <- coef(fitfs)[-1]
    beta <- matrix(beta, ncol = 1)
    nbeta <- beta/sqrt(sum(beta * beta))
    Xfar <- Xfar - Xfar %*% nbeta %*% t(nbeta)
  }
  farout <- compFarness(type = "pca", testdata = FALSE,
                         yint = yintv, nlab = nlab, X = Xfar, fig = NULL, d = NULL,
                         figparams = NULL, PCAfits = NULL, keepPCA = TRUE)
  fig <- matrix(rep(0, n * nlab), ncol = nlab)
  fig[indsv, ] <- farout$fig
  farness <- ofarness <- rep(NA, n)
  farness[indsv] <- farout$farness
  ofarness[indsv] <- farout$ofarness
  figparams <- farout$figparams
  figparams$ortho <- ortho
  figparams$nbeta <- nbeta
  figparams$Xloadings <- Xloadings
  figparams$transfmat <- transfmat
  figparams$PCAfits <- farout$PCAfits
  return(list(yint = yint, y = levels[yint], levels = levels,
              predint = predint, pred = levels[predint], altint = altint,
              altlab = levels[altint], PAC = PAC, figparams = figparams,
              fig = fig, farness = farness, ofarness = ofarness, svfit = svfit,
              X = X))
}


vcr.svm.newdata = function (Xnew, ynew = NULL, vcr.svm.train.out){
  Xnew <- as.matrix(Xnew)
  if (nrow(Xnew) == 1) Xnew <- t(Xnew)
  n <- nrow(Xnew)
  if (is.null(vcr.svm.train.out$figparams$PCAfits)) {
    stop(paste0("\nFor test data, the farness computation ",
                "requires the trained object", "\nto contain ",
                "the value figparams$PCAfits."))
  }
  PCAfits <- vcr.svm.train.out$figparams$PCAfits
  if (sum(is.na(as.vector(Xnew))) > 0)
    stop("The data matrix Xnew contains NA's.")
  levels <- vcr.svm.train.out$levels
  nlab <- length(levels)
  if (is.null(ynew))
    ynew <- factor(rep(NA, n), levels = levels)
  checked <- checkLabels(ynew, n, training = FALSE, levels)
  lab2int <- checked$lab2int
  indsv <- checked$indsv
  nvis <- length(indsv)
  yintnew <- rep(NA, n)
  yintnew[indsv] <- lab2int(ynew[indsv])
  yintv <- yintnew[indsv]
  svfit <- vcr.svm.train.out$svfit
  k <- svfit$kernel
  preds <- predict(object = svfit, newdata = data.frame(X = Xnew),
                   decision.values = TRUE, probability = TRUE)
  predint <- lab2int(preds)
  altint <- PAC <- rep(NA, n)
  if (nvis > 0) {
    altintv <- PACv <- rep(NA, nvis)
    if (nlab == 2) {
      altintv <- 3 - yintv
      probs <- attr(preds, "probabilities")[indsv,
      ]
      probs <- probs[, order(lab2int(colnames(probs)))]
      for (i in seq_len(nvis)) PACv[i] <- probs[i, altintv[i]]
    }
    if (nlab > 2) {
      probs <- attr(preds, "probabilities")[indsv,
      ]
      probs <- probs[, order(lab2int(colnames(probs)))]
      for (i in seq_len(nvis)) {
        ptrue <- probs[i, yintv[i]]
        others <- (seq_len(nlab))[-yintv[i]]
        palt <- max(probs[i, others])
        altintv[i] <- others[which.max(probs[i, others])]
        PACv[i] <- palt/(ptrue + palt)
      }
    }
    PAC[indsv] <- PACv
    altint[indsv] <- altintv
  }
  Zfar <- Xnew
  rm(Xnew)
  # dim(Zfar) # 75 2
  if (k == 0) {
    Xloadings <- vcr.svm.train.out$figparams$Xloadings
    if (!is.null(Xloadings)) {
      Zfar <- Zfar %*% Xloadings
    }
  }
  else {
    if (is.null(vcr.svm.train.out$figparams$transfmat)) {
      stop("vcr.svm.train.out$figparams$transfmat is missing")
    }
    Zfar <- makeFV(makeKernel(Zfar, vcr.svm.train.out$X,
                               svfit = vcr.svm.train.out$svfit), vcr.svm.train.out$figparams$transfmat)$Xf
    # dim(Zfar) # 75 196
  }
  if (vcr.svm.train.out$figparams$ortho == TRUE) {
    nbeta <- vcr.svm.train.out$figparams$nbeta
    Zfar <- Zfar - Zfar %*% nbeta %*% t(nbeta)
  }
  farout <- compFarness(type = "pca", testdata = TRUE,
                         yint = yintnew, nlab = nlab, X = Zfar, fig = NULL, d = NULL,
                         figparams = vcr.svm.train.out$figparams, PCAfits = PCAfits,
                         keepPCA = FALSE)
  return(list(yintnew = yintnew, ynew = levels[yintnew], levels = levels,
              predint = predint, pred = levels[predint], altint = altint,
              altlab = levels[altint], PAC = PAC, fig = farout$fig,
              farness = farout$farness, ofarness = farout$ofarness))
}

vcr.neural.train <- function(X, y, probs, estmethod = meancov) {
  #
  # Using the outputs of a neural network for classification
  # applied to the training data, this function prepares for
  # graphical displays.
  #
  # Arguments:
  #   X         : the coordinates of the n objects of the training
  #               data, in the layer chosen by the user. Missing
  #               values are not allowed.
  #   y         : factor with the given class labels of the objects.
  #               Make sure that the levels are in the same order as
  #               used in the neural net, i.e. the columns of its
  #               binary "once-hot-encoded" response vectors.
  #   probs     : posterior probabilities obtained by the neural
  #               net, e.g. in keras. For each case (row of X),
  #               the classes have probabilities that add up to 1.
  #               Each row of the matrix probs contains these
  #               probabilities. The columns of probs must be in
  #               the same order as the levels of y.
  #   estmethod : function for location and covariance estimation.
  #               Should return a list with $m and $S. Can be meancov
  #               (classical mean and covariance matrix) or DetMCD.
  #               If one or more classes have a singular covariance
  #               matrix, the function automatically switches to
  #               the PCA-based farness used in vcr.svm.train().
  #
  # Returns:
  #   yint      : given labels as integer 1, 2, 3, ...
  #   y         : given labels
  #   levels    : levels of y
  #   predint   : predicted class number. For each case this
  #               is the class with the highest probability,
  #               that is, which.max(probs[i,]).
  #   pred      : predicted label of each object.
  #   altint    : alternative class as integer, i.e. the non-given
  #               class number with the highest mixture density.
  #   altlab    : alternative class of each object.
  #   classMS   : list with center and covariance matrix of each class
  #   PAC       : probability of alternative class for each object
  #   PCAfits   : if not NULL, PCA fits to each class, estimated from
  #               the training data but also useful for new data.
  #   figparams : parameters for computing fig, can be used for
  #               new data.
  #   fig       : distance of each object i from each class g.
  #   farness   : farness of each object to its given class.
  #   ofarness  : For each object i, its lowest farness to any
  #               class, including its own. Always exists.
  #
  keepPCA <- TRUE; prec <- 1e-10

  X <- as.matrix(X) # in case it is a data frame
  if (nrow(X) == 1) X <- t(X)
  if (sum(is.na(as.vector(X))) > 0) {
    stop("The coordinate matrix X has NA's.")
  }
  n <- nrow(X)
  d <- ncol(X)
  if (n < 2) stop("The training data should have more than one case.")
  # Check whether y and its levels are of the right form:
  checked <- checkLabels(y, n, training = TRUE)
  # If it did not stop: yint has length n, and its values are in
  # 1, ..., nlab without gaps. It is NA where y is: outside indsv.
  lab2int <- checked$lab2int # is a function
  indsv   <- checked$indsv
  levels  <- checked$levels
  nlab    <- length(levels)
  yint    <- lab2int(y)
  yintv   <- yint[indsv]
  classSizes <- rep(0, nlab)
  for (g in seq_len(nlab)) {classSizes[g] <- sum(yintv == g)}
  # classSizes
  #
  # Check matrix of posterior probabilities:
  #
  probs <- as.matrix(probs)
  if (length(dim(probs)) != 2) stop("probs should be a matrix.")
  if (nrow(probs) != n) stop(paste0(
    "The matrix probs should have ", n, " rows"))
  if (ncol(probs) != nlab) stop(paste0(
    "The matrix probs should have ", nlab, " columns"))
  if (any(is.na(probs))) stop("probs should not have any NA's.")
  #
  # Compute prediction for all objects in the training data:
  #
  predint <- apply(probs[, , drop = FALSE], 1, which.max)
  #
  # Compute ptrue and palt for all objects with available y:
  #
  ptrue <- palt <- altint <- PAC <- rep(NA, n)
  for (g in seq_len(nlab)) { # g=1
    clinds <- indsv[which(yintv == g)] # indices in 1, ..., n
    others <- (seq_len(nlab))[-g] # alternative classes
    ptrue[clinds]  <- probs[clinds, g]
    palt[clinds]   <- apply(probs[clinds, others, drop = FALSE], 1, max)
    altint[clinds] <- others[apply(probs[clinds, others, drop = FALSE],
                                   1, which.max)]
  }
  #
  # Compute PAC:
  #
  PAC[indsv] <- palt[indsv] / (ptrue[indsv] + palt[indsv])
  # (PAC and altint stay NA outside indsv)
  #
  # Compute farness:
  #
  computeMD <- TRUE
  tinyclasses <- which(classSizes < (d + 1))
  if (length(tinyclasses) > 0) {
    wnq(paste0("\nThere is at least one class with fewer",
               " than the ", d + 1, "\nmembers required to compute a ",
               "nonsingular covariance matrix, ", "\nso farness ",
               "will be computed from PCA."))
    computeMD <- FALSE
  } else {
    classMS <- list()
    # Compute the center and covariance matrix of each class:
    iCN <- rep(NA, nlab) # for numerical conditions
    for (g in seq_len(nlab)) {
      clinds       <- indsv[which(yintv == g)]
      # class indices in 1, ..., n
      class.out    <- estmethod(X[clinds, ]) # contains mu and Sigma
      iCN[g]       <- numericalCond(class.out$S)
      classMS[[g]] <- class.out # now has $m, $S
    }
    tinyCN <- which(iCN < prec) # also ran with 1e-10 to test PCA
    if (length(tinyCN) > 0) {
      wnq(paste0("\nThere is at least one class with ",
                 "(near-) singular covariance matrix, ",
                 "\nso farness will be computed from PCA."))
      computeMD <- FALSE
    }
  }
  if (computeMD == TRUE) {
    # compute MD2 of each object to all classes:
    initfig <- matrix(0, n, nlab) # for initial farness
    for (g in seq_len(nlab)) { # g=1
      M <- classMS[[g]]$m
      S <- classMS[[g]]$S # now S is nonsingular
      initfig[, g]  <- sqrt(mahalanobis(X, M, S))
      # Computed for all objects, even when response y is NA.
    }
    farout <- compFarness(type = "affine", testdata = FALSE, yint = yint,
                          nlab = nlab, X = NULL, fig = initfig,
                          d = d, figparams = NULL)
  } else {
    classMS <- NULL
    farout <- compFarness(type = "pca", testdata = FALSE, yint = yint,
                          nlab = nlab, X = X, fig = NULL,
                          d = NULL, figparams = NULL,
                          PCAfits = NULL, keepPCA = keepPCA)
  }

  figparams <- farout$figparams
  figparams$ncolX <- d
  figparams$computeMD <- computeMD
  figparams$classMS <- classMS
  figparams$PCAfits <- farout$PCAfits
  return(list(X = X,
              yint = yint,
              y = levels[yint],
              levels = levels,
              predint = predint,
              pred = levels[predint],
              altint = altint,
              altlab = levels[altint],
              PAC = PAC,
              figparams = figparams,
              fig = farout$fig,
              farness = farout$farness,
              ofarness = farout$ofarness))
}


vcr.neural.newdata <- function(Xnew, ynew = NULL, probs,
                               vcr.neural.train.out){
  #
  # Prepares graphical display of new data fitted by a neural
  # net that was modeled on the training data, using the output
  # of vcr.neural.train() on the training data.
  #
  # Arguments:
  #   Xnew                 : data matrix of the new data, with the
  #                          same number of columns d as in the
  #                          training data.
  #                          Missing values are not allowed.
  #   ynew                 : factor with class membership of each new
  #                          case. Can be NA for some or all cases.
  #                          If NULL, is assumed to be NA everywhere.
  #   probs                : posterior probabilities obtained by
  #                          running the neural net on the new data.
  #   vcr.neural.train.out : output of vcr.neural.train() on the
  #                          training data.
  #
  # Returns:
  #   yintnew   : given labels as integers 1, 2, 3, ..., nlab.
  #               Can have NA's.
  #   ynew      : labels if yintnew is available, else NA.
  #   levels    : levels of the response, from vcr.neural.train.out
  #   predint   : predicted label as integer, always exists.
  #   pred      : predicted label of each object
  #   altint    : alternative label as integer if yintnew was given,
  #               else NA.
  #   altlab    : alternative label if yintnew was given, else NA.
  #   classMS   : list with center and covariance matrix of each class,
  #               from vcr.neural.train.out
  #   PAC       : probability of alternative class for each object
  #               with non-NA yintnew.
  #   figparams : (from training) parameters used to compute fig
  #   fig       : farness of each object i from each class g.
  #               Always exists.
  #   farness   : farness of each object to its given class, for
  #               objects with non-NA yintnew.
  #   ofarness  : For each object i, its lowest farness to any
  #               class, including its own. Always exists.
  #
  Xnew <- as.matrix(Xnew) # in case it is a data frame
  if (nrow(Xnew) == 1) Xnew <- t(Xnew)
  n <- nrow(Xnew)
  d <- vcr.neural.train.out$figparams$ncolX
  if (ncol(Xnew) != d) {
    stop(paste0("Xnew should have ", d,
                " columns, like the training data."))
  }
  if (sum(is.na(as.vector(Xnew))) > 0) {
    stop("The coordinate matrix Xnew contains NA's.")
  }
  levels <- vcr.neural.train.out$levels # as in training data
  nlab   <- length(levels) # number of classes
  #
  if (is.null(ynew)) ynew <- factor(rep(NA, n), levels = levels)
  checked <- checkLabels(ynew, n, training = FALSE, levels)
  lab2int <- checked$lab2int # is a function
  indsv   <- checked$indsv   # INDiceS of cases we can Visualize,
  #                         # can even be empty, is OK.
  nvis    <- length(indsv)   # can be zero.
  yintnew <- rep(NA, n)       # needed to overwrite "new" levels.
  yintnew[indsv] <- lab2int(ynew[indsv])
  # Any "new" levels are now set to NA.
  #
  probs <- as.matrix(probs)
  if (length(dim(probs)) != 2) stop("probs should be a matrix.")
  if (ncol(probs) == 1) probs <- t(probs) # if new data is 1 object
  if (ncol(probs) != nlab) stop(paste0(
    "The matrix probs should have ", nlab, " columns"))
  if (any(is.na(probs))) stop("probs should not have any NA's.")
  #
  # Compute prediction for all objects in the new data:
  #
  predint <- apply(probs[, , drop = FALSE], 1, which.max)
  #
  # Compute PAC for all objects with available y:
  #
  ptrue <- palt <- altint <- PAC <- rep(NA, n)
  if (nvis > 0) { # if there are new data with labels
    yintv  <- yintnew[indsv] # yint of cases we will Visualize
    ayintv <- sort(unique(yintv)) # available yintv values
    # ayintv can be a proper subset of 1, ..., nlab for new data.
    for (g in ayintv) {
      clinds <- indsv[which(yintv == g)] # indices in 1, ..., n
      others <- (seq_len(nlab))[-g] # non-self classes
      ptrue[clinds]  <- probs[clinds, g]
      palt[clinds]   <- apply(probs[clinds, others, drop  = FALSE], 1, max)
      altint[clinds] <- others[apply(probs[clinds, others, drop = FALSE],
                                     1, which.max)]
    }
    PAC[indsv] <- palt[indsv] / (ptrue[indsv] + palt[indsv])
    # (PAC and altint stay NA outside indsv)
  }
  #
  # Compute farness:
  #
  if (vcr.neural.train.out$figparams$computeMD == TRUE) {
    # Compute MD2S
    initfig <- matrix(0, n, nlab) # for Mahalanobis distances
    classMS <- vcr.neural.train.out$figparams$classMS
    for (g in seq_len(nlab)) {
      M <- classMS[[g]]$m
      S <- classMS[[g]]$S
      initfig[, g] <- sqrt(mahalanobis(Xnew, M, S))
      # Computed for all objects, even when ynew is NA.
    }
    farout <- compFarness(type = "affine", testdata = TRUE,
                          yint = yintnew, nlab = nlab, X = NULL,
                          fig = initfig, d = d,
                          figparams = vcr.neural.train.out$figparams)
  } else {
    farout <- compFarness(type = "pca", testdata = TRUE, yint = yintnew,
                          nlab = nlab, X = Xnew, fig = NULL, d = NULL,
                          figparams = vcr.neural.train.out$figparams,
                          PCAfits = vcr.neural.train.out$figparams$PCAfits,
                          keepPCA = FALSE)
  }
  return(list(yintnew = yintnew,
              ynew = levels[yintnew],
              levels = levels,
              predint = predint,
              pred = levels[predint],
              altint = altint,
              altlab = levels[altint],
              PAC = PAC,
              fig = farout$fig,
              farness = farout$farness,
              ofarness = farout$ofarness))
}

## ----echo = FALSE-------------------------------------------------------------
knitr::opts_chunk$set(
 fig.width  = 5 ,
 fig.height = 3.5,
 fig.align  = 'center'
)
oldpar <- list(mar = par()$mar, mfrow = par()$mfrow)

## -----------------------------------------------------------------------------
library(randomForest)
library(classmap)

## -----------------------------------------------------------------------------
data("data_instagram")
traindata <- data_instagram[which(data_instagram$dataType == "train"), -13]
str(traindata)


# The variable names and their interpretation are
colnames(traindata)

# profile.pic: binary, indicates whether profile has picture
# nums.length.username: ratio of number of numerical chars in username to its length
# fullname.words: number of words in full name
# nums.length.fullname: ratio of number of numerical characters in full name to its length
# name..username: binary, indicates whether name == username of the profile
# description.length: length of the description/biography of the profile (in number of characters)
# external.URL: binary, indicates whether profile has external url
# private: binary, indicates whether profile is private or not
# X.posts: number of posts made by profile
# X.followers: number of followers
# X.follows: numbers of follows
# y: whether profile is fake or not.

x_train <- traindata[, -12]
y_train <- traindata[, 12]

dim(traindata)
table(traindata$y) # 50/50 split of genuine/fake accounts:

## -----------------------------------------------------------------------------
set.seed(71) 
rfout <- randomForest(y ~ ., data = traindata, keep.forest = TRUE)

## -----------------------------------------------------------------------------
mytype <- list(symm = c(1, 5, 7, 8)) 

## -----------------------------------------------------------------------------
vcrtrain <- vcr.forest.train(X = x_train, y = y_train,
                            trainfit = rfout, type = mytype)

names(vcrtrain)
vcrtrain$predint[c(1:10, 301:310)] # prediction as integer
vcrtrain$pred[c(1:10, 301:310)]    # prediction as label
vcrtrain$altint[c(1:10, 301:310)]  # alternative label as integer
vcrtrain$altlab[c(1:10, 301:310)]  # alternative label

# Probability of Alternative Class (PAC) of each object:
vcrtrain$PAC[1:3] 
#
summary(vcrtrain$PAC)

# f(i, g) is the distance from case i to class g:
vcrtrain$fig[1:3, ] # for the first 3 objects:

# The farness of an object i is the f(i, g) to its own class: 
vcrtrain$farness[1:3]
#
summary(vcrtrain$farness)

# The "overall farness" of an object is defined as the 
# lowest f(i, g) it has to any class g (including its own):
summary(vcrtrain$ofarness)

sum(vcrtrain$ofarness > 0.99, na.rm = TRUE) 
# With the default cutoff = 0.99 we find 6 outliers,
# also shown in the last column of the confusion matrix:

confmat.vcr(vcrtrain) 

# If we do not want to show the outliers:
confmat.vcr(vcrtrain, showOutliers = FALSE)

# Note that the accuracy is computed before any objects
# are flagged, so it does not depend on the cutoff.
# Here the accuracy is `perfect' due to overfitting. 
# The out-of-box prediction accuracy is about 92%.
cols <- c("blue", "red3")

## -----------------------------------------------------------------------------
stackedplot(vcrtrain, classCols = cols, main =
              "Instagram training data")

# Silhouette plot:
silplot(vcrtrain, classCols = cols)
# Here all the s(i) are nonnegative (due to overfitting).

# Class maps:
classmap(vcrtrain, "genuine", classCols = cols) #, identify = TRUE)

# farness outliers from furthest to closer: 45, 25, 41
x_train[c(45, 25, 41), ] # they have huge numbers of followers.

classmap(vcrtrain, "fake", classCols = cols) #, identify = TRUE)
# only case 261 is borderline far.

## -----------------------------------------------------------------------------
testdata <- data_instagram[which(data_instagram$dataType == "test"), -13]
Xnew <- testdata[, -12]
ynew <- testdata[, 12]

## -----------------------------------------------------------------------------
vcrtest <- vcr.forest.newdata(Xnew, ynew, vcrtrain)

confmat.vcr(vcrtest)

## -----------------------------------------------------------------------------
stackedplot(vcrtest, classCols = cols, 
            main = "RF on Instagram test data")

# Silhouette plot:
silplot(vcrtest, classCols = cols, main =
          "Silhouettes of RF on Instagram test data") # now some s(i) are negative

## -----------------------------------------------------------------------------

## Class of genuine accounts:

classmap(vcrtest, "genuine", classCols = cols) #, identify = TRUE)

# one farness outlier:
Xnew[c(30), ]
# has very lengthy bio/description
# has large number of X.posts
# has very large number of followers and follows

# genuine misclassified as fake: from highest PAC to lowest
Xnew[c(21, 29, 51), ] # and 2 more borderline cases
# They have some unusual characteristics for their class:
# * 21, 29 have a (very) high nums.length.username, i.e. the
#   percentage of numerical characters in the username.
# * 21, 29 have a full name of only 1 word.
# * 21, 29 and 51 have description.length = 0, i.e. no 
#   description/biography of their profile.
# * they all have low X.posts (even 0 for case 21), i.e.
#   relatively few previous posts.
# All of these characteristics are more common for fake profiles
# than for genuine profiles, as we can see below:

trcols <- cols[as.numeric(y_train)]

plot(x_train[, 1], col = trcols, main = "profile.pic")
# fakes are less likely to have a profile picture
plot(x_train[, 2], col = trcols, main = "nums.length.username")
# is higher for fakes
plot(x_train[, 3], col = trcols, main = "fullname.words")
# is lower for fakes
plot(x_train[, 4], col = trcols, main = "nums.length.fullname")
# is a bit higher for fakes
plot(x_train[, 5], col = trcols, main = "name..username")
# mostly 0 for genuine; fakes have a few values 1
plot(x_train[, 6], col = trcols, main = "description.length")
# fakes are typically lower, and more often zero
plot(x_train[, 7], col = trcols, main = "external.URL")
# fakes never had them, genuines sometimes did
plot(x_train[, 8], col = trcols, main = "private")
# no visible difference
plot((x_train[, 9])^0.1, col = trcols, main = "X.posts")
# fakes have fewer posts, and often none
plot((x_train[, 10])^0.1, col = trcols, main = "X.followers")
# fakes have fewer followers, sometimes none
plot((x_train[, 11])^0.1, col = trcols, main = "X.follows")
# fakes follow a bit fewer, but the difference is small.


## Class of fake accounts:

classmap(vcrtest, "fake", classCols = cols) #, identify = TRUE)

# Fake identified as genuine, from highest PAC to lower:
# c(27, 51, 34, 23, 58)
Xnew[which(ynew == "fake")[c(27, 34, 51, 23, 58)], ]

# These have a number of characteristics which are more common 
# for genuine profiles:
#
# all have profile pictures
# none have numerical characters in username
# none have numerical characters in fullname
# 27 has a lengthy bio description
# all have a relatively high number of followers
# all have a relatively high number of follows.

vcr.forest.train <- function(X, y, trainfit, type = list(),
                             k = 5, stand = TRUE) {
  #
  # Constructs a vcr object from a trained random forest.
  #
  # Arguments:
  # X        : A rectangular matrix or data frame, where the
  #            columns (variables) may be of mixed type.
  # y        : factor with the given class labels.
  #            It is crucial that X and y are EXACTLY the same
  #            as in the call to randomForest().
  #            Some entries of y may be missing, since randomForest
  #            has the argument na.action which is na.fail by
  #            default but can be set to na.omit .
  #            Still, y must contain some non-NAs.
  # k        : the number of nearest neighbors used in the
  #            farness computation.
  # trainfit : the output of a randomForest() training cycle.
  # other arguments: development
  #
  # Values include:
  #   yint        : given labels as integer 1, 2, 3, ...
  #   y           : given labels
  #   levels      : levels of y
  #   predint     : predicted class number. For each case this
  #                 is the class with the highest mixture density.
  #   pred        : predicted label of each object.
  #   altint      : competing class as integer, i.e. the non-given
  #                 class number with the highest mixture density.
  #   altlab      : competing class of each object.
  #   PAC         : PAC of each object
  #   farness     : farness of each object to its given class
  #   ofarness    : For each object i, its lowest farness to any
  #                 class, including its own. Always exists.

  # Auxiliary function:
  #
  importanceWeights <- function(varimp, cnames) {
    # Takes variable importance from the package rpart or the
    # package randomForest.
    # In varimp the variables are listed from highest to lowest
    # importance. The function rearranges them in the order of
    # the original variables, where cnames contains the column
    # names of the training data.
    # Variables not occurring in varimp are assigned weight zero.
    #
    varimp <- pmax(varimp, 0)
    varimp <- drop(varimp / sum(varimp))
    namesimp <- names(varimp)
    nonzeroVars <- which(cnames %in% namesimp)
    varweights <- rep(0, length(cnames))
    for (j in nonzeroVars) { # j=2
      varnamej <- cnames[j]
      wgt <- varimp[which(namesimp == varnamej)]
      varweights[j] <- wgt
    }
    return(varweights)
  }

  # Here the main code starts
  n <- nrow(X)
  if (n < 2) stop("The training data should have >= 2 cases.")
  checked <- checkLabels(y, n, training = TRUE)
  # If it did not stop: yint has length n, and its values are in
  # 1, ..., nlab without gaps. It is NA where y is: outside indsv.
  #
  # Test here, before doing long computations:
  varimp  <- importance(trainfit, type = 2, scale = TRUE)
  lab2int <- checked$lab2int # is a function
  levels  <- checked$levels
  nlab    <- length(levels)
  yint    <- lab2int(y)
  indsv   <- checked$indsv # cases in indsv can be visualized
  nvis    <- length(indsv)
  yintv   <- yint[indsv]
  #

  probs <- predict(trainfit, newdata = X, type = "prob")
  # Fortunately, in predict.randomForest() type="response"
  # is consistent with type="prob", so we only need to run
  # with type="prob" above.
  #
  # Note that trainfit$predicted is the out of bag (OOB)
  # prediction, but predict.randomForest() is not. The
  # latter is more overfitted than the former, but we use
  # it anyway because it allows us to deal with NA's in y.
  # This is also needed for consistency in the class maps.
  #
  probs <- probs[, order(lab2int(colnames(probs)))]
  # rowSums(probs) # all 1
  predint <- apply(probs, 1, which.max)
  #
  # Compute PAC:
  #
  # We can only compute PAC etc. for cases with available y,
  # so from here on we restrict ourselves to cases in indsv:
  probs <- probs[indsv, ]
  PAC <- altint <- rep(NA, n)
  altintv <- PACv <- rep(NA, nvis)
  for (i in seq_len(nvis)) {
    ptrue      <- probs[i, yintv[i]]
    others     <- (seq_len(nlab))[-yintv[i]]
    palt       <- max(probs[i, others])
    altintv[i] <- others[which.max(probs[i, others])]
    PACv[i]    <- palt/(ptrue + palt)
  }
  PAC[indsv]    <- PACv    # the other entries of PAC stay NA
  altint[indsv] <- altintv # the other entries of altint stay NA
  if (nlab == 2) altint <- 3 - yint # for consistency with svm etc.
  #
  # Compute farness based on knn
  if (is.null(k)) {k <- 5} # since k=1 failed on Titanic
  #
  # Construct initial daisy matrix:
  #
  varweights <- importanceWeights(varimp, colnames(X))
  X <- X[indsv, ]
  daisy.out <- daisy_vcr(X, type = type, weights = varweights,
                         stand = stand)
  dismat    <- as.matrix(daisy.out$disv)
  meandis   <- mean(as.vector(dismat), na.rm = TRUE)
  dismat[which(is.na(dismat))] <- meandis
  #
  # Compute neighbors by sorting dissimilarities:
  sortNgb <- t(apply(dismat, 1, order))[, -1]
  sortDis <- t(apply(dismat, 1, sort))[, -1]
  #
  # Compute initial fig[i, g] from new cases to training classes:
  #
  fig <- matrix(rep(NA, nvis * nlab), ncol = nlab)
  for (i in seq_len(nvis)) { # loop over all cases in indsv
    for (g in seq_len(nlab)) { # loop over classes
      ngbg <- which(yintv[sortNgb[i, ]] == g)
      if (length(ngbg) > k) {ngbg <- ngbg[seq_len(k)]}
      fig[i, g] <- median(sortDis[i, ngbg])
    }
  }
  #
  # Compute farness:
  #
  farout <- compFarness(type = "knn", testdata = FALSE, yint = yintv,
                        nlab = nlab, X = NULL, fig = fig,
                        d = NULL, figparams = NULL)
  fig <- matrix(rep(0, n * nlab), ncol = nlab)
  fig[indsv, ] <- farout$fig
  farness <- ofarness <- rep(NA, n)
  farness[indsv] <- farout$farness
  ofarness[indsv] <- farout$ofarness
  figparams <- farout$figparams
  figparams$daisy.out <- daisy.out
  figparams$meandis <- meandis
  figparams$k <- k
  #
  return(list(X = X,
              yint = yint,
              y = levels[yint],
              levels = levels,
              predint = predint,
              pred = levels[predint],
              altint = altint,
              altlab = levels[altint],
              PAC = PAC,
              figparams = figparams,
              fig = fig,
              farness = farness,
              ofarness = ofarness,
              trainfit = trainfit))
}


vcr.forest.newdata <- function(Xnew, ynew = NULL,
                               vcr.forest.train.out,
                               LOO = FALSE) {
  #
  # Predicts class labels for new data by rpart, using the output
  # of vcr.forest.train() on the training data.
  # For new data cases whose label in ynew is non-missing, additional
  # output is produced for constructing graphical displays.
  #
  # Arguments:
  #   Xnew               : data matrix of the new data, with the same
  #                        number of columns d as in the training data.
  #                        Missing values are not allowed.
  #   ynew               : factor with class membership of each new
  #                        case. Can be NA for some or all cases.
  #                        If NULL, is assumed to be NA everywhere.
  #   vcr.forest.train.out : output of vcr.forest.train() on training
  #                        data. This also contains nlab, and levels.
  #   LOO                : leave one out. Only used when testing this
  #                        function on a subset of the training data.
  #                        Default is LOO=F.
  #
  # Values include:
  #   yintnew   : given labels as integer 1, 2, 3, ..., nlab.
  #               Can be NA.
  #   ynew      : labels if yintnew is available, else NA.
  #   levels    : levels of the response, from vcr.forest.train.out
  #   predint   : predicted class number, always exists.
  #   pred      : predicted label of each object
  #   altint    : alternative label as integer if yintnew was given,
  #               else NA
  #   altlab    : alternative label if yintnew was given, else NA
  #   PAC       : probability of alternative class for each object
  #               with non-NA yintnew
  #   farscales : (from training data) scales to use in fig
  #   fig       : distance of each object i from each class g.
  #               Always exists.
  #   farness   : farness of each object to its given class, for
  #               objects with non-NA yintnew.
  #   ofarness  : For each object i, its lowest farness to any
  #               class, including its own. Always exists.
  #
  X <- vcr.forest.train.out$X # training data
  trainfit <- vcr.forest.train.out$trainfit
  # namesInModel <- rownames(trainfit$importance)
  # varsInModel <- which(colnames(X) %in% namesInModel)
  varsInModel <- attr(trainfit$terms, "term.labels")
  checkXnew(Xtrain = X, Xnew =  Xnew, allowNA = FALSE,
            varsInModel = varsInModel)
  # From here on we know that the variable names match, but they
  # might be in a different order. To correct for this:
  namesmatch <- match(colnames(X), colnames(Xnew))
  Xnew <- Xnew[, namesmatch]
  #
  yint   <- vcr.forest.train.out$yint # class membership in training data
  indold <- which(!is.na(yint))
  yint   <- yint[indold]
  ntrain <- length(indold)
  n      <- nrow(Xnew) # number of new cases
  levels <- vcr.forest.train.out$levels # same names as in training data
  nlab   <- length(levels) # number of classes
  #
  if (is.null(ynew)) ynew <- factor(rep(NA, n), levels = levels)
  checked <- checkLabels(ynew, n, training = FALSE, levels)
  lab2int <- checked$lab2int # is a function
  indsv   <- checked$indsv   # INDiceS of cases we can Visualize,
  #                         # can even be empty, is OK.
  nvis    <- length(indsv)   # can be zero.
  yintnew <- rep(NA,n)       # needed to overwrite "new" levels:
  yintnew[indsv] <- lab2int(ynew[indsv])
  # Any "new" levels are now set to NA.
  yintv   <- yintnew[indsv]  # entries of yintnew that can be visualized
  # From here on yintnew has length n, with values in seq_len(nlab).
  #
  newdata <- as.data.frame(Xnew)
  probs <- predict(trainfit, newdata = newdata, type = "prob")
  probs <- probs[, order(lab2int(colnames(probs)))]
  predint <- as.numeric(apply(probs, 1, which.max))
  #
  # Compute PAC:
  #
  # We can only compute PAC etc. for cases with available y,
  # so from here on we restrict ourselves to cases in indsv:
  probs <- probs[indsv, ]
  PAC <- altint <- rep(NA, n)
  if (nvis > 0) { # if there are cases to visualize:
    altintv <- PACv <- rep(NA, nvis)
    for (i in seq_len(nvis)) {
      ptrue      <- probs[i, yintv[i]]
      others     <- (seq_len(nlab))[-yintv[i]]
      palt       <- max(probs[i, others])
      altintv[i] <- others[which.max(probs[i, others])]
      PACv[i]    <- palt / (ptrue + palt)
    }
    PAC[indsv]    <- PACv    # the other entries of PAC stay NA
    altint[indsv] <- altintv # the other entries of altint stay NA
  }
  if (nlab == 2) altint <- 3 - yintnew # for consistency with svm etc.
  #
  # Compute farness based on knn
  figparams <- vcr.forest.train.out$figparams
  k         <- vcr.forest.train.out$figparams$k
  #
  # Construct dissimilarity matrix
  #
  Xall <- rbind(X, Xnew) # has ntrain + n rows
  # Here we have used that the variables of Xnew are in the
  # same order as in X.
  indn <- (ntrain + 1):(ntrain + n) # range of new rows
  daisynew <- daisy_vcr(Xall, daisy.out = figparams$daisy.out)
  dismatnew <- as.matrix(daisynew$disv)[indn, -indn, drop = FALSE]
  dismatnew[is.na(dismatnew)] <- figparams$meandis
  sortNgb <- t(apply(dismatnew, 1, order))
  sortDis <- t(apply(dismatnew, 1, sort))
  if (LOO) sortNgb <- sortNgb[, -1] # leave one out, only for testing
  if (LOO) sortDis <- sortDis[, -1] # leave one out, only for testing
  #
  # Compute initial fig[i, g] from all new cases to classes:
  #
  fignew <- matrix(rep(NA, n * nlab), ncol = nlab)
  for (i in seq_len(n)) {
    for (g in seq_len(nlab)) {
      ngbg <- which(yint[sortNgb[i, ]] == g)
      if (length(ngbg) > k) {
        ngbg <- ngbg[seq_len(k)]
      }
      fignew[i, g] <- median(sortDis[i, ngbg])
    }
  }
  farout <- compFarness(type = "knn", testdata = TRUE, yint = yintnew,
                       nlab = nlab, X = NULL, fig = fignew,
                       figparams = vcr.forest.train.out$figparams)
  return(list(yintnew = yintnew,
              ynew = levels[yintnew],
              levels = levels,
              predint = predint,
              pred = levels[predint],
              altint = altint,
              altlab = levels[altint],
              PAC = PAC,
              fig = farout$fig,
              farness = farout$farness,
              ofarness = farout$ofarness))
}

vcr.da.train <- function(X, y, rule = "QDA",
                         estmethod = "meancov") {
  #
  # Custom DA function which prepares for graphical displays.
  # The disciminant analysis itself is carried out by the maximum
  # a posteriori rule which maximizes the density of the mixture.
  #
  # Arguments:
  #   X         : data matrix. Missing values are not allowed.
  #   y         : vector with the given class labels.
  #               Missing y are not allowed, such cases should be
  #               removed first.
  #   rule      : either "QDA" for quadratic discriminant analysis
  #               or "LDA" for linear discriminant analysis.
  #   estmethod : function for location and covariance estimation.
  #               Should return a list with $m and $S. Can be "meancov"
  #               (classical mean and covariance matrix) or "DetMCD".
  #
  # Returns:
  #   yint      : given labels as integer 1, 2, 3, ...
  #   y         : given labels
  #   levels    : levels of y
  #   predint   : predicted class number. For each case this
  #               is the class with the highest mixture density.
  #   pred      : predicted label of each object.
  #   altint    : alternative class as integer, i.e. the non-given
  #               class number with the highest mixture density.
  #   altlab    : alternative class of each object.
  #   fig       : Mahalanobis distance of each object i to each class g
  #   classMS   : list with center and covariance matrix of each class
  #   lCurrent  : log(mixture density) in object's given class.
  #   lPred     : log(mixture density) in object's predicted class.
  #   lAlt      : log(mixture density) in object's alternative class.
  #   PAC       : probability of alternative class for each object
  #   figparams : parameters for computing fig, can be used for
  #               new data.
  #   fig       : distance of each object i from each class g.
  #   farness   : farness of each object to its given class.
  #   ofarness  : For each object i, its lowest farness to any class.
  #               If much above 1, we suspect i may be an outlier.
  #
  X <- as.matrix(X) # in case it is a data frame
  if (nrow(X) == 1) X <- t(X)
  if (sum(is.na(as.vector(X))) > 0) {
    stop("The data matrix X has NA's.")
  }
  n <- nrow(X)
  d <- ncol(X)
  if (n < 2) stop("The training data should have more than one case.")
  if (!(rule %in% c("LDA", "QDA"))) {
    stop(" rule should be either LDA or QDA")
  }
  if (!is.function(estmethod)) {
    if (estmethod == "meancov") {
      estmethod <- meancov
    } else if (estmethod == "DetMCD") {
      estmethod <- DetMCD
    }
  }

  # Check whether y and its levels are of the right form:
  checked <- checkLabels(y, n, training = T)
  lab2int <- checked$lab2int # is a function
  indsv   <- checked$indsv
  levels  <- checked$levels
  nlab    <- length(levels)
  yint    <- lab2int(y)
  yintv   <- yint[indsv]
  classSizes <- rep(0, nlab)
  for (g in seq_len(nlab)) {classSizes[g] <- sum(yintv == g)}
  # classSizes
  MD2s    <- matrix(0, n, nlab) # squared mahalanobis distances
  mixf    <- matrix(0, n, nlab) # component of mixture density
  lmixf   <- matrix(0, n, nlab) # the logarithm of that density
  classMS <- list()
  #
  if (rule == "QDA") {
    tinyclasses <- which(classSizes < (d + 1)) # was 2*(d+1)
    if (length(tinyclasses) > 0) {
      stop(paste0(" The following classes have too few members",
                  " to compute a covariance matrix:",
                  levels[tinyclasses]))
    }
    # Compute the center and covariance matrix of each class:
    iCN <- rep(NA, nlab) # for inverse condition numbers
    for (g in seq_len(nlab)) {
      clinds       <- indsv[which(yintv == g)]
      # class indices in 1,...,n
      class.out    <- estmethod(X[clinds, ]) # contains mu and Sigma
      iCN[g]       <- numericalCond(class.out$S)
      classMS[[g]] <- class.out # now has $m, $S
    }
    tinyCN <- which(iCN < 1e-06)
    if (length(tinyCN) > 0) {
      stop(paste0("\n The following classes have an ill-conditioned",
                  " covariance matrix: \n", levels[tinyCN],
                  "\n You may want to try LDA or another classifier"))
    } # From here onward we can use the inverse of each S.
  } # end of QDA
  #
  if (rule == "LDA") { # Then we compute a single S instead:
    Xv <- X[indsv, ] # only the rows with available y
    for (g in seq_len(nlab)) {
      clinds  <- which(yintv == g) # indices in 1, ..., indsv
      classMS[[g]] <- meancov(Xv[clinds, ]) # colMeans
      Xv[clinds, ] <- sweep(Xv[clinds, ], 2, classMS[[g]]$m) # centered
    }
    pool.out <- estmethod(Xv) # estimates m and S on pooled data.
    if (numericalCond(pool.out$S) < 1e-06) {
      stop(paste0("\n The pooled covariance matrix is ill-conditioned.",
                  "\n You may want to try a different classifier")) }
    for (g in seq_len(nlab)) {
      classMS[[g]]$m  <- classMS[[g]]$m + pool.out$m
      classMS[[g]]$S  <- pool.out$S
    }
  }
  # From here on everything is the same for LDA and QDA.
  #
  # compute MD2 of each object to all classes:
  for (g in seq_len(nlab)) { # g=1
    class.out <- classMS[[g]] # now nonsingular S
    MD2s[, g] <- mahalanobis(X, class.out$m, class.out$S)
  }
  # Compute prior probabilities:
  for (g in seq_len(nlab)) { # g=1
    clinds <- which(yintv == g) # indices in 1, ..., indsv
    classMS[[g]]$prior <- length(clinds) / length(indsv)
  }
  #
  for (g in seq_len(nlab)) { # g=1
    S          <- classMS[[g]]$S
    prior      <- classMS[[g]]$prior
    lmixf[, g] <- -0.5*(log(det(S)) + MD2s[, g]) + log(prior)
    mixf[, g]  <- exp(lmixf[, g])
  }
  # Compute predictions far all objects in 1, ..., n
  predint <- lPred <- rep(NA, n)
  lPred   <- apply(lmixf[, , drop = FALSE], 1, max)
  predint <- apply(lmixf[, , drop = FALSE], 1, which.max)
  #
  # Compute fCurrent and fAlt for all objects with available y:
  lCurrent <- fCurr <- fAlt <- lAlt <- altint <- rep(NA, n)
  for (g in seq_len(nlab)) { # g=1
    clinds <- indsv[which(yintv == g)] # indices in 1, ..., n
    others <- (seq_len(nlab))[-g] # alternative classes
    fCurr[clinds]    <- mixf[clinds, g]
    lCurrent[clinds] <- lmixf[clinds, g]
    fAlt[clinds]     <- apply(mixf[clinds, others, drop = FALSE],
                              1, max)
    lAlt[clinds]     <- apply(lmixf[clinds, others, drop = FALSE],
                              1, max)
    altint[clinds]   <- others[apply(lmixf[clinds, others, drop = FALSE],
                                     1, which.max)]
  }
  #
  # Compute PAC:
  #
  PAC <- rep(NA, n)
  PAC[indsv] <- fAlt[indsv] / (fCurr[indsv] + fAlt[indsv])
  #
  # Compute farness:
  #
  farout <- compFarness(type = "affine", testdata = F, yint = yint,
                        nlab = nlab, X = NULL, fig = sqrt(MD2s),
                        d = d, figparams = NULL)
  return(list(yint = yint,
              y = levels[yint],
              levels = levels,
              predint = predint,
              pred = levels[predint],
              altint = altint,
              altlab = levels[altint],
              PAC = PAC,
              figparams = farout$figparams,
              fig = farout$fig,
              farness = farout$farness,
              ofarness = farout$ofarness,
              classMS = classMS,
              lCurrent = lCurrent,
              lPred = lPred,
              lAlt = lAlt))
}


vcr.da.newdata <- function(Xnew, ynew = NULL, vcr.da.train.out) {
  #
  # Predicts class labels for new data by discriminant analysis,
  # using the output of vcr.da.train() on the training data.
  # For new data cases whose label in yintnew is non-missing,
  # additional output is produced for constructing graphical
  # displays.
  #
  # Arguments:
  #   Xnew             : data matrix of the new data, with the same
  #                      number of columns d as in the training data.
  #                      Missing values are not allowed.
  #   ynew             : factor with class membership of each new
  #                      case. Can be NA for some or all cases.
  #                      If NULL, is assumed to be NA everywhere.
  #   vcr.da.train.out : output of vcr.da.train() on the training
  #                      data. This also contains d, nlab, and levels.
  #
  # Returns:
  #   yintnew   : given labels as integer 1, 2, 3, ..., nlab.
  #               Can be NA.
  #   ynew      : labels if yintnew is available, else NA.
  #   levels    : levels of the response, from vcr.da.train.out
  #   predint   : predicted label as integer, always exists.
  #   pred      : predicted label of each object
  #   altint    : alternative label as integer if yintnew was given,
  #               else NA
  #   altlab    : alternative label if yintnew was given, else NA
  #   fig       : Mahalanobis distance of each object i to each class g
  #   classMS   : list with center and covariance matrix of each class,
  #               from vcr.da.train.out
  #   lCurrent  : log of density in object's given class, if yintnew
  #               was given, else NA
  #   lPred     : log of density in object's predicted class
  #   lAlt      : log of density in object's alternative class if
  #               yintnew was given, else NA
  #   PAC       : probability of alternative class for each object
  #               with non-NA yintnew
  #   figparams : (from training) parameters used to compute fig
  #   fig       : farness of each object i from each class g.
  #               Always exists.
  #   farness   : farness of each object to its given class, for
  #               objects with non-NA yintnew.
  #   ofarness  : For each object i, its lowest farness to any class.
  #               Always exists. If much above 1, we suspect i may be
  #               an outlier.
  #
  Xnew <- as.matrix(Xnew) # in case it is a data frame
  if (nrow(Xnew) == 1) Xnew <- t(Xnew)
  n <- nrow(Xnew)
  d <- length(vcr.da.train.out$classMS[[1]]$m) # = ncol(X)
  if (ncol(Xnew) != d) {
    stop(paste0("Xnew should have ", d,
                " columns, like the training data."))
  }
  ### We could also check whether the colnames are the same...
  if (sum(is.na(as.vector(Xnew))) > 0) {
    stop("The data matrix Xnew contains NA's.")
  }
  levels <- vcr.da.train.out$levels # same names as in training data
  nlab   <- length(levels) # number of classes
  #
  if (is.null(ynew)) ynew <- factor(rep(NA, n), levels = levels)
  checked <- checkLabels(ynew, n, training = F, levels)
  lab2int <- checked$lab2int # is a function
  indsv   <- checked$indsv   # INDiceS of cases we can Visualize,
  #                         # can even be empty, is OK.
  nvis    <- length(indsv)   # can be zero.
  yintnew <- rep(NA, n)       # needed to overwrite "new" levels.
  yintnew[indsv] <- lab2int(ynew[indsv])
  # Any "new" levels are now set to NA.
  yintv   <- yintnew[indsv]  # entries of yintnew that can be visualized
  #
  MD2s <- matrix(0, n, nlab) # for squared Mahalanobis distances
  mixf <- lmixf <- matrix(0, n, nlab) # mixture density and its log
  classMS <- vcr.da.train.out$classMS
  for (g in seq_len(nlab)) { # g=1
    M          <- classMS[[g]]$m
    S          <- classMS[[g]]$S
    prior      <- classMS[[g]]$prior
    MD2s[, g]  <- mahalanobis(Xnew, M, S)
    lmixf[, g] <- -0.5*(log(det(S)) + MD2s[, g]) + log(prior)
    mixf[, g]  <- exp(lmixf[, g])
  }
  # Compute predictions far all objects in 1, ..., n
  predint <- lPred <- rep(NA, n)
  lPred   <- apply(lmixf[, , drop = FALSE], 1, max)
  predint <- apply(lmixf[, , drop = FALSE], 1, which.max)
  #
  # For PAC:
  #
  lCurrent <- fCurr <- fAlt <- lAlt <- altint <- rep(NA, n)
  PAC <- rep(NA, n) # will remain NA for cases outside indsv
  # We can only compute PAC for cases with available yint.
  if (nvis > 0) { # if there are new data with labels
    yintv  <- yintnew[indsv] # yint of cases we will Visualize
    ayintv <- sort(unique(yintv)) # available yintv values
    # We cannot assume that ayintv = 1, ..., nlab for test data.
    for (g in ayintv) { # g=1 # g=3
      clinds <- indsv[which(yintv == g)] # indices in 1, ..., n
      others <- (seq_len(nlab))[-g]
      fCurr[clinds]    <- mixf[clinds, g]
      lCurrent[clinds] <- lmixf[clinds, g]
      fAlt[clinds]     <- apply(mixf[clinds, others, drop = FALSE],
                                1, max)
      lAlt[clinds]     <- apply(lmixf[clinds, others, drop = FALSE],
                                1, max)
      altint[clinds]   <- others[apply(lmixf[clinds, others, drop = FALSE],
                                       1, which.max)]
    }
    #
    # Compute PAC:
    #
    PAC[indsv] <- fAlt[indsv] / (fCurr[indsv] + fAlt[indsv])
  }
  #
  # Compute farness:
  #
  farout <- compFarness(type = "affine", testdata = T,
                        yint = yintnew,
                        nlab = nlab, X = NULL,
                        fig = sqrt(MD2s), d = d,
                        figparams = vcr.da.train.out$figparams)
  return(list(yintnew = yintnew,
              ynew = levels[yintnew],
              levels = levels,
              predint = predint,
              pred = levels[predint],
              altint = altint,
              altlab = levels[altint],
              PAC = PAC,
              fig = farout$fig,
              farness = farout$farness,
              ofarness = farout$ofarness,
              classMS = classMS,
              lCurrent = lCurrent,
              lPred = lPred,
              lAlt = lAlt))
}

vcr.knn.train <- function(X, y, k){
  #
  # Carries out a k-nearest neighbor classification on the
  # training data. Each case is predicted from the labels of
  # all cases except its own.
  # Various additional output is produced for the purpose
  # of constructing graphical displays.
  #
  # X        : This can be a rectangular matrix or data frame
  #            of (already standardized) measurements, or a
  #            dist object obtained from dist() or daisy().
  #            Missing values are not allowed.
  # y        : factor with the given (observed) class labels.
  #            There need to be non-missing y in order to be
  #            able to train the classifier.
  # k        : the number of nearest neighbors used. It can
  #            be selected by running cross-validation using
  #            a different package.
  #
  # Values include:
  #   yint      : given labels as integer 1, 2, 3, ...
  #   y         : given labels
  #   levels    : levels of y
  #   predint   : predicted class number. For each case this
  #               is the class with the highest mixture density.
  #   pred      : predicted label of each object.
  #   altint    : alternative class as integer, i.e. the non-given
  #               class number with the highest mixture density.
  #   altlab    : alternative class of each object.
  #   PAC       : probability of alternative class for each object
  #   fig       : distance of each object i from each class g.
  #   farness   : farness of each object to its given class.
  #   ofarness  : For each object i, its lowest farness to any class.
  #               If much above 1, we suspect i may be an outlier.

  # Auxiliary function for kNN:
  #
  predDis <- function(nlab, labsi, dissi) {
    # Computes the kNN prediction and its average distance.
    # When the highest frequency is tied, the predicted class
    # is the one with lowest average distance from i.
    freqz <- rep(0, times = nlab)
    for (j in seq_len(nlab)) { # freqz = frequencies of labels
      freqz[j] <- sum(labsi == j, na.rm = TRUE)
    }
    ranking <- order(freqz, decreasing = TRUE)
    # positions of most to least frequent labels
    sfreq <- freqz[ranking] # frequencies in that order
    predi <- NA
    if (sfreq[1] > sfreq[2]) { # frequency has only 1 maximizer
      predi <- ranking[1] # predicted class
      averdis <- sum(dissi[labsi == predi]) / sfreq[1]
      seconddis <- NA
    } else {# frequency has several maximizers
      multipreds <- ranking[which(sfreq == sfreq[1])]
      multi <- length(multipreds); multi
      avdis <- rep(0, times = multi)
      for (j in seq_len(multi)) { # adds at each tied solution
        avdis[j] <- sum(dissi[labsi == multipreds[j]]) / sfreq[1]
      }
      bestInMulti <- which.min(avdis) # avdis minimizer
      averdis <- avdis[bestInMulti] # smallest average distance
      # compute second smallest average distance:
      seconddis <- min(avdis[-bestInMulti])
      predi <- ranking[bestInMulti] # predicted class
    }
    freq <- sfreq[1]
    return(list(predi = predi, averdis = averdis, freq = freq,
                seconddis = seconddis))
  }

  # Here the main function starts:
  if (inherits(X, "dist")) { # input is a dist object
    n <- cluster::sizeDiss(X); n # 150
    # turn it into a symmetric square matrix:
    X <- as.matrix(X); dim(X)
    dismat <- X; rm(X) # R's renaming, to save space
    if (sum(is.na(as.vector(dismat))) > 0) {
      stop("The input dissimilarity matrix X has NA's.")
    }
    X <- NULL # we don't need to return this
  } else {
    # Assumes X is already standardized
    X <- as.matrix(X) # in case it is a data frame
    if (nrow(X) == 1) X <- t(X)
    if (sum(is.na(as.vector(X))) > 0) {
      stop(paste0("The data matrix X has NA's. You could either",
                  " remove them", "\nor make a dissimilarity",
                  " matrix without NA's, e.g. with daisy()."))
    }
    n <- nrow(X)
    dismat <- as.matrix(dist(X, method = "euclidean"))
    # If you want a different metric, call dist() or daisy() first
    # and then input the result into vcr.knn.train instead of
    # the coordinate matrix.
  }
  if (n < 2) stop("The training data should have more than one case.")
  # Check whether y and its levels are of the right form:
  checked <- checkLabels(y, n, training = TRUE)
  # If it did not stop: yint has length n, and its values are in
  # 1, ..., nlab without gaps. It is NA where y is: outside indsv.
  lab2int <- checked$lab2int
  levels  <- checked$levels
  nlab    <- length(levels)
  yint    <- lab2int(y)
  indsv   <- checked$indsv # cases in indsv can be visualized
  nvis    <- length(indsv)
  yintv   <- yint[indsv]
  #
  # Only neighbors with known class are useful for training:
  # dim(dismat)
  dismat <- dismat[, indsv] # has the right colnames
  # The k nearest neighbors of each i, as indices in 1, ..., n are:
  sortNgb <- t(apply(dismat, 1, order))[, -1] # excludes i itself,
  # this is the usual convention for knn on training data.
  # To make the entries case indices in 1, ..., n we do:
  sortNgb <- apply(sortNgb, 2, function(x) indsv[x])
  dismat <- t(apply(dismat, 1, sort))[, -1] # excludes d(i, i)=0
  sortDis <- dismat; rm(dismat) # R's renaming, to save space

  #
  # Make neighborhoods:
  #
  # First take care of situations where the k-th distance
  # is tied with (at least) the (k+1)-th. For such an object
  # we include all those tied distances.
  prec <- 1e-12
  ktrues <- rep(NA, times = n)
  for (i in seq_len(n)) { # i=1
    rowdis <- sortDis[i, ]
    ktrues[i] <- length(which(rowdis < rowdis[k] + prec))
  }
  #
  # Make predictions:
  #
  preds  <- matrix(rep(NA, times = 2 * n), ncol = 2)
  freqs  <- matrix(rep(0, times = 3 * n), ncol = 3) # no NA's, OK
  avedis <- matrix(rep(NA, times = 3 * n), ncol = 3)
  #
  for (i in seq_len(n)) { # loop over cases to predict # i=7
    #
    # Compute statistics for the true label, if it exists:
    #
    ktrue <- ktrues[i]
    labsi <- yint[sortNgb[i, seq_len(ktrue)]]
    # labels of neighbors exist
    dissi <- sortDis[i, seq_len(ktrue)] # dissimilarities from i
    if (!is.na(yint[i])) {
      freqs[i, 3] <- sum(labsi == yint[i])
      # Note that freqs[i, 3] can be zero, but not NA
      if (freqs[i, 3] > 0) {
        avedis[i, 3] <- sum(dissi[labsi == yint[i]]) / freqs[i, 3]
      }
    }
    #
    # Compute predicted label:
    #
    out1       <- predDis(nlab, labsi, dissi)
    # calls the auxiliary function predDis
    preds[i, 1] <- out1$predi
    freqs[i, 1] <- out1$freq
    if (!is.na(out1$averdis)) {
      avedis[i, 1] <- out1$averdis
    } else {
      avedis[i, 1] <- min(sortDis[i, yint[sortNgb[i, ]] == out1$predi])
    }
    #
    # Compute second best prediction if there is one:
    #
    if (out1$freq < ktrues[i]) { # else nothing to do
      remNgb      <- which(labsi != out1$predi) # remaining neighbors
      labsi       <- labsi[remNgb]
      dissi       <- dissi[remNgb]
      out2        <- predDis(nlab, labsi, dissi)
      preds[i, 2]  <- out2$predi   # was initialized as NA
      freqs[i, 2]  <- out2$freq    # was initialized as 0
      avedis[i, 2] <- out2$averdis # was initialized as NA
    }
  } # ends loop over cases to predict
  #
  counts <- freqs[, c(3, 1, 2)]
  # given class, predicted class, alternative class
  predint <- preds[, 1] # integer predictions for all cases,
  #                   # even those with missing y.
  freq1v <- freqs[indsv, 1] # highest frequency (= of prediction)
  freq2v <- freqs[indsv, 2] # second highest frequency
  freq3v <- freqs[indsv, 3] # frequency of self
  pred1v <- preds[indsv, 1] # prediction
  pred2v <- preds[indsv, 2] # best non-prediction
  #
  # Compute PAC:
  #
  PAC <- altint <- rep(NA, n)
  freqalt <- freq1v
  altintv <- pred1v # preliminary
  matched <- which(pred1v == yintv) # prediction == selfclass
  freqalt[matched] <- freq2v[matched]
  altintv[matched] <- pred2v[matched]
  PACv <- freqalt / (freq3v + freqalt)
  PAC[indsv] <- PACv
  altint[indsv] <- altintv
  #
  ### Compute initial fig
  #
  fig <- matrix(rep(NA, nvis * nlab), ncol = nlab)
  # compute fig[i, g] of object i to class g:
  for (i in seq_len(nvis)) { # loop over all cases in indsv
    for (g in seq_len(nlab)) { # loop over classes
      ngbg <- which(yint[sortNgb[indsv[i], ]] == g)
      if (length(ngbg) > k) {
        ngbg <- ngbg[seq_len(k)]
      }
      fig[i, g] <- median(sortDis[indsv[i], ngbg])
    }
  }
  #
  # Compute farness:
  #
  farout <- compFarness(type = "knn", testdata = FALSE, yint = yintv,
                        nlab = nlab, X = NULL, fig = fig, d = NULL,
                        figparams = NULL)
  fig <- matrix(rep(0, n * nlab), ncol = nlab)
  fig[indsv, ] <- farout$fig
  farness <- ofarness <- rep(NA, n)
  farness[indsv] <- farout$farness
  ofarness[indsv] <- farout$ofarness
  if (nlab == 2) altint <- 3 - yint # for consistency with svm etc.
  # This way altint exists, even if it has frequency zero.
  return(list(yint = yint,
              y = levels[yint],
              levels = levels,
              predint = predint,
              pred = levels[predint],
              altint = altint,
              altlab = levels[altint],
              PAC = PAC,
              figparams = farout$figparams,
              fig = fig,
              farness = farness,
              ofarness = ofarness,
              k = k,
              ktrues = ktrues,
              counts = counts,
              X = X))
}

vcr.knn.newdata <- function(Xnew, ynew = NULL,
                            vcr.knn.train.out, LOO = FALSE) {
  #
  # Predicts class labels for new data by k nearest neighbors,
  # using the output of vcr.knn.train() on the training data.
  # For new data with available labels in ynew, additional output
  # is produced for constructing graphical displays.
  #
  #   Xnew              : If the training data was a matrix of
  #                       coordinates, Xnew must be such a matrix
  #                       with the same number of columns.
  #                       If the training data was a set of
  #                       dissimilarities, Xnew must be a rectangular
  #                       matrix of dissimilarities, with each row
  #                       containing the dissmilarities of a new
  #                       case to all training cases.
  #                       Missing values are not allowed.
  #   ynew              : factor with class membership of each new
  #                       case. Can be NA for some or all cases.
  #                       If NULL, is assumed to be NA everywhere.
  #   vcr.knn.train.out : output of vcr.knn.train() on the training
  #                       data. This contains nlab, levels and k.
  #   LOO               : leave one out. Only used when testing this
  #                       function on a subset of the training data.
  #                       Default is LOO=FALSE.
  #
  # Values include:
  #   yintnew   : given labels as integer 1, 2, 3, ..., nlab.
  #               Can be NA.
  #   ynew      : labels if yintnew is available, else NA.
  #   levels    : levels of the response, from vcr.da.train.out
  #   predint   : predicted class number, always exists.
  #   pred      : predicted label of each object
  #   altint    : alternative label as integer if yintnew was given,
  #               else NA
  #   altlab    : alternative label if yintnew was given, else NA
  #   PAC       : probability of alternative class for each object
  #               with non-NA yintnew
  #   fig       : distance of each object i from each class g.
  #               Always exists.
  #   farness   : farness of each object to its given class, for
  #               objects with non-NA yintnew.
  #   ofarness  : For each object i, its lowest farness to any class.
  #               Always exists. If much above 1, we suspect i may be
  #               an outlier.

  # Auxiliary function for kNN:
  #
  predDis <- function(nlab, labsi, dissi) {
    # Computes kNN prediction and its average distance.
    # This works even if the data contains < nlab labels.
    freqz <- rep(0, times = nlab)
    for (j in seq_len(nlab)) { # freqz = frequencies of labels
      freqz[j] <- sum(labsi == j, na.rm = TRUE)
    }
    ranking <- order(freqz, decreasing = TRUE)
    # positions of most to least frequent labels
    sfreq <- freqz[ranking] # frequencies in that order
    predi <- NA
    if (sfreq[1] > sfreq[2]) { # frequency has only 1 maximizer
      predi <- ranking[1] # predicted class
      averdis <- sum(dissi[labsi == predi]) / sfreq[1]
      seconddis <- NA
    } else {# frequency has several maximizers
      multipreds <- ranking[which(sfreq == sfreq[1])]
      multi <- length(multipreds); multi
      avdis <- rep(0, times = multi)
      for (j in seq_len(multi)) { # adds at each tied solution
        avdis[j] <- sum(dissi[labsi == multipreds[j]]) / sfreq[1]
      }
      bestInMulti <- which.min(avdis) # avdis minimizer
      averdis <- avdis[bestInMulti] # smallest average distance
      # compute second smallest average distance:
      seconddis <- min(avdis[-bestInMulti])
      predi <- ranking[bestInMulti] # predicted class
    }
    freq <- sfreq[1]
    return(list(predi = predi, averdis = averdis, freq = freq,
                seconddis = seconddis))
  }

  # Here the main function starts:
  Xnew <- as.matrix(Xnew)
  if (is.null(vcr.knn.train.out$X)) {
    # The training data were given as a dissimilarity matrix.
    # Therefore Xnew must be a dissimilarity matrix too.
    if (is.vector(Xnew)) {Xnew <- matrix(Xnew, nrow = 1)}
    Xnew <- as.matrix(Xnew)
    if (sum(is.na(as.vector(Xnew))) > 0) {
      stop("The dissimilarity matrix Xnew contains NA's.")
    }
    dismat <- Xnew; rm(Xnew) # R's version of rename, to save space.
    ntrain <- length(vcr.knn.train.out$yint)
    if (ncol(dismat) != ntrain) {
      stop(paste0("\nThe dissimilarity matrix Xnew should have ", ntrain,
                  "\ncolumns, as many as there were training cases."))
    }
    n <- nrow(dismat) # number of new cases
  } else {
    # The training data was a matrix of coordinates.
    Xnew <- as.matrix(Xnew) # in case it is a data frame
    if (nrow(Xnew) == 1) Xnew <- t(Xnew)
    d <- ncol(vcr.knn.train.out$X)
    if (ncol(Xnew) != d) {
      stop(paste0("Xnew should have ", d,
                  " columns, like the training data."))
    }
    # Should check column names here (also in da)
    # first vcr.da.train should output $colnames
    if (sum(is.na(as.vector(Xnew))) > 0) {
      stop("The data matrix Xnew contains NA's.")
    }
    ntrain <- nrow(vcr.knn.train.out$X)
    n      <- nrow(Xnew) # number of new cases
    Xall   <- rbind(vcr.knn.train.out$X, Xnew) # has ntrain + n rows
    indn   <- (ntrain + 1):(ntrain + n) # range of new rows
    dismat <- as.matrix(dist(Xall, method = "euclidean")
    )[indn, seq_len(ntrain), drop = FALSE]
  }
  levels <- vcr.knn.train.out$levels # same names as in training data
  nlab   <- length(levels) # number of classes
  #
  if (is.null(ynew)) ynew <- factor(rep(NA, n), levels = levels)
  checked <- checkLabels(ynew, n, training = FALSE, levels)
  lab2int <- checked$lab2int # is a function
  indsv   <- checked$indsv   # INDiceS of cases we can Visualize,
  #                         # can even be empty, is OK.
  nvis    <- length(indsv)   # can be zero.
  yintnew <- rep(NA, n)       # needed to overwrite "new" levels.
  yintnew[indsv] <- lab2int(ynew[indsv])
  # Any "new" levels are now set to NA.
  yintv   <- yintnew[indsv]  # entries of yintnew that can be visualized
  #
  yint <- vcr.knn.train.out$yint # class membership in training data
  k    <- vcr.knn.train.out$k
  # For each new case, find its neighbors in the training data:
  sortNgb <- t(apply(dismat, 1, order))
  if (LOO) sortNgb <- sortNgb[, -1] # leave one out, only for testing
  colnames(sortNgb) <- NULL # colnames were irrelevant
  # Obtain their dissimilarities:
  dismat  <- t(apply(dismat, 1, sort)) # to save space
  if (LOO) dismat <- dismat[, -1] # leave one out, only for testing
  sortDis <- dismat; rm(dismat) # to save space
  #
  # Make neighborhoods:
  #
  # First take care of situations where the k-th distance
  # is tied with (at least) the (k+1)-th. For such an object
  # we include all those tied distances.
  prec <- 1e-12
  ktrues <- rep(NA, n)
  for (i in seq_len(n)) { # i=1
    rowdis <- sortDis[i, ]
    ktrues[i] <- length(which(rowdis < rowdis[k] + prec))
  }
  ktrues
  #
  # Make predictions:
  #
  preds  <- matrix(rep(NA, times = 2 * n), ncol = 2)
  freqs  <- matrix(rep(0, times = 3 * n), ncol = 3)
  avedis <- matrix(rep(NA, times = 3 * n), ncol = 3)
  #
  for (i in seq_len(n)) { # loop over cases to predict
    #
    # Compute statistics for the true label, if it exists:
    #
    ktrue <- ktrues[i]
    labsi <- yint[sortNgb[i, seq_len(ktrue)]]
    # labels in training data exist
    dissi <- sortDis[i, seq_len(ktrue)]
    # dissim. between i and training data
    if (is.na(yintnew[i])) { # if i has no given label
      freqs[i, 3] <- 0
    } else {# if i does have a given label
      freqs[i, 3] <- sum(labsi == yintnew[i])
    }
    if (freqs[i, 3] > 0) {
      avedis[i, 3] <- sum(dissi[labsi == yintnew[i]]) / freqs[i, 3]
    } else {
      avedis[i, 3] <- NA
    }
    #
    # Compute predicted label:
    #
    out1       <- predDis(nlab, labsi, dissi)
    preds[i, 1] <- out1$predi
    freqs[i, 1] <- out1$freq
    if (!is.na(out1$averdis)) {
      avedis[i, 1] <- out1$averdis } else {
        avedis[i, 1] <- min(sortDis[i, yint[sortNgb[i, ]] == out1$predi]) }
    #
    # Compute second best prediction if there is one:
    #
    if (out1$freq < ktrues[i]) { # else nothing to do
      remNgb      <- which(labsi != out1$predi) # remaining neighbors
      labsi       <- labsi[remNgb]
      dissi       <- dissi[remNgb]
      out2        <- predDis(nlab, labsi, dissi)
      preds[i, 2]  <- out2$predi   # was initialized as NA
      freqs[i, 2]  <- out2$freq    # was initialized as 0
      avedis[i, 2] <- out2$averdis # was initialized as NA
    }
  } # ends loop over cases to predict
  #
  counts <- freqs[, c(3, 1, 2)]
  # given class, predicted class, alternative class
  predint <- preds[, 1] # integer predictions for all cases,
  #                   # even those with missing ynew.
  #
  PAC <- altint <- rep(NA, n)
  if (nvis > 0) { # there are new data with labels
    # We can only compute PAC etc. for cases with available ynew.
    freq1v <- freqs[indsv, 1]
    freq2v <- freqs[indsv, 2] # can have (many) zeroes, no NAs
    freq3v <- freqs[indsv, 3]
    pred1v <- preds[indsv, 1] # prediction
    pred2v <- preds[indsv, 2] # best non-prediction
    #
    # Compute PAC:
    #
    freqalt <- freq1v # frequency of prediction
    altintv <- pred1v # preliminary
    matched <- which(pred1v == yintv) # prediction == selfclass
    freqalt[matched] <- freq2v[matched] # frequency of alternative
    altintv[matched] <- pred2v[matched]
    PACv <- freqalt / (freq3v + freqalt)
    PAC[indsv] <- PACv # the other entries of PAC remain NA
    altint[indsv] <- altintv
    if (nlab == 2) altint <- 3 - yintnew # for consistency with svm etc.
    # When ynew is NA, also altint will be
  }
  #
  # Compute initial fig[i, g] from new cases to training classes:
  #
  fig <- matrix(rep(NA, n * nlab), ncol = nlab)
  for (i in seq_len(n)) { # loop over all cases in newdata
    for (g in seq_len(nlab)) { # loop over classes in training data
      ngbg <- which(yint[sortNgb[i, ]] == g)
      if (length(ngbg) > k) {
        ngbg <- ngbg[seq_len(k)]
      }
      fig[i, g] <- median(sortDis[i, ngbg])
    }
  }
  #
  # Compute farness:
  #
  farout <- compFarness(type = "knn", testdata = TRUE, yint = yintnew,
                        nlab = nlab, X = NULL, fig = fig, d = NULL,
                        figparams = vcr.knn.train.out$figparams)
  return(list(yintnew = yintnew,
              ynew = levels[yintnew],
              levels = levels,
              predint = predint,
              pred = levels[predint],
              altint = altint,
              altlab = levels[altint],
              PAC = PAC,
              fig = farout$fig,
              farness = farout$farness,
              ofarness = farout$ofarness,
              k = k,
              ktrues = ktrues,
              counts = counts))
}

predscomp <- function(fit, maxnpreds = 8,
                      sort.by.stdev = TRUE,
                      adj.order = FALSE,
                      casetoshow = NULL,
                      verbose = TRUE){
  # This is an auxiliary function which is called by
  # both predsplot() and predscor(). It does not
  # need to be visible to the casual user of the
  # package.
  #
  # Arguments:
  #
  # fit           : an output object of lm() or glm().
  # maxnpreds     : the maximal number of prediction terms to
  #                 plot. When there are more prediction terms
  #                 than this, those with smallest
  #                 standard deviations are combined.
  # sort.by.stdev : if TRUE, sorts the prediction terms by
  #                 decreasing standard deviation.
  # adj.order     : if TRUE, this modifies the order
  #                 of the prediction terms in an attempt
  #                 to bring highly correlated
  #                 prediction terms close to each other
  #                 in the predscor() display. If
  #                 sort.by.stdev is TRUE, this happens
  #                 after that sorting.
  # casetoshow    : if not NULL, the particular case
  #                 to be displayed by predsplot().
  #                 This can be a case number or
  #                 row name of a case in the fitted
  #                 dataset, or a list or vector with
  #                 input values of a new case.
  # verbose       : if TRUE, some computational results are
  #                 shown on the console.

  # Check whether fit is an lm or glm object:
  myclass = class(fit)
  mymodeltype = "lm"
  # check whether the model only contains "lm" and "glm"
  if(sum(!(myclass %in% c("glm","lm"))) > 0) stop(paste0(
    "Your fit object is of class ",myclass," but this",
    " tool requires the fit to be from lm() or glm(). \n"))
  #
  # Compute total prediction:
  #
  if("glm" %in% myclass) { # then is a glm
    totpred = predict(fit, type="link", na.action = na.omit)
    # this is the linear total prediction
    mymodeltype = "glm"
  } else {
    totpred = predict(fit, type="response", na.action = na.omit)
  }
  if(is.vector(totpred) == FALSE){
    if(is.matrix(totpred) == TRUE){
      stop(paste0("This model predicts a multivariate response.\n",
                  "Please rerun the fit for one response at a time."))
    } else {
      stop(paste0("Running predict(fit, type=\"response\"",
                  " did not return a vector."))
    }
  }
  centercept = mean(totpred, na.rm = TRUE)
  totpred = totpred - centercept
  totpredscale = sd(totpred, na.rm = TRUE)

  # Extract response name:
  # attformul = attributes(terms.formula(fit$call$formula))
  # yposit    = attformul$response
  # yname     = rownames(attformul$factors)[yposit]
  yname=deparse(as.list(fit$call$formula)[[2]])

  # Compute prediction terms:

  mycoeffs = fit$coefficients
  if(sum(is.na(mycoeffs)) > 0) stop(paste0(
    "This fit contains coefficients that are NA.\n",
    "  This implies that one or more coefficients are aliased",
    " in your model."))
  intercept = NULL
  if("(Intercept)" %in% names(mycoeffs)){
    jj = which(names(mycoeffs) == "(Intercept)")
    intercept = mycoeffs[jj]
  }

  dclasses = attributes(fit$terms)$dataClasses
  if (yname %in% names(dclasses)) {
    jj = which(names(dclasses) == yname)
    dclasses = dclasses[-jj]
  }
  if(any(grepl("nmatrix", dclasses)) == TRUE){
    cat("\nRegressor(s):\n")
    print(dclasses)
    cat("\n")
    stop(paste0(
      "\n  The formula type you used in your call to ",
      mymodeltype,"()"," is",
      "\n  not supported by predsplot() and predscor(),",
      "\n  because it contains a matrix as regressor.",
      "\n  Please rerun your ", mymodeltype, "()",
      " with a formula specifying ",
      "\n  all individual regressors, like ",
      "y ~ x1 + x2 + x3",
      "\n  and run predsplot() or predscor() on its",
      " output.", "\n "
    ))
  }
  predterms=predict(fit,type="terms",na.action=na.omit)
  regressortypes=attributes(fit$terms)$dataClasses[-1]

  npreds=ncol(predterms)
  colm=apply(predterms,2,mean,na.rm=T)
  # We submean the prediction terms, which is
  # needed for regression without intercept:
  predterms=scale(predterms,center=colm,scale=FALSE)
  predscales=apply(predterms,2,sd,na.rm=T)

  combinedpreds = FALSE
  predtermnames = colnames(predterms)
  if(!(is.numeric(maxnpreds))) stop("maxnpreds must be numeric")
  if(length(maxnpreds)>1) stop("maxnpreds must have length 1")
  if(maxnpreds < 2) stop("maxnpreds must be at least 2")
  if(maxnpreds < npreds){ # for npreds <= 2 we indeed do nothing
    rankpreds = rank(predscales, ties.method = "last")
    tocombine = which(rankpreds < (npreds - maxnpreds + 2))
    remterm   = rowSums(predterms[,tocombine])
    tokeep    = which(rankpreds > (npreds - maxnpreds + 1))
    predterms = cbind(predterms[,tokeep],remterm)
    regressortypes = c(regressortypes[tokeep],"combination")
    colnames(predterms)[maxnpreds] = "remainder"
    predtermnames = colnames(predterms) # reordered
    predscales = apply(predterms, 2, sd, na.rm=T)
    npreds     = ncol(predterms)
    combinedpreds = TRUE
  }

  # determine the directions:
  directions = myslopes = rep(0, npreds)
  predtypes  = rep("none", npreds)
  names(predtypes) = predtermnames
  for (jj in seq_len(npreds)){ # jj=4
    # Note that interactions aren't in the regressortypes
    # list, only in the list of prediction terms. So for
    # each prediction term, we first check whether it is
    # in the regressortypes list.
    varname = predtermnames[jj]
    if(varname %in% names(regressortypes)){
      ll = which(names(regressortypes) == varname)
      if(regressortypes[ll] == "factor"){
        predtypes[jj] = "factor"
      }
      if(regressortypes[ll] == "numeric"){
        if(varname %in% names(mycoeffs)){
          predtypes[jj] = "numeric" # it was already
          kk = which(names(mycoeffs) == varname)
          myslopes[jj] = mycoeffs[kk]
          if(mycoeffs[kk] > 0) directions[jj] = 1
          if(mycoeffs[kk] < 0) directions[jj] = -1
          # any zero coeffs keep directions = 0
        }
      }
    }
  }
  names(predscales) = predtermnames
  names(directions) = predtermnames
  names(myslopes)   = predtermnames
  updown = rep("",npreds)
  for(jj in seq_len(npreds)){
    if(directions[jj] ==  1) updown[jj] = "up"
    if(directions[jj] == -1) updown[jj] = "down"
  }
  if(mymodeltype == "lm"){
    tpredname = paste0("Total prediction of ", yname)
  } else {
    tpredname = paste0("Total linear prediction of ", yname)
  }
  termnames = c(predtermnames, tpredname)
  sdevs = c(predscales, totpredscale)
  sdevs = signif(sdevs,4)
  updown = c(updown, "")
  predsummary = data.frame(termnames,
                           sdevs,
                           updown,
                           row.names = NULL)
  colnames(predsummary) = c("prediction term",
                            "stdev", "up/down")
  if(verbose == TRUE & is.null(casetoshow)){
    cat("\n")
    cat("Prediction terms in the model:\n")
    print(predsummary, row.names = FALSE)
    cat("\n")
  }

  ## Prediction for one case:
  ##
  showcasename = ""
  casesummary = NULL
  casetotpred = 0.0; casepredterms = predscales*0.0
  casenames = rownames(predterms)
  if(!is.null(casetoshow)){
    showcasename = " for the new case"
    if(!is.data.frame(casetoshow) & !is.list(casetoshow)){
      # then it is not a set of input variables
      if(length(casetoshow) != 1) stop(paste0(
        "casetoshow is not a list or dataframe of input variables,\n",
        "  a single case number, or a single case name."))
      if(casetoshow %in% casenames){
        # (This can be a case number or a name.)
        ii = which(casenames == casetoshow)
        showcasename = paste0(" for case ",casetoshow)
        casetotpred = totpred[ii]
        casepredterms = predterms[ii,,drop=F]
        # These prediction terms were already submeaned before.
        if(mymodeltype == "glm") {
          casetpred2 = predict(fit, type="response",
                               na.action = na.omit)[ii]
        }
      } else { stop(paste0(
        "casetoshow = ", casetoshow,
        " is not a valid case in the dataset."))
      }
    } else { # here it should be a set of input variables,
      # either in the form of a dataframe or a list.
      if(verbose == TRUE){
        cat("\n The case to show is: \n")
        print(casetoshow)
      }
      if(mymodeltype == "glm") {
        casetotpred = predict(fit, casetoshow, type="link",
                              na.action = na.omit)
        # = linear totpred
        casetpred2 = predict(fit, casetoshow, type="response",
                             na.action = na.omit)
        # = totpred in response units
      } else {
        casetotpred = predict(fit, casetoshow, type="response",
                              na.action = na.omit)
      }
      # This will correctly stop when some input variables
      # are missing that are needed in the model.
      casetotpred = casetotpred - centercept
      casepredterms = predict(fit, casetoshow, type="terms",
                              na.action = na.pass)
      # Here we do need to submean:
      casepredterms = casepredterms - colm
    }
    colnames(casepredterms) = colnames(predterms)
    listtermnames = c(predtermnames, "SUM", "centercept",
                      tpredname)
    listpreds1 = c(casepredterms, casetotpred)
    listpreds2 = c(centercept,casetotpred+centercept)
    if(mymodeltype == "glm") {
      listpreds2 = c(listpreds2, casetpred2)
      totprdname = paste0("Total prediction of ", yname,
                          " in response units")
      listtermnames = c(listtermnames, totprdname)
    }
    listpredterms = c(formatC(listpreds1, format="f", digits=5, flag="+"), formatC(listpreds2, format="f", digits=5))
    casesummary = data.frame(listtermnames,
                             listpredterms,
                             row.names = NULL)
    colnames(casesummary) = c("prediction term","value")
    # rownames(casesummary) = NULL
    if(verbose == TRUE){
      cat("\n ")
      cat(paste0("Prediction terms",showcasename,":\n"))
      print(casesummary, row.names = FALSE)
      cat("\n")
    }
  }

  ## Reordering the prediction terms:
  ##
  orderpreds = seq_len(npreds)
  if(sort.by.stdev == TRUE){
    if(combinedpreds == FALSE){
      orderpreds = order(predscales, decreasing = TRUE)
    } else { # we need to keep the remainder at the end:
      orderpreds = order(predscales[-npreds], decreasing = TRUE)
      orderpreds = c(orderpreds, npreds)
    }
  }
  regressortypes = regressortypes[orderpreds]
  predtypes      = predtypes[orderpreds]
  predscales     = predscales[orderpreds]
  predterms      = predterms[, orderpreds, drop=F]
  casepredterms  = casepredterms[orderpreds]
  directions     = directions[orderpreds]
  myslopes       = myslopes[orderpreds]

  dendro = NULL
  orderpreds = seq_len(npreds)
  if(adj.order == TRUE){
    if(combinedpreds == FALSE){
      if(npreds > 2){ # otherwise nothing to do
        cormat = cor(predterms)
        dendro <- as.dendrogram(hclust(as.dist(1 - abs(cormat)),
                                       method = "average"))
        # hclust does method = "complete" by default, but we
        # can also take "single", "average" (= UPGMA) etc.
        dendro <- reorder(dendro, seq_len(npreds))
        # Due to seq_len(npreds) this reorders the dendrogram
        # in a way that is closest to the previous order of
        # the variables.
        orderpreds <- order.dendrogram(dendro)
        if(length(orderpreds) != npreds)
          stop("dendrogram ordering gave wrong order length")
      }
    } else { # we need to keep the remainder last
      if(npreds > 3){ # otherwise nothing to do
        cormat = cor(predterms[-npreds])
        dendro <- as.dendrogram(hclust(as.dist(1 - abs(cormat)),
                                       method = "average"))
        dendro <- reorder(dendro, seq_len(npreds))
        orderpreds <- c(order.dendrogram(dendro), npreds)
        if(length(orderpreds) != npreds)
          stop("dendrogram ordering gave wrong order length")
      }
    }
  }
  regressortypes = regressortypes[orderpreds]
  predtypes      = predtypes[orderpreds]
  predscales     = predscales[orderpreds]
  predterms      = predterms[, orderpreds, drop=F]
  casepredterms  = casepredterms[orderpreds]
  directions     = directions[orderpreds]
  myslopes       = myslopes[orderpreds]
  attr(predterms, "names") = NULL

  outpt = list(mymodeltype = mymodeltype,
               yname = yname,
               totpred = totpred,
               centercept = centercept,
               intercept = intercept,
               totpredscale = totpredscale,
               predterms = predterms,
               myslopes = myslopes,
               predtypes = predtypes,
               regressortypes = regressortypes,
               predscales = predscales,
               directions = directions,
               predsummary = predsummary,
               showcasename = showcasename,
               casetotpred = casetotpred,
               casepredterms = casepredterms,
               casesummary = casesummary)
  return(outpt)
}


predsplot = function(fit,
                     maxnpreds = 8,
                     sort.by.stdev = TRUE,
                     displaytype = "histogram",
                     totalpred.type = "response",
                     trunc.totalpred = TRUE,
                     casetoshow = NULL,
                     staircase = FALSE,
                     verbose = TRUE,
                     nfact = 8,
                     main = NULL,
                     cex.main = 1,
                     xlab = NULL,
                     ylab = NULL,
                     cex.labs = 1,
                     cex.varnames = 1,
                     cex.values = 1,
                     cex.levelnames = 1,
                     maxchar.level = 5,
                     plotcentercept = FALSE, # TRUE,
                     vlinewidth = 0.25,
                     hlinewidth = 0.25,
                     drawborder = TRUE,
                     borderwidth = 1.0,
                     densitycolors = NULL,
                     predupcolor = "red",
                     preddowncolor = "blue",
                     predpointsize = 3,
                     draw.segments = TRUE,
                     predsegmentwidth = 0.8,
                     profile = FALSE,
                     bw = "nrd0",
                     adjust = 1){

  # Arguments:
  #
  # fit              : an output object of lm() or glm().
  # maxnpreds        : the maximal number of prediction terms
  #                    to plot. When there are more prediction
  #                    terms than this, those with smallest
  #                    standard deviations are combined.
  # sort.by.stdev    : if TRUE, sorts the prediction terms by
  #                    decreasing standard deviation.
  # displaytype      : when "histogram", the distributions
  #                    of the numerical prediction terms and
  #                    the total prediction are displayed as
  #                    histograms. When "density", the
  #                    default density estimate of R is
  #                    plotted instead.
  # totalpred.type   : when fit is a glm object, option
  #                    "response" plots labels of the total
  #                    prediction that are obtained by the
  #                    inverse link function. Option
  #                    "linear" plots labels of the total
  #                    linear prediction. When fit is an
  #                    lm object, argument totalpred.type
  #                    is ignored.
  # trunc.totalpred  : if TRUE, the default, the range of
  #                    the total prediction is truncated
  #                    so ymin and ymax are determined by
  #                    the individual prediction terms. FALSE
  #                    may make the prediction terms look small
  #                    in the plot.
  # casetoshow       : if not NULL, the particular case
  #                    to be displayed. This can be a case
  #                    number or row name of a case in the
  #                    dataset, or a list or vector with
  #                    input values of a new case.
  # staircase        : if TRUE and casetoshow is not
  #                    NULL, the prediction for the case
  #                    is shown in staircase style.
  # verbose          : if TRUE, some intermediate results are
  #                    shown on the console.
  # nfact            : if a numeric input variable has at most
  #                    nfact unique values, it will be
  #                    displayed as if it were a factor.
  #                    This may be useful since R considers
  #                    a binary variable to be numeric.
  # main             : main title of the grill plot.
  # cex.main         : its size.
  # xlab, ylab       : horizontal and vertical legend.
  # cex.labs         : their size.
  # cex.varnames     : size of the variable names. It may help
  #                    to shorten the names in the data frame
  #                    before running the regression.
  # cex.values       : size of the values of the numeric variables.
  # cex.levelnames   : size of the level names of non-numeric
  #                    variables.
  # maxchar.level    : only the first maxchar.level
  #                    characters of the categorical levels
  #                    are displayed.
  # plotcentercept   : logical, whether to list the centercept in the plot.
  # vlinewidth       : width of the vertical lines.
  # hlinewidth       : width of the horizontal line.
  # drawborder       : if TRUE, draws a box outside the entire
  #                    plot.
  # densitycolors    : a vector with 4 colors. The first is
  #                    for numeric input variables pointing up,
  #                    the second for numeric input variables
  #                    pointing down, the third for
  #                    prediction terms without orientation
  #                    and the total prediction, and the
  #                    fourth for factors. If NULL, the
  #                    default colors are used.
  # predupcolor      : the color of a positive prediction
  #                    for casetoshow.
  # preddowncolor    : the color of a negative prediction
  #                    for casetoshow.
  # predpointsize    : the size of the points displaying the
  #                    predictions for casetoshow.
  # draw.segments    : if TRUE, also plots a line segment
  #                    from the center of each prediction term
  #                    to the prediction term for casetoshow,
  #                    in the same color as the prediction.
  # predsegmentwidth : the width of that line segment.
  # profile          : when casetoshow is not NULL and staircase
  #                    is FALSE, this plots the profile of the
  #                    case with a feint grey line.
  # bw               : the bandwidth of the density estimation,
  #                    only used when displaytype = "density".
  #                    This is the argument 'bw' of the function
  #                    density().
  # adjust           : multiplier of the bandwidth of the density
  #                    estimation, only used when displaytype =
  #                    "density". This is the argument 'adjust'
  #                    of the function density().
  #
  # Values:
  #
  # p               : the predictions plot, which is a ggplot2
  #                   object.
  # totpred         : vector with the total linear
  #                   prediction of all cases for which
  #                   it can be computed.
  # centercept      : the centercept, which is the total
  #                   linear prediction when all prediction
  #                   terms have their average value.
  # predterms       : matrix of cases by prediction terms.
  # predsummary     : data frame with the standard deviation of
  #                   each prediction term and the total linear
  #                   prediction.
  # casetotpred     : a number, the total linear prediction
  #                   for casetoshow if one is given.
  # casepredterms   : a vector with the values of the
  #                   prediction terms for casetoshow.
  # casesummary     : data frame which shows all prediction
  #                   terms for casetoshow together with
  #                   the centercept, total linear
  #                   prediction, and for a glm fit also
  #                   the total prediction in response units.

  draw_densities <- function(plotdata, predterms, xcoord,
                             displaytype, height = 0.7,
                             bw = "nrd0", adjust = 1) {
    # Auxiliary function
    varnames <- plotdata$varnames
    nvar <- length(varnames)
    centers <- plotdata$centers
    scales <- plotdata$scales
    directions <- plotdata$directions
    distcolor <- plotdata$distcolor
    out <- as.list(rep(NA, nvar))
    for (jj in seq_len(nvar)) {
      miny = plotdata$yminvals[jj]
      maxy = plotdata$ymaxvals[jj]
      if(displaytype == "histogram" ||
         plotdata$showtypes[jj] %in% c("factor","num2fac")){
        histdata <- predterms[,jj] + centers[jj]
        #
        # If categorical variable and few levels,
        # draw a kind of barplot:
        upreds = unique(histdata)
        if(length(upreds) < 8){
          upreds = sort(upreds)
          delta = min(abs(diff(upreds)))/3
          delta = max(delta, (maxy-miny)/500)
          delta = min(delta, (maxy-miny)/200)
          breaks = sort(c(upreds - delta, upreds + delta))
          myhist = hist(histdata, plot = FALSE,
                        breaks = breaks)
          eps = 0
        } else {
          myhist = hist(histdata, plot = FALSE, breaks=50)
          eps    = (maxy-miny)/1000
        }
        breaks = myhist$breaks
        nbr    = length(breaks)
        allbreaks = c(breaks, breaks[2:(nbr-1)] - eps)
        allbreaks = sort(allbreaks)
        allbreaks = c(min(allbreaks)-eps, allbreaks,
                      max(allbreaks)+eps)
        mygrid = c(miny, allbreaks, maxy)
        densit = myhist$density
        maxd = max(densit)
        for(ii in seq_along(densit)){
          # Hack to make outliers visible:
          if(densit[ii] > 0) densit[ii] = max(densit[ii],maxd/6)
        }
        alldensity = rep(NA, 2*(nbr-1))
        for(ii in seq_along(densit)){
          alldensity[2*ii - 1] = densit[ii]
          alldensity[2*ii] = densit[ii]
        }
        curve = c(0, 0, alldensity, 0, 0)
        length(mygrid); length(curve)
      }
      if(displaytype == "density" &
         showtypes[jj] == "numeric"){
        predvals <- predterms[,jj] + centers[jj]
        dens = stats::density(predvals, from = miny, to = maxy,
                              bw = bw, adjust = adjust)
        mygrid = c(miny, dens$x, maxy)
        curve = c(0, dens$y, 0)
      }
      ngrid = length(mygrid)
      predterm = rep(varnames[jj], ngrid)
      denscolor = rep(distcolor[jj], ngrid)
      out[[jj]] <- data.frame(predterm = predterm,
                              denscolor = denscolor,
                              gridvals = mygrid,
                              curve = curve)
      cmax <- max(out[[jj]]$curve)
      out[[jj]]$curve <- out[[jj]]$curve*height/cmax + xcoord[jj]
    }
    out[[nvar]]$curve <- 2*xcoord[nvar] - out[[nvar]]$curve
    out <- do.call("rbind", out)
    return(out)
  }

  ## Here the main function starts
  #
  # Check whether fit$model is available:
  exist = !(is.null(fit$model))
  if(exist == FALSE) stop(paste0(
    "There is no $model in your fit object. Please rerun",
    "\n  the fit with the default option model = TRUE."))
  #
  if (!is.numeric(vlinewidth)) stop("vlinewidth must be a number.")
  if (!is.numeric(hlinewidth)) stop("hlinewidth must be a number.")
  if(!(displaytype) %in% c("histogram","density")) stop(
    "displaytype must be either \"histogram\" ot \"density\"")

  out = predscomp(fit=fit, maxnpreds = maxnpreds,
                  sort.by.stdev = sort.by.stdev,
                  casetoshow = casetoshow,
                  verbose = verbose)
  mymodeltype    = out$mymodeltype
  yname          = out$yname
  regressortypes = out$regressortypes
  predtypes      = out$predtypes
  totpred        = out$totpred
  centercept     = out$centercept
  predterms      = out$predterms
  scales         = out$predscales
  directions     = out$directions
  myslopes       = out$myslopes
  nvar           = length(scales)
  varnames       = names(scales)
  xcoord         = seq_len(nvar)
  xcasepos       = xcoord
  predsummary    = out$predsummary
  showcasename   = out$showcasename
  casepredterms  = out$casepredterms
  casesummary    = out$casesummary

  showregmat <- predterms
  showtypes = predtypes
  modmat <- model.matrix(fit) # these are all numbers
  # Obtain the original values of the input variables
  # (possibly after log or X^2 or x2*x3) before
  # they were multiplied by the coefficients:
  datset = fit$model # may contain character strings
  classes = sapply(datset, class)
  datacolnames = colnames(datset)
  valid.rows = rownames(predterms)
  datset = datset[valid.rows,,drop=F]
  truelevels = as.list(rep(NA, nvar))
  #
  for (jj in seq_len(nvar)){ # jj=4
    varname = varnames[jj]
    # Note that interactions aren't in the regressortypes
    # list, only in the list of prediction terms. So for
    # each prediction term, we first check whether it is
    # in the regressortypes list.
    if(varname %in% names(regressortypes)){
      ll = which(names(regressortypes) == varname)
      if(predtypes[ll] == "numeric"){
        # This excludes "none","remainder" as it should.
        if(varname %in% colnames(modmat)){
          kk = which(colnames(modmat) == varname)
          showregmat[,jj] = as.vector(modmat[,kk])
          truelevels[[jj]] = "none"
        }
      }
      if(predtypes[ll] != "numeric"){
        if(varname %in% datacolnames){
          kk = which(datacolnames == varname)
          if("factor" %in% classes[[kk]]) {
            # this includes ordinal factors!
            showtypes[jj] = "factor"
            showregmat[,jj]  = datset[[kk]]
            truelevels[[jj]] = levels(datset[[kk]])
          }
          if(classes[[kk]] == "character") {
            levs = unique(datset[[kk]])
            datset[[kk]] = factor(datset[[kk]], levels = levs)
            showtypes[jj] = "factor"
            showregmat[,jj]  = datset[[kk]]
            truelevels[[jj]] = levs
          }
          if(classes[[kk]] == "logical") {
            showtypes[jj] = "factor"
            showregmat[,jj][datset[[kk]] == FALSE] = 1
            showregmat[,jj][datset[[kk]] == TRUE] = 2
            truelevels[[jj]] = c("F","T")
          }
        }
      }
    } else {
      showtypes[jj] = "none"
      truelevels[[jj]] = "none"
    }
  }

  for (jj in seq_len(nvar)){ # jj=4
    # If a numeric prediction term takes fewer than nfact
    # values, we will plot it like a factor.
    if(showtypes[jj] %in% c("numeric","none")){
      # so not "factor"
      uniq = unique(as.vector(showregmat[,jj]))
      if(length(uniq) <= nfact){
        showtypes[jj] = "num2fac"
        truelevels[[jj]] = sort(uniq)
      }
    }
  }

  if(is.null(casetoshow) == TRUE) staircase = FALSE

  if(is.null(xlab)){
    xlab = "prediction terms and total prediction"
    xlab = paste0(xlab,showcasename)
  }

  if(is.null(ylab)){
    if(mymodeltype == "lm"){
      ylab = "contributions to the prediction"
      if(staircase == TRUE){
        ylab = "cumulative prediction" }
    } else {
      ylab = "contributions to the linear prediction"
      if(staircase == TRUE){
        ylab = "cumulative linear prediction" }
    }
    ylab = paste0(ylab, " of ", yname)
  }

  totalpredlabel = "linear"
  ymaxpreds = apply(predterms, 2, max, na.rm=T)
  yminpreds = apply(predterms, 2, min, na.rm=T)
  # Append total prediction below and subtract the
  # centercept (we give the right labels later).
  # Instead of its ymaxvals take the max of the
  # ymaxvals of the prediction terms (so when a case
  # is shown, we can get wider xlim and ylim.)
  nvar                 = nvar + 1
  varnames             = c(varnames, "totalpred")
  scales               = c(scales, out$totpredscale)
  predterms            = cbind(predterms, totpred)
  showtypes            = c(showtypes, "numeric")
  casepredterms        = c(casepredterms, out$casetotpred)
  directions           = c(directions, 0)
  names(scales)        = varnames
  names(predterms)     = varnames
  names(casepredterms) = varnames
  names(directions)    = varnames
  xcoord         = c(xcoord, nvar + 0.6)
  xcasepos       = xcoord
  xcasepos[nvar] = xcoord[nvar] - 0.008*nvar
  if(mymodeltype == "glm"){
    if(!(totalpred.type %in% c("linear","response"))){
      stop("totalpred.type must be \"linear\" or \"response\".")}
    totalpredlabel = totalpred.type
  }

  centers  = rep(0, nvar)
  ymaxvals = apply(predterms, 2, max, na.rm=T)
  yminvals = apply(predterms, 2, min, na.rm=T)

  if(is.null(densitycolors)){
    densitycolors = c("olivedrab3", "chocolate",
                      "grey60", "black")
  }
  distcolor = rep(densitycolors[3], nvar)
  for(jj in seq_len(nvar-1)){
    if(directions[jj] == 1) distcolor[jj] = densitycolors[1]
    if(directions[jj] == -1) distcolor[jj] = densitycolors[2]
    if(directions[jj] == 0) distcolor[jj] = densitycolors[3]
    if(showtypes[jj] == "factor") distcolor[jj] = densitycolors[4]
  }
  distcolor[nvar] = densitycolors[3]
  if(!(is.null(casetoshow) == TRUE)) distcolor = rep("grey60", nvar)

  casepos = as.vector(casepredterms) # all 0 if no casetoshow
  predpointcolor = rep(predupcolor,nvar)
  if(!is.null(preddowncolor)){
    for(j in seq_len(nvar)){
      if(casepos[j] < 0) predpointcolor[j] = preddowncolor
    }
  }

  if(staircase == TRUE){
    partsum = cumsum(c(0.0,casepos))
    centers = centers + partsum[-nvar]
    centers[nvar] = 0.0
    casepos   = casepos + centers
    yminvals  = yminvals + centers
    ymaxvals  = ymaxvals + centers
    yminpreds = yminpreds + centers[seq_along(yminpreds)]
    ymaxpreds = ymaxpreds + centers[seq_along(yminpreds)]
  }

  names(centers) = varnames
  names(casepos) = varnames
  if(trunc.totalpred == TRUE){
    globalmax = max(c(ymaxpreds,casepos))
    globalmin = min(c(yminpreds,casepos))
  } else {
    globalmax = max(c(ymaxvals,casepos))
    globalmin = min(c(yminvals,casepos))
  }
  globalran = globalmax - globalmin # range
  ymaxvals  = ymaxvals + 0.06*globalran
  yminvals  = yminvals - 0.06*globalran
  globalmax = globalmax + 0.06*globalran
  globalmin = globalmin - 0.25*globalran
  allymax = rep(globalmax, nvar)
  allymin = rep(globalmin, nvar)
  vlinewidth = rep(vlinewidth, nvar)
  vlinewidth[nvar] = 0.0
  eps = globalran/50
  arrowstart = arrowend = rep(0, nvar)
  for(jj in seq_len(nvar)){
    if(directions[jj] == 1){
      arrowstart[jj] = ymaxvals[jj] - eps
      arrowend[jj]   = ymaxvals[jj]
    }
    if(directions[jj] == -1){
      arrowstart[jj] = yminvals[jj] + eps
      arrowend[jj]   = yminvals[jj]
    }
  }

  plotdata = data.frame(varnames, centers, scales, showtypes,
                        yminvals, ymaxvals, directions,
                        casepos, xcoord, allymin, allymax,
                        vlinewidth, arrowstart, arrowend,
                        distcolor, predpointcolor,
                        row.names = NULL, check.rows = TRUE)
  height = 0.6
  densities <- draw_densities(plotdata = plotdata,
                              predterms = predterms,
                              xcoord = xcoord,
                              displaytype = displaytype,
                              height = height, bw = bw,
                              adjust = adjust)
  height = rep(height, nvar)
  height[nvar] = -height[nvar]
  plotdata$height = height

  p <- ggplot(data = plotdata,
              aes(x = xcoord,
                  y = centers,
                  ymin = yminvals,
                  ymax = ymaxvals)) + theme_classic() +
    theme(axis.line = element_line(color='white'),
          plot.background = element_blank(),
          panel.grid.major = element_blank(),
          panel.grid.minor = element_blank(),
          panel.border = element_blank())

  p <- p + geom_polygon(data = densities, aes(y = .data$gridvals, x = .data$curve, group = .data$predterm, fill = .data$denscolor), alpha = 0.7, show.legend = FALSE, inherit.aes = FALSE) + scale_fill_identity()

  ## Add vertical line segments:
  p = p + ggplot2::geom_segment(aes(x = xcoord, xend = xcoord, y = yminvals, yend = ymaxvals), colour = "black", linewidth = vlinewidth, show.legend = FALSE)

  ## Add arrowheads in the right directions.
  # (If directions==0 no arrowhead is drawn.)
  p = p + ggplot2::geom_segment(aes(x = xcoord, xend = xcoord, y = arrowstart, yend = arrowend), colour = "black", linewidth = vlinewidth, arrow = arrow(angle = 20, length = unit(abs(directions)*0.2, "cm"), ends = "last", type="open"), show.legend = FALSE)

  # Draw the horizontal line through zero:
  if(!(staircase == TRUE) | nvar == 2){
    p <- p + geom_hline(yintercept = 0, linetype = 1, linewidth = hlinewidth, color = "black") }

  # Draw the profile if requested:
  if(!(staircase == TRUE) & profile == TRUE & nvar > 2){
    p <- p + geom_segment(aes(x = c(xcoord[1],xcoord[1:nvar-1]), y = c(casepos[1],casepos[1:nvar-1]), xend = c(xcoord[2],xcoord[2:(nvar-1)],xcoord[nvar-1]), yend = c(casepos[2],casepos[2:(nvar-1)],casepos[nvar-1])), linewidth = 2*hlinewidth, colour = "grey65", linetype=1) }

  ## Draw horizontal line segments for prediction staircase:
  if(staircase == TRUE){
    # Plot a point at the center:
    p <- p + geom_point(aes(x = xcasepos[1:nvar], y = centers[1:nvar]), size = 1.5, colour = "black")
    # Plot the steps of the staircase:
    p <- p + geom_segment(aes(x = c(xcoord[1],xcoord[1:nvar-1]), y = c(casepos[1],casepos[1:nvar-1]), xend = c(xcoord[1],xcoord[2:nvar]), yend = c(casepos[1],casepos[1:nvar-1])), linewidth = hlinewidth, colour = "black", linetype=1) }

  ## Add colored points at predictions:
  if(!is.null(casetoshow)){
    bla = c("red","blue")
    if(draw.segments){
      p <- p + ggplot2::geom_pointrange(aes(x = xcasepos, y = casepos, ymin = pmin(casepos,centers), ymax = pmax(casepos,centers)), position = ggplot2::position_dodge(width = 0.5), fill = "white", fatten = predpointsize, colour = predpointcolor, linewidth = predsegmentwidth, show.legend = FALSE)
    } else {
      p <- p + ggplot2::geom_pointrange(aes(x = xcoord, y = casepos, ymin = casepos, ymax = casepos), position = ggplot2::position_dodge(width = 0.5), fill = "white", fatten = predpointsize, colour = predpointcolor, linewidth = 0, show.legend = FALSE)
    }
  }

  ## add vertical label
  p <- p + ylab(ylab) + theme(axis.title.y = element_text(size = 12*cex.labs))

  ## Add variable names and overall axis label
  p <- p + scale_x_continuous(breaks = plotdata$xcoord, labels = NULL, name = xlab)
  p <- p + xlab(xlab) + theme(axis.title.x = element_text(size = 12*cex.labs))

  ## Set the xlim and ylim:
  xleft = min(xcoord)-0.7; xright = max(xcoord)
  ybottom = 1.01*globalmin; ytop = 1.01*globalmax
  p <- p + coord_cartesian(xlim = c(xleft, xright),
                           ylim = c(ybottom,ytop),
                           expand = FALSE)

  ## Remove x-ticks
  p <- p + theme(axis.ticks.x = element_blank())

  ## draw the vertical axis and set size of break labels.
  myxrange = max(xcoord) - min(xcoord)
  mysize  = 2.4*min(c(21/(myxrange), 4.5))
  p <- p + theme(axis.line.y = element_line(linewidth = vlinewidth, colour = "black"), axis.text.y = element_text(size = mysize, color = "black") )

  ## Add second axis:
  if(totalpredlabel == "linear"){
    if(mymodeltype == "lm"){
      tpredname = "total prediction"
    } else {
      tpredname = "total linear prediction"
    }
    if(isTRUE(plotcentercept)){
      tpredname = paste0(tpredname," (including centercept = ", signif(centercept,3),")")
    }
    p <- p + scale_y_continuous(sec.axis = ggplot2::sec_axis(~. + centercept, name = tpredname, breaks = scales::breaks_extended(n = 9), guide = guide_axis(theme=theme(axis.line = element_line(linewidth = 0, colour = "white")))))
  }
  #
  if(totalpredlabel == "response"){
    # Now we have to transform the linear totpred
    # by the inverse of the link function.
    myident   = function(y) { y + centercept }
    myexpit   = function(y) { 1/(1 + exp(-(y + centercept))) }
    myinverse = function(y) { 1/(y + centercept) }
    myexp     = function(y) { exp(y + centercept) }
    my1osqr   = function(y) { 1/sqrt(y + centercept) }
    #
    mylink = fit$family$link
    if(mylink == "identity") mytransf = myident
    if(mylink == "logit")    mytransf = myexpit
    if(mylink == "inverse")  mytransf = myinverse
    if(mylink == "log")      mytransf = myexp
    if(mylink == "1/mu^2")   mytransf = my1osqr
    if(!(mylink %in% c("identity","logit","inverse",
                       "log","1/mu^2"))) stop(paste0(
                         "The link ",mylink," has not been implemented,",
                         "\n  only identity, logit, inverse, log, 1/mu^2."))
    cat(paste0("The glm fit used the ",mylink," link function. \n"))
    p <- p + scale_y_continuous(sec.axis = ggplot2::sec_axis(transform = mytransf, name = "total prediction in response units", breaks = scales::breaks_extended(n = 9), guide = guide_axis(theme=theme(axis.line = element_line(linewidth = 0, color ="white")),check.overlap = TRUE)))
  }
  # flip the right y-axis title to match the left y-axis:
  p <- p + theme(axis.title.y.right = element_text(angle=90))

  ## Annotations on the vertical lines:
  lennames = max(nchar(varnames))+2
  varsize = cex.varnames * 1.2 * min(c(30/lennames, 3))
  for (jj in seq_len(ncol(showregmat))){
    varname = varnames[jj]
    p <- p + annotate("text", x = xcoord[jj], angle=90,
                      y = yminvals[jj],
                      label = paste0(varname,"  "),
                      hjust = 1, vjust = 0.3,
                      size = varsize)
    if(showtypes[jj] == "numeric"){
      showvals = as.numeric(showregmat[,jj])
      mybreaks = pretty(showvals, n=5,
                        min.n = 4, bounds = FALSE)
      rawmin = min(showvals); rawmax = max(showvals)
      tokeep = which(mybreaks >= rawmin & mybreaks <= rawmax)
      mybreaks = mybreaks[tokeep]
      mylabels = mybreaks
      for(pp in seq_along(mylabels)){
        mylabels[pp] = paste0(mylabels[pp],"-") }
      if(length(mybreaks) < 2) stop("under 2 breaks left")
      mymean  = mean(showvals, na.rm = TRUE)
      myslope = myslopes[jj]
      lchar   = max(nchar(mybreaks))
      mysize  = cex.values * 0.9*min(c(21/(myxrange), 14/lchar, 4.5))
      ypos    = myslope*(mybreaks - mymean) + centers[jj]
      p <- p + annotate("text", x = xcoord[jj],
                        y = ypos, label = mylabels,
                        hjust = 1, size = mysize)
    }
    if(showtypes[jj] == "factor"){
      levls = truelevels[[jj]]
      unt = sort(unique(predterms[,jj]))
      mypreds = mylabs = rep(NA,length(levls))
      for(kk in seq_along(unt)){
        mm = which(predterms[,jj] == unt[kk])[1]
        numlevl = showregmat[mm, jj]
        mylabs[kk] = levls[numlevl]
        mylabs[kk] = strtrim(mylabs[kk], maxchar.level)
        mylabs[kk] = paste0(mylabs[kk],"-")
      }
      lchar = max(nchar(mylabs))
      mysize = cex.levelnames * 0.9*min(c(21/(myxrange), 14/lchar, 4.5))
      ypos = unt + centers[jj]
      p <- p + annotate("text", x = xcoord[jj],
                        y = ypos, label = mylabs,
                        hjust = 1, size = mysize)
    }
    if(showtypes[jj] == "num2fac"){
      levls = truelevels[[jj]]
      unt = rep(NA,length(levls))
      mypreds = mylabs = rep(NA,length(levls))
      for(kk in seq_along(levls)){
        mm = which(showregmat[,jj] == levls[kk])[1]
        unt[kk] = predterms[mm,jj]
        mylabs[kk] = levls[kk]
        mylabs[kk] = strtrim(mylabs[kk], maxchar.level)
        mylabs[kk] = paste0(mylabs[kk],"-")
      }
      lchar = max(nchar(mylabs))
      mysize = 0.9*min(c(21/(myxrange), 14/lchar, 4.5))
      ypos = unt + centers[jj]
      p <- p + annotate("text", x = xcoord[jj],
                        y = ypos, label = mylabs,
                        hjust = 1, size = mysize)
    }
  }

  if(!is.null(main)){
    p <- p + theme(plot.title = element_text(hjust = 0.5, size = rel(1.2*cex.main)))
    p <- p + ggtitle(main)
  }

  if(drawborder == TRUE){
    p <- p + theme(plot.background = element_rect(color = "black", fill = NA, linewidth = borderwidth))
  }
  plot(p)
  attr(predterms, "names") = NULL
  outp = list(p               = p,
              totpred         = totpred,
              centercept      = centercept,
              predterms       = predterms,
              predsummary     = out$predsummary,
              casetotpred     = out$casetotpred,
              casepredterms   = casepredterms,
              casesummary     = out$casesummary)
  invisible(outp)
}


predscor <- function(fit,
                     maxnpreds = 8,
                     sort.by.stdev = TRUE,
                     adj.order = FALSE,
                     cell.length = "stdev",
                     plot.abs.cor = FALSE,
                     palette = NULL,
                     diagonalcolor = "black"){
  #
  # The function predscor() computes the correlations
  # between the prediction terms in a regression fit, and
  # displays them graphically in a way that takes the
  # standard deviations of the prediction terms into account.
  #
  # Arguments:
  #
  # fit           : an output object of lm() or glm().
  # maxnpreds     : the maximal number of prediction terms
  #                 to plot. When there are more prediction
  #                 terms than this, those with smallest
  #                 standard deviations are combined.
  # sort.by.stdev : if TRUE, sorts the prediction terms by
  #                 decreasing standard deviation.
  # adj.order     : if TRUE, this modifies the order
  #                 of the prediction terms in an attempt
  #                 to bring highly correlated
  #                 prediction terms close to each other
  #                 in the predscor() display. If
  #                 sort.by.stdev is TRUE, this happens
  #                 after that sorting.
  # cell.length   : if "stdev", the sides of the square
  #                 cells on the diagonal of the correlation
  #                 matrix are proportional to the standard
  #                 deviation of their prediction term.
  #                 If "sqrt" they are proportional to the
  #                 square root of the standard deviation.
  #                 If "equal" all sides are the same.
  # plot.abs.cor  : if FALSE, the default, positive and
  #                 negative correlations are shown in
  #                 different colors, typically red and
  #                 blue. If TRUE the absolute values
  #                 of the correlations are shown.
  # palette       : a vector with colors to display
  #                 correlations ranging from -1 to 1.
  #                 If NULL, the default palette shows
  #                 positive correlations in red,
  #                 negative correlations in blue, and
  #                 uses white for correlation zero.
  # diagonalcolor : color of the cells on the diagonal
  #                 of the correlation matrix. The
  #                 default is "black".
  #
  # Values:
  #
  # cormat      : the correlation matrix of the
  #               prediction terms.
  # predterms   : matrix of cases by prediction terms.
  # predsummary : data frame with the standard deviation of each
  #               prediction term and the total linear
  #               prediction.

  if(!(cell.length) %in% c("stdev","sqrt","equal")) stop(
    "cell.length must be \"stdev\", \"sqrt\" or \"equal\"")
  out = predscomp(fit = fit,
                  maxnpreds = maxnpreds,
                  sort.by.stdev = sort.by.stdev,
                  adj.order = adj.order,
                  casetoshow = NULL,
                  verbose = FALSE)
  predterms = out$predterms
  scales    = out$predscales
  nvar      = length(scales)
  varnames  = names(scales)
  cormat    = cor(predterms)
  cat("Correlation matrix of the prediction terms:\n")
  print(round(cormat,2))
  if(nvar < 2){ cat(paste0(
    "\nNo correlation plot is drawn because there is only",
    " one prediction term."))
  } else {
    if(is.null(palette)) palette = colorRampPalette(c("blue", "white", "red"))(500)
    palette[length(palette)] = diagonalcolor
    # The palette is long on purpose, so that we only
    # get black if corr > 0.996.
    lmat = lwid = lhei = 1
    layout(lmat, widths = lwid, heights = lhei, respect = TRUE)
    par(mar = c(5,10,10,5))
    z <- cormat[, nvar:1] # had to flip for image().
    # transformation to emphasize big abs(cor):
    z = z*(abs(z)<=0.5) + (1.5*z^2+0.125)*sign(z)*(abs(z)>0.5)
    if(plot.abs.cor == TRUE) z <- abs(z)
    mywidths = c(0,scales)
    if(cell.length == "sqrt") mywidths = c(0,sqrt(scales))
    if(cell.length == "equal") mywidths = c(0,rep(1,nvar))
    mywidths = pmax(mywidths, 0.02*max(mywidths))
    # print(mywidths)
    hbreaks = cumsum(mywidths)
    hbreaks = 15*hbreaks/max(hbreaks)
    vbreaks = max(hbreaks) - rev(hbreaks)
    #
    # Compute required width of margins:
    char_width_inch <- par("cin")[1]
    # = width of a single character in inches
    label_width_inch = max(strwidth(varnames,
                                    units = "inches"))
    margin_lines = label_width_inch/char_width_inch
    par(mar = c(0.3, margin_lines, margin_lines, 0.3))
    #        bottom       left         top      right
    #
    image(x=hbreaks, y=vbreaks, z=z,
          xlim = range(hbreaks), ylim = range(vbreaks),
          zlim = c(-1.625,1.625), axes = FALSE,
          xlab = "", ylab = "", col = palette)
    hmidpoints = (hbreaks[-(nvar+1)] + hbreaks[-1])/2
    vmidpoints = (vbreaks[-(nvar+1)] + vbreaks[-1])/2
    axis(side = 3, at = hmidpoints, labels = varnames,
         tick = FALSE, line = -0.7, cex.axis = 1,
         gap.axis = -1, las = 2)
    axis(side = 2, at = vmidpoints, labels = rev(varnames),
         tick = FALSE, line = -0.7, cex.axis = 1,
         gap.axis = -1, las = 2)
    abline(v = hbreaks)
    abline(h = vbreaks)
  }
  invis = list(cormat = cormat,
               predterms = predterms,
               predsummary = out$predsummary)
  invisible(invis)
}

daisy_vcr <- function(x, daisy.out = NULL, type = NULL,
                     weights = NULL, mybounds = NULL,
                     warnBin = FALSE, stand = TRUE) {
  #
  # To do:
  # - let daisy_vcr take X, Xnew and then compute the rectangular
  #   matrix of dissimilarities that we need, saving time and
  #   storage space. Jakob suggests that the data.matrix()
  #   mechanism needs to be replaced for this.
  #
  # Function that computes dissimilarities, based on cluster::daisy.
  # In case daisy.out is not NULL, x is considered to be new data,
  # and the information on the variables is extracted from daisy.out.
  # First, compatibility of x with daisy.out$x is checked.
  #
  # Type should be a list with elements:
  # $asymm, $symm, $ordratio, $logratio, $nominal.
  # Variable types not declared explicitly are assumed to be numeric.
  # [Peter] In fact, "$ordratio" is a strange name, it should just
  #   be "$ordinal". I modified the code to accept "$ordinal" too.
  #   Also, $logratio really just means take the logarithm, and this
  #   is superfluous in R where such a pretransformation is trivial.
  #   In general I think that type2 is much easier to work with than
  #   type, but we have to be backward compatible here. Maybe not in
  #   the side project later, that function will get a different
  #   name anyway.
  #
  # Options that were removed: metric (you can use dist() for
  # "euclidean" and "manhattan"), stand (you can standardize before
  # calling daisy_vcr), and most warn* (Jakob wants the warnings).
  # [Peter] I've put back one warning option (warnBin=F) as this
  #   warning would confuse some users, and it is a harmless
  #   occurrence since the result is the same anyway.
  #
  # Arguments:
  # x         : data matrix, in which the variables may be of
  #             mixed types.
  # daisy.out : if not NULL, the output of daisy_vcr on training
  #             data. If NULL, x is treated as training data.
  # type      : for training data, a list with variable types
  #             containing variable names (or their column numbers).
  #             If NULL, all variables are treated as numeric.
  #             For new data, the types of the training variables
  #             are used.
  # weights   : For training data, these are the variable weights,
  #             and by default they are all 1.
  #             For new data, the weights of the training variables
  #             are used.
  # mybounds  : a numeric vector c(a, b) of length two, with a < b.
  #             Only used for new data. After scaling a numeric
  #             variable (by the parameters it has in the training
  #             data), there may be values below 0 or above 1.
  #             These are then truncated to lie in the interval
  #             [a, b]. By default no truncation takes place.
  # warnBin   : if TRUE, a warning is given whenever a numeric
  #             variable takes on only 2 values.
  #
  if (length(dim(x)) != 2 && !(is.data.frame(x))) {
    stop("x is not a dataframe or a matrix.")
    }
  xorig <- x
  n <- nrow(x)
  p <- ncol(x)
  #
  if (is.null(daisy.out)) {
    # Checking the weights:
    if (is.null(weights)) weights <- 1
    if (length(weights) == 1) {
      weights <- rep.int(weights, p)
    } else if (length(weights) != p) {
      stop(paste0("'weights' must be of length ", p," (or 1)."))
    }
    if (sum(weights < 0) > 0) stop("There are negative weights.")
    if (is.null(type)) type <- list()
    #
  } else {
    # calculate dissimilarities on new data using
    # the result daisy.out from the training data:
    if (!is.null(weights)) warning(paste0(
      "\nFor new data, the weights of the training data are used.",
      "\nThe argument 'weights' in this call will not be used."))
    weights <- daisy.out$weights
    if (!is.null(type)) warning(paste0(
      "\nFor new data, the types of the training data are used.",
      "\nThe argument 'type' in this call will not be used."))
    type  <- daisy.out$type
    x_old <- daisy.out$x
    #
    # Check: is new data format compatible with training format?
    #
    allowNA <- TRUE # is NA allowed in test data?
    varsInModel <- which(daisy.out$weights > 0)
    checkXnew(x_old, x, allowNA = allowNA,
              varsInModel = varsInModel)
    # Match colnames to make sure they are in the same order, so
    # we can use the old weights, types, ranges, etc.:
    namesmatch <- match(colnames(x_old), colnames(x))
    x <- x[, namesmatch]
  }
  #
  # Now perform checks on the variable types:
  #
  varnms <- dimnames(x)[[2]]
  pColl <- function(n) paste(n, collapse = ", ")
  if (length(type)) {
    if (!is.list(type) || is.null(ntyp <- names(type)) ||
        any(ntyp == ""))
      stop(gettextf("invalid %s; must be named list",
                    sQuote("type")))
    typelist <- c("asymm", "symm", "nominal",
                 "logratio", "ordratio", "ordinal")
    for (nt in ntyp) { # e.g. nt = "nominal"
      if (!(nt %in% typelist)) stop(paste0(
        "type ",nt," is not allowed."))
      cvec <- type[[nt]]
      ct <- paste0("type$", nt)
      if (is.character(cvec)) {
        if (!is.null(varnms) && !all(cvec %in% varnms))
          stop(gettextf("%s has invalid column names",
                        ct))
      }
      else if (is.numeric(cvec)) {
        if (!all(1 <= cvec & cvec <= p))
          stop(gettextf("%s must be between 1 and ncol(x)",
                        ct))
      }
      else stop(gettextf("%s must contain column names or numbers",
                         ct))
    }
    tA <- type$asymm
    tS <- type$symm
    if (!is.null(tA) || !is.null(tS)) {
      d.bin <- cbind(as.data.frame(x[, tA, drop = FALSE]),
                     x[, tS, drop = FALSE])
      lenB <- sapply(lapply(d.bin, function(y) levels(as.factor(y))),
                     length)
      if (any(lenB > 2))
        stop("at least one binary variable has more than 2 levels.")
      if (!is.null(daisy.out) && any(lenB < 2))
        warning("at least one binary variable has < 2 levels.")
      if (any(is.f <- sapply(d.bin, is.factor)))
        d.bin[is.f] <- lapply(d.bin[is.f], function(f)
          as.integer(as.character(f)))
      if (!all(sapply(d.bin, function(y) is.logical(y) ||
                      all(sort(unique(as.numeric(y[!is.na(y)]))) %in%
                          0:1))))
        stop("at least one binary variable has values not in {0,1,NA}.")
    }
  }
  #
  # Convert the data to a numeric matrix, and create type2:
  #
  if (is.data.frame(x)) {
    type2 <- sapply(x, data.class)
    x <- data.matrix(x) # turns characters into numeric values.
  } else {
    type2 <- rep("numeric", p)
    names(type2) <- colnames(x)
  }
  if (length(type)) {
    tT <- type$ordratio
    if (is.null(tT)) tT <- type$ordinal # added
    x[, names(type2[tT])] <- unclass(as.ordered(x[, names(type2[tT])]))
    type2[tT] <- "T"
    tL <- type$logratio
    if (sum(as.vector(x[, names(type2[tL])]) <= 0) > 0) stop(
      "\nVariable(s) to be log-transformed are not always > 0.")
    x[, names(type2[tL])] <- log(x[, names(type2[tL])])
    type2[tL] <- "L"
    type2[tA] <- "A"
    type2[tS] <- "S"
    tN <- type$nominal
    type2[tN] <- "N"
  }
  type2[tI <- type2 %in% c("numeric", "integer")] <- "I"
  # here "I" stands for Interval-scaled.
  if (is.null(daisy.out) && warnBin && n > 9 && any(tI) &&
      any(iBin <- apply(x[, tI, drop = FALSE],
                        2, function(v) length(table(v)) == 2))) {
    warning(gettextf("\nNumeric variable(s) %s take only two values.",
                     pColl(which(tI)[iBin]))) }
  type2[type2 == "ordered"]   <- "O"
  type2[type2 == "factor"]    <- "N"
  type2[type2 == "character"] <- "N"
  if (any(ilog <- type2 == "logical")) {
    warning(sprintf(ngettext(
      sum(ilog),
      "setting 'logical' variable %s to type 'asymm'",
      "setting 'logical' variables %s to type 'asymm'"),
      pColl(which(ilog))), domain = NA)
    type2[ilog] <- "A"
  }
  #
  # Compute location and scale of the variables:
  #
  if (is.null(daisy.out)) {
    if (stand) {# classical Daisy standardization using the range
      colR   <- apply(x, 2, range, na.rm = TRUE)
      colmin <- colR[1, ]
    } else {# no standardization
      colR   <- rep(1, ncol(x))
      colmin <- rep(0, ncol(x))
    }

  } else {
    colR   <- daisy.out$colR
    colmin <- daisy.out$colmin
  }
  sx <- colR[2, ] - colmin
  if (any(sx == 0)) {sx[sx == 0] <- 1}
  #
  #
  for (j in seq_len(p)) { # iterate over the variables
    if (weights[j] > 0) { # skip variables with zero weight
      tempType <- type2[j]
      if (tempType %in% c("I", "T", "L")) {
        # Only Interval/Ordinal/Log variables are scaled:
        x[, j] <- scale(x[, j], center = colmin[j], scale = sx[j])

      } # ends if(type)
    } # ends if(weight > 0)
  } # ends loop over variables
  #
  # Actual distance computations:
  #
  nums   <- matrix(0, n, n) # numerator of the final dissim. matrix
  denoms <- matrix(0, n, n) # denominator of the final dissim. matrix
  for (j in seq_len(p)) { # iterate over the variables
    if (weights[j] > 0) { # skip variables with zero weight
      tempVar    <- x[, j]
      tempType   <- type2[j]
      denomstemp <- matrix(1, n, n) # by default, the denominator
      # is increased by 1 (we will multiply by weights[j] later).
      #
      # compute contribution to the numerator:
      if (tempType %in% c("A", "S", "N")) { # binary and nominal
        numtemp <- outer(tempVar, tempVar, "!=") + 0
      } else {# Interval/Ordinal/Log: Manhattan distance
        numtemp <- abs(outer(tempVar, tempVar, "-"))
      }
      # adjust denominator for asymmetric binary
      if (tempType == "A") {
        denomstemp <- (outer(tempVar, tempVar, "+") > 0) + 0
      }
      # handle NA's: the distance is 0, but the denominator does
      # not get incremented:
      numtemp[which(is.na(numtemp))] <- 0
      denomstemp[which(is.na(denomstemp))] <- 0
      #
      # use weights:
      numtemp    <- weights[j] * numtemp
      denomstemp <- weights[j] * denomstemp
      #
      # add to the overall distance matrix:
      nums   <- nums + numtemp
      denoms <- denoms + denomstemp
    }
  }
  disv <- as.dist(nums / denoms) # makes dissimilarity object
  #
  if (anyNA(disv)) attr(disv, "NA.message") <-
    "NA-values in the dissimilarity matrix !"
  class(disv) <-  c("dissimilarity","dist")
  attr(disv, "Labels") <- dimnames(x)[[1]]
  attr(disv, "Size")   <- n
  attr(disv, "Metric") <- "mixed"
  attr(disv, "Types")  <- type2
  #
  result <- list(disv = disv, # the dissimilarity object
                type = type, # types of the variables
                type2 = type2, # simpler version, as a check
                colR = colR, # column ranges (scales)
                colmin = colmin, # minimum (location) of each column
                weights = weights, # variable weights
                x = xorig)
  return(result)
}

#######################
# Auxiliary functions #
#######################


# Hidden from the user:
wnq <- function(string, qwrite = TRUE) { # auxiliary function
  # writes a line without quotes
  if (qwrite) write(noquote(string), file = "", ncolumns = 100)
}


# Hidden from the user:
pnq <- function(string, qwrite = TRUE) { # auxiliary function
  # prints without quotes
  if (qwrite) print(noquote(string))
}


# Hidden from the user:
checkLabels <- function(y, n, training = TRUE, levels = NULL) {
  if (training == TRUE) { # for training data
    if (!is.null(levels)) {
      stop("For training data, the levels argument is not used.")
    }
    whichy <- "y"
  } else {# for new data
    if (is.null(levels)) {
      stop("For new data, the argument levels is required")
    }
    whichy <- "ynew"
  }
  if (!(is.factor(y))) {
    stop(paste0("\n The response ", whichy, " is not a factor."))
  }
  if (is.null(attr(y, "levels"))) {
    stop(paste0("\n The response factor ", whichy, " has no levels.",
                "\n Please use as.factor() or factor() first."))
  }
  levelsattr <- attr(y, "levels")
  if (sum(is.na(levelsattr)) > 0) {
    stop(paste0("\n The levels attribute of ", whichy,
                " has at least one NA."))
  }
  dupls <- duplicated(levelsattr)
  if (sum(dupls) > 0) stop(paste0(
    "\n The label ", levelsattr[dupls == TRUE], " occurs more than ",
    " once in the levels attribute of ", whichy))
  if (length(y) != n) {
    stop(paste0("\n The response y should have length ", n,
                ", the number of cases."))
  }
  yv <- as.character(y) # this yv is not a factor any more.
  # When the factor y is malformed, the function stops here.
  # This happens when not all entries of y belong to levelsattr.
  indsv <- which(!is.na(yv)) # INDiceS of y that can be Visualized
  if (training == TRUE) { # for training data:
    if (length(indsv) == 0) {
      stop(paste0("The response factor y only contains NA's,"
                  , "\n so training is not possible."))
    }
    yv <- yv[indsv] # thus has at least 1 entry
    uniqy <- sort(unique(yv))
    for (g in seq_len(length(uniqy))) { # g = 1
      if (!(uniqy[g] %in% levelsattr)) {
        stop(paste0("\n The response label ", uniqy[g], " does not",
                    " appear in the levels attribute of y."))
      }
    }
    # From here on we know that all the yv are in levelsattr.
    if (length(uniqy) < 2) {
      stop(paste0("\n Only a single level occurs ( ", uniqy[1],
                  " ) so training is not possible."))
    }
    for (g in seq_len(length(levelsattr))) { # g = 1
      if (!(levelsattr[g] %in% uniqy)) {
        wnq(paste0("\nThe level ", levelsattr[g], " does not occur",
                   " among the training cases. A model trained",
                   "\non these data will not be able to",
                   " assign any objects to this class.\n"))
      }
    }
    # Create array "levels" with the labels that actually occur:
    levels <- levelsattr[which(levelsattr %in% uniqy)]; levels
    #
  } else {# for new data:
    if (length(indsv) > 0) {
      yv <- yv[indsv] # thus has at least 1 entry
      uniqy <- sort(unique(yv))
      for (g in seq_len(length(uniqy))) { # g = 1
        if (!(uniqy[g] %in% levelsattr)) {
          stop(paste0("\n The response label ", uniqy[g], " does not",
                      " appear in the levels attribute of ynew."))
        }
      }
      # From here on we know that all the yv are in levelsattr.
      badlevels <- uniqy[which(!(uniqy %in% levels))]; badlevels
      if (length(badlevels) > 0) { wnq(paste0(
        "\n The level ", badlevels, " occurs in ynew but",
        " was not in the training data.",
        "\n Such cases will be treated as if their ynew is NA",
        " so their response", "\n can be predicted, but",
        " these cases will not be in the class map.\n"))
        indsv <- indsv[which(!(yv %in% badlevels))]
        # the resulting indsv may again be empty
      }
    }
  }
  xlevelz <- levels # for use in the following function:
  lab2int <- function(labs){
    # labs is a vector of labels, that may contain NA's
    ints <- rep(NA, length(labs))
    indv <- which(!is.na(labs))
    labv <- labs[indv] # uses the final indsv above
    for (g in seq_len(length(xlevelz))) { # g = 1
      clinds <- indv[which(labv == xlevelz[g])] # in all of labs
      ints[clinds] <- g
    }
    ints
  }
  return(list(lab2int = lab2int, levels = levels, indsv = indsv))
}


# Hidden from the user:
meancov <- function(X) {
  # Wrapper around colMeans and classical covariance matrix
  return(list(m = colMeans(X), S = cov(X)))
}


# Hidden from the user:
DetMCD <- function(X) {
  # Wrapper around robustbase::covMCD
  detmcd.out <- robustbase::covMcd(X, nsamp = "deterministic", alpha = 0.5)
  return(list(m = detmcd.out$center, S = detmcd.out$cov))
}


# Hidden from the user:
numericalCond <- function(S) {
  # Computes the inverse condition number of a matrix, which
  # is the ratio: smallest eigenvalue/largest eigenvalue .
  # Unlike the CN, this avoids division by (near) zero.
  # I call this the "numerical condition" (NC).
  #
  S <- as.matrix(S)
  if (length(which(is.na(S))) > 0) stop(" S contains NA's.")
  d <- nrow(S)
  if (ncol(S) != d) stop(" S is not square.")
  maxS <- max(abs(as.vector(S)))
  if (!isSymmetric(S, tol = 1e-10 * maxS)) stop(" S is not symmetric.")
  eigvals <- eigen(S, only.values = TRUE)$values
  eigvals[d] / eigvals[1]
}


# Visible to the user (is called in example script):
makeKernel <- function(X1, X2 = NULL, svfit) {
  # Computes kernel value or kernel matrix, where the
  # kernel type is extracted from svfit.
  #
  # Arguments:
  # X1    : first data (coordinate) matrix or vector.
  # X2    : if not NULL, second data matrix or vector.
  # svfit : an object returned by e1071::svm(), or a list
  #         with the members $scaled, $kernel, $degree,
  #         $gamma, and $coef0 .
  #         Here $kernel may be in c(0, 1, 2, 3) or in
  #         c("linear", "polynomial", "radial", "sigmoid").
  #
  # A heuristic choice for gamma in radial and sigmoid is
  #    0.5/median(as.vector(as.matrix(dist(Xmat)))^2)
  #
  # Value:
  # kmat  : the kernel matrix or kernel value.
  #
  if (is.vector(X1)) {
    X1 <- t(as.matrix(X1))
  } else {
    X1 <- as.matrix(X1)
  }
  d <- ncol(X1)
  if (!is.null(X2)) {
    if (is.vector(X2)) {
      X2 <- t(as.matrix(X2))
    } else {
      X2 <- as.matrix(X2)
    }
    if (ncol(X2) != d) stop(" X1 and X2 have different dimensions")
  }
  if (is.null(svfit$kernel)) {
    stop(" svfit$kernel is missing")
  } else {
    k <- svfit$kernel
  }
  if (!(k == 0 | k == "linear" | k == 1 | k == "polynomial" |
        k == 2 | k == "radial" | k == 3 | k == "sigmoid")) {
    stop(paste(" Unknown kernel type: ", k, sep = ""))
  }
  if (is.null(svfit$scaled)) {
    stop(" svfit$scaled is missing")
  } else {
    scaled <- svfit$scaled
  }
  if (is.null(svfit$degree)) {
    stop(" svfit$degree is missing")
  } else {
    degree <- svfit$degree
  }
  if (is.null(svfit$gamma)) {
    stop(" svfit$gamma is missing")
  } else {
    gamma <- svfit$gamma
  }
  if (is.null(svfit$coef0)) {
    stop(" svfit$coef0 is missing")
  } else {
    coef0 <- svfit$coef0
  }

  #

  if (length(scaled) != d) {
    stop(paste(" svfit$scaled should have length ", d, sep = ""))
  }
  # Now we scale the corresponding columns:
  sind <- which(scaled == TRUE)
  if (length(sind) > 0) {
    if (is.null(svfit$x.scale)) {
      stop(" svfit$x.scale is missing")
    } else {
      ctrs <- svfit$x.scale[[1]]
      scls <- svfit$x.scale[[2]]
      X1[, sind] <- scale(X1[, sind], center = ctrs, scale = scls)
      if (!is.null(X2)) {
        X2[, sind] <- scale(X2[, sind], center = ctrs, scale = scls)
      }
    }
  }
  #
  # kernel definitions
  #
  if (k == 0 | k == "linear") {
    kern <- function(xi, yj) {sum(xi * yj)}
  }
  if (k == 1 | k == "polynomial") {
    kern <- function(xi, yj) {(gamma * sum(xi * yj) + coef0)^degree}
  }
  if (k == 2 | k == "radial") {
    kern <- function(xi, yj) {exp(-gamma * sum((xi - yj) * (xi - yj)))}
  }
  if (k == 3 | k == "sigmoid") {
    kern <- function(xi, yj) {tanh(gamma * sum(xi * yj) + coef0)}
  }
  n1 <- nrow(X1)
  if (is.null(X2)) { # compute n1 by n1 matrix kmat
    kmat <- matrix(rep(0, n1 * n1), nrow = n1)
    dim(kmat)
    # Only fill triangular matrix and then put in mirror
    # image to obtain whole symmetric matrix:
    for (i in seq_len(n1)) {
      for (j in seq_len(i)) {
        kmat[i, j] <- kern(X1[i, ], X1[j, ])
      }
    }
    kmat <- kmat + t(kmat)
    diag(kmat) <- 0.5 * diag(kmat)
  }
  if (!is.null(X2)) { # compute n1 by n2 matrix kmat
    n2 <- nrow(X2)
    kmat <- matrix(rep(0, n1 * n2), nrow = n1)
    for (i in seq_len(n1)) {
      for (j in seq_len(n2)) {kmat[i, j] <- kern(X1[i, ], X2[j, ])
      }
    }
  }
  if (nrow(kmat) == 1 & ncol(kmat) == 1) kmat <- kmat[1, 1]
  return(kmat)
}


# Visible to the user:
makeFV <- function(kmat, transfmat = NULL, precS = 1e-12)
{
  kmat <- as.matrix(kmat)
  if (length(which(is.na(kmat))) > 0) {
    stop(" The kernel matrix kmat contains NA's.")
  }
  n <- nrow(kmat)
  maxK <- max(abs(as.vector(kmat)))
  if (is.null(transfmat)) { # so kmat contains training data
    if (ncol(kmat) != n) {
      stop(" The training kernel matrix kmat is not square.")
    }
    if (maxK < 0.001) {
      stop(" The training kernel matrix is too close to zero.")
    }
    if (max(abs(as.vector(kmat - t(kmat)))) > 1e-10 * maxK) {
      stop(" The training kernel matrix kmat is not symmetric.")
    }
    defaultPrecS <- 1e-16
    if (is.null(precS)) {
      precS <- defaultPrecS
    }
    if (precS < defaultPrecS) {
      precS <- defaultPrecS
    }
    eig <- eigen(kmat)
    eigvals <- pmax(eig$values, 0)
    eig <- eig$vectors
    V <- eig; rm(eig) # R's rename, to save memory
    keep <- length(which(eigvals > precS))
    if (keep < ncol(kmat)) {
      wnq(paste0(" The kernel matrix has ", ncol(kmat) -
                              keep, " eigenvalues below precS = ", precS,
                            "."))
      wnq(" Will take this into account.")
    }
    V <- V[,seq_len(keep)]
    colnames(V) <- NULL
    flipcolumn <- function(x) {
      if (x[which.max(abs(x))] < 0) {
        -x
      }
      else {
        x
      }
    }
    V <- apply(V, 2L, FUN = flipcolumn)
    scals <- sqrt(eigvals[seq_len(keep)])
    # Are the root sum of squares of the feature variables.
    Xf <- V %*% diag(scals) # This is safe because we do not
    # have to invert scals.
    V <- V %*% diag(1/scals)
    transfmat <- V; rm(V) # R's rename, to save memory
  }
  else { # so kmat contains new data
    dim <- nrow(transfmat)
    if (ncol(kmat) != dim) {
      stop(paste(" The matrix kmat has ", ncol(kmat), " columns",
                 " but the argument transfmat was made for ",
                 dim, " columns.", sep = ""))
    }
    Xf <- kmat %*% transfmat
    transfmat <- NULL # no need to output it again
  }
  return(list(Xf = Xf, transfmat = transfmat))
}

# Hidden from the user:
compFarness = function (type = "affine", testdata = FALSE, yint, nlab,
                         X = NULL, fig = NULL, d = NULL, PCAfits = NULL, keepPCA = FALSE,
                         figparams = NULL){
  far2probability <- function(farness, trainInds = NULL) {
    YJ <- function(y, lambda, chg = NULL, stdToo = TRUE) {
      indlow <- which(y < 0)
      indhigh <- which(y >= 0)
      if (lambda != 0) {
        y[indhigh] <- ((1 + y[indhigh])^(lambda) - 1)/lambda
      }
      else {
        y[indhigh] <- log(1 + y[indhigh])
      }
      if (lambda != 2) {
        y[indlow] <- -((1 - y[indlow])^(2 - lambda) -
                         1)/(2 - lambda)
      }
      else {
        y[indlow] <- -log(1 - y[indlow])
      }
      if (stdToo) {
        if (length(y) > 1) {
          locScale <- cellWise::estLocScale(matrix(y,
                                                   ncol = 1), type = "hubhub")
          zt <- (y - locScale$loc)/locScale$scale
        }
        else {
          zt <- y
        }
      }
      else {
        zt <- NULL
      }
      return(list(yt = y, zt = zt))
    }
    origfarness <- farness
    if (!is.null(trainInds)) {
      farness <- farness[trainInds]
    }
    indnz <- which(farness > 1e-10)
    farnz <- farness[indnz]
    farloc <- median(farnz, na.rm = TRUE)
    farsca <- mad(farnz, na.rm = TRUE)
    if (farsca < 1e-10) {
      farsca <- sd(farnz, na.rm = TRUE)
    }
    sfar <- scale(farnz, center = farloc, scale = farsca)
    YJ.out <- cellWise::transfo(X = sfar, robust = TRUE, standardize = FALSE,
                                checkPars = list(silent = TRUE))
    xt <- YJ.out$Y
    tfarloc <- median(xt, na.rm = TRUE)
    tfarsca <- mad(xt, na.rm = TRUE)
    lambda <- YJ.out$lambdahats
    origIndnz <- which(origfarness > 1e-10)
    origFarnz <- origfarness[origIndnz]
    zt <- scale(YJ(scale(origFarnz, farloc, farsca), lambda = lambda,
                   stdToo = FALSE)$yt, tfarloc, tfarsca)
    probs <- rep(0, length(origfarness))
    probs[origIndnz] <- pnorm(zt)
    tfunc <- function(qs) {
      qsIndnz <- which(qs > 1e-10)
      qsnz <- qs[qsIndnz]
      zn <- scale(YJ(scale(qsnz, farloc, farsca), lambda = lambda,
                     stdToo = FALSE)$yt, tfarloc, tfarsca)
      pn <- rep(0, length(qs))
      pn[qsIndnz] <- pnorm(zn)
      return(pn)
    }
    return(list(probs = probs, far2prob = tfunc))
  }

  transformYJ <- function(X, lambda) {
    if (!is.vector(X)) {
      stop("transformYJ: the data should be a vector or a number")
    }
    if (sum(is.na(X)) > 0)
      stop("transformYJ: the data contains NA's")
    Xt <- rep(NA, length(X))
    indlow <- which(X < 0)
    indhigh <- which(X >= 0)
    if (lambda != 0) {
      Xt[indhigh] <- ((1 + X[indhigh])^(lambda) - 1)/lambda
    }
    else {
      Xt[indhigh] <- log(1 + X[indhigh])
    }
    if (lambda != 2) {
      Xt[indlow] <- -((1 - X[indlow])^(2 - lambda) - 1)/(2 -
                                                           lambda)
    }
    else {
      Xt[indlow] <- -log(1 - X[indlow])
    }
    return(Xt)
  }
  mednz <- function(x, prec = 1e-12) {
    if (!(prec > 0))
      stop("mednz: prec = ", prec, " should be > 0")
    xx <- x[which(!is.na(x))]
    xx <- xx[which(xx > prec)]
    if (length(xx) == 0) {
      mednzx <- prec
    }
    else {
      mednzx <- median(xx)
    }
    mednzx
  }

  if (!(type %in% c("affine", "knn", "pca"))) {
    stop(" type must be either \"affine\" or \"knn\" or \"pca\".")
  }
  n <- length(yint)
  indsv <- which(!is.na(yint))
  yintv <- yint[indsv]
  ayint <- sort(unique(yint[indsv]))
  farness <- rep(NA, n)
  classSizes <- rep(0, nlab)
  for (g in seq_len(nlab)) {
    classSizes[g] <- sum(yintv == g)
  }
  medOD <- medSD <- SDs <- ODs <- NULL
  if (type == "affine") {
    if (is.null(fig)) {
      stop("argument fig is missing")
    }
    if (testdata == FALSE) {
      for (g in ayint) {
        clinds <- which(yint == g)
        farness[clinds] <- fig[clinds, g]
      }
      omedian <- mednz(farness, prec = 1e-08)
      farscales <- rep(NA, nlab)
      for (g in ayint) {
        clinds <- which(yint == g)
        gfarness <- farness[clinds]
        farscales[g] <- mednz(gfarness, prec = 1e-08)/omedian
        farness[clinds] <- gfarness/farscales[g]
      }
      mfar <- median(farness, na.rm = TRUE)
      sfar <- max(mad(farness, na.rm = TRUE), 1e-08)
      farness <- as.vector(scale(farness, center = mfar,
                                 scale = sfar))
      tout <- cellWise::transfo(farness, type = "YJ", robust = TRUE,
                                standardize = FALSE,
                                checkPars = list(silent = TRUE))
      lambda <- tout$lambdahats[1]
      muhat <- tout$muhat
      sigmahat <- tout$sigmahat
      figparams <- list(farscales = farscales, mfar = mfar,
                        sfar = sfar, lambda = lambda, muhat = muhat,
                        sigmahat = sigmahat)
    }
    else {
      if (is.null(figparams))
        stop("Argument figparams is missing")
      farscales <- figparams$farscales
      if (is.null(farscales))
        stop("figparams$farscales is missing")
      mfar <- figparams$mfar
      sfar <- figparams$sfar
      lambda <- figparams$lambda
      muhat <- figparams$muhat
      sigmahat <- figparams$sigmahat
    }
    fig <- scale(fig, center = FALSE, scale = farscales)
    for (g in seq_len(nlab)) {
      fig[, g] <- as.vector(scale(fig[, g], center = mfar,
                                  scale = sfar))
      fig[, g] <- transformYJ(fig[, g], lambda)
      fig[, g] <- as.vector(scale(fig[, g], center = muhat,
                                  scale = sigmahat))
      fig[, g] <- pnorm(fig[, g])
    }
    for (g in ayint) {
      clinds <- which(yint == g)
      farness[clinds] <- fig[clinds, g]
    }
  }# ends "affine"
  if (type == "knn") {
    if (is.null(fig)) {
      stop("argument fig is missing")
    }
    if (testdata == FALSE) {
      for (g in ayint) {
        clinds <- which(yint == g)
        farness[clinds] <- fig[clinds, g]
      }
      omedian <- mednz(farness, prec = 1e-08)
      farscales <- rep(NA, nlab)
      for (g in ayint) {
        clinds <- which(yint == g)
        gfarness <- farness[clinds]
        farscales[g] <- mednz(gfarness, prec = 1e-08)/omedian
      }
      fig <- scale(fig, center = FALSE, scale = farscales)
      farness <- rep(NA, n)
      far2probFuncs <- list()
      probs <- matrix(NA, nrow = nrow(fig), ncol = ncol(fig))
      for (g in seq_len(nlab)) {
        clinds <- which(yint == g)
        ftemp <- fig[, g]
        far2prob.out <- far2probability(ftemp, trainInds = clinds)
        fig[, g] <- far2prob.out$probs
        farness[clinds] <- fig[clinds, g]
        far2probFuncs[[g]] <- far2prob.out$far2prob
      }
      figparams <- list(farscales = farscales, far2probFuncs = far2probFuncs)
    }
    else {
      if (is.null(figparams))
        stop("Argument figparams is missing")
      farscales <- figparams$farscales
      if (is.null(farscales))
        stop("figparams$farscales is missing")
      fig <- scale(fig, center = FALSE, scale = farscales)
      far2probFuncs <- figparams$far2probFuncs
      for (g in seq_len(nlab)) {
        far2probTemp <- far2probFuncs[[g]]
        fig[, g] <- far2probTemp(fig[, g])
      }
      for (g in ayint) {
        clinds <- which(yint == g)
        farness[clinds] <- fig[clinds, g]
      }
    }
  } # ends "knn"
  if (type == "pca") {
    if (is.null(X)) {
      stop("argument X is missing")
    }
    if (is.vector(X)) {
      X <- matrix(X, ncol = 1)
    }
    X <- as.matrix(X)
    dim(X) # 75 196
    d <- ncol(X)
    if (nrow(X) != n) {
      stop(" X and yint have incompatible sizes")
    }
    fig <- SDs <- ODs <- matrix(rep(NA, n * nlab), ncol = nlab)
    if (testdata == FALSE) {
      if (n < 2)
        stop(" X should have more than one row")
      PCAfits <- NULL
      if (keepPCA)
        PCAfits <- list()
      medSD <- medOD <- rep(NA, nlab)
      medscores <- madscores <- list()
      for (j in seq_len(nlab)) {
        clinds <- which(yint == j)
        Xj <- X[clinds, , drop = FALSE]
        nXju <- sum(duplicated(Xj) == FALSE)
        cntr <- colMeans(Xj)
        outpca <- prcomp(sweep(Xj, 2, cntr), scale = FALSE)
        if (nXju == 1) {
          outpca$rotation <- outpca$rotation * 0
        }
        sdev <- outpca$sdev
        tokeep <- length(which(sdev > 1e-08))
        ncomps <- max(tokeep, 1)
        sdev <- sdev[seq_len(ncomps)]
        loadings <- outpca$rotation[, seq_len(ncomps),
                                    drop = FALSE]
        if (keepPCA)
          PCAfits[[j]] <- list(lj = loadings, cj = cntr,
                               sj = sdev)
        dim(Xj)
        length(cntr)
        dim(loadings)
        pX <- sweep(X, 2, cntr) %*% loadings
        pX <- matrix(pX, nrow = n)
        dim(pX)
        if (tokeep == d) {
          ODs[, j] <- rep(0, n)
        }
        else {
          Xdiffs <- sweep(X, 2, cntr) - pX %*% t(loadings)
          ODs[, j] <- sqrt(rowSums(Xdiffs^2))
          ODs[clinds, j] <- rep(0, length(clinds))
        }
        medOD[j] <- mednz(ODs[setdiff(indsv, clinds),
                              j], prec = 1e-08)
        ODs[, j] <- ODs[, j]/medOD[j]
        scores <- pX[clinds, , drop = FALSE]
        dim(scores)
        medscores[[j]] <- apply(scores, 2, median)
        madscores[[j]] <- apply(scores, 2, mad)
        madscores[[j]] <- pmax(madscores[[j]], 1e-08)
        scscores <- scale(pX, center = medscores[[j]],
                          scale = madscores[[j]])
        SDs[, j] <- sqrt(rowSums(scscores^2))
        medSD[j] <- mednz(SDs[clinds, j], prec = 1e-08)
        SDs[, j] <- SDs[, j]/medSD[j]
        if (tokeep == 0)
          SDs[, j] <- rep(1, length(SDs[, j]))
        fig[, j] <- sqrt(SDs[, j]^2 + ODs[, j]^2)
      }
      for (g in seq_len(nlab)) {
        clinds <- which(yint == g)
        farness[clinds] <- fig[clinds, g]
      }
      omedian <- mednz(farness[indsv], prec = 1e-08)
      farscales <- rep(NA, nlab)
      for (g in ayint) {
        clinds <- which(yint == g)
        gfarness <- farness[clinds]
        farscales[g] <- mednz(gfarness, prec = 1e-08)/omedian
        farness[clinds] <- gfarness/farscales[g]
      }
      mfar <- median(farness[indsv], na.rm = TRUE)
      sfar <- mad(farness[indsv], na.rm = TRUE)
      sfar <- max(sfar, 1e-08)
      farness <- as.vector(scale(farness, center = mfar,
                                 scale = sfar))
      tout <- cellWise::transfo(farness[indsv], type = "YJ", robust = TRUE,
                                standardize = FALSE, checkPars = list(silent = TRUE))
      lambda <- tout$lambdahats[1]
      muhat <- tout$muhat
      sigmahat <- tout$sigmahat
      figparams <- list(medOD = medOD, medscores = medscores,
                        madscores = madscores, medSD = medSD, farscales = farscales,
                        mfar = mfar, sfar = sfar, lambda = lambda, muhat = muhat,
                        sigmahat = sigmahat)
    } # ends if(training)
    else { # if(newdata)
      if (is.null(PCAfits)) {
        stop(paste0("\nFor test data, the farness computation ",
                    "requires the trained object", "\nto contain ",
                    "the value $PCAfits. Rerun the training ",
                    "with keepPCA = TRUE"))
      }
      if (is.null(figparams))
        stop("Argument figparams is missing")
      medOD <- figparams$medOD
      madscores <- figparams$madscores
      medscores <- figparams$medscores
      medSD <- figparams$medSD
      farscales <- figparams$farscales
      for (j in seq_len(nlab)) { # j=1
        loadings <- PCAfits[[j]]$lj
        dim(loadings) # 20 20
        cntr <- PCAfits[[j]]$cj
        sdev <- PCAfits[[j]]$sj
        tokeep <- length(which(sdev > 1e-08))
        ncomps <- ncol(loadings)
        clinds <- which(yint == j)
        Xj = X
        if(nrow(loadings) < ncol(Xj)) {
          Xj = Xj[, seq_len(nrow(loadings))]
        }
        pX <- sweep(Xj, 2, cntr) %*% loadings
        Xdiffs <- sweep(Xj, 2, cntr) - pX %*% t(loadings)
        ODs[, j] <- sqrt(rowSums(Xdiffs^2))
        ODs[, j] <- ODs[, j]/medOD[j]
        scscores <- scale(pX, center = medscores[[j]],
                          scale = madscores[[j]])
        SDs[, j] <- sqrt(rowSums(scscores^2))
        SDs[, j] <- SDs[, j]/medSD[j]
        if (tokeep == 0)
          SDs[, j] <- rep(1, length(SDs[, j]))
        fig[, j] <- sqrt(SDs[, j]^2 + ODs[, j]^2)
      }
    }
    fig <- scale(fig, center = FALSE, scale = farscales)
    mfar <- figparams$mfar
    sfar <- figparams$sfar
    lambda <- figparams$lambda
    muhat <- figparams$muhat
    sigmahat <- figparams$sigmahat
    for (g in seq_len(nlab)) {
      fig[, g] <- as.vector(scale(fig[, g], center = mfar,
                                  scale = sfar))
      fig[, g] <- transformYJ(fig[, g], lambda)
      fig[, g] <- as.vector(scale(fig[, g], center = muhat,
                                  scale = sigmahat))
      fig[, g] <- pnorm(fig[, g])
    }
    for (g in ayint) {
      clinds <- which(yint == g)
      farness[clinds] <- fig[clinds, g]
    }
  }
  ofarness <- apply(fig, 1, min)
  if (!keepPCA)
    PCAfits <- NULL
  return(list(fig = fig, farness = farness, ofarness = ofarness,
              figparams = figparams, PCAfits = PCAfits, medSD = medSD,
              medOD = medOD, SDs = SDs, ODs = ODs))
}

###############################################################################


checkXnew <- function(Xtrain, Xnew, varsInModel = NULL,
                      allowNA = FALSE) {
  # This function checks whether Xnew follows the same format
  # as Xtrain. It checks variable names, variable types,
  # data.frame vs. matrix, and factor levels.
  #
  # varsInModel is a vector of indices indicating which
  # variables are used in the model, so these should be
  # present in Xnew in the same format as in the training data.
  # If NULL, it is assumed that all variables should be checked.
  #
  #
  # Check object type: data frame vs. matrix
  if (is.data.frame(Xtrain)) {
    if (!is.data.frame(Xnew)) stop(paste0(
      "\nThe test data should be a data frame, ",
      " like the training data."))
  }
  if (is.matrix(Xtrain)) {
    if (!is.matrix(Xnew)) stop(paste0(
      "\nThe test data should be a matrix, ",
      " like the training data."))
  }
  #
  # Check number of variables:
  if (is.null(varsInModel) | identical(varsInModel,
                                       seq_len(ncol(Xtrain)))) {
    varsInModel <- seq_len(ncol(Xtrain))
    if (ncol(Xtrain) != ncol(Xnew)) stop(paste0(
      "\nThe test data should have ", ncol(Xtrain), " variables, ",
      " like the training data."))
  }
  #
  # Only compare the variables actually used in the model:
  Xtrain <- Xtrain[, varsInModel]
  #
  # Check variable names:
  #
  namesmatch <- match(colnames(Xtrain), colnames(Xnew))
  # namesmatch
  if (anyNA(namesmatch)) {
    message("\nThe following variable name(s) are missing in Xnew:")
    print(colnames(Xtrain)[is.na(namesmatch)])
    stop(paste0("\nThe variable names of the test data should",
                "\nmatch those of the training data."))
  } else {
    Xnew <- Xnew[, namesmatch] # puts the columns of Xnew in the
    # same order, and only keeps those columns we actually need.
  }
  # From here on the variable names match, and are in the same order.
  #
  if (!allowNA) {
    if (anyNA(Xnew)) stop(paste0(
      "\nXnew contains at least one NA in a",
      " variable used in the model."))
  }
  #
  # Check variable types:
  #
  if (is.data.frame(Xtrain)) {
    # if (!identical(sapply(Xtrain, data.class),
    #                sapply(Xnew, data.class) )) { }
    differ <- which(sapply(Xtrain, data.class) !=
                      sapply(Xnew, data.class))
    if (length(differ) > 0) {
      message("\nThe type of the following variable(s) in Xnew:")
      print(colnames(Xnew)[differ])
      message("differs from their type in the training data.")
      stop(paste0("\nThe variable types of the test data should",
                  "\nmatch those of the training data."))
    }
  }
  if (is.matrix(Xtrain)) {
    if (mode(Xnew) != mode(Xtrain)) stop(paste0(
      "\nThe mode of the test data matrix is ", mode(Xnew),
      ", \nbut that of the training data was ", mode(Xtrain), "."))
  }
  #
  # Check levels of factors and of character variables:
  #
  if (is.data.frame(Xtrain)) {
    varTypes <- sapply(Xtrain, data.class)
    #
    # First the factors:
    #
    facVars  <- which(varTypes == "factor")
    for (j in seq_len(length(facVars))) {
      charvec <- as.character(Xnew[, facVars[j]])
      # if the test factor is malformed, this throws an error.
      # It is not clear how to make the code report which
      # factor is responsible before the error is thrown.

      # check whether there are any new levels in the test factor
      levelsTrain <- attr(Xtrain[, facVars[j]], "levels")
      levelsTest  <- attr(Xnew[, facVars[j]], "levels")
      # check for new levels in the test variable:
      if (any(!(levelsTest %in% levelsTrain))) stop(
        paste0("\nThe factor \"", colnames(Xnew)[facVars[j]],
               "\" of the test data has at least one level",
               "\nthat does not appear in the training data."))
      # check for duplicate levels in the test variable:
      if (any(duplicated(levelsTest))) stop(
        paste0("\nThe factor \"", colnames(Xnew)[facVars[j]],
               "\" of the test data has duplicate levels."))
    }
    #
    # Now for the character variables. Note that this function
    # should be called with the variables actually used in the
    # modeling, and not e.g. "name" variables.
    #
    charVars <- which(varTypes == "character")
    for (j in seq_len(length(charVars))) {
      # Check whether there are any new values in the factor:
      valsTrain <- unique(Xtrain[, charVars[j]])
      valsTest  <- unique(Xnew[, charVars[j]])
      if (allowNA) {valsTrain <- c(valsTrain, NA)}
      #
      # check for new levels in the test variable
      if (any(!(valsTest %in% valsTrain))) stop(paste0(
        "\nThe character variable \"", colnames(Xnew)[charVars[j]],
        "\" of the test data has", "\nat least one value",
        " that does not appear in the training data."))
    }
  }
  #
  if (is.matrix(Xtrain)) {
    if ((mode(Xtrain) == "character")) {
      # a matrix of character variables
      for (j in seq_len(ncol(Xtrain))) {
        # Check whether there are any new values in the factor:
        valsTrain <- unique(Xtrain[, j])
        valsTest  <- unique(Xnew[, j])
        if (allowNA) {
          valsTrain <- c(valsTrain, NA)
        }
        # check for new levels in test variable
        if (any(!(valsTest %in% valsTrain))) stop(paste0(
          "\nThe character variable \"", colnames(Xnew)[j],
          "\" of the test data has", "\nat least one value",
          " that does not appear in the training data."))
      }
    }
  }
}

bixplot = function(...,
                   names = NA,
                   add = FALSE,
                   at = NULL,
                   horizontal = FALSE,
                   col = "gray60", bodyCol = NULL,
                   bodyOpaque = 0.5,
                   bodyW = 0.80,
                   bodysize = "area_from_count",
                   modeCol = c("cadetblue3", "hotpink2",
                               "lawngreen", "orange","cyan3",
                               "coral2","gray60", "darkgoldenrod2",
                               "slateblue2", "brown2", "khaki3"),
                   curveCol = "black",
                   curveLwd = 1,
                   border = "black", boxCol = NULL,
                   boxOpaque = 0.5,
                   boxW = 0.32,
                   boxLwd = 2,
                   makewhiskers = TRUE,
                   innerwhiskers = TRUE,
                   rugCol = "black",
                   rugoutCol = par("fg"), # "black",
                   rugNumeric = NULL,
                   rugNumericColors = c("blue", "green"),
                   colorbarW = 0.12,
                   rugFactor = NULL,
                   rugFactorColors = c("red", "blue", "forestgreen"),
                   rugLwd = par("lwd"),
                   rugoutLwd = 2*rugLwd,
                   rugOpaque = 0.4,
                   jittering = TRUE,
                   jitteramount = NULL,
                   rugW = 0.12,
                   stickCol = "black",
                   stickLwd = 2,
                   boxwex = 1,
                   width = NULL, # leftover from boxplot
                   side = "no",
                   tick = TRUE,
                   diplevel = 0.01,
                   minN = 15,
                   kmax = NULL,
                   bigN = 500,
                   clusMinN = 3,
                   bw = "SJ-dpi",
                   kernel = "gaussian",
                   cut = 3,
                   cutmin = -Inf,
                   cutmax = Inf,
                   xlim = NULL,
                   ylim = NULL,
                   cex.axis = 1,
                   main = "bixplot",
                   cex.main = 1, line.main = 1,
                   cex.colorbar = 1,
                   xlab = "",
                   ylab = "",
                   cex.lab = 1, line.lab = 2.2,
                   xaxs = "r", yaxs = "r", las = 1,
                   ann = !add,
                   plot = TRUE,
                   log = NULL)

{ # bixplot for data input and formula input
  #
  # Arguments
  #
  # ... : a vector of numeric values to plot,
  #       or a dataframe with variables to plot, or a
  #       matrix with columns to plot, or a single list of
  #       vectors that may be of different lengths. The
  #       data may contain NAs, which will be left out.
  #       One can also provide a formula, such as y ~ grp,
  #       where y is a numeric vector of data values to be
  #       split into groups according to the grouping
  #       variable grp (usually a factor). Note that
  #       ~ g1 + g2 is equivalent to ~ g1:g2. If a formula
  #       is given, it may be necessary to also specify
  #       an argument data = with the data.frame containing
  #       the variables in the formula.
  # names	: group labels that will be printed next to
  #       each variable (= group). [Argument from
  #       graphics::boxplot.] If NA, the default, these are
  #       extracted from the provided input data, and if
  #       the data does not contain names the labels
  #       1, 2, 3,... are used.
  # add : logical, if true then the bixplot is added to
  #       the current plot. Defaults to FALSE. [Argument
  #       from boxplot.]
  # at : numeric vector giving the locations where the
  #       bixplots should be drawn, particularly when
  #       add = TRUE. [Argument from boxplot.]
  #       Defaults to 1:p where p is the number of variables.
  # horizontal : logical indicating if the bixplots should
  #       be horizontal. [Argument from boxplot.]
  #       The default FALSE means vertical boxes.
  # bodyW : the width of the body, which is the region
  #       inside the estimated density.
  # col : contains the color(s) of the bodies for the p
  #       variables. col can be a single color or a
  #       vector, and will be recycled as needed. [Argument
  #       from boxplot]. It is only applied to those
  #       variables the method deems unimodal. When col is
  #       NA, the body is not drawn.
  # bodyCol : if not NULL, it overrides col.
  # bodyOpaque: opacity of the body.
  # bodyW : the width of the body. When zero, the body
  #       is not drawn. When two or more modes are shown
  #       for the same variable, it is the width of the
  #       widest body.
  # bodysize: either "area_is_constant", "area_from_count",
  #       or "width_is_constant".
  #       Determines the relative sizes of bodies of modes
  #       in the same variable.
  #       When "area_is_constant", the area of each density
  #       is the same.
  #       When "area_from_count", the default, the area of
  #       each density is proportional to the number of
  #       members in its cluster (mode).
  #       When "width_is_constant", all bodies have the same
  #       width.
  #       In all three settings, the width of the widest
  #       body is bodyW[j], possibly modified by argument
  #       width if it is not null.
  #       The width of the box changes proportionally
  #       with that of the body.
  # modeCol : contains the color(s) of the bodies of the
  #       modes (clusters) of the variables deemed not
  #       unimodal. Can be a single color or a vector,
  #       and will be recycled as needed. If NULL, all
  #       modes of a variable will receive the same color
  #       as provided in col for that variable.
  # curveCol : contains the color(s) of the density
  #       curves of each variable, that is, the boundary
  #       of the body. Can be a single color or a vector,
  #       and will be recycled as needed. If NA, no
  #       curve boundary is drawn. Defaults to "black".
  # curveLwd : a single number gving the line width ("lwd")
  #       of the density curves.
  # border : contains the color(s) of the boxes and
  #       whiskers [argument from boxplot].
  #       It can be a single color or a vector, and will
  #       be recycled as needed. A variable deemed
  #       unimodal will have a single box, otherwise
  #       each mode has its own box. When border is NA,
  #       the box is not drawn.
  # boxCol : if not NULL, it overrides border.
  # boxOpaque : opacity of the box.
  # boxW : the width of the box. When zero, the box is
  #       not drawn.
  # boxLwd : a single number giving the line width ("lwd")
  #       of the box and whiskers.
  # makewhiskers : logical specifying whether whiskers
  #       should be drawn.
  # innerwhiskers : logical. If TRUE and there is more than
  #       one mode, also the whiskers _between_ the modes
  #       are drawn. If FALSE, the default, no whiskers
  #       between modes are not drawn, because they do not
  #       have their usual meaning due to possible
  #       overlapping of densities.
  # rugCol : contains the color(s) of the rug plot of
  #       the variables. The rug consists of small lines
  #       perpendicular to the variable axis, one for
  #       each data value. It can be a single color or a
  #       vector, and will be recycled as needed. If NA,
  #       the rug is not plotted.
  # rugoutCol : if not NULL or NA, and the body of a
  #       variable or its modes are drawn, this specifies
  #       the color(s) of the part of the ruglines outside
  #       of the body. If NULL or NA, the same color rugCol
  #       is used as inside the body.
  # rugNumeric : external numeric variable for coloring the rug.
  # rugNumericColors : two or three colors to which
  #       colorRampPalette can be applied, to obtain a smooth
  #       palette of colors for rugNumeric.
  # colorbarW : the width of the color bar with values of
  #       rugNumeric, relative to the width of the main plot.
  #       Defaults to 0.1 . Only has an effect if rugNumeric
  #       is specified.
  # rugFactor : external factor variable for coloring the rug.
  # rugFactorColors : colors of the factor levels of
  #       rugFactor.
  # rugLwd : a single number specifying the width of the
  #       lines in the rug plot. Defaults to par("lwd").
  # rugoutLwd : a single number specifying the width of
  #       the part of the ruglines outside of the body.
  #       Defaults to 2*rugLwd to make isolated points
  #       stand out visually.
  # rugOpaque: opacity of the rug.
  # jittering : logical indicating whether the values at
  #       which the rug lines are drawn are to be jittered
  #       by base::jitter, to make tied values more visible.
  #       Defaults to TRUE.
  # jitteramount : the amount of jittering (see ?jitter).
  # rugW : the width of the rug. When zero, the rug is
  #       not drawn.
  # stickCol : when side = "both" this contains the color(s)
  #       of the `stick', which is a typically thin line
  #       separating the bodies of two variables plotted
  #       side by side. It goes from the smallest to the
  #       largest value of these variables. When the
  #       density is computed, it contains the range of
  #       the density as well. stickCol can be
  #       a single color or a vector, and will be recycled
  #       as needed. If NA, the default, the stick is not
  #       plotted.
  # stickLwd : a single number giving the line width ("lwd")
  #       of the stick. When zero, the stick is not drawn.
  # boxwex : a scale factor to be applied to the width
  #       of the body, box, and rug of all bixplots
  #       [Argument from boxplot].
  # width : a vector giving the relative widths of the p
  #       bixplots, one for each variable. [Argument from
  #       boxplot]. The entries should be strictly
  #       positive numbers, that will be divided by the
  #       largest of them. The resulting ratios multiply
  #       the widths of the body, box, and rug.
  # side : the side of the variable axis where the body,
  #       box and rug will be plot. [Argument from
  #       beanplot::beanplot]. The default is "no",
  #       meaning that each bixplot is symmetric about
  #       its axis. Options "first" and "second" plot
  #       all "half" bixplots on that given side. Option
  #       "both" plots variables on alternate sides of
  #       each axis, so fewer variable axes are needed.
  # tick : boolean indicating whether to draw tick marks
  #       on the plot axis with the variable (group)
  #       names. Only has an effect when add = FALSE.
  # diplevel : the level of Hartigan's dip test for
  #       unimodality of a variable. Only when this test
  #       rejects unimodality, that is, its p-value is
  #       at most diplevel, do we search for clusters.
  #       Defaults to 0.01. Setting it higher may lead to
  #       more clusters, and setting it lower to fewer
  #       clusters.
  # minN : the minimum number of observations required per
  #       potential cluster. Defaults to 15. The maximum
  #       number of clusters searched is limited to n/minN,
  #       where n is the number of non-missing values.
  #       If n < 2*minN, clustering is not attempted to
  #       avoid spurious clusters from small samples.
  # kmax : the maximal number of clusters. It will be
  #       truncated to 5 internally. If NULL, it is set
  #       to min(floor(n/minN), 5). When setting kmax=1
  #       all variables are considered as single clusters,
  #       making the display resemble a violin plot.
  # bigN : when a variable has over bigN non-NA values,
  #       we sample bigN values from it without
  #       replacement, to save computation time.
  #       Defaults to 500. If a number below 300 is
  #       given, it is set to 300 internally.
  # clusMinN : the minimum number of unique values in a
  #       cluster. The clustering is constrained so that
  #       any cluster must contain at least clusMinN
  #       unique values. Defaults to 3.
  # bw : the bandwidth for use by stats::density which
  #       constructs the body of a variable (group). It
  #       can be a number or a formula, and defaults to
  #       "SJ-dpi" (see ?density).
  # kernel : the kernel for use by stats::density.
  #       Defaults to "gaussian".
  # cut : denoting a variable by y, the density is only
  #       constructed from min(y) - cut*bw to
  #       max(y) + cut*bw where bw is the numeric
  #       bandwidth. Defaults to 3. [Argument from
  #       beanplot.]
  # cutmin : if specified, the densities of all variables
  #       and modes cannot go below the value cutmin.
  #       Defaults to -Inf which has no effect. [Argument
  #       from beanplot.]
  # cutmax : if specified, the densities of all variables
  #       and modes cannot go above the value cutmax.
  #       Defaults to Inf which has no effect. [Argument
  #       from beanplot.]
  #
  # Graphical arguments, only when add = FALSE:
  # xlim : vector with limits of the axis with the groups
  #       (variables), whether horizontal is TRUE or FALSE.
  #       This is a convention from graphics::boxplot.
  # ylim : vector with limits of the axis with the numeric
  #       values taken, whether horizontal is TRUE or FALSE.
  #       This is a convention from graphics::boxplot.
  # cex.axis, main, cex.main, line.main, xlab, ylab,
  # cex.lab, line.lab, xaxs, yaxs, las : work as usual.
  # ann : logical indicating if the plot should be
  #       annotated by xlab, ylab, and main title.
  #       [Argument from boxplot]. Defaults to !add.
  # plot : if TRUE (the default) then the numerical
  #       summary on which the plot is based is
  #       returned invisibly. If FALSE it is also
  #       output visibly on the terminal. [Argument
  #       from boxplot.]
  # log : this argument from boxplot has to be NULL,
  #       because a log transform can change the
  #       clusters (modes). If not NULL, a warning is
  #       given but the argument has no further effect.
  #       If you wish to use a logarithm it needs to be
  #       put explicitly in the formula, or you can
  #       transform the data first.
  #
  # There is no need to declare unused arguments from
  # boxplot, vioplot, beanplot for backward compatibility
  # (and then to ignore them): they are absorbed by the ...
  #
  # Value
  #
  # A list with the following components:
  # call : the call to bixplot
  # p    : the number of variables
  # <variable1>, <variable2>,... : for each variable, this
  #        is a list.
  #        When no clusters were found in that variable,
  #        it contains the non-NA values of that variable
  #        with their names if any, and their five-number
  #        summary from stats::fivenum that was used to
  #        draw the box.
  #        When clusters were found in that variable, it
  #        contains the clustering (a vector of integers),
  #        and for each cluster its values and five-number
  #        summary.

  wnq = function (string, qwrite = TRUE) {
    if (qwrite)
      write(noquote(string), file = "", ncolumns = 100)
  }

  checkFactor = function(y){
    # Auxiliary function for use of rugFactor.
    # Argument:
    # y : the factor variable
    #
    # Results:
    # ints   : the integer version of the factor variable
    # levels : the levels of the factor variable
    # indsv  : the indices of the non-NA entries of the factor variable
    #
    if (!(is.factor(y))) {
      stop(paste0("\n The rugFactor variable is not a factor.",
                  "\n Please use as.factor() or factor() first."))
    }
    if (is.null(attr(y, "levels"))) {
      stop(paste0("\n The rugFactor variable has no levels.",
                  "\n Please use as.factor() or factor() first."))
    }
    levelsattr <- attr(y, "levels")
    if (sum(is.na(levelsattr)) > 0) {
      stop(paste0("\n The levels attribute of the rugFactor variable",
                  "\n has at least one NA."))
    }
    dupls <- duplicated(levelsattr)
    if (sum(dupls) > 0)
      stop(paste0("\n The label ", levelsattr[dupls == TRUE],
                  " occurs more than once in",
                  "\n the levels attribute of the rugFactor variable."))
    yv <- as.character(y)
    indsv <- which(!is.na(yv))
    if (length(indsv) == 0) {
      stop(paste0("The rugFactor variable only contains NA's."))
    }
    yv <- yv[indsv]
    uniqy <- sort(unique(yv))
    for (g in seq_len(length(uniqy))) {
      if (!(uniqy[g] %in% levelsattr)) {
        stop(paste0("\n The label ", uniqy[g], " does not appear among",
                    "\n the levels attribute of the rugFactor variable."))
      }
    }
    if (length(uniqy) < 2) {
      stop(paste0("\n The rugFactor variable has only a single level ( ",
                  uniqy[1], " )", "\n so we cannot color by level."))
    }
    for (g in seq_len(length(levelsattr))) {
      if (!(levelsattr[g] %in% uniqy)) {
        wnq(paste0("\nThe level ", levelsattr[g], " does not occur",
                   " in the rugFactor variable.\n"))
      }
    }
    levels <- levelsattr[which(levelsattr %in% uniqy)]
    xlevelz <- levels
    lab2int <- function(labs, xlevelz) {
      ints <- rep(NA, length(labs))
      indv <- which(!is.na(labs))
      labv <- labs[indv]
      for (g in seq_len(length(xlevelz))) {
        clinds <- indv[which(labv == xlevelz[g])]
        ints[clinds] <- g
      }
      ints
    }
    ints = lab2int(y, xlevelz)
    return(list(ints = ints, levels = levels, indsv = indsv))
  }

  drawpoints = function(y, coord=1, horizontal = FALSE,
                        bokw = 0.36, side = "no",
                        cex = 1, pointcol = par()$fg){
    if(!(side %in% c("no","first","second"))) stop("invalid side")
    if(side == "first")  coord = coord - 0.04*bokw
    if(side == "second") coord = coord + 0.04*bokw
    coords = rep(coord,length(y))
    if(horizontal==FALSE){
      points(coords, y, col=pointcol, pch=16, cex=cex)
    } else {
      points(y, coords, col=pointcol, pch=16, cex=cex)
    }
  }

  drawbox = function(y, coord = 1, horizontal = FALSE,
                     bokw = 0.36, lwd = 1,
                     bcol = "blue", side = "no",
                     whiskers = "none"){
    # Auxiliary function to draw box at (Q1, median, Q3)
    # y       : a data vector
    # coord   : coordinate of the box
    # bokw    : box width if side = "no"
    # side    : side on which to draw the box
    # whisker : whether to draw whiskers
    #
    if(!(side %in% c("no","first","second"))) stop("invalid side")
    if(!(whiskers %in% c("both","up","down","none")))
      stop("invalid whiskers")
    bstat = grDevices::boxplot.stats(y, coef = 1.53)$stats
    nuniq = length(unique(y))
    miny = min(y, na.rm = TRUE); maxy = max(y, na.rm = TRUE)
    if(nuniq == 3) bstat = c(miny, miny, bstat[3], maxy, maxy)
    spine = coord
    Left  = coord - 0.50*bokw
    Right = coord + 0.50*bokw
    if(side == "second") {
      Left  = coord + 0.06*bokw
      spine = coord + 0.28*bokw
    }
    if(side == "first") {
      Right = coord - 0.06*bokw
      spine = coord - 0.28*bokw
    }
    Low = bstat[2]; Mid = bstat[3]; Hi = bstat[4]
    if(horizontal == FALSE){
      if(nuniq > 1) lines(c(Left,Left,Right,Right,Left),
                          c(Low,Hi,Hi,Low,Low),
                          lwd = lwd, col = bcol)
      if(nuniq > 2) lines(c(Left,Right), c(Mid,Mid),
                          lwd = lwd, col = bcol)
      if(length(y) > 3){
        if(whiskers %in% c("both","up")){
          lines(c(spine,spine),c(Hi,bstat[5]),
                lwd = lwd, col = bcol) }
        if(whiskers %in% c("both","down")){
          lines(c(spine,spine),c(Low,bstat[1]),
                lwd = lwd, col = bcol) }
      }
    } else {
      if(nuniq > 1) lines(c(Low,Hi,Hi,Low,Low),
                          c(Left,Left,Right,Right,Left),
                          lwd = lwd, col = bcol)
      if(nuniq > 2) lines(c(Mid,Mid), c(Left,Right),
                          lwd = lwd, col = bcol)
      if(length(y) > 3){
        if(whiskers %in% c("both","up")){
          lines(c(Hi,bstat[5]), c(spine,spine),
                lwd = lwd, col = bcol) }
        if(whiskers %in% c("both","down")){
          lines(c(Low,bstat[1]), c(spine,spine),
                lwd = lwd, col = bcol) }
      }
    }
  }

  compdens <- function(x, bwnum, kernel, cut, cutmin, cutmax) {
    # Auxiliary function for drawdensity
    if (length(x) > 0) {
      from <- max(cutmin, (min(x, na.rm = TRUE) - cut * bwnum))
      to <- min(cutmax, max(x, na.rm = TRUE) + cut * bwnum)
      density(x, bw = bwnum, kernel = kernel, from = from,
              to = to)[c("x", "y")]
    }
    else list(x = numeric(), y = numeric())
    # if no non-NA data, this empty density comes out.
    # But we only run this function for n > 0.
  }

  drawdensity = function(side, dens, at, bodcol,
                         curcol, curlwd = 1, horizontal){
    # if curcol is NA we don't draw the curve.
    if (side == "first") {
      x1 <- NULL
      y1 <- NULL
    } else {
      x1 <- dens$y + at
      y1 <- dens$x
    }
    if (side == "second") {
      x2 <- NULL
      y2 <- NULL
    } else {
      x2 <- rev(dens$y) * -1 + at
      y2 <- rev(dens$x)
    }
    if (length(x1) > 0) {
      x1 <- c(at, x1, at)
      y1 <- c(y1[1], y1, y1[length(y1)])
    }
    if (length(x2) > 0) {
      x2 <- c(at, x2, at)
      y2 <- c(y2[1], y2, y2[length(y2)])
    }
    if (horizontal) {
      polygon(c(y1, y2), c(x1, x2), col = bodcol, border = NA)
      if (!is.na(curcol))
        lines(c(y1, y2), c(x1, x2), col = curcol, lwd = curlwd)
    } else {
      polygon(c(x1, x2), c(y1, y2), col = bodcol, border = NA)
      if (!is.na(curcol))
        lines(c(x1, x2), c(y1, y2), col = curcol, lwd = curlwd)
    }
  }

  drawrug2sided = function(y, side, jittering, jitteramount, dens,
                           at, rugw, colin, colout, rugLwd,
                           rugoutLwd, horizontal)
  { # only for side == "no"
    if(side != "no") stop('side must be \"no\" here')
    if (jittering == TRUE) {
      rugy <- list(jitter(y, amount = jitteramount),
                   rep(1, length(y)))
    } else {
      rugy <- list(y, rep(1, length(y)))
      # this is poor when there are many ties
    }
    rugy[[2]] <- rugy[[2]] * rugw/2 # 1-sided width of ruglines
    if (!isTRUE(all.equal(colin,colout))) {
      crossings <- approx(dens$x, dens$y, rugy[[1]])$y
      crossings <- pmin(crossings, rugy[[2]], na.rm = TRUE)
      extra <- unlist(rugy[2]) - crossings
    }
    else {
      crossings <- rugy[[2]]
      extra <- 0
    }
    ## First we plot the part of the rug that sticks out of the body:
    if (max(extra) > 0) { # if at least one segment exceeds density
      w <- (extra > 0) # which segments exceed density
      if (horizontal) {
        if (side %in% c("no", "first"))
          segments(rugy[[1]][w], -extra[w] - crossings[w] + at,
                   rugy[[1]][w], -crossings[w] + at,
                   col = colout, lwd = rugoutLwd)
        if (side %in% c("no", "second"))
          segments(rugy[[1]][w], crossings[w] + at,
                   rugy[[1]][w], extra[w] + crossings[w] + at,
                   col = colout, lwd = rugoutLwd)
      }
      else {
        if (side %in% c("no", "first"))
          segments(-extra[w] - crossings[w] + at, rugy[[1]][w],
                   -crossings[w] + at, rugy[[1]][w],
                   col = colout, lwd = rugoutLwd)
        if (side %in% c("no", "second"))
          segments(crossings[w] + at, rugy[[1]][w],
                   extra[w] + crossings[w] + at, rugy[[1]][w],
                   col = colout, lwd = rugoutLwd)
      }
    }
    ## Now plot the central part, inside the body:
    x1 = -crossings + at
    x2 =  crossings + at
    if (horizontal)
      segments(rugy[[1]], x1, rugy[[1]], x2, col = colin)
    else segments(x1, rugy[[1]], x2, rugy[[1]], col = colin)
  }

  drawrug1sided = function(y, side, jittering, jitteramount, at,
                           rugw, rugcenter, colin, rugLwd, horizontal)
  { # only use this function when side != "no", so here it
    # has to be either "first" or "second".
    if(!(side %in% c("first", "second"))) stop(
      'side must be \"first\" or \"second\" here')
    if (jittering == TRUE) {
      rugy <- jitter(y, amount = jitteramount)
    } else {
      rugy <- y # this is poor when there are many ties
    }
    spine = rugcenter
    if(side == "first") spine = -rugcenter
    x1 <- at + spine - rugw/2
    x2 <- at + spine + rugw/2
    if (horizontal)
      segments(rugy, x1, rugy, x2, col = colin)
    else segments(x1, rugy, x2, rugy, col = colin)
  }

  # Here main function starts
  if(!(side %in% c("no", "first", "second", "both")))
    stop("invalid side")
  if(!(bodysize %in% c("area_is_constant",
                       "area_from_count",
                       "width_is_constant"))) stop("invalid bodysize")
  if(!is.null(rugNumeric) & !is.null(rugFactor)) stop(
    "You cannot specify rugNumeric and rugFactor simultaneously.")
  if(!is.null(log)) warning(paste0(
    "The argument 'log' has no effect here. If you wish\n",
    "to use a logarithm it needs to be put explicitly in\n",
    "the formula, or you can transform the data first.\n",
    "Note that a log transform can change the modes."))
  #
  # Getting the y variables to plot
  args <- match.call()
  cat("\n")
  print(args)
  mcall <- as.list(args)
  noname = rep(FALSE, length(args))
  for(ii in seq_len(length(args))){
    noname[ii] = (is.null(base::names(args[ii])))
  }
  whichnoname = which(noname == TRUE)
  isadatafr = isamatrix = FALSE
  if(length(whichnoname) == 2){
    tempmat = eval(mcall[[whichnoname[2]]], envir = parent.frame())
    isamatrix = is.matrix(tempmat)
    isadatafr = is.data.frame(tempmat)
  }
  if(isamatrix == TRUE || isadatafr == TRUE){
    p = NCOL(tempmat)
    nams = colnames(tempmat)
    if(is.null(nams) || missing(nams)) nams = paste0("X",seq_len(p))
    n = NROW(tempmat); n
    rownams = rownames(tempmat)
    if(is.null(rownams) || missing(rownams)) rownams = seq_len(n)
    ylist = list()
    for(j in seq_len(p)){
      tempvec = tempmat[,j]
      names(tempvec) = rownams
      ylist[[nams[j]]] = tempvec
    }
  } else {
    envir = parent.frame(2)
    if (is.null(base::names(args)))
      vars <- 1:length(args)
    else vars <- c(1, which(base::names(args)[2:length(args)] %in%
                              c("formula", "x", "data", "")) + 1)
    if (length(vars) < 2)
      return(list())
    options <- which(base::names(args) %in%
                       c("subset", "na.action",
                         "drop.unused.levels", "xlev"))
    args <- args[c(vars, options)]
    args[[1]] <- quote(model.frame)
    hashad <- rep(FALSE, length(vars))
    ylist <- list()
    notnamed <- 0
    subsetno <- numeric()
    naactno <- numeric()
    dulno <- numeric()
    xlevno <- numeric()
    datano <- numeric()
    argsvals <- lapply(as.list(args[2:length(vars)]), eval, envir)
    islists <- lapply(argsvals, function(x) {
      is.list(x) || is.null(x)
    })
    for (i in 2:length(vars)) {
      if (hashad[i])
        next
      x <- argsvals[[i - 1]]
      if (inherits(x, "formula")) {
        datanonext <- match(TRUE, islists[
          max(datano, i):length(islists)]) +
          max(datano, i)
        if (!is.na(datanonext)) {
          hashad[datanonext] <- TRUE
          datano <- datanonext
        }
        nextargpos <- function(name, pos) {
          if (any(pos == length(args))) return(pos)
          posnext <- match(name, base::names(args[
            max(pos + 1, 3):length(args)])) +
            max(pos + 1, 3) - 1
          if (!is.na(posnext)) pos <- posnext
          pos
        }
        subsetno <- nextargpos("subset", subsetno)
        naactno <- nextargpos("na.action", naactno)
        dulno <- nextargpos("drop.unused.levels", dulno)
        xlevno <- nextargpos("xlev", xlevno)
        attr(args, "names")[i] <- "formula"
        m <- args[c(1, i, datano, subsetno, naactno, dulno,
                    xlevno)]
        mf <- eval(m, envir)
        response <- attr(attr(mf, "terms"), "response")
        ylist <- c(ylist, split(mf[[response]], mf[-response]))
      }
      else if (is.list(x)) {
        ylist <- c(ylist, x)
      }
      else {
        x <- list(x)
        notnamed <- notnamed + 1
        attr(x, "names") <- notnamed
        ylist <- c(ylist, x)
      }
    }
  }
  p <- length(ylist)
  if(p == 0) stop("no data found to plot")
  constantN = TRUE
  nn = length(ylist[[1]])
  for(j in seq_len(p)){
    if(length(ylist[[j]]) != nn) constantN = FALSE
  }
  if(is.null(col)) col = NA
  if(is.null(bodyCol)) bodyCol = col
  colormodes = TRUE
  if(is.null(modeCol)) colormodes = FALSE
  if(is.null(curveCol)) curveCol = NA # no curves
  if(is.null(boxCol)) boxCol = border
  if(is.null(rugCol)) rugCol = par("fg")
  if(is.null(rugoutCol)) rugoutCol = rugCol
  if(is.na(rugoutCol)) rugoutCol = rugCol
  #
  # adjusting transparency of colors:
  indx = which(!is.na(bodyCol))
  if(length(indx) >= 1) bodyCol[indx] =
    adjustcolor(bodyCol[indx], alpha.f = bodyOpaque)
  if(colormodes == TRUE) {
    indx = which(!is.na(modeCol))
    if(length(indx) >= 1) modeCol[indx] =
        adjustcolor(modeCol[indx], alpha.f = bodyOpaque)
  }
  indx = which(!is.na(boxCol))
  if(length(indx) >= 1) boxCol[indx] =
    adjustcolor(boxCol[indx], alpha.f = boxOpaque)
  indx = which(!is.na(rugCol))
  if(length(indx) >= 1) rugCol[indx] =
    adjustcolor(rugCol[indx], alpha.f = rugOpaque)
  indx = which(!is.na(rugoutCol))
  if(length(indx) >= 1) rugoutCol[indx] =
    adjustcolor(rugoutCol[indx], alpha.f = 1)
  #
  # Making vectors longer if needed:
  cycle2size = function(x, size){
    if(length(x) < size) x = rep(x, ceiling(size/length(x)))
    return(x)
  }
  bodyCol   = cycle2size(bodyCol,   p)
  bodyW     = cycle2size(bodyW,     p)
  modeCol   = cycle2size(modeCol,   5*p) # assumes mykmax <= 5
  boxCol    = cycle2size(boxCol,    p)
  boxW      = cycle2size(boxW,      p)
  curveCol  = cycle2size(curveCol,  p)
  rugCol    = cycle2size(rugCol,    p)
  rugoutCol = cycle2size(rugoutCol, p)
  rugW      = cycle2size(rugW,      p)
  stickCol  = cycle2size(stickCol,  p)
  #
  if(!is.null(width)){
    if(length(width) != p) stop(paste0(
      "Argument width should be a vector of length" ,p))
    if(min(width) <= 0) stop(
      "Argument width should only contain positive entries")
    bodyW = bodyW * width/(max(width))
    boxW  = boxW * width/(max(width))
    rugW  = rugW * width/(max(width))
  }
  bodyW = boxwex * abs(bodyW)
  boxW  = boxwex * abs(boxW)
  rugW  = boxwex * abs(rugW)
  #
  if (missing(names)) {
    if (is.null(base::names(ylist)))
      attr(ylist, "names") = 1:displayp
    names <- base::names(ylist)
  } else { attr(ylist, "names") <- names }

  if(!(side == "both")){
    displayp = p
    xpos = seq_len(p)
    mysides = rep(side, p)
    if(!is.null(at)){
      lat = length(at)
      if(lat != p)
        stop(paste0("argument at has length ",lat,
                    " which should be ", p))
      xpos = at
    }
    xcoords = xpos
  }
  if(side == "both"){
    # if plot on both sides: "half" as many variable axes
    displayp = ceiling(p/2)
    xpos = seq_len(displayp)
    if(!is.null(at)){
      lat = length(at)
      if(lat != displayp)
        stop(paste0("argument at has length ", lat,
                    " which should be ", displayp))
      xpos = at
    }
    if ((side == "both") && (length(names) > displayp)) {
      combinenames <- function(string1, string2)
      { if (is.na(string2)) return(string1)
        if (string1 == string2) return(string1)
        s1 <- unlist(strsplit(string1, NULL))
        s2 <- unlist(strsplit(string2, NULL))
        fromleft <- 0
        gfromleft <- 0
        while ((fromleft < length(s1)) && (fromleft < length(s2)) &&
               (s1[fromleft + 1] == s2[fromleft + 1])) {
          if (any(grep("[. \t]", s1[fromleft + 1])))
            gfromleft <- fromleft
          fromleft <- fromleft + 1
        }
        fromright <- 0
        gfromright <- 0
        while ((fromright < length(s1)) && (fromright < length(s2)) &&
               (s1[length(s1) - fromright] == s2[length(s2) - fromright])) {
          if (any(grep("[. \t]", s1[length(s1) - fromright])))
            gfromright <- fromright
          fromright <- fromright + 1
        }
        if (gfromleft > gfromright)
          result <- substr(string1, 1, gfromleft)
        else if (gfromright == 0)
          result <- paste(string1, string2, sep = "+")
        else result <- substr(string1, nchar(string1) - gfromright +
                                1, nchar(string1))
        result
      }
      for (i in 1:displayp) {
        names[i] <- combinenames(names[i*2 - 1], names[i*2])
      }
      length(names) <- displayp
    }
    xcoords = c(matrix(rep(xpos, 2), nrow=2, byrow = TRUE))
    mysides = rep(c("first","second"), displayp)
  }
  ## From here on, each mysides[j] is either "no", "first", or "second".
  ## It cannot be "both" anymore.
  #
  rugfact = FALSE
  if(!is.null(rugFactor)){
    rugfact = TRUE
    if(constantN == FALSE) {
      warning(
        "We cannot color the rug by rugFactor because not all",
        "\n  y variables have the same length.")
      rugfact = FALSE
    } else {
      if(length(rugFactor) != nn) {
        warning(paste0(
          "We cannot color the rug by rugFactor,",
          "\n  because rugFactor has length ", length(rugFactor),
          "\n  and the y variables have length ", nn, ".\n"))
        rugfact = FALSE
      }
    }
    if(rugfact == TRUE){
      out = checkFactor(rugFactor)
      # If this errors, also bixplot stops.
      # So from here on we can assume it didn't.
      ints = out$ints
      mylevels = out$levels
      indsv = out$indsv
      nlevels = length(mylevels)
      ncolors = length(rugFactorColors)
      if(nlevels > ncolors) {
        rugFactorColors = cycle2size(rugFactorColors, nlevels) }
    }
    cat("The variable rugFactor has the levels:\n")
    print(mylevels)
  }
  #
  rugnumer = FALSE
  if(add == FALSE){
    if(!is.null(rugNumeric)) {
      rugnumer = TRUE
      if(constantN == FALSE) {
        warning(
          "We cannot color the rug by rugNumeric because not all",
          "\n  y variables have the same length.")
        rugnumer = FALSE
      } else {
        if(length(rugNumeric) != nn) {
          warning(paste0(
            "We cannot color the rug by rugNumeric,",
            "\n  because rugNumeric has length ", length(rugNumeric),
            "\n  and the y variables have length ", nn, ".\n"))
          rugnumer = FALSE
        }
      }
    }
  }
  #
  # Determine limits of x and y
  if(is.null(xlim)){
    xlim = c(min(xcoords) - 0.5, max(xcoords) + 0.5)
  }
  ymin = Inf; ymax = -Inf
  rugNum = rugFac = list()
  if(bigN < 300) bigN = 300
  for(j in seq_len(p)){
    y = ylist[[j]]
    if(length(y) >= 1){
      if(!(is.numeric(y))) stop(paste0("variable ",j,
                                       " is not numeric"))
      if(rugnumer == TRUE) rugNum[[j]] = rugNumeric
      if(rugfact  == TRUE) rugFac[[j]] = ints
      if(length(y) > bigN){
        set.seed(1)
        inds = sample(seq_len(length(y)), bigN, replace = FALSE)
        # In order to display the worst outliers:
        inds[1] = which.min(y)
        inds[2] = which.max(y)
        y = y[inds]
        if(rugnumer == TRUE) rugNum[[j]] = rugNum[[j]][inds]
        if(rugfact  == TRUE) rugFac[[j]] = rugFac[[j]][inds]
      }
      noNA = which(!is.na(y))
      y = y[noNA]
      if(rugnumer == TRUE) rugNum[[j]] = rugNum[[j]][noNA]
      if(rugfact  == TRUE) rugFac[[j]] = rugFac[[j]][noNA]
      ord = order(y)
      y = y[ord] # This sort is necessary. The names are permuted too.
      if(rugnumer == TRUE) rugNum[[j]] = rugNum[[j]][ord]
      if(rugfact  == TRUE) rugFac[[j]] = rugFac[[j]][ord]
      ylist[[j]] = y
      indy  = which(!is.na(y))
      tempy = y[indy]
      indy  = which(is.finite(tempy))
      tempy = tempy[indy]
      bwnum = 0
      if(length(tempy) > 1 && !(all(is.na(modeCol)) && all(is.na(bodyCol)))){
        bwnum = if(!is.numeric(bw)) density(unlist(tempy), bw = bw)$bw else bw
      }
      ymin  = min(ymin, min(tempy) - cut*bwnum)
      ymax  = max(ymax, max(tempy) + cut*bwnum)
    }
  }
  if(is.null(ylim)) { ylim = c(ymin, ymax) }
  if(add == FALSE){
    if(rugnumer == TRUE){ # make room for color bar
      layout(matrix(c(1, 2), ncol = 2), widths = c(1, colorbarW))
      maruser = par("mar")
      mymar = maruser
      mymar[4] = 1
      par(mar = mymar)
      colorpalette <- colorRampPalette(rugNumericColors)(100)
      ran = range(rugNumeric, na.rm = TRUE)
    }
    if(horizontal == FALSE){
      plot(NA, NA, type = "n",
           xaxt = "n", xlim = xlim, ylim = ylim,
           xaxs = xaxs, yaxs = yaxs,
           xlab = "", ylab = "", cex.axis = cex.axis)
      axis(side = 1, at = xpos, labels = names, las = las,
           cex.axis = cex.axis, tick = tick)
    } else {
      plot(NA, type = "n",
           yaxt = "n", xlim = ylim, ylim = xlim,
           xaxs = xaxs, yaxs = yaxs,
           xlab = "", ylab = "", cex.axis = cex.axis)
      axis(side = 2, at = xpos, labels = names, las = las,
           cex.axis = cex.axis, tick = tick, ylim = xlim)
    }
    if(ann == TRUE){
      title(main = main, line = line.main, cex.main = cex.main)
      title(xlab = xlab, line = line.lab, cex.lab = cex.lab)
      title(ylab = ylab, line = line.lab, cex.lab = cex.lab)
    }
  }
  outp = list()
  outp[["call"]] = match.call()
  outp[["p"]] = p
  #
  jj = 0 # counter for modeCol
  for(j in seq_len(p)){
    listj = list() # to add to outp list
    y = ylist[[j]]
    n = length(y)
    cat(paste0("\nVariable ", j, " has length = ", n, ":\n"))
    if(n < 1) { cat("we are skipping it.\n")
    } else { # draw this plot
      nuniq = length(unique(y))
      if((side == "both") && (j%%2 == 0)){
        ymintemp = min(c(y, ymintemp))
        ymaxtemp = max(c(y, ymaxtemp))
      } else{
        ymintemp = min(y)
        ymaxtemp = max(y)
      }
      # If bandwidth bw is a formula, turn it into a number:
      bwnum = bw
      if(n == 1) { bwnum = 0.5
      } else {
        if (!is.numeric(bw)) {
          bwnum = density(y, kernel = kernel, bw = bw)$bw
        }
        if(is.nan(bwnum)) bwnum <- 0.5
      }
      mykmax = floor(n/minN)
      if(!is.null(kmax)) mykmax = min(kmax, mykmax)
      mykmax = min(mykmax, 5)
      reject0 = FALSE
      if(mykmax > 1){ # else we don't need the dip test
        # Is there more than 1 mode:
        pval = round(diptest::dip.test(y)$p.value,4)
        cat(paste0("pvalue(dip.test) = ",pval,"\n"))
        if(pval <= diplevel) reject0 = TRUE
      }
      kopt = 1
      if(reject0 == TRUE){
        mykmax = max(1, min(mykmax, floor(nuniq/clusMinN)))
        cat(paste0("mykmax = ",mykmax,"\n"))
        if(mykmax >= 2){
          silwidths = rep(-2, mykmax)
          cluslist = list()
          cluslist[[1]] = NA
          for(kk in 2:mykmax){
            out = pamc1d(y, kk, clusMinN, maxit = 100,
                         verbose = FALSE)
            cluslist[[kk]] = out$clustering
            silh = cluster::silhouette(cluslist[[kk]], dist(y))
            silwidths[kk] = mean(silh[,3], na.rm = TRUE)
          }
          cat("Silhouette widths: ")
          print(round(silwidths[-1],3))
          if(max(silwidths) > 0){ # else we keep kopt = 1
            kopt = which.max(silwidths) # always >= 2
            clusvec = cluslist[[kopt]]
          }
        } # ends mykmax >= 2
      }
      cat(paste0("Selected k = ", kopt,".\n"))
      if(kopt > 1){
        mymodes = list()
        if(rugnumer == TRUE | rugfact == TRUE) myrugs = list()
        tempval = rep(NA, kopt)
        for(m in seq_len(kopt)){ # m=1
          mymodes[[m]] = y[which(clusvec == m)]
          if(rugnumer == TRUE) myrugs[[m]] = rugNum[[j]][which(clusvec == m)]
          if(rugfact  == TRUE) myrugs[[m]] = rugFac[[j]][which(clusvec == m)]
          names(mymodes[[m]]) = names(y)[which(clusvec == m)]
          tempval[m] = (mymodes[[m]])[1]
        }
        ordr = order(tempval) # to sort the modes
        cat("Cluster sizes:\n")
        print(t(as.matrix(table(clusvec))))
        sortedmodes = list()
        if(rugnumer == TRUE | rugfact == TRUE) sortedrugs = list()
        clustnew = rep(NA, n)
        for(kk in seq_len(kopt)){
          kkk = ordr[kk]
          sortedmodes[[kk]] = as.vector(mymodes[[kkk]])
          names(sortedmodes[[kk]]) = names(mymodes[[kkk]])
          if(rugnumer == TRUE) sortedrugs[[kk]] = as.vector(myrugs[[kkk]])
          if(rugfact  == TRUE) sortedrugs[[kk]] = as.vector(myrugs[[kkk]])
          clustnew[clusvec == kkk] = kk
        }
        clusvec = clustnew; rm(clustnew)
        names(clusvec) = names(y)
        listj[["clustering"]] = clusvec
        modelist = list()
        #
        # We have to compute the densities first, since their
        # relative widths will depend on the argument bodysize.
        densities = list()
        areas = heights = counts = rep(NA, kopt)
        for(m in seq_len(kopt)){ # loop over the modes
          ym = sortedmodes[[m]]
          counts[m] = length(ym)
          if(length(unique(ym)) < clusMinN){
            # This should never happen due to pamc
            densities[[m]] = NA
          } else { # for this mode:
            dens = NA
            # compute density of mode m:
            dens = compdens(ym, bwnum = bwnum, kernel = kernel,
                            cut = cut, cutmin = cutmin,
                            cutmax = cutmax)
            # we reuse the overall bandwidth for all modes!
            ymintemp = min(ymintemp, min(dens$x))
            ymaxtemp = max(ymaxtemp, max(dens$x))
            xrange = max(dens$x) - min(dens$x)
            area   = xrange * mean(dens$y, na.rm = T)
            dens$y = dens$y/area # so area is exactly 1
            areas[m] = xrange * mean(dens$y, na.rm = T)
            heights[m] = max(dens$y)
            densities[[m]] = dens
          } # ends density computation for this mode
        } # ends loop over modes
        factors = rep(1, kopt)
        if(bodysize == "width_is_constant"){
          factors = 1/heights
        }
        if(bodysize == "area_is_constant"){
          factors = rep(1, kopt)/max(heights)
        }
        if(bodysize == "area_from_count"){
          factors = (counts/max(counts))
          factors = factors/max(factors*heights)
        }
        factors[which(is.na(factors))] = 0
        ratios = heights*factors # for width of boxes and rug
        ratios = ratios/max(ratios)
      } # ends if(kopt > 1)
      #
      # Start plotting
      if(kopt == 1){ # Plot as a single cluster
        listj[["values"]] = y
        my5 = fivenum(y, na.rm=TRUE)
        names(my5) = NULL
        listj[["fivenumbersummary"]] = my5
        whiskers = "none"
        if(makewhiskers == TRUE) whiskers = "both"
        if(length(unique(y)) < clusMinN){ # should never happen
          drawpoints(y, coord = xcoords[j], bokw = boxW[j],
                     side = mysides[j], cex=0.6,
                     horizontal = horizontal, pointcol = par()$fg)
          if(boxW[j] > 0) {
            drawbox(y, coord = xcoords[j], horizontal = horizontal,
                    side = mysides[j], bokw = boxW[j],
                    bcol = boxCol[j], whiskers = "none") }
        } else { # For this variable:
          colin = rugCol[j]
          colout = rugoutCol[j]
          if(rugnumer == TRUE){
            mycols = round(99*(rugNum[[j]]-ran[1])/(ran[2]-ran[1])) + 1
            mycols[is.na(mycols)] = 0
            mycols = colorpalette[mycols]
            colin = colout = mycols
          }
          if(rugfact == TRUE){
            mycols = rugFac[[j]]
            mycols[is.na(mycols)] = 0
            mycols = rugFactorColors[mycols]
            colin = colout = mycols
          }
          if((!is.na(bodyCol[j]) | !is.na(curveCol[j])) && (bodyW[j] > 0)) {
            # Compute the density
            dens = compdens(y, bwnum = bwnum, kernel = kernel,
                            cut = cut, cutmin = cutmin,
                            cutmax = cutmax)
            ymintemp = min(ymintemp, min(dens$x))
            ymaxtemp = max(ymaxtemp, max(dens$x))
            dens$y = 0.5*bodyW[j]*dens$y/max(dens$y)
            drawdensity(side = mysides[j], dens, at = xcoords[j],
                        bodcol = bodyCol[j], curcol = curveCol[j],
                        curlwd = curveLwd, horizontal = horizontal)
          } else {
            colout = colin # whether fixed or variable
          }
          if (sum(is.na(colin)) == 0 && (rugW[j] > 0)) {
            set.seed(1) # so jittering always gives same result
            if(mysides[j] == "no"){
              drawrug2sided(y, side = mysides[j], jittering = jittering,
                            jitteramount = jitteramount, dens = dens,
                            at = xcoords[j], rugw = rugW[j],
                            colin = colin, colout = colout,
                            rugLwd = rugLwd, rugoutLwd = rugoutLwd,
                            horizontal = horizontal)
            } else { # mysides[j] is either "first" or "second"
              if(boxW[j] > 0){
                drawrug1sided(y, side = mysides[j], jittering = jittering,
                              jitteramount = jitteramount, at = xcoords[j],
                              rugw = rugW[j]/2, rugcenter = 0.27*boxW[j],
                              colin = colin, rugLwd = rugLwd,
                              horizontal = horizontal)
              }
            }
          }
          if(boxW[j] > 0) {
            drawbox(y, coord = xcoords[j], horizontal = horizontal,
                    side = mysides[j], bokw = boxW[j], lwd = boxLwd,
                    bcol = boxCol[j], whiskers = whiskers) }
        }
      } else { # kopt > 1
        for(m in seq_len(kopt)){ # plot all the modes
          ym = sortedmodes[[m]] # modes in increasing order
          my5 = fivenum(ym, na.rm=TRUE)
          names(my5) = NULL
          modelist[[paste0("cluster_",m)]] =
            list(members = ym, fivenumbersummary = my5)
          whiskers = "none"
          if(makewhiskers == TRUE){
            whiskers = "both"
            if(innerwhiskers == FALSE){
              whiskers = "none"
              if(m == 1) whiskers = "down"
              if(m == kopt) whiskers = "up"
            }
          }
          if(length(unique(ym)) < clusMinN){ # should never happen
            drawpoints(ym, coord = xcoords[j], bokw = boxW[j],
                       side = mysides[j], cex=0.6,
                       horizontal = horizontal, pointcol = par()$fg)
            if(boxW[j] > 0) drawbox(ym, coord = xcoords[j],
                                    horizontal = horizontal,
                                    side = mysides[j], bokw = boxW[j],
                                    bcol = boxCol[j], whiskers = "none")
          } else { # for this mode:
            thismodecol = bodyCol[j]
            if(colormodes == TRUE){
              jj = jj+1
              thismodecol = modeCol[jj] # can be NA
            }
            colin = rugCol[j]
            colout = rugoutCol[j]
            if(rugnumer == TRUE){
              mycols = round(99*(sortedrugs[[m]]-ran[1])/(ran[2]-ran[1])) + 1
              mycols[is.na(mycols)] = 0
              mycols = colorpalette[mycols]
              colin = colout = mycols
            }
            if(rugfact == TRUE){
              mycols = sortedrugs[[m]]
              mycols[is.na(mycols)] = 0
              mycols = rugFactorColors[mycols]
              colin = colout = mycols
            }
            if((!is.na(thismodecol) | !is.na(curveCol[j])) && (bodyW[j] > 0)) {
              dens = densities[[m]]
              dens$y = 0.5*bodyW[j]*factors[m]*dens$y
              drawdensity(side = mysides[j], dens, at = xcoords[j],
                          bodcol = thismodecol, curcol = curveCol[j],
                          curlwd = curveLwd, horizontal = horizontal)
            } else { # if we don't draw a density, the outside
              # rug color should coincide with the inside
              colout = colin # whether fixed or variable
            }
            if (sum(is.na(colin)) == 0 && (rugW[j] > 0)) {
              set.seed(1) # so jittering always gives same result
              if(mysides[j] == "no"){
                drawrug2sided(ym, side = mysides[j], jittering = jittering,
                              jitteramount = jitteramount, dens = dens,
                              at = xcoords[j], rugw = rugW[j],
                              colin = colin, colout = colout,
                              rugLwd = rugLwd, rugoutLwd = rugoutLwd,
                              horizontal = horizontal)
              } else { # mysides[j] is either "first" or "second"
                if(boxW[j] > 0){
                  drawrug1sided(ym, side = mysides[j], jittering = jittering,
                                jitteramount = jitteramount, at = xcoords[j],
                                rugw = rugW[j]/2,
                                rugcenter = 0.27*ratios[m]*boxW[j],
                                colin = colin, rugLwd = rugLwd,
                                horizontal = horizontal)
                }
              }
            }
            bokw = ratios[m]*boxW[j]
            if(bokw > 0) drawbox(ym, coord = xcoords[j],
                                 horizontal = horizontal,
                                 side = mysides[j], bokw = bokw,
                                 lwd = boxLwd, bcol = boxCol[j],
                                 whiskers = whiskers)
          }
        } # ends loop over modes m
        listj = c(listj, modelist)
      } # ends kopt > 1
    } # ends if(n >= 1)
    if(!is.na(stickCol[j]) && (stickLwd > 0) && (side == "both")){
      if((ymintemp < Inf) && (ymaxtemp > -Inf)){
        if(horizontal == FALSE){
          lines(c(xcoords[j], xcoords[j]), c(ymintemp, ymaxtemp),
                col = stickCol[j], lwd = stickLwd)
        } else {
          lines(c(ymintemp, ymaxtemp), c(xcoords[j], xcoords[j]),
                col = stickCol[j], lwd = stickLwd)
        }
      }
    } # ends stick
    name = names[j]
    outp[[name]] = listj # noquote(paste0("Variable ",j))
  } # ends loop over j
  if(add == FALSE){
    if(rugnumer == TRUE){ # plot color bar:
      mymar = maruser
      mymar[2] = 0
      par(mar = mymar)
      color_bar_matrix <- matrix(seq(0, 1, length.out = 100), nrow = 1)
      image(color_bar_matrix, col = colorpalette, axes=FALSE)
      # Add a y-axis to the color bar to show the values:
      # Use pretty() to get "nice" tick mark values:
      tickvalues <- pretty(ran, n = 5)
      # Calculate the corresponding positions on the 0-1
      # scale of the color bar
      tickpositions <- (tickvalues - ran[1])/(ran[2] - ran[1])
      # Use these positions and values to create the axis
      axis(4, at = tickpositions, labels = tickvalues,
           cex.axis = cex.colorbar, las = 1)
      box() # Add a box around the color bar
      # Reset the layout to default:
      layout(1)
      # Reset the margins to before:
      par(mar = maruser)
    }
  }
  if(plot == TRUE) { invisible(outp) } else { return(outp) }
}


pamc1d <- function(y, k, minsize = 4, countwhat = "unique",
                   stand = TRUE, maxit = 100, verbose = TRUE){
  # Constrained version of k-medoids clustering (partitioning
  # around medoids) for univariate data y.
  # This is a modification of an algorithm for constrained
  # k-means by Bradley-Bennett-Demiriz (2000, Microsoft report
  # MSR-TR-2000-65). The modifications include working with
  # medians instead of means, and providing a minsize constraint
  # on unique cases instead of any cases, to prevent placing
  # tied cases in separate clusters.
  #
  # Arguments
  # y       : univariate data
  # k       : specified number of clusters
  # minsize : lower bound on the size of all clusters
  # maxit   : maximum number of iterations in reassignment loop
  # verbose : TRUE will print intermediate steps.

  objL1 = function(yy, centers, clusvec){
    mycmat = abs(outer(as.vector(yy),centers,"-"))
    sum(mycmat*outer(clusvec, 1:length(centers), "=="))/length(yy)
  }

  if(!(countwhat %in% c("any", "unique"))) stop("invalid countwhat")
  y = sort(na.omit(y))
  if(length(unique(y)) < 2) stop("y has < 2 non-NA unique values, so no clustering")
  n <- length(y)
  if(is.null(names(y))) names(y) = seq_len(length(y))
  yraw = y
  if(stand == TRUE) y = scale(y) # possible since sd(y) > 0
  if(countwhat == "unique"){
    ydup = duplicated(y)
    nuniq = sum(ydup == FALSE)
    if(nuniq == n) countwhat = "any"
  }
  kminsize = k*minsize
  if(countwhat == "any"){
    if(kminsize > n) stop(paste0(
      "A partition with k = ",k," clusters that each\n",
      "have at least minsize = ", minsize, " members\n",
      "would require ", kminsize," data points, but\n",
      "there are only ", n, " data points."))
  } else { # if(countwhat == "unique")
    if(kminsize > nuniq) stop(paste0(
      "A partition with k = ",k," clusters that each\n",
      "have at least minsize = ", minsize, " unique members\n",
      "would require ", kminsize," unique data points, but\n",
      "there are only ", nuniq, " unique data points."))
  }
  objectives = list()
  set.seed(1)
  clusvec = cluster::pam(y, k, cluster.only = TRUE,
                         keep.diss=F, keep.data=F)
  if(verbose == TRUE){
    cat("Initial clustering vector:\n")
    print(clusvec) # for all n cases (whether unique or not)
    cat("Initial clustering table:\n")
    print(as.data.frame(t(as.matrix(table(clusvec)))),
          row.names = FALSE)
  }
  centers = rep(NA, k)
  bigenough = TRUE
  for(h in seq_len(k)){
    clush = which(clusvec == h)
    yh = y[clush]
    if(countwhat == "any"){
      if(length(yh) < minsize) bigenough = FALSE
    }
    if(countwhat == "unique"){
      if(length(unique(yh)) < minsize) bigenough = FALSE
    }
    centers[h] = median(yh)
  }
  obj = objL1(y, centers, clusvec)
  if(verbose == TRUE){
    cat(paste0("Unconstrained objective = ", obj, "\n"))
  }
  objectives = c(objectives, obj)
  num = 0
  iter = 1
  if(bigenough == FALSE){
    if(countwhat == "any"){
      num = 1
      while (iter <= maxit & num > 0){
        # The second condition says: stop when no assignment changes
        if(verbose) cat(paste0("\nIteration Number = ", iter,"\n"))
        cost.mat = abs(outer(as.vector(y),centers,"-"))
        # = matrix with distances of all cases to each center
        trans <- lpSolve::lp.transport(cost.mat = cost.mat,
                              direction = "min",
                              row.signs = rep("=", n),
                              row.rhs = rep(1, n),
                              col.signs = rep(">=", k),
                              col.rhs = rep(minsize, k))
        clusnew <- apply(trans$solution, 1, which.max)
        num <- sum(clusnew != clusvec)
        if(verbose == TRUE){
          cat(paste0("Number of assignment changes = ", num,"\n"))
        }
        if(num > 0){ # if not, we are done
          clusvec <- clusnew # update clustering
          if(verbose == TRUE){
            print(clusvec) # new clustering vector
            print(as.data.frame(t(as.matrix(table(clusvec)))),
                  row.names = FALSE)
          }
          # recompute centers:
          for(h in seq_len(k)){
            clush = which(clusvec == h)
            centers[h] = median(y[clush])
          }
          obj = objL1(y, centers, clusvec)
          if(verbose == TRUE){
            cat(paste0("objective = ", obj, "\n"))
          }
          objectives = c(objectives, obj)
          iter <- iter + 1
        }
      } # ends while
      if(min(table(clusvec)) < minsize){ # should never happen
        warning(paste0("Failed to satisfy the minsize condition for\n",
                       "all clusters. Try with lower k or minsize."))
      }
    } # ends countwhat == any
    if(countwhat == "unique"){
      mult = rep(1, nuniq)
      if(verbose == TRUE){
        cat(paste0("\nThere are ", nuniq, " unique cases out of ",
                   n," cases.\n"))
      }
      whichuni = which(ydup == FALSE)
      yuniq = y[whichuni]
      yrawuniq = yraw[whichuni]
      if(verbose == TRUE){
        cat("Unique cases:\n")
        print(clusvec[whichuni]*0 + yrawuniq)
      }
      clusuni = clusvec[whichuni]
      whichdup = which(ydup == TRUE)
      if(verbose == TRUE){
        cat("Duplicate cases:\n")
        print(clusvec[whichdup]*0 + yraw[whichdup])
      }
      dupofwhat = rep(NA, length(whichdup))
      for(i in seq_len(length(whichdup))){ # i = 10
        yi = y[whichdup[i]]
        dupofwhat[i] = min(which.min(abs(y[whichuni] - yi)))
        mult[dupofwhat[i]] = mult[dupofwhat[i]] + 1
      }
      if(verbose == TRUE){
        cat("They are duplicates of the cases:\n")
        whichof = whichuni[dupofwhat]
        print(clusvec[whichof]*0 + yraw[whichof])
      }
      num = 1
      while (iter <= maxit & num > 0){
        if(verbose) cat(paste0("\nIteration Number = ", iter,"\n"))
        cost.mat = abs(outer(as.vector(yuniq),centers,"-"))
        trans <- lp.transport(cost.mat = cost.mat,
                              direction = "min",
                              row.signs = rep("=", nuniq),
                              row.rhs = rep(1, nuniq),
                              col.signs = rep(">=", k),
                              col.rhs = rep(minsize, k))
        clusuninew <- apply(trans$solution, 1, which.max)
        num <- sum(clusuninew != clusuni) # number of assignment changes
        if(verbose == TRUE){
          cat(paste0("Number of assignment changes = ", num,"\n"))
        }
        if(num > 0){
          clusuni <- clusuninew # update clustering
          # turn clusuni back into clusvec:
          if(nuniq < n){
            clusvec[whichuni] = clusuni
            clusvec[whichdup] = clusuni[dupofwhat]
          } else { clusvec = clusuni }
          if(verbose == TRUE){
            print(clusvec) # true clustering vector
            print(as.data.frame(t(as.matrix(table(clusvec)))),
                  row.names = FALSE)
          }
          for(h in 1:k){
            clush = which(clusvec == h)
            centers[h] = median(y[clush])
          }
          obj = objL1(y, centers, clusvec)
          if(verbose == TRUE){
            cat(paste0("objective = ", obj, "\n"))
          }
          objectives = c(objectives, obj)
          iter <- iter + 1
        }
      } # ends while
      if(min(table(clusuni)) < minsize){ # should never happen
        warning(paste0("Failed to satisfy the minsize condition for\n",
                       "all clusters. Try with lower k or minsize."))
      }
    } # ends countwhat == unique
  } # ends bigenough == FALSE
  clustable = clusvec
  clustable = table(clustable)
  if(verbose == TRUE){
    cat("\nObjective function in iteration steps: \n")
    print(matrix(objectives, ncol=1))
    cat("\nFinal clustering vector:\n")
    print(clusvec)
    cat("\nFinal clustering table:\n")
    print(as.data.frame(t(as.matrix(clustable))), row.names = FALSE)
  }
  result <- list(iter = iter, converged = (num==0),
                 clustering = clusvec, obj = obj,
                 centers = centers, clustable = clustable)
  return(result)
}

confmat.vcr <- function(vcrout, cutoff = 0.99,
                        showClassNumbers = FALSE,
                        showOutliers = TRUE, silent = FALSE) {
  # Builds a confusion matrix from a vcr object.
  #
  # Arguments:
  # vcrout           : output of vcr.*.train() or vcr.*.newdata().
  # cutoff           : the objects with overall farness
  #                    ofarness > cutoff are flagged as outliers.
  # showClassNumbers : if TRUE, the row and column names are the
  #                    number of each level instead of the level
  #                    itself. Useful for long level names.
  # showOutliers     : if TRUE and some points were flagged as
  #                    outliers, it adds an extra column on the
  #                    right of the confusion matrix for these
  #                    outliers, with label "outl".
  # silent           : if FALSE, the confusion matrix and
  #                    accuracy are shown on the screen.
  #
  # Returns:
  # confMat          : the confusion matrix.
  #
  levels <- vcrout$levels
  nlab   <- length(levels)
  if (is.null(vcrout$yintnew)) {
    if (is.null(vcrout$yint)) {
      stop("there is no vcrout$yint or vcrout$yintnew")
    }
    given <- vcrout$yint
    training <- TRUE
  } else {
    given <- vcrout$yintnew
    training <- FALSE
  }
  whichyint <- if (training) {"yint"} else {"yintnew"}
  indsv <- which(!is.na(given))
  if (length(indsv) == 0) {
    stop(paste0(whichyint, " has no available values, ",
                " so no confusion matrix can be made."))
  }
  given <- given[indsv]
  if (sum(!(given %in% seq_len(nlab))) > 0) {
    stop(paste0("Not all ", whichyint, "[indsv] lie in 1:", nlab))
    }
  avobs <- sort(unique(given)) # available given values
  predicted <- vcrout$predint[indsv] # corresponding predictions
  if (sum(is.na(predicted)) > 0) {
    stop("\n predint[indsv] has missing values.")}
  if (sum(!(predicted %in% seq_len(nlab))) > 0) {
    stop(paste0("Not all predint[indsv] lie in 1:", nlab))
    }
  # compute the accuracy before taking outliers into account:
  accuracy <- sum(given == predicted) / length(indsv)
  myrownames <- levels[avobs]
  ofarness <- vcrout$ofarness[indsv]
  outliers <- (ofarness > cutoff)
  if (showOutliers & sum(outliers) > 0) {
    predicted[which(outliers)] <- nlab + 1
    confMat <- table(given, predicted)
    mycolints <- sort(unique(predicted))
    mycolnames <- c(levels, "outl")[mycolints]
    mycolints[which(mycolints == (nlab + 1))] <- "outl"
  } else {
    confMat <- table(given, predicted)
    mycolints <- sort(unique(predicted))
    mycolnames <- levels[mycolints]
  }
  if (showClassNumbers) {
    rownames(confMat) <- avobs
    colnames(confMat) <- mycolints
  } else {
    rownames(confMat) <- myrownames
    colnames(confMat) <- mycolnames
  }
  if (silent != TRUE) {
    wnq(paste0("\nConfusion matrix:"))
    pnq(confMat)
    wnq(paste0("\nThe accuracy is ", round(100 * accuracy, 2), "%."))
  }
  invisible(confMat)
}



stackedplot <- function(vcrout, cutoff = 0.99, classCols = NULL,
                        classLabels = NULL, separSize = 1,
                        minSize = 1.5, showOutliers = TRUE,
                        showLegend = FALSE, main = NULL,
                        htitle = NULL, vtitle = NULL)
{
  # Argument added:
  # main              : title for the plot
  #
  if (is.null(vcrout$yintnew)) {
    if (is.null(vcrout$yint)) {
      stop("there is no vcrout$yint or vcrout$yintnew")
    }
    whichdata <- "[training]"
  }
  else {
    whichdata <- "[newdata]"
  }
  if (is.null(htitle))
    htitle <- paste0(whichdata, " given class")
  if (is.null(vtitle))
    vtitle <- paste0(whichdata, " predicted class")
  nlab <- length(vcrout$levels)
  if (is.null(classLabels)) {
    lvls <- vcrout$levels
    if ("outl" %in% lvls)
      stop("Please rename the level currently called 'outl'.")
  }
  else {
    if (!is.vector(classLabels)) {
      stop("\n classLabels should be a vector")
    }
    lvls <- classLabels
    if ("outl" %in% lvls)
      stop("Do not use 'outl' as a class label.")
    if (length(lvls) != nlab) {
      stop(paste0("\n The number of classLabels should equal",
                  " length(vcrout$levels) = ", nlab, "."))
    }
  }
  if (is.null(classCols)) {
    mosaicColors <- rainbow(nlab)
  }
  else {
    if (!is.vector(classCols)) {
      stop("\n classCols should be a vector")
    }
    if (length(classCols) < nlab) {
      stop(paste0("\n The number of classCols should be at",
                  " least length(vcrout$levels) = ", nlab,
                  "."))
    }
    mosaicColors <- classCols[seq_len(nlab)]
  }
  confusi <- confmat.vcr(vcrout, cutoff = cutoff, showOutliers = showOutliers,
                         showClassNumbers = TRUE, silent = TRUE)
  pconfusi <- confusi / rowSums(confusi) * 100
  ayint <- as.numeric(rownames(confusi))
  nlaba <- length(ayint)
  if (nlaba < nlab) {
    wnq(paste0("\nNot all classes occur in these data.",
               " The classes to plot are:"))
    pnq(ayint)
  }
  colints <- colnames(confusi)
  colints[which(colints == "outl")] <- nlab + 1
  for (j in seq_len(nlaba)) {
    confvec <- pconfusi[j, ]
    tooSmall <- which((confvec > 0) & (confvec < minSize))
    if (length(tooSmall) > 0) {
      confvec[tooSmall] <- minSize
      confvec <- confvec/sum(confvec) * 100
      pconfusi[j, ] <- confvec
    }
  }
  hasOutliers <- 0
  if (showOutliers) {
    if ("outl" %in% colnames(confusi)) {
      lvls <- c(lvls, "outliers")
      hasOutliers <- 1
      mosaicColors <- c(mosaicColors[seq_len(nlab)], "gray20")
    }
  }
  confRow <- rep(0, length(lvls))
  colints <- as.numeric(colints)
  confRow[colints] <- pconfusi[1, ]
  df <- data.frame(g1_fill = factor(lvls[seq_len(nlab + hasOutliers)],
                                    lvls), g1_count = confRow)
  if (nlaba > 1) {
    for (j in 2:nlaba) {
      g <- ayint[j]
      confRow <- rep(0, length(lvls))
      colints <- as.numeric(colints)
      confRow[colints] <- pconfusi[j, ]
      fillname <- paste0("g", j, "_fill")
      perm <- c(g, (seq_len(nlab + hasOutliers))[-g])
      df[[fillname]] <- factor(lvls[perm], lvls[perm])
      countname <- paste0("g", j, "_count")
      df[[countname]] <- confRow[perm]
    }
  }
  widths <- table(vcrout$yint)/length(vcrout$yint)
  pos <- 0.5 * (cumsum(widths) + cumsum(c(0, widths[-length(widths)])))
  gg <- ggplot(df)
  for (i in seq_len(nlaba)) {
    fillname <- paste0("g", i, "_fill")
    countname <- paste0("g", i, "_count")
    gg <- gg + geom_bar(mapping = aes_string(x = pos[i], y = countname,
                                             fill = fillname),
                        width = widths[i], stat = "identity",
                        size = separSize, colour = "white",
                        show.legend = showLegend)
  }
  gg <- gg + scale_x_continuous(htitle, breaks = pos,
                                labels = lvls[ayint],
                                expand = c(0.02, 0))
  gg <- gg + scale_y_continuous(vtitle, trans = "reverse",
                                expand = c(0.02, 0))
  gg <- gg + theme(axis.text.y = element_blank(),
                   axis.ticks.y = element_blank(),
                   panel.grid.major.y = element_blank(),
                   panel.grid.minor.y = element_blank())
  if (showLegend) {
    gg <- gg + labs(fill = "Classes")
  }
  gg <- gg + theme(axis.ticks.x = element_blank(),
                   panel.grid.major.x = element_blank(),
                   panel.grid.minor.x = element_blank())
  gg <- gg + scale_fill_manual(values = mosaicColors)
  if (!is.null(main)) {
    gg <- gg + labs(title = paste0(main))
    gg <- gg + theme(plot.title = element_text(hjust = 0.5)) # center title
  }
  return(gg)
}




classmap <- function(vcrout, whichclass, classLabels = NULL,
                     classCols = NULL, main = NULL, cutoff = 0.99,
                     plotcutoff = TRUE, identify = FALSE, cex = 1,
                     cex.main = 1.2, cex.lab = NULL, cex.axis = NULL,
                     opacity = 1, squareplot = TRUE,
                     maxprob = NULL, maxfactor = NULL) {
  #
  # Draws class maps to visualize classification results.
  #
  # Arguments:
  # vcrout       : the result of a function vcr.*.*()
  # whichclass   : the class to be displayed. Required. The colors
  #                of the points are those of their predicted
  #                labels. Required.
  # classLabels  : the labels (levels) of the classes. If NULL, they
  #                are taken from vcrout.
  # classCols    : a list of colors for the class labels 1,... There
  #                should be at least as many as there are levels.
  #                If NULL the classCols are taken as 2,...
  # main         : title for the plot.
  # cutoff       : objects with overall farness > cutoff will be
  #                flagged as outliers and plotted with a black
  #                border around them.
  # plotcutoff   : if TRUE, a dashed vertical line is drawn at
  #                the cutoff.
  # identify     : if TRUE, left-click on a point to get its number,
  #                then ESC to exit.
  # cex          : passed on to plot().
  # cex.main     : same, for title.
  # cex.lab      : same, for labels on horizontal and vertical axes.
  # cex.axis     : same, for axes.
  # opacity      : determines opacity of plotted dots.
  #                Value between 0 and 1, where 0 is transparent
  #                and 1 is opaque.
  # squareplot   : if TRUE, the horizontal and vertical axis will
  #                get the same length.
  # maxprob      : draws the farness axis at least upto probability
  #                maxprob.
  #                If NULL, the limits are obtained automatically.
  # maxfactor    : if not NULL, a number slightly higher than 1
  #                to increase the space at the right hand side of
  #                the plot, to make room for marking points.
  #
  # Returns:
  # coordinates  : a matrix with 2 columns containing the
  #                coordinates of the plotted points. The
  #                first coordinate is the quantile of the
  #                farness probability.

  # Auxiliary function:
  qfunc <- function(probs) {
    # quantile function of N(0,1) restricted to [0,4].
    if (min(probs) < 0) stop("There are negative probs")
    if (max(probs) > 1) stop("there are probs > 1")
    qnorm(probs * (pnorm(4) - 0.5) + 0.5)
  }

  # The main function starts here:
  nlab <- length(vcrout$levels)
  if (is.null(classLabels)) { # no classLabels are given
    levels <- vcrout$levels
  } else {# classLabels are given
    if (!is.vector(classLabels)) {
      stop("\n classlabels should be a vector")
    }
    levels <- classLabels
    if (length(levels) != nlab) {
      stop(paste0("\n The number of classLabels should equal",
                  " length(vcrout$levels) = ", nlab, "."))
    }
  }
  if (is.null(classCols)) {
    classCols <- seq_len(nlab) + 1 # red, green, blue,...
  } else {
    if (!is.vector(classCols)) {
      stop("\n classCols should be a vector")
    }
    if (length(classCols) < nlab) {
      stop(paste0("\n The number of classCols should be at",
                  " least length(vcrout$levels) = ", nlab, "."))
    }
    classCols <- classCols[seq_len(nlab)]
  }
  if (is.null(vcrout$yintnew)) {
    if (is.null(vcrout$yint)) stop("vcrout$yint is missing")
    yint <- vcrout$yint
    training <- TRUE
  } else {
    yint <- vcrout$yintnew
    training <- FALSE
  }
  whichdata <- if (training) {"[training]"} else {"[newdata]"}
  whichyint <- if (training) {"yint"} else {"yintnew"}
  indsv <- which(!is.na(yint))
  if (length(indsv) == 0) {
    stop(paste0(whichyint, " has no available values, ",
                " so the plot cannot be made.")) }
  yint     <- vcrout$yint[indsv]
  predint  <- vcrout$predint[indsv]
  PAC      <- vcrout$PAC[indsv]
  ofarness <- vcrout$ofarness[indsv]
  outliers <- (ofarness > cutoff)
  # In the classmap we will plot farness at quantile positions:
  qfarness <- qfunc(vcrout$farness[indsv])
  coordinates <- cbind(qfarness, PAC)
  rownames(coordinates) <- indsv # case numbers in entire set.
  #
  fardashed <- qfunc(cutoff)
  if (!is.null(maxprob)) maxfar <- min(maxprob, 1)
  if (is.null(maxfactor)) maxfactor <- 1
  maxfar <- max(max(qfarness), fardashed)
  if (sum(!(yint %in% seq_len(nlab))) > 0) {
    stop(paste0("Not all ", whichyint, " lie in 1:", nlab)) }
  ayint <- sort(unique(yint)) # available yint values
  #
  if (sum(!(predint %in% seq_len(nlab))) > 0) {
    stop(paste0(whichdata, " Not all predint lie in 1:", nlab))
  }
  #
  if (is.numeric(whichclass)) {
    if (!(whichclass %in% seq_len(nlab))) stop(paste0(
      "whichclass must be in 1:", nlab, ", the number of classes."))
  } else {#
    classnumber <- match(whichclass, vcrout$levels)
    if (is.na(classnumber)) stop(paste0(
      'level "', whichclass, '" does not exist.'))
    whichclass <- classnumber
  }
  if (!(whichclass %in% ayint)) stop(paste0(
    "Class number ", whichclass, " with label ",
    levels[whichclass], " has no objects to visualize."))
  if (is.null(main)) {
    main <- paste0(whichdata, " predictions of class ",
                   levels[whichclass])
  }
  clinds   <- which(yint == whichclass) # class indices
  predint  <- predint[clinds]
  casecols <- classCols[predint]
  coordinates <- coordinates[clinds, ]
  outliers <- which(outliers[clinds])
  #
  # Prepare for the plot:
  #
  PAClim <- PACsolid <- c(0, 1)
  mypch <- rep(19, length(predint)) # disk (16 is a bit smaller)
  mypch[outliers] <- 21 # filled circle
  mycex <- rep(cex, length(predint))
  mycex[outliers] <- 1.1*cex
  mycol <- mybg <- adjustcolor(casecols, alpha.f = opacity)
  mycol[outliers] <- "black"
  mycex.main <- 1
  if (!is.null(cex.main)) {mycex.main <- cex.main}
  mycex.lab <- 1
  if (!is.null(cex.lab)) {mycex.lab <- cex.lab}
  mycex.axis <- 1
  if (!is.null(cex.axis)) {mycex.axis <- cex.axis}
  #
  # Make actual plot
  #
  if (squareplot == TRUE) par(pty = "s")
  # makes the axes of the plot equally long
  farlim <- c(0, maxfar * maxfactor)
  plot(0, 0, type = "n", ann = FALSE, axes = FALSE,
       xlim = farlim, ylim = PAClim)
  # makes an empty plot
  title(main = main, line = 1, cex.main = mycex.main)
  title(ylab =  "P[alternative class]", line = 2.3,
        cex.lab = mycex.lab)
  title(xlab = "farness from given class",
        line = 2.3, cex.lab = mycex.lab)
  rect(farlim[1] - farlim[2], -2,farlim[2] * 2, 0.5, col = "lightgray",
       border = FALSE) # adds rectangle
  # col can be modified to other gray intensities
  par(new = TRUE) # overlay next plot on this one
  plot(coordinates, xlab = "", ylab = "", xlim = farlim,
       ylim = PAClim, col = mycol, bg = mybg, pch = mypch,
       cex = mycex, cex.axis = mycex.axis, xaxt = "n")
  probs <- c(0, 0.5, 0.75, 0.9, 0.99, 0.999, 1)
  axis(1, at = qfunc(probs), labels = probs,
       cex.axis = mycex.axis)
  if (plotcutoff) abline(v = fardashed, lty = 4)
  abline(h = PACsolid)
  attr(coordinates, "ids") <- integer(0)
  if (identify) {
    message("Press the escape key to stop identifying.")
    iout <- identify(coordinates, order = TRUE)
    ids <- iout$ind[order(iout$order)]
    if (length(ids) > 0) {
      wnq("Identified point(s): ")
      pnq(ids)
      attr(coordinates, "ids") <- ids
    } else {wnq("No points were identified.")}
  }
  invisible(coordinates)
}



silplot <- function(vcrout, classLabels = NULL, classCols = NULL,
                    showLegend = TRUE, showClassNumbers = FALSE,
                    showCases = FALSE, drawLineAtAverage = FALSE,
                    topdown = TRUE, main = NULL, summary = TRUE)
{
  nlab <- length(vcrout$levels)
  if (is.null(classLabels)) {
    lvls <- vcrout$levels
  }
  else {
    if (!is.vector(classLabels)) {
      stop("\n classLabels should be a vector")
    }
    lvls <- classLabels
    if (length(lvls) != nlab) {
      stop(paste0("\n The number of classLabels should equal",
                  " length(vcrout$levels) = ", nlab, "."))
    }
  }
  if (is.null(classCols)) {
    classCols <- rainbow(nlab)
  }
  else {
    if (!is.vector(classCols)) {
      stop("\n classCols should be a vector")
    }
    if (length(classCols) < nlab) {
      stop(paste0("\n The number of classCols should be at",
                  " least length(vcrout$levels) = ", nlab,
                  "."))
    }
    classCols <- classCols[seq_len(nlab)]
  }
  if (is.null(vcrout$yintnew)) {
    if (is.null(vcrout$yint)) {
      stop("there is no vcrout$yint or vcrout$yintnew")
    }
    yintv <- vcrout$yint
    training <- TRUE
  }
  else {
    yintv <- vcrout$yintnew
    training <- FALSE
  }
  whichdata <- if (training) {
    "[training]"
  }
  else {
    "[newdata]"
  }
  whichyint <- if (training) {
    "yint"
  }
  else {
    "yintnew"
  }
  indsv <- which(!is.na(yintv))
  if (length(indsv) == 0) {
    stop(paste0(whichyint, " has no available values, ",
                " so no silhouette plot can be made."))
  }
  if (length(indsv) < 2) {
    stop(paste0("At least 2 cases with non-missing ",
                whichyint, " are required."))
  }
  yintv <- yintv[indsv]
  ayint <- sort(unique(yintv))
  if (sum(!(yintv %in% seq_len(nlab))) > 0) {
    stop(paste0("Not all ", whichyint, "[indsv] lie in 1:",
                nlab))
  }
  PAC <- vcrout$PAC[indsv]
  if (sum(is.na(PAC)) > 0)
    stop("PAC[indsv] has missing values.")
  si <- 1 - 2 * PAC
  if (is.null(main)) {
    main <- paste0(whichdata, " Silhouette plot of classification")
  }
  df <- as.data.frame(cbind(class = yintv, si, name = indsv))
  if (topdown) {
    df <- df[order(-df$class, df$si), ]
  }
  else {
    df <- df[order(df$class, -df$si), ]
  }
  avswidth <- mean(df$si)
  df$name  <- factor(df$name, levels = df$name)
  df$class <- as.factor(df$class)
  if (showClassNumbers) {
    mapping <- aes_string(x = "name", y = "si",
                          color = "class", fill = "class")
  }
  else {
    df$label <- factor(lvls[df$class], levels = lvls[seq_len(nlab)])
    mapping <- aes_string(x = "name", y = "si",
                          color = "label", fill = "label")
  }
  gg <- ggplot(df, mapping) + geom_bar(stat = "identity",
                                       show.legend = showLegend,
                                       size = 0.05, width = 0.75) +
    labs(y = paste0("Silhouette width s(i)"),
         caption = paste0("\nOverall average silhouette width: ",
                          round(avswidth, 2)), x = "",
         title = paste0(main))
  silwidths_perclass <- round(sapply(lvls,
                                     function(y) mean(df$si[which(df$label == y)])), 2)
  templabels <- c()
  # ?sprintf
  for (i in seq_len(length(unique(df$class)))) {
    sbar <- sprintf("%1.2f", silwidths_perclass[i])
    templabel <- paste('paste(bar(s), " = ", ',
                       '"', sprintf("%1.2f", silwidths_perclass[i]),'"',
                       ', " ", ', '"', lvls[i], '"', ")")
    templabels <- c(templabels, templabel)
  }
  gg <- gg + scale_fill_manual(values = classCols[ayint],
                               aesthetics = c("colour", "fill"),
                               labels = parse(text = templabels),
                               name = "Classes") +
    theme(legend.text.align = 0)
  if (topdown)
    gg <- gg + coord_flip()
  if (drawLineAtAverage) {
    gg <- gg + geom_hline(yintercept = avswidth, linetype = "dashed",
                          color = "red")
  }
  if (!showCases) {
    if (topdown) {
      gg <- gg + theme(axis.text.y = element_blank(),
                       axis.ticks.y = element_blank())
    }
    else {
      gg <- gg + theme(axis.text.x = element_blank(),
                       axis.ticks.x = element_blank())
    }
  }
  else {
    if (topdown) {
      gg <- gg + theme(axis.text.y = element_text(angle = 0))
    }
    else {
      gg <- gg + theme(axis.text.x = element_text(angle = 90))
    }
  }
  gg <- gg + theme(plot.title = element_text(hjust = 0.5))
  gg <- gg + theme(axis.line.x = element_line(size = 0.25))
  gg <- gg + scale_x_discrete(expand = c(0, 0.013 * length(indsv)))
  gg <- gg + scale_y_continuous(expand = c(0, 0), limits = c(-1, 1))
  gg <- gg + theme(panel.background = element_blank())
  gg <- gg + theme(plot.title.position = "plot")
  gg <- gg + theme(axis.title = element_text(size = 10))
  gg <- gg + theme(plot.caption = element_text(size = 10, hjust = 0,
                                               margin = ggplot2::margin(0, 0, 0, 0)))
  if (summary) {
    ave <- tapply(df$si, df$class, mean)
    n <- tapply(df$class, df$class, length)
    sil.sum <- data.frame(classNumber = names(ave), classLabel = lvls[as.numeric(names(ave))],
                          classSize = n, classAveSi = round(ave, 2), stringsAsFactors = TRUE)
    print(sil.sum, row.names = FALSE)
  }
  return(gg)
}



qresplot <- function(PAC, feat, xlab = NULL, xlim = NULL,
                     main = NULL, identify = FALSE, gray = TRUE,
                     opacity = 1, squareplot = FALSE, plotLoess = FALSE,
                     plotErrorBars = FALSE, plotQuantiles = FALSE,
                     grid = NULL, probs = c(0.5, 0.75),
                     cols = NULL, fac = 1, cex = 1,
                     cex.main = 1.2, cex.lab = 1,
                     cex.axis = 1, pch = 19){
  #
  # Draws a quasi residual plot of PAC versus a data feature
  #
  # Arguments:
  # PAC           : vector with the PAC values of a classification,
  #                 typically the $PAC of a function vcr.*.*()
  # feat          : the PAC will be plotted versus this data
  #                 feature. Note that feat does not have to be
  #                 one of the explanatory variables of the model.
  #                 It can be another variable, a combination of
  #                 variables (like a sum or a principal component
  #                 score), the row number of the cases if they
  #                 were recorded succesively, etc.
  # xlab          : label for the horizontal axis, i.e. the name
  #                 of variable feat.
  # xlim          : limits for the horizontal axis. If NULL, the
  #                 range of feat is used.
  # main          : title for the plot.
  # identify      : if TRUE, left-click on a point to get its number,
  #                 then ESC to exit.
  # gray          : logical, if TRUE (the default) the plot region
  #                 where PAC < 0.5 gets a light gray background.
  #                 Points in this region were classified into
  #                 their given class, and the points above this
  #                 region were misclassified.
  # opacity       : determines opacity of plotted dots.
  #                 Value between 0 and 1, where 0 is transparent
  #                 and 1 is opaque.
  # squareplot    : if TRUE, the horizontal and vertical axis will
  #                 get the same length.
  # plotLoess     : if TRUE, a standard loess curve is fitted and
  #                 superimposed on the plot. May not work well if
  #                 feat is discrete with few values.
  #                 At most one of the options plotLoess,
  #                 plotErrorbars, or plotQuantiles can be selected.
  # plotErrorBars : if TRUE, the average PAC and its standard
  #                 error are computed on the intervals of a grid
  #                 (see option grid). Then a red curve connecting
  #                 the averages is plotted, as well as two blue
  #                 curves corresponding to the average plus or
  #                 minus one standard error. At most one of the
  #                 options plotLoess, plotErrorbars, or
  #                 plotQuantiles can be selected.
  # plotQuantiles : if TRUE, one or more quantiles of the PAC
  #                 are computed on the intervals of a grid
  #                 (see option grid). The quantiles correspond
  #                 the probabilities in option probs.
  #                 Then the curves connecting the quantiles
  #                 are plotted. At most one of the options plotLoess,
  #                 plotErrorbars, or plotQuantiles can be selected.
  # grid          : only used when plotErrorBars or plotQuantiles
  #                 are selected. This is a vector with increasing
  #                 feat values, forming the grid. If NULL, the
  #                 grid consists of the minimum and the maximum
  #                 of feat, with 9 equispaced points between them.
  # probs         : only used when plotQuantiles is selected. This
  #                 is a vector with probabilities determining the
  #                 quantiles. If NULL, defaults to c(0.5, 0.75).
  # cols          : only used when plotquantiles is selected.
  #                 A vector with the colors of the quantile curves.
  #                 If NULL the cols are taken as 2,...
  # fac           : only used when ploTloess, plotErrorBars or
  #                 plotQuantiles are selected. A real number to
  #                 multiply the resulting curves. A value fac > 1
  #                 can be useful to better visualize the curves
  #                 when they would be too close to zero.
  #                 By default (fac=1) this is not done.
  # cex           : passed on to plot().
  # cex.main      : same, for title.
  # cex.lab       : same, for labels on horizontal and vertical axes.
  # cex.axis      : same, for axes.
  # pch           : plot character for the points, defaults to 19.

  # Returns:
  # coordinates  : a matrix with 2 columns containing the
  #                coordinates of the plotted points. This makes it
  #                easier to add text next to interesting points.
  #                If identify = T, the attribute ids of coordinates
  #                contains the row numbers of the identified points
  #                in the matrix coordinates.

  if (!is.vector(PAC)) stop(
    "The PAC on the vertical axis should be a vector.")
  n <- length(PAC)
  if (!is.vector(feat)) stop(
    "The feature on the horizontal axis should be a vector.")
  if (length(feat) != n) stop(paste0("The feature should have ",
                                     "length ", n, ", like PAC."))
  if (sum(c(plotLoess, plotErrorBars, plotQuantiles)) > 1) {
    stop(paste0("Only one of the options plotLoess, ",
                "plotErrorBars, or plotQuantiles can be selected.")) }
  if (is.null(main)) {
    mymain <- "quasi residual plot"
  } else {
    mymain <- main
  }
  if (is.null(xlab)) {
    myxlab <- "feature"
  } else {
    myxlab <- xlab
  }
  xran <- range(feat, na.rm = TRUE)
  myxlim <- xran
  if (!is.null(xlim)) {myxlim <- xlim}
  if (is.null(gray)) {gray <- TRUE}
  myopacity <- 1
  if (!is.null(opacity)) {myopacity <- opacity}
  if (squareplot) {par(pty = "s")} else {par(pty = "m")}
  # The "else" part is still missing in classmap,
  # but there the default is sqauareplot=T anyway.
  if (is.null(fac)) {fac <- 1}
  mycex <- 1
  if (!is.null(cex)) {mycex <- cex}
  mycex.main <- 1.2
  if (!is.null(cex.main)) {mycex.main <- cex.main}
  mycex.lab <- 1
  if (!is.null(cex.lab)) {mycex.lab <- cex.lab}
  mycex.axis <- 1
  if (!is.null(cex.axis)) {mycex.axis <- cex.axis}
  mypch <- 19 # disk (16 is a bit smaller)
  if (!is.null(pch)) {mypch <- pch}
  # Start the actual plot:
  plot(0, 0, type = "n", xlim = myxlim, ylim = c(0, 1),
       ann = FALSE, axes = FALSE)
  # makes an empty plot
  title(main = mymain, line = 1, cex.main = cex.main)
  title(ylab = "P[alternative class]", line = 2.3,
        cex.lab = 1)
  title(xlab = myxlab, line = 2.3, cex.lab = 1)
  if (gray) { # adds gray region
    rect(myxlim[1] - myxlim[2], -2, myxlim[2] * 2, 0.5,
         col = "lightgray", border = FALSE)
  }
  par(new = TRUE) # overlay next plot on this one
  mycol <- adjustcolor(1, alpha.f = myopacity)
  coordinates <- cbind(feat, PAC)
  plot(coordinates, xlab = "", ylab = "", xlim = myxlim,
       ylim = c(0, 1), pch = mypch, col = mycol,
       cex = mycex, cex.axis = mycex.axis)
  if (plotLoess) {# add a loess curve
    lofit <- loess(PAC ~ feat)
    grid <- seq(from = myxlim[1], to = myxlim[2], length.out = 201)
    lines(grid, fac * predict(lofit, grid), col = "red", lwd = 2)
  }
  if (plotErrorBars) { # add average plus/minus standard error
    if (is.null(grid)) grid <-
        seq(from = xran[1], to = xran[2], length.out = 11)
    nintervals <- length(grid) - 1
    midpoints <- avers <- sders <- rep(NA, nintervals)
    for (j in seq_len(nintervals)) {
      midpoints[j] <- (grid[j] + grid[j + 1]) / 2
      clinds <- which((grid[j] <= feat) & (feat < grid[j + 1]))
      avers[j] <- mean(PAC[clinds], na.rm = TRUE)
      sders[j] <- sd(PAC[clinds], na.rm = TRUE) / sqrt(length(clinds))
    }
    lines(midpoints, fac * (avers + sders), col = "blue", lwd = 2)
    lines(midpoints, fac * avers, col = "red", lwd = 2)
    lines(midpoints, fac * (avers - sders), col = "blue", lwd = 2)
  }
  if (plotQuantiles) { # add quantile-based curves
    if (is.null(grid)) grid <-
        seq(from = xran[1], to = xran[2], length.out = 11)
    nintervals <- length(grid) - 1
    midpoints <- rep(NA, nintervals)
    if (is.null(probs)) probs <- c(0.5, 0.75)
    nprobs <- length(probs)
    quants <- matrix(NA, ncol = nintervals, nrow = nprobs)
    for (j in seq_len(nintervals)) {
      midpoints[j] <- (grid[j] + grid[j + 1]) / 2
      clinds <- which((grid[j] <= feat) & (feat < grid[j + 1]))
      quants[, j] <- quantile(PAC[clinds], probs)
    }
    if (is.null(cols)) {
      cols <- seq_len(nprobs) + 1 # red, green, blue,...
    } else {
      if (!is.vector(cols)) {
        stop("\n cols should be a vector")
      }
      if (length(cols) < nprobs) {
        stop(paste0("\n The number of cols should be at",
                    " least length(probs) = ", nprobs, "."))
      }
      cols <- cols[seq_len(nprobs)]
    }
    for (k in seq_len(nprobs)) {
      lines(midpoints, fac * quants[k, ], col = cols[k], lwd = 2)
    }
  }
  attr(coordinates, "ids") <- integer(0)
  if (identify) {
    message("Press the escape key to stop identifying.")
    iout <- identify(coordinates, order = TRUE)
    ids <- iout$ind[order(iout$order)]
    if (length(ids) > 0) {
      write(noquote("Identified point(s): "))
      print(noquote(ids))
      attr(coordinates, "ids") <- ids
    } else {
      write(noquote("No points were identified."))
    }
  }
  invisible(coordinates)
}

---
title: "Random_forest_examples"
author: "Raymaekers, J. and Rousseeuw, P.J."
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Random_forest_examples}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, echo = FALSE}
knitr::opts_chunk$set(
 fig.width  = 5 ,
 fig.height = 3.5,
 fig.align  = 'center'
)
oldpar <- list(mar = par()$mar, mfrow = par()$mfrow)
```

# Introduction

This vignette visualizes classification results from a random forest, using tools from the package.

```{r}
library(randomForest)
library(classmap)
```

# Instagram training data

We use the Instagram data to illustrate the visualization of a
random forest classification. The data is on the identification of genuine/fake (spam) accounts on Instagram. The original data source is:
https://www.kaggle.com/free4ever1/instagram-fake-spammer-genuine-accounts from Bardiya Bakhshandeh.



First we load and inspect the data.

```{r}
data("data_instagram")
traindata <- data_instagram[which(data_instagram$dataType == "train"), -13]
str(traindata)


# The variable names and their interpretation are
colnames(traindata)

# profile.pic: binary, indicates whether profile has picture
# nums.length.username: ratio of number of numerical chars in username to its length
# fullname.words: number of words in full name
# nums.length.fullname: ratio of number of numerical characters in full name to its length
# name..username: binary, indicates whether name == username of the profile
# description.length: length of the description/biography of the profile (in number of characters)
# external.URL: binary, indicates whether profile has external url
# private: binary, indicates whether profile is private or not
# X.posts: number of posts made by profile
# X.followers: number of followers
# X.follows: numbers of follows
# y: whether profile is fake or not.

x_train <- traindata[, -12]
y_train <- traindata[, 12]

dim(traindata)
table(traindata$y) # 50/50 split of genuine/fake accounts:
```


Now we train a random forest. We set the seed as it is not deterministic.

```{r}
set.seed(71) 
rfout <- randomForest(y ~ ., data = traindata, keep.forest = TRUE)
```

Now we create a list called mytype which describes the 
types of the variables in the data. The variables that are not listed will be interval-scaled by default. The Instagram data contains
mostly numeric variables and 4 symmetric binary variables. 

```{r}
mytype <- list(symm = c(1, 5, 7, 8)) 
```

Now we prepare for the visualization of the random forest classification.

```{r}
vcrtrain <- vcr.forest.train(X = x_train, y = y_train,
                            trainfit = rfout, type = mytype)

names(vcrtrain)
vcrtrain$predint[c(1:10, 301:310)] # prediction as integer
vcrtrain$pred[c(1:10, 301:310)]    # prediction as label
vcrtrain$altint[c(1:10, 301:310)]  # alternative label as integer
vcrtrain$altlab[c(1:10, 301:310)]  # alternative label

# Probability of Alternative Class (PAC) of each object:
vcrtrain$PAC[1:3] 
#
summary(vcrtrain$PAC)

# f(i, g) is the distance from case i to class g:
vcrtrain$fig[1:3, ] # for the first 3 objects:

# The farness of an object i is the f(i, g) to its own class: 
vcrtrain$farness[1:3]
#
summary(vcrtrain$farness)

# The "overall farness" of an object is defined as the 
# lowest f(i, g) it has to any class g (including its own):
summary(vcrtrain$ofarness)

sum(vcrtrain$ofarness > 0.99, na.rm = TRUE) 
# With the default cutoff = 0.99 we find 6 outliers,
# also shown in the last column of the confusion matrix:

confmat.vcr(vcrtrain) 

# If we do not want to show the outliers:
confmat.vcr(vcrtrain, showOutliers = FALSE)

# Note that the accuracy is computed before any objects
# are flagged, so it does not depend on the cutoff.
# Here the accuracy is `perfect' due to overfitting. 
# The out-of-box prediction accuracy is about 92%.
cols <- c("blue", "red3")
```

Now we can use the visualization tools from this package.

```{r}
stackedplot(vcrtrain, classCols = cols, main =
              "Instagram training data")

# Silhouette plot:
silplot(vcrtrain, classCols = cols)
# Here all the s(i) are nonnegative (due to overfitting).

# Class maps:
classmap(vcrtrain, "genuine", classCols = cols) #, identify = TRUE)

# farness outliers from furthest to closer: 45, 25, 41
x_train[c(45, 25, 41), ] # they have huge numbers of followers.

classmap(vcrtrain, "fake", classCols = cols) #, identify = TRUE)
# only case 261 is borderline far.
```

The classification of the training data is not very realistic due to overfitting, so let us look at the test data.


# Instagram test data

Now we consider the test data. First we load the data.

```{r}
testdata <- data_instagram[which(data_instagram$dataType == "test"), -13]
Xnew <- testdata[, -12]
ynew <- testdata[, 12]
```

We can now prepare for visualization:

```{r}
vcrtest <- vcr.forest.newdata(Xnew, ynew, vcrtrain)

confmat.vcr(vcrtest)
```

First we visualize using the stacked plot and the silhouette plot:

```{r}
stackedplot(vcrtest, classCols = cols, 
            main = "RF on Instagram test data")

# Silhouette plot:
silplot(vcrtest, classCols = cols, main =
          "Silhouettes of RF on Instagram test data") # now some s(i) are negative
```

Now we make the class maps

```{r}

## Class of genuine accounts:

classmap(vcrtest, "genuine", classCols = cols) #, identify = TRUE)

# one farness outlier:
Xnew[c(30), ]
# has very lengthy bio/description
# has large number of X.posts
# has very large number of followers and follows

# genuine misclassified as fake: from highest PAC to lowest
Xnew[c(21, 29, 51), ] # and 2 more borderline cases
# They have some unusual characteristics for their class:
# * 21, 29 have a (very) high nums.length.username, i.e. the
#   percentage of numerical characters in the username.
# * 21, 29 have a full name of only 1 word.
# * 21, 29 and 51 have description.length = 0, i.e. no 
#   description/biography of their profile.
# * they all have low X.posts (even 0 for case 21), i.e.
#   relatively few previous posts.
# All of these characteristics are more common for fake profiles
# than for genuine profiles, as we can see below:

trcols <- cols[as.numeric(y_train)]

plot(x_train[, 1], col = trcols, main = "profile.pic")
# fakes are less likely to have a profile picture
plot(x_train[, 2], col = trcols, main = "nums.length.username")
# is higher for fakes
plot(x_train[, 3], col = trcols, main = "fullname.words")
# is lower for fakes
plot(x_train[, 4], col = trcols, main = "nums.length.fullname")
# is a bit higher for fakes
plot(x_train[, 5], col = trcols, main = "name..username")
# mostly 0 for genuine; fakes have a few values 1
plot(x_train[, 6], col = trcols, main = "description.length")
# fakes are typically lower, and more often zero
plot(x_train[, 7], col = trcols, main = "external.URL")
# fakes never had them, genuines sometimes did
plot(x_train[, 8], col = trcols, main = "private")
# no visible difference
plot((x_train[, 9])^0.1, col = trcols, main = "X.posts")
# fakes have fewer posts, and often none
plot((x_train[, 10])^0.1, col = trcols, main = "X.followers")
# fakes have fewer followers, sometimes none
plot((x_train[, 11])^0.1, col = trcols, main = "X.follows")
# fakes follow a bit fewer, but the difference is small.


## Class of fake accounts:

classmap(vcrtest, "fake", classCols = cols) #, identify = TRUE)

# Fake identified as genuine, from highest PAC to lower:
# c(27, 51, 34, 23, 58)
Xnew[which(ynew == "fake")[c(27, 34, 51, 23, 58)], ]

# These have a number of characteristics which are more common 
# for genuine profiles:
#
# all have profile pictures
# none have numerical characters in username
# none have numerical characters in fullname
# 27 has a lengthy bio description
# all have a relatively high number of followers
# all have a relatively high number of follows.
```

***

---
title: "Support_vector_machine_examples"
author: "Raymaekers, J. and Rousseeuw, P.J."
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Support_vector_machine_examples}
  %\VignetteEngine{knitr::knitr}
  %\VignetteEncoding{UTF-8}
---

```{r, echo = FALSE}
knitr::opts_chunk$set(
 fig.width  = 5 ,
 fig.height = 3.5,
 fig.align  = 'center'
)
oldpar <- list(mar = par()$mar, mfrow = par()$mfrow)
```

# Introduction

This vignette visualizes classification results from support vector machines, using tools from the package.

```{r}
library("e1071")
library("classmap")
```


# Small toy example

The example from the paper, the Amazon book reviews data, takes a long time to run. As a simpler illustration we first run a small toy example.

The data is artificial and copied verbatim from Section  9.6.2 (page 363) of "An Introduction to Statistical Learning" by G. James, D. Witten, T. Hastie and R. Tibshirani (Springer eBook, eighth printing, 2017).


We first generate the data and inspect it. We see that class 1 is disconnected, and class 2 lies in the middle. In the original space, linear separation is not possible.

```{r}
set.seed(1)
X <- matrix(rnorm(200 * 2), ncol = 2)
X[1:100, ] <- X[1:100, ] + 2
X[101:150, ] <- X[101:150, ] - 2
y <- as.factor(c(rep("blue", 150), rep("red", 50))) # two classes
table(y)
dim(X) # 200 2
length(y) # 200
cols <- c("deepskyblue3", "red")
col <- cols[as.numeric(y)]
plot(X, col = col, pch = 19) 
```


We now fit an SVM with radial basis kernel to the data. We fix the seed in order
to make the result of svm() reproducible:
```{r}
dat <- data.frame(X = X, y = y)
set.seed(1) 
svmfit <- svm(y~., data = dat, scale = FALSE, kernel = "radial", cost = 10, gamma = 1, probability = TRUE)
```

Now we inspect the fitted svm. We also plot the decision values and the fit. We see that the decision value separates the classes reasonably well. The decision boundary is far from linear, but it is linear in the feature space:

```{r}
names(svmfit)

plot(svmfit$decision.values, col = cols[as.numeric(y)]); abline(h = 0)

plot(svmfit, data = dat, X.2~X.1,  col = cols)
```

Now we build the vcr object for the trained svm:

```{r}
vcr.train <- vcr.svm.train(X, y, svfit = svmfit)
```

We inspect the output, which contains among other things the prediction as integer, the prediction as label, the alternative label as integer and the alternative label:

```{r}
names(vcr.train)
vcr.train$predint 
vcr.train$pred[c(1:20, 151:170)]
vcr.train$altint  
vcr.train$altlab[c(1:20, 151:170)]
```

Now inspect the Probability of Alternative Class (PAC) of each object, which
is one of the ingredients for the class map:

```{r}
vcr.train$PAC[1:3] 
summary(vcr.train$PAC)
```

The $fig element of the output contains the distance from case i to class g. Let's look at it for the first 5 objects:

```{r}
vcr.train$fig[1:5, ] 
```
From the fig, the farness of each object can be computed. The farness of an object i is the f(i, g) to its own class: 

```{r}
vcr.train$farness[1:5]
summary(vcr.train$farness)
```


The "overall farness" of an object is defined as the lowest f(i, g) it has to any class g (including its own):

```{r}
summary(vcr.train$ofarness)
```


Objects with ofarness > cutoff are flagged as "outliers". These can be included in a separate column in the confusion matrix. This confusion matrix can be computed using confmat.vcr, which also returns the accuracy. To illustrate this we choose a rather low cutoff, so a few cases are flagged:

```{r}
confmat.vcr(vcr.train, cutoff = 0.975)
```

With the default cutoff = 0.99 no objects are flagged in this example:

```{r}
confmat.vcr(vcr.train)
```

Note that the accuracy is computed before any objects are flagged, so it does not depend on the cutoff.

The confusion matrix can also be constructed showing class numbers instead of labels. This option can be useful for long level names.

```{r}
confmat.vcr(vcr.train, showClassNumbers = TRUE)
```

A stacked mosaic plot made with the stackedplot() function can be used to visualize the confusion matrix. The outliers, if there are any, appear as grey areas on top.

```{r}
stackedplot(vcr.train, classCols = cols, separSize = 1.5,
            minSize = 1, showLegend = TRUE, cutoff = 0.975,
            main = "Stacked plot of SVM on artifical data")

stplot <- stackedplot(vcr.train, classCols = cols, 
                     separSize = 1.5, minSize = 1,
                     main = "Stacked plot of SVM on artifical data")
stackedplot(vcr.train, classCols = cols)
```

The silhouette plot is made below:
```{r}
# pdf("Artif_SVM_silhouettes.pdf", width = 5.0, height = 4.6)
silplot(vcr.train, classCols = cols, 
        main = "Silhouette plot of SVM on artificial data") 
# dev.off()
```

We now make the class maps based on the vcr object. This can be done using the classmap() function. We make a separate class map for each of the two classes. We see that for class 1 (blue), most of the points have a low PAC, with some
exceptions that are assigned to the red class. For class 2, the PAC is typically higher than in class 1.

```{r}
par(mfrow = c(1,1))
classmap(vcr.train, 1, classCols = cols)
classmap(vcr.train, 2, classCols = cols)
par(oldpar)
```
***

To illustrate the use of new data we create a fake dataset which is a subset of the training data, where not all classes occur, and ynew has NA's.


```{r}
Xnew <- X[c(1:25, 151:175), ]; dim(Xnew) # 50 200
ynew <- y
ynew[c(21:25, 171:175)] <- NA
(ynew <- ynew[c(1:25, 151:175)]) 
length(ynew) # 50
```


Now we build the vcr object on the training data.

```{r}
vcr.test <- vcr.svm.newdata(Xnew, ynew, vcr.train)
```


Inspect some of the output in the vcr object:

```{r}
vcr.test$predint 
length(vcr.test$predint) 
vcr.test$altint 
length(vcr.test$altint) 
vcr.test$PAC 
length(vcr.test$PAC) 
vcr.test$farness 
length(vcr.test$farness) 
vcr.test$ofarness # This exists even for cases whose true label is NA:
length(vcr.test$ofarness) 
```

Now perform some check to confirm that it corresponds with what we would expect, given that the test data is a subset of the training data:

```{r}  
plot(vcr.test$yintnew[c(1:20, 26:45)],
     vcr.train$yint[c(1:20, 151:170)]); abline(0, 1)
plot(vcr.test$predint[c(1:20, 26:45)],
     vcr.train$predint[c(1:20, 151:170)]); abline(0, 1)
plot(vcr.test$altint[c(1:20, 26:45)],
     vcr.train$altint[c(1:20, 151:170)]); abline(0, 1)
plot(vcr.test$PAC[c(1:20, 26:45)],
     vcr.train$PAC[c(1:20, 151:170)]); abline(0, 1)
plot(as.vector(vcr.test$fig[c(1:20, 26:45), ]),
     as.vector(vcr.train$fig[c(1:20, 151:170), ])); abline(0, 1)
plot(vcr.test$farness[c(1:20, 26:45)],
     vcr.train$farness[c(1:20, 151:170)]); abline(0, 1)
plot(vcr.test$ofarness[c(1:20, 26:45)],
     vcr.train$ofarness[c(1:20, 151:170)]); abline(0, 1)
```

We also make the silhouette plot on the artificial subset:

```{r}
# pdf("Artif_subset_SVM_silhouettes.pdf", width = 5.0, height = 4.6)
silplot(vcr.test, classCols = cols, 
        main = "Silhouette plot of SVM on subset of data")  
# dev.off()
```

Now we redo the analysis in feature space. What do these data look like in feature space? In order to answer the question, we first compute the gaussian kernel used in svmfit:

```{r}  
Kxx <- makeKernel(X, svfit = svmfit)
dim(Kxx) 
```

Now we can compute a representation of these points in feature space. The function MakeFV() takes a kernel, and from it creates a feature space with the desired inner products. The kernel matrix has a few eigenvalues below the precision precS, and this is taken into account:

```{r} 
outFV <- makeFV(Kxx)
Xf <- outFV$Xf 
```

As a sanity check, we make sure that the kernel matrix and
the inner products of Xf match (almost exactly):

```{r} 
max(abs(as.vector(Kxx - Xf %*% t(Xf)))) # tiny, this is good.
```


The rows of Xf are in 196-dimensional space, do we need all of those dimensions? It turns out that we do need many. More precisely, we need 39 components to explain 95% of the variance. That's many when starting from bivariate data.

```{r} 
explvar <- cumsum(apply(Xf, 2, var))
plot(explvar / explvar[ncol(Xf)]); abline(h = 0.95)
min(which(explvar > 0.95 * explvar[ncol(Xf)])) 
```

We inspect the data in feature space a little bit more.
We see that all points in Xf lie on the unit sphere. This is always true for the radial basis kernel.

```{r} 
range(rowSums(Xf^2)) 
```

In some of the pairwise plots, we can also see spherical effects. The data does look more separable here than in the original two-dimensional space.

```{r} 
pairs(Xf[, 1:5], col = col)
plot(Xf[, 1], Xf[, 5], col = col, pch = 19)
```


Now we run SVM on Xf with **linear** kernel and the same cost. This yields the same result as above:

```{r}
Xfdat <- data.frame(X = Xf, y = y)
set.seed(1)
svmfit1 <- svm(y~., data = Xfdat, scale = FALSE, kernel = "linear", cost = 10, probability = TRUE)
plot(svmfit1$decision.values, svmfit$decision.values); abline(0, 1)
vcr.trainf <- vcr.svm.train(X = Xf, y, svfit = svmfit1)
```

We see that everything matches with the original svm which was trained using a gaussian kernel on the original data:

```{r}
plot(vcr.train$yint, vcr.trainf$yint); abline(0, 1)
plot(vcr.train$predint, vcr.trainf$predint); abline(0, 1)
plot(vcr.train$altint, vcr.trainf$altint); abline(0, 1)
plot(vcr.train$PAC, vcr.trainf$PAC); abline(0, 1)
plot(vcr.train$fig, vcr.trainf$fig); abline(0, 1)
vcr.train$figparams$farscales
plot(vcr.train$farness, vcr.trainf$farness); abline(0, 1)
plot(vcr.train$ofarness, vcr.trainf$ofarness); abline(0, 1)
```

We now make a test data with NA's in ynew, but in feature space:

```{r}
Xnew <- Xf[c(1:25, 151:175), ]; dim(Xnew) 
```

Using this test data, we build a new vcr object:

```{r}
vcr.testf <- vcr.svm.newdata(Xnew, ynew, vcr.trainf)
```

Again, we see an exact match with original version:

```{r}
plot(vcr.testf$yintnew, vcr.test$yintnew); abline(0, 1)
plot(vcr.testf$predint, vcr.test$predint); abline(0, 1)
plot(vcr.testf$altint, vcr.test$altint); abline(0, 1)
plot(vcr.testf$PAC, vcr.test$PAC); abline(0, 1)
plot(vcr.testf$fig, vcr.test$fig); abline(0, 1)
plot(vcr.testf$farness, vcr.test$farness); abline(0, 1)
```
***



# Amazon book review data


```{r}
data(data_bookReviews)
X <- data_bookReviews[1:500, 1]; y <- data_bookReviews[1:500, 2] 
Xnew  <- data_bookReviews[501:1000, 1]; ynew <- data_bookReviews[501:1000, 2] 
length(X); length(Xnew); length(y); length(ynew) 
```

This is a subset of the data used in the paper, which
was assembled by Prettenhofer and Stein (2010), see
https://arxiv.org/abs/1008.0716

The full dataset has been used for a variety of things, including classification using svm.
A list of citations:
https://scholar.google.com/scholar?hl=en&as_sdt=0%2C5&
  sciodt=0%2C5&cites=4738650459241064524&scipsc=1&q=svm+&btnG=

The subset was chosen small enough to keep the computation
time low, while still containing the examples in the paper.

First we define the colors and inspect an example of a book review:

```{r}
cols <- c("deepskyblue3", "red")
col <- cols[as.numeric(y)]
X[5]
```

The next question is what kernel to use. There are many string kernels. 
kernlab provides several options:
The "spectrum" kernel matches substrings of an exact size.
The "boundrange" kernel matches substrings of a length up 
  to a given size.
The "constant" kernel matches all substrings.
The "exponential" kernel matches all substrings but gives 
  different weights depending on the length of the matched
  string.

The best kernel for these data in terms of accuracy turned 
out to be the spectrum kernel with length 7:

```{r}
strdot <- kernlab::stringdot(type = "spectrum", length = 7)
strdot
ptm <- proc.time()
Kr_train <- kernlab::kernelMatrix(kernel = strdot, x = X)
(proc.time() - ptm)[3] 
dim(Kr_train) 
```

Using the makeFV() function, we can compute the corresponding points in feature space. It takes into account almost-zero eigenvalues. The resulting Xf can be seen as the scores of a kernel PCA without centering
(mean subtraction).

```{r}
outFV <- makeFV(Kr_train)
Xf <- outFV$Xf 
dim(Xf) 
```


To see how well these feature vectors and Kr_train match, we can compute the inner products in Xf, and verify that they are close to Kr_train:

```{r}
inprod <- Xf %*% t(Xf) 
max(abs(as.vector(Kr_train - inprod))) 
```

We can also verify that the points of Xf lie on the unit sphere in 500 dimensions:

```{r}
range(rowSums(Xf^2)) # all 1
```


By construction, Xf is expressed in an orthogonal coordinate system.
In order to explain 95% of the variance in the data, we need 463 out of the 500 components, so we are in a truly high dimensional situation:

```{r}
explvar <- cumsum(apply(Xf, 2, var))
plot(explvar / explvar[ncol(Xf)]); abline(h = 0.95)
min(which(explvar > 0.95 * explvar[ncol(Xf)]))
```


Now we can train the support vector classifier on these data.
On Xf we train the SVM with the LINEAR kernel, since
we are in feature space:

```{r}
Xfdat <- data.frame(X = Xf, y = y)
set.seed(1) # we have to fix the seed!
svmfit <- svm(y~., data = Xfdat, scale = FALSE, kernel = "linear", cost = 2, probability = TRUE)
```

We plot the decision values. the two horizontal lines
contain the support vectors. There are a lot of them,
since the dimension is high.

```{r}
plot(svmfit$decision.values) 
```

Now create the VCR object of the training data:

```{r}
vcr.train <- vcr.svm.train(Xf, y, svfit = svmfit)
```

Using the VCR object, we can build the confusion matrix. We see that the accuracy is 100%.
This classification is perfect due to overfitting!

```{r}
confmat.vcr(vcr.train, showOutliers = FALSE)
```

The classification on the test data is more interesting.
First construct the kernel matrix we need to classify the test data, which takes roughly a minute:

```{r}
ptm <- proc.time()
Kr_test <- kernlab::kernelMatrix(kernel = strdot, x = Xnew, y = X)
(proc.time() - ptm)[3] 
```

Now we compute feature vectors of the test set. These features live in the same space as the Xf of the training data.

```{r}
FVtest <- makeFV(kmat = Kr_test, transfmat = outFV$transfmat)
Zf <- FVtest$Xf 
dim(Zf) 
```

We compare the inner products with the kernel matrix to make sure that they are the same:

```{r}
inprod <- Zf %*% t(Xf) 
max(abs(as.vector(Kr_test - inprod))) # tiny, good.
```

Now create the VCR object of the test data:

```{r}
vcr.test <- vcr.svm.newdata(Zf, ynew, vcr.train)
```

Compute the confusion matrix from the VCR object.  In the full dataset, the accuracy was 82 %. Here we achieve roughly 74% accuracy, which is lower since training was done on 500 texts instead of 2000.

```{r}
confmat.vcr(vcr.test)
```

The stacked mosaic plot of the classification is presented below:

```{r}
stackedplot(vcr.test, classCols = cols, separSize = 1.5, minSize = 1.5, main = "SVM on the book review test data")
```

The silhouette plot of the classification is presented below:

```{r}
# pdf("Bookreview_test_SVM_silhouettes.pdf", width = 5.0, height = 4.6)
silplot(vcr.test, classCols = cols, main = 
           "Silhouette plot of SVM on book review test data")      
# dev.off()
```


The overall average silhouette width is quite poor, as was the classification accuracy. Now we build the class maps for the test set.
First the class map of the negative reviews:

```{r}
# To identify the atypical points, uncomment the next line:
# classmap(vcr.test, 1, classCols = cols, identify = TRUE)
# Identified point(s): 
# [1]  67 127 118 145  94  45
#
# pdf("Bookreviews_500_classmap_test_negative.pdf")
par(mar = c(3.5, 3.5, 2.0, 0.2))
coords <- classmap(vcr.test, 1, classCols = cols, cex = 1.5,
                  main = "predictions of the negative book reviews", cex.lab = 1.5, cex.axis = 1.5, cex.main = 1.5, maxfactor = 1.05, identify = FALSE)
indstomark <- c(67, 127, 118, 145, 94, 45) # from identify=TRUE above
labs  <- c("a", "b", "x", "c", "d", "e")
xvals <- coords[indstomark, 1] +
   c(-0.08, 0, 0, 0.09, 0.09, 0.09) # visual fine tuning
yvals <- coords[indstomark, 2] +
   c(0, 0.033, 0.030, 0, 0, 0)
text(x = xvals, y = yvals, labels = labs, cex = 1.5)
legend("topright", fill = cols,
       legend = c("negative", "positive"), cex = 1.4,
       ncol = 1, bg = "white")
# dev.off()
par(oldpar)
```

We marked the following points:

```{r}
(Xnew[which(ynew == 1)])[67] # (a) in the paper.

(Xnew[which(ynew == 1)])[127] # (b) in the paper.

(Xnew[which(ynew == 1)])[118] # (x) 
# Negative review, but highly recommends another book!

(Xnew[which(ynew == 1)])[145] # (c) in the paper.
# Contains some grammatical errors
# Is negative, but expressed very indirectly.

(Xnew[which(ynew == 1)])[94] # (d) in the paper. 
# Definitely negative, but quite short.

(Xnew[which(ynew == 1)])[45] # (e) in the paper.
```

Now the class map of the positive reviews in the test set:

```{r}
#
# To identify the atypical points:
# classmap(vcr.test, 2, classCols = cols, identify = TRUE)
# Identified point(s): 
# [1] 48  32 18 168 113
#
# pdf("Bookreviews_500_classmap_test_positive.pdf")
par(mar = c(3.5, 3.5, 2.0, 0.2))
coords <- classmap(vcr.test, 2, classCols = cols, cex = 1.5,
                  main = "predictions of the positive book reviews", cex.lab = 1.5, cex.axis = 1.5, cex.main = 1.5,
                  maxfactor = 1.05, identify = FALSE)
indstomark <- c(48, 32, 18, 168, 113) # from identify = TRUE above
coords[indstomark, ]
labs  <- letters[6:10]
xvals <- coords[indstomark, 1] +
   c(0, 0, 0.09, 0.09, 0.09) # visual fine tuning
yvals <- coords[indstomark, 2] +
   c(0.032, 0.032, 0, 0, 0)
text(x = xvals, y = yvals, labels = labs, cex = 1.5)
legend("left", fill = cols,
       legend = c("negative", "positive"), cex = 1.4,
       ncol = 1, bg = "white")
# dev.off()
par(oldpar)
```

We marked the following points:

```{r}
(Xnew[which(ynew == 2)])[48] # (f) in the paper.

(Xnew[which(ynew == 2)])[32] # (g) in the paper.

(Xnew[which(ynew == 2)])[18] # (h) in the paper.

(Xnew[which(ynew == 2)])[168] # (i) in the paper.

(Xnew[which(ynew == 2)])[113] # (j) in the paper.
# very short (and spelling error).
```

# Sweets data

We first load the nutrients data set out of the robCompositions package.

```{r}
# Load the entire nutrients_branded data set:
library("robCompositions")
data(nutrients_branded, package = "robCompositions")

X <- as.matrix(nutrients_branded)
colnames(X)
```

We now preprocess the data, removing uninformative variables and those with many NAs:

```{r}
X            <- X[, -c(1:7, 10, 18)]
productnames <- nutrients_branded$name_D
hasNA        <- apply(X, 1, function(y) sum(is.na(y)) > 0)
X            <- X[-which(hasNA), ]
productnames <- productnames[-which(hasNA)]
```

There are a lot of food types in this dataset. We will select a subset.

```{r}
type  <- nutrients_branded$category_E 
type  <- type[-which(hasNA)]
inds1 <- which(type == "Sweets/Cookies/Biscuits")
inds2 <- which(type == "Sweets/Milk based ice cream")
inds3 <- which(type == "Sweets/Cakes and tarts")
inds4 <- which(type == "Sweets/Creams and puddings")

X <- X[c(inds1, inds2, inds3, inds4), ]
X <- as.matrix(apply(X, 2, as.numeric))
prod.names <- productnames[c(inds1, inds2, inds3, inds4)]
X <- as.matrix(scale(X))
head(X)

y <- factor(c(rep("biscuits", length(inds1)),
             rep("ice cream", length(inds2)),
             rep("cakes", length(inds3)),
             rep("puddings", length(inds4))),
            levels = c("biscuits", "ice cream", 
                      "cakes","puddings"))

table(y)
```

We now proceed by training the multiclass SVM, and constructing the VCR object:

```{r}
Xdat <- data.frame(X = X, y = y)
set.seed(1) # needed, otherwise not deterministic.
svmfit <- svm(y~., data = Xdat, scale = FALSE,
              kernel = "linear", cost = 10,
              probability = TRUE)
names(svmfit)

vcrobj <- vcr.svm.train(X, y, svfit = svmfit)
```

Based on the output of vcr.svm.train, we can now analyze and visualize the classification results. First we look at the confusion matrix and the stacked plot.


```{r}
confmat.vcr(vcrobj, showOutliers = FALSE)


cols <- c("firebrick4", "violet", "chocolate3", "orange")

# Stacked plot in the paper:
# pdf("Sweets_stackplot_with_outliers.pdf",width=5,height=4.6)
stackedplot(vcrobj, classCols = cols, separSize = 0.5, minSize = 1.5,
            main = "Stacked plot of SVM on the sweets data",
            htitle = "given class", vtitle = "predicted class")
# dev.off()

# With legend:
stackedplot(vcrobj, classCols = cols, separSize = 0.5, minSize = 1.5,
            main = "Stacked plot of SVM on the sweets data",
            htitle = "given class", vtitle = "predicted class",
            showLegend = TRUE)

# Without showing outliers:
stackedplot(vcrobj, classCols = cols, separSize = 0.5, minSize = 1.5,
            main = "Stacked plot of SVM on the sweets data",
            showOutliers = FALSE)
```

Now we make the silhouette plot:


```{r}
# pdf("Sweets_SVM_silhouettes.pdf", width=5.0, height=4.6)
silplot(vcrobj, classCols = cols,
        main = "Silhouette plot of SVM on the sweets data")
# dev.off()
```

And finally, the class maps:

```{r}
labels <- c("biscuits", "ice cream", "cakes", "puddings")
par(mfrow = c(1, 1))

# To identify special points in class 1:
# classmap(vcrobj, 1, classCols = cols, identify=T)
# Press the escape key to stop identifying.
indstomark <- c(122, 1, 184)
#
mymaxf <- 1.05
coords <- classmap(vcrobj, 1,classCols = cols,
         main = paste0("predictions of ", labels[1]),
         cex = 1.5, cex.lab = 1.5, cex.axis = 1.5,
         cex.main = 1.5, maxfactor = mymaxf)
coords[indstomark, ]
labs  <- c("a", "b", "c")
xvals <- c(2.380, 2.73, 4)
yvals <- c(0.825, 0.105, 0.065)
text(x = xvals, y = yvals, labels = labs, cex = 1.5)
legend("right", fill = cols, legend = labels,
       cex = 1.15, ncol = 1, bg = "white")

# Analogously for classes 2, 3, 4.

# Figure in the paper with all four class maps together:
#
labels <- c("biscuits", "ice cream", "cakes", "puddings")
# pdf(file="Sweets_all_class_maps.pdf",width=9,height=9)
par(mfrow = c(2, 2)) # (nrows,ncols)
par(mar = c(3.6, 3.5, 2.8, 1.0))
mymaxf <- 1.05
classmap(vcrobj, 1, classCols = cols,
         main = paste0("predictions of ", labels[1]),
         cex = 1.5, cex.lab = 1.5, cex.axis = 1.5,
         cex.main = 1.5, maxfactor = mymaxf)
labs  <- c("a", "b", "c") # points 122, 1, 184
xvals <- c(2.380, 2.73, 4)
yvals <- c(0.825, 0.105, 0.065)
text(x = xvals, y = yvals, labels = labs, cex = 1.5)
legend("right", fill = cols, legend = labels,
       cex = 1.15, ncol = 1, bg = "white")
#
par(mar = c(3.6, 2.0, 2.8, 0.3))
classmap(vcrobj, 2,classCols = cols,
         main = paste0("predictions of ", labels[2]),
         cex = 1.5, cex.lab = 1.5, cex.axis = 1.5,
         cex.main = 1.5, maxfactor = mymaxf)
labs  <- c("d", "e", "f", "g", "h-l")
xvals <- c(2.73, 3.21, 3.86, 3.83, 3.9)
yvals <- c(0.78, 0.96, 0.67, 0.57, 0.16)
text(x = xvals, y = yvals, labels = labs, cex = 1.5)
legend("topleft", fill = cols, legend = labels,
       cex = 1.15, ncol = 1, bg = "white")
#
par(mar = c(3.6, 3.5, 2.8, 1.0))
classmap(vcrobj, 3, classCols = cols,
         main = paste0("predictions of ", labels[3]),
         cex = 1.5, cex.lab = 1.5, cex.axis = 1.5,
         cex.main = 1.5, maxfactor = mymaxf)
labs  <- c("m", "n", "o", "p")
xvals <- c(1.05, 3.12, 2.86, 4)
yvals <- c(0.893, 0.95, 0.24, 0.09)
text(x = xvals, y = yvals, labels = labs, cex = 1.5)
legend("right", fill = cols, legend = labels,
       cex = 1.15, ncol = 1, bg = "white")
#
par(mar = c(3.6, 2.0, 2.8, 0.3))
classmap(vcrobj, 4,classCols = cols,
         main = paste0("predictions of ", labels[4]),
         cex = 1.5, cex.lab = 1.5, cex.axis = 1.5,
         cex.main = 1.5, maxfactor = mymaxf)
labs  <- c("q", "r", "s-u", "v", "w", "x")
xvals <- c(3.00, 3.6, 4.1, 3.82, 4.04, 3.9)
yvals <- c(0.95, 0.97, 0.926, 0.875, 0.795, 0.588)
text(x = xvals, y = yvals, labels = labs, cex = 1.5)
legend("topleft", fill = cols, legend = labels,
       cex = 1.16, ncol = 1, bg = "white")
# dev.off()
```

***

---
title: "Support_vector_machine_examples"
author: "Raymaekers, J. and Rousseeuw, P.J."
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Support_vector_machine_examples}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, echo = FALSE}
knitr::opts_chunk$set(
 fig.width  = 5 ,
 fig.height = 3.5,
 fig.align  = 'center'
)
oldpar <- list(mar = par()$mar, mfrow = par()$mfrow)
```

# Introduction

This vignette visualizes classification results from support vector machines, using tools from the package.

```{r}
library("e1071")
library("classmap")
```


# Small toy example

The example from the paper, the Amazon book reviews data, takes a long time to run. As a simpler illustration we first run a small toy example.

The data is artificial and copied verbatim from Section  9.6.2 (page 363) of "An Introduction to Statistical Learning" by G. James, D. Witten, T. Hastie and R. Tibshirani (Springer eBook, eighth printing, 2017).


We first generate the data and inspect it. We see that class 1 is disconnected, and class 2 lies in the middle. In the original space, linear separation is not possible.

```{r}
set.seed(1)
X <- matrix(rnorm(200 * 2), ncol = 2)
X[1:100, ] <- X[1:100, ] + 2
X[101:150, ] <- X[101:150, ] - 2
y <- as.factor(c(rep("blue", 150), rep("red", 50))) # two classes
table(y)
dim(X) # 200 2
length(y) # 200
cols <- c("deepskyblue3", "red")
col <- cols[as.numeric(y)]
plot(X, col = col, pch = 19) 
```


We now fit an SVM with radial basis kernel to the data. We fix the seed in order
to make the result of svm() reproducible:
```{r}
dat <- data.frame(X = X, y = y)
set.seed(1) 
svmfit <- svm(y~., data = dat, scale = FALSE, kernel = "radial", cost = 10, gamma = 1, probability = TRUE)
```

Now we inspect the fitted svm. We also plot the decision values and the fit. We see that the decision value separates the classes reasonably well. The decision boundary is far from linear, but it is linear in the feature space:

```{r}
names(svmfit)

plot(svmfit$decision.values, col = cols[as.numeric(y)]); abline(h = 0)

plot(svmfit, data = dat, X.2~X.1,  col = cols)
```

Now we build the vcr object for the trained svm:

```{r}
vcr.train <- vcr.svm.train(X, y, svfit = svmfit)
```

We inspect the output, which contains among other things the prediction as integer, the prediction as label, the alternative label as integer and the alternative label:

```{r}
names(vcr.train)
vcr.train$predint 
vcr.train$pred[c(1:20, 151:170)]
vcr.train$altint  
vcr.train$altlab[c(1:20, 151:170)]
```

Now inspect the Probability of Alternative Class (PAC) of each object, which
is one of the ingredients for the class map:

```{r}
vcr.train$PAC[1:3] 
summary(vcr.train$PAC)
```

The $fig element of the output contains the distance from case i to class g. Let's look at it for the first 5 objects:

```{r}
vcr.train$fig[1:5, ] 
```
From the fig, the farness of each object can be computed. The farness of an object i is the f(i, g) to its own class: 

```{r}
vcr.train$farness[1:5]
summary(vcr.train$farness)
```


The "overall farness" of an object is defined as the lowest f(i, g) it has to any class g (including its own):

```{r}
summary(vcr.train$ofarness)
```


Objects with ofarness > cutoff are flagged as "outliers". These can be included in a separate column in the confusion matrix. This confusion matrix can be computed using confmat.vcr, which also returns the accuracy. To illustrate this we choose a rather low cutoff, so a few cases are flagged:

```{r}
confmat.vcr(vcr.train, cutoff = 0.975)
```

With the default cutoff = 0.99 no objects are flagged in this example:

```{r}
confmat.vcr(vcr.train)
```

Note that the accuracy is computed before any objects are flagged, so it does not depend on the cutoff.

The confusion matrix can also be constructed showing class numbers instead of labels. This option can be useful for long level names.

```{r}
confmat.vcr(vcr.train, showClassNumbers = TRUE)
```

A stacked mosaic plot made with the stackedplot() function can be used to visualize the confusion matrix. The outliers, if there are any, appear as grey areas on top.

```{r}
stackedplot(vcr.train, classCols = cols, separSize = 1.5,
            minSize = 1, showLegend = TRUE, cutoff = 0.975,
            main = "Stacked plot of SVM on artifical data")

stplot <- stackedplot(vcr.train, classCols = cols, 
                     separSize = 1.5, minSize = 1,
                     main = "Stacked plot of SVM on artifical data")
stackedplot(vcr.train, classCols = cols)
```

The silhouette plot is made below:
```{r}
# pdf("Artif_SVM_silhouettes.pdf", width = 5.0, height = 4.6)
silplot(vcr.train, classCols = cols, 
        main = "Silhouette plot of SVM on artificial data") 
# dev.off()
```

We now make the class maps based on the vcr object. This can be done using the classmap() function. We make a separate class map for each of the two classes. We see that for class 1 (blue), most of the points have a low PAC, with some
exceptions that are assigned to the red class. For class 2, the PAC is typically higher than in class 1.

```{r}
par(mfrow = c(1,1))
classmap(vcr.train, 1, classCols = cols)
classmap(vcr.train, 2, classCols = cols)
par(oldpar)
```
***

To illustrate the use of new data we create a fake dataset which is a subset of the training data, where not all classes occur, and ynew has NA's.


```{r}
Xnew <- X[c(1:25, 151:175), ]; dim(Xnew) # 50 200
ynew <- y
ynew[c(21:25, 171:175)] <- NA
(ynew <- ynew[c(1:25, 151:175)]) 
length(ynew) # 50
```


Now we build the vcr object on the training data.

```{r}
vcr.test <- vcr.svm.newdata(Xnew, ynew, vcr.train)
```


Inspect some of the output in the vcr object:

```{r}
vcr.test$predint 
length(vcr.test$predint) 
vcr.test$altint 
length(vcr.test$altint) 
vcr.test$PAC 
length(vcr.test$PAC) 
vcr.test$farness 
length(vcr.test$farness) 
vcr.test$ofarness # This exists even for cases whose true label is NA:
length(vcr.test$ofarness) 
```

Now perform some check to confirm that it corresponds with what we would expect, given that the test data is a subset of the training data:

```{r}  
plot(vcr.test$yintnew[c(1:20, 26:45)],
     vcr.train$yint[c(1:20, 151:170)]); abline(0, 1)
plot(vcr.test$predint[c(1:20, 26:45)],
     vcr.train$predint[c(1:20, 151:170)]); abline(0, 1)
plot(vcr.test$altint[c(1:20, 26:45)],
     vcr.train$altint[c(1:20, 151:170)]); abline(0, 1)
plot(vcr.test$PAC[c(1:20, 26:45)],
     vcr.train$PAC[c(1:20, 151:170)]); abline(0, 1)
plot(as.vector(vcr.test$fig[c(1:20, 26:45), ]),
     as.vector(vcr.train$fig[c(1:20, 151:170), ])); abline(0, 1)
plot(vcr.test$farness[c(1:20, 26:45)],
     vcr.train$farness[c(1:20, 151:170)]); abline(0, 1)
plot(vcr.test$ofarness[c(1:20, 26:45)],
     vcr.train$ofarness[c(1:20, 151:170)]); abline(0, 1)
```

We also make the silhouette plot on the artificial subset:

```{r}
# pdf("Artif_subset_SVM_silhouettes.pdf", width = 5.0, height = 4.6)
silplot(vcr.test, classCols = cols, 
        main = "Silhouette plot of SVM on subset of data")  
# dev.off()
```

Now we redo the analysis in feature space. What do these data look like in feature space? In order to answer the question, we first compute the gaussian kernel used in svmfit:

```{r}  
Kxx <- makeKernel(X, svfit = svmfit)
dim(Kxx) 
```

Now we can compute a representation of these points in feature space. The function MakeFV() takes a kernel, and from it creates a feature space with the desired inner products. The kernel matrix has a few eigenvalues below the precision precS, and this is taken into account:

```{r} 
outFV <- makeFV(Kxx)
Xf <- outFV$Xf 
```

As a sanity check, we make sure that the kernel matrix and
the inner products of Xf match (almost exactly):

```{r} 
max(abs(as.vector(Kxx - Xf %*% t(Xf)))) # tiny, this is good.
```


The rows of Xf are in 196-dimensional space, do we need all of those dimensions? It turns out that we do need many. More precisely, we need 39 components to explain 95% of the variance. That's many when starting from bivariate data.

```{r} 
explvar <- cumsum(apply(Xf, 2, var))
plot(explvar / explvar[ncol(Xf)]); abline(h = 0.95)
min(which(explvar > 0.95 * explvar[ncol(Xf)])) 
```

We inspect the data in feature space a little bit more.
We see that all points in Xf lie on the unit sphere. This is always true for the radial basis kernel.

```{r} 
range(rowSums(Xf^2)) 
```

In some of the pairwise plots, we can also see spherical effects. The data does look more separable here than in the original two-dimensional space.

```{r} 
pairs(Xf[, 1:5], col = col)
plot(Xf[, 1], Xf[, 5], col = col, pch = 19)
```


Now we run SVM on Xf with **linear** kernel and the same cost. This yields the same result as above:

```{r}
Xfdat <- data.frame(X = Xf, y = y)
set.seed(1)
svmfit1 <- svm(y~., data = Xfdat, scale = FALSE, kernel = "linear", cost = 10, probability = TRUE)
plot(svmfit1$decision.values, svmfit$decision.values); abline(0, 1)
vcr.trainf <- vcr.svm.train(X = Xf, y, svfit = svmfit1)
```

We see that everything matches with the original svm which was trained using a gaussian kernel on the original data:

```{r}
plot(vcr.train$yint, vcr.trainf$yint); abline(0, 1)
plot(vcr.train$predint, vcr.trainf$predint); abline(0, 1)
plot(vcr.train$altint, vcr.trainf$altint); abline(0, 1)
plot(vcr.train$PAC, vcr.trainf$PAC); abline(0, 1)
plot(vcr.train$fig, vcr.trainf$fig); abline(0, 1)
vcr.train$figparams$farscales
plot(vcr.train$farness, vcr.trainf$farness); abline(0, 1)
plot(vcr.train$ofarness, vcr.trainf$ofarness); abline(0, 1)
```

We now make a test data with NA's in ynew, but in feature space:

```{r}
Xnew <- Xf[c(1:25, 151:175), ]; dim(Xnew) 
```

Using this test data, we build a new vcr object:

```{r}
vcr.testf <- vcr.svm.newdata(Xnew, ynew, vcr.trainf)
```

Again, we see an exact match with original version:

```{r}
plot(vcr.testf$yintnew, vcr.test$yintnew); abline(0, 1)
plot(vcr.testf$predint, vcr.test$predint); abline(0, 1)
plot(vcr.testf$altint, vcr.test$altint); abline(0, 1)
plot(vcr.testf$PAC, vcr.test$PAC); abline(0, 1)
plot(vcr.testf$fig, vcr.test$fig); abline(0, 1)
plot(vcr.testf$farness, vcr.test$farness); abline(0, 1)
```
***



# Amazon book review data


```{r}
data(data_bookReviews)
X <- data_bookReviews[1:500, 1]; y <- data_bookReviews[1:500, 2] 
Xnew  <- data_bookReviews[501:1000, 1]; ynew <- data_bookReviews[501:1000, 2] 
length(X); length(Xnew); length(y); length(ynew) 
```

This is a subset of the data used in the paper, which
was assembled by Prettenhofer and Stein (2010), see
https://arxiv.org/abs/1008.0716

The full dataset has been used for a variety of things, including classification using svm.
A list of citations:
https://scholar.google.com/scholar?hl=en&as_sdt=0%2C5&sciodt=0%2C5&cites=4738650459241064524&scipsc=1&q=svm+&btnG=

The subset was chosen small enough to keep the computation
time low, while still containing the examples in the paper.

First we define the colors and inspect an example of a book review:

```{r}
cols <- c("deepskyblue3", "red")
col <- cols[as.numeric(y)]
X[5]
```

The next question is what kernel to use. There are many string kernels. 
kernlab provides several options:
The "spectrum" kernel matches substrings of an exact size.
The "boundrange" kernel matches substrings of a length up 
  to a given size.
The "constant" kernel matches all substrings.
The "exponential" kernel matches all substrings but gives 
  different weights depending on the length of the matched
  string.

The best kernel for these data in terms of accuracy turned 
out to be the spectrum kernel with length 7:

```{r}
strdot <- kernlab::stringdot(type = "spectrum", length = 7)
strdot
ptm <- proc.time()
Kr_train <- kernlab::kernelMatrix(kernel = strdot, x = X)
(proc.time() - ptm)[3] 
dim(Kr_train) 
```

Using the makeFV() function, we can compute the corresponding points in feature space. It takes into account almost-zero eigenvalues. The resulting Xf can be seen as the scores of a kernel PCA without centering
(mean subtraction).

```{r}
outFV <- makeFV(Kr_train)
Xf <- outFV$Xf 
dim(Xf) 
```


To see how well these feature vectors and Kr_train match, we can compute the inner products in Xf, and verify that they are close to Kr_train:

```{r}
inprod <- Xf %*% t(Xf) 
max(abs(as.vector(Kr_train - inprod))) 
```

We can also verify that the points of Xf lie on the unit sphere in 500 dimensions:

```{r}
range(rowSums(Xf^2)) # all 1
```


By construction, Xf is expressed in an orthogonal coordinate system.
In order to explain 95% of the variance in the data, we need 463 out of the 500 components, so we are in a truly high dimensional situation:

```{r}
explvar <- cumsum(apply(Xf, 2, var))
plot(explvar / explvar[ncol(Xf)]); abline(h = 0.95)
min(which(explvar > 0.95 * explvar[ncol(Xf)]))
```


Now we can train the support vector classifier on these data.
On Xf we train the SVM with the LINEAR kernel, since
we are in feature space:

```{r}
Xfdat <- data.frame(X = Xf, y = y)
set.seed(1) # we have to fix the seed!
svmfit <- svm(y~., data = Xfdat, scale = FALSE, kernel = "linear", cost = 2, probability = TRUE)
```

We plot the decision values. the two horizontal lines
contain the support vectors. There are a lot of them,
since the dimension is high.

```{r}
plot(svmfit$decision.values) 
```

Now create the VCR object of the training data:

```{r}
vcr.train <- vcr.svm.train(Xf, y, svfit = svmfit)
```

Using the VCR object, we can build the confusion matrix. We see that the accuracy is 100%.
This classification is perfect due to overfitting!

```{r}
confmat.vcr(vcr.train, showOutliers = FALSE)
```

The classification on the test data is more interesting.
First construct the kernel matrix we need to classify the test data, which takes roughly a minute:

```{r}
ptm <- proc.time()
Kr_test <- kernlab::kernelMatrix(kernel = strdot, x = Xnew, y = X)
(proc.time() - ptm)[3] 
```

Now we compute feature vectors of the test set. These features live in the same space as the Xf of the training data.

```{r}
FVtest <- makeFV(kmat = Kr_test, transfmat = outFV$transfmat)
Zf <- FVtest$Xf 
dim(Zf) 
```

We compare the inner products with the kernel matrix to make sure that they are the same:

```{r}
inprod <- Zf %*% t(Xf) 
max(abs(as.vector(Kr_test - inprod))) # tiny, good.
```

Now create the VCR object of the test data:

```{r}
vcr.test <- vcr.svm.newdata(Zf, ynew, vcr.train)
```

Compute the confusion matrix from the VCR object.  In the full dataset, the accuracy was 82 %. Here we achieve roughly 74% accuracy, which is lower since training was done on 500 texts instead of 2000.

```{r}
confmat.vcr(vcr.test)
```

The stacked mosaic plot of the classification is presented below:

```{r}
stackedplot(vcr.test, classCols = cols, separSize = 1.5, minSize = 1.5, main = "SVM on the book review test data")
```

The silhouette plot of the classification is presented below:

```{r}
# pdf("Bookreview_test_SVM_silhouettes.pdf", width = 5.0, height = 4.6)
silplot(vcr.test, classCols = cols, main = 
           "Silhouette plot of SVM on book review test data")      
# dev.off()
```


The overall average silhouette width is quite poor, as was the classification accuracy. Now we build the class maps for the test set.
First the class map of the negative reviews:

```{r}
# To identify the atypical points, uncomment the next line:
# classmap(vcr.test, 1, classCols = cols, identify = TRUE)
# Identified point(s): 
# [1]  67 127 118 145  94  45
#
# pdf("Bookreviews_500_classmap_test_negative.pdf")
par(mar = c(3.5, 3.5, 2.0, 0.2))
coords <- classmap(vcr.test, 1, classCols = cols, cex = 1.5,
                  main = "predictions of the negative book reviews", cex.lab = 1.5, cex.axis = 1.5, cex.main = 1.5, maxfactor = 1.05, identify = FALSE)
indstomark <- c(67, 127, 118, 145, 94, 45) # from identify=TRUE above
labs  <- c("a", "b", "x", "c", "d", "e")
xvals <- coords[indstomark, 1] +
   c(-0.08, 0, 0, 0.09, 0.09, 0.09) # visual fine tuning
yvals <- coords[indstomark, 2] +
   c(0, 0.033, 0.030, 0, 0, 0)
text(x = xvals, y = yvals, labels = labs, cex = 1.5)
legend("topright", fill = cols,
       legend = c("negative", "positive"), cex = 1.4,
       ncol = 1, bg = "white")
# dev.off()
par(oldpar)
```

We marked the following points:

```{r}
(Xnew[which(ynew == 1)])[67] # (a) in the paper.

(Xnew[which(ynew == 1)])[127] # (b) in the paper.

(Xnew[which(ynew == 1)])[118] # (x) 
# Negative review, but highly recommends another book!

(Xnew[which(ynew == 1)])[145] # (c) in the paper.
# Contains some grammatical errors
# Is negative, but expressed very indirectly.

(Xnew[which(ynew == 1)])[94] # (d) in the paper. 
# Definitely negative, but quite short.

(Xnew[which(ynew == 1)])[45] # (e) in the paper.
```

Now the class map of the positive reviews in the test set:

```{r}
#
# To identify the atypical points:
# classmap(vcr.test, 2, classCols = cols, identify = TRUE)
# Identified point(s): 
# [1] 48  32 18 168 113
#
# pdf("Bookreviews_500_classmap_test_positive.pdf")
par(mar = c(3.5, 3.5, 2.0, 0.2))
coords <- classmap(vcr.test, 2, classCols = cols, cex = 1.5,
                  main = "predictions of the positive book reviews", cex.lab = 1.5, cex.axis = 1.5, cex.main = 1.5,
                  maxfactor = 1.05, identify = FALSE)
indstomark <- c(48, 32, 18, 168, 113) # from identify = TRUE above
coords[indstomark, ]
labs  <- letters[6:10]
xvals <- coords[indstomark, 1] +
   c(0, 0, 0.09, 0.09, 0.09) # visual fine tuning
yvals <- coords[indstomark, 2] +
   c(0.032, 0.032, 0, 0, 0)
text(x = xvals, y = yvals, labels = labs, cex = 1.5)
legend("left", fill = cols,
       legend = c("negative", "positive"), cex = 1.4,
       ncol = 1, bg = "white")
# dev.off()
par(oldpar)
```

We marked the following points:

```{r}
(Xnew[which(ynew == 2)])[48] # (f) in the paper.

(Xnew[which(ynew == 2)])[32] # (g) in the paper.

(Xnew[which(ynew == 2)])[18] # (h) in the paper.

(Xnew[which(ynew == 2)])[168] # (i) in the paper.

(Xnew[which(ynew == 2)])[113] # (j) in the paper.
# very short (and spelling error).
```

# Sweets data

We first load the nutrients data set out of the robCompositions package.

```{r}
# Load the entire nutrients_branded data set:
library("robCompositions")
data(nutrients_branded, package = "robCompositions")

X <- as.matrix(nutrients_branded)
colnames(X)
```

We now preprocess the data, removing uninformative variables and those with many NAs:

```{r}
X            <- X[, -c(1:7, 10, 18)]
productnames <- nutrients_branded$name_D
hasNA        <- apply(X, 1, function(y) sum(is.na(y)) > 0)
X            <- X[-which(hasNA), ]
productnames <- productnames[-which(hasNA)]
```

There are a lot of food types in this dataset. We will select a subset.

```{r}
type  <- nutrients_branded$category_E 
type  <- type[-which(hasNA)]
inds1 <- which(type == "Sweets/Cookies/Biscuits")
inds2 <- which(type == "Sweets/Milk based ice cream")
inds3 <- which(type == "Sweets/Cakes and tarts")
inds4 <- which(type == "Sweets/Creams and puddings")

X <- X[c(inds1, inds2, inds3, inds4), ]
X <- as.matrix(apply(X, 2, as.numeric))
prod.names <- productnames[c(inds1, inds2, inds3, inds4)]
X <- as.matrix(scale(X))
head(X)

y <- factor(c(rep("biscuits", length(inds1)),
             rep("ice cream", length(inds2)),
             rep("cakes", length(inds3)),
             rep("puddings", length(inds4))),
            levels = c("biscuits", "ice cream", 
                      "cakes","puddings"))

table(y)
```

We now proceed by training the multiclass SVM, and constructing the VCR object:

```{r}
Xdat <- data.frame(X = X, y = y)
set.seed(1) # needed, otherwise not deterministic.
svmfit <- svm(y~., data = Xdat, scale = FALSE,
              kernel = "linear", cost = 10,
              probability = TRUE)
names(svmfit)

vcrobj <- vcr.svm.train(X, y, svfit = svmfit)
```

Based on the output of vcr.svm.train, we can now analyze and visualize the classification results. First we look at the confusion matrix and the stacked plot.


```{r}
confmat.vcr(vcrobj, showOutliers = FALSE)


cols <- c("firebrick4", "violet", "chocolate3", "orange")

# Stacked plot in the paper:
# pdf("Sweets_stackplot_with_outliers.pdf",width=5,height=4.6)
stackedplot(vcrobj, classCols = cols, separSize = 0.5, minSize = 1.5,
            main = "Stacked plot of SVM on the sweets data",
            htitle = "given class", vtitle = "predicted class")
# dev.off()

# With legend:
stackedplot(vcrobj, classCols = cols, separSize = 0.5, minSize = 1.5,
            main = "Stacked plot of SVM on the sweets data",
            htitle = "given class", vtitle = "predicted class",
            showLegend = TRUE)

# Without showing outliers:
stackedplot(vcrobj, classCols = cols, separSize = 0.5, minSize = 1.5,
            main = "Stacked plot of SVM on the sweets data",
            showOutliers = FALSE)
```

Now we make the silhouette plot:


```{r}
# pdf("Sweets_SVM_silhouettes.pdf", width=5.0, height=4.6)
silplot(vcrobj, classCols = cols,
        main = "Silhouette plot of SVM on the sweets data")
# dev.off()
```

And finally, the class maps:

```{r}
labels <- c("biscuits", "ice cream", "cakes", "puddings")
par(mfrow = c(1, 1))

# To identify special points in class 1:
# classmap(vcrobj, 1, classCols = cols, identify=T)
# Press the escape key to stop identifying.
indstomark <- c(122, 1, 184)
#
mymaxf <- 1.05
coords <- classmap(vcrobj, 1,classCols = cols,
         main = paste0("predictions of ", labels[1]),
         cex = 1.5, cex.lab = 1.5, cex.axis = 1.5,
         cex.main = 1.5, maxfactor = mymaxf)
coords[indstomark, ]
labs  <- c("a", "b", "c")
xvals <- c(2.380, 2.73, 4)
yvals <- c(0.825, 0.105, 0.065)
text(x = xvals, y = yvals, labels = labs, cex = 1.5)
legend("right", fill = cols, legend = labels,
       cex = 1.15, ncol = 1, bg = "white")

# Analogously for classes 2, 3, 4.

# Figure in the paper with all four class maps together:
#
labels <- c("biscuits", "ice cream", "cakes", "puddings")
# pdf(file="Sweets_all_class_maps.pdf",width=9,height=9)
par(mfrow = c(2, 2)) # (nrows,ncols)
par(mar = c(3.6, 3.5, 2.8, 1.0))
mymaxf <- 1.05
classmap(vcrobj, 1, classCols = cols,
         main = paste0("predictions of ", labels[1]),
         cex = 1.5, cex.lab = 1.5, cex.axis = 1.5,
         cex.main = 1.5, maxfactor = mymaxf)
labs  <- c("a", "b", "c") # points 122, 1, 184
xvals <- c(2.380, 2.73, 4)
yvals <- c(0.825, 0.105, 0.065)
text(x = xvals, y = yvals, labels = labs, cex = 1.5)
legend("right", fill = cols, legend = labels,
       cex = 1.15, ncol = 1, bg = "white")
#
par(mar = c(3.6, 2.0, 2.8, 0.3))
classmap(vcrobj, 2,classCols = cols,
         main = paste0("predictions of ", labels[2]),
         cex = 1.5, cex.lab = 1.5, cex.axis = 1.5,
         cex.main = 1.5, maxfactor = mymaxf)
labs  <- c("d", "e", "f", "g", "h-l")
xvals <- c(2.73, 3.21, 3.86, 3.83, 3.9)
yvals <- c(0.78, 0.96, 0.67, 0.57, 0.16)
text(x = xvals, y = yvals, labels = labs, cex = 1.5)
legend("topleft", fill = cols, legend = labels,
       cex = 1.15, ncol = 1, bg = "white")
#
par(mar = c(3.6, 3.5, 2.8, 1.0))
classmap(vcrobj, 3, classCols = cols,
         main = paste0("predictions of ", labels[3]),
         cex = 1.5, cex.lab = 1.5, cex.axis = 1.5,
         cex.main = 1.5, maxfactor = mymaxf)
labs  <- c("m", "n", "o", "p")
xvals <- c(1.05, 3.12, 2.86, 4)
yvals <- c(0.893, 0.95, 0.24, 0.09)
text(x = xvals, y = yvals, labels = labs, cex = 1.5)
legend("right", fill = cols, legend = labels,
       cex = 1.15, ncol = 1, bg = "white")
#
par(mar = c(3.6, 2.0, 2.8, 0.3))
classmap(vcrobj, 4,classCols = cols,
         main = paste0("predictions of ", labels[4]),
         cex = 1.5, cex.lab = 1.5, cex.axis = 1.5,
         cex.main = 1.5, maxfactor = mymaxf)
labs  <- c("q", "r", "s-u", "v", "w", "x")
xvals <- c(3.00, 3.6, 4.1, 3.82, 4.04, 3.9)
yvals <- c(0.95, 0.97, 0.926, 0.875, 0.795, 0.588)
text(x = xvals, y = yvals, labels = labs, cex = 1.5)
legend("topleft", fill = cols, legend = labels,
       cex = 1.16, ncol = 1, bg = "white")
# dev.off()
```

***

---
title: "Discriminant_analysis_examples"
author: "Raymaekers, J. and Rousseeuw, P.J."
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Discriminant_analysis_examples}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, echo = FALSE}
knitr::opts_chunk$set(
 fig.width  = 5 ,
 fig.height = 3.5,
 fig.align  = 'center'
)
oldpar <- list(mar = par()$mar, mfrow = par()$mfrow)
```

# Introduction

This vignette visualizes classification results from discriminant analysis, using tools from the package.

```{r}
library("classmap")
library("ggplot2")
library("gridExtra")
```

# Iris data

As a first small example, we consider the Iris data.
We first load the data and inspect it.

```{r}
data(iris)
X <- iris[, 1:4]
y <- iris[, 5]
is.factor(y)
table(y)

pairs(X, col = as.numeric(y) + 1, pch = 19)
```

Now we carry out quadratic discriminant analysis and inspect the output. Note that we can also do linear discriminant analysis by choosing rule = "LDA".

```{r}
vcr.train <- vcr.da.train(X, y, rule = "QDA")
names(vcr.train)
```

We now inspect the output in detail. First look at the prediction as integer, the prediction as label, the alternative label as integer and the alternative label:

```{r}
vcr.train$predint 
vcr.train$pred[c(1:10, 51:60, 101:110)]
vcr.train$altint  
vcr.train$altlab[c(1:10, 51:60, 101:110)]
```

The Probability of Alternative Class (PAC) of each object is found in the $PAC element of the output:

```{r}
vcr.train$PAC[1:3] 
summary(vcr.train$PAC)
```

The $fig element of the output contains the distance from case i to class g. Let's look at it for the first 5 objects:

```{r}
vcr.train$fig[1:5, ]
```

From the fig, the farness of each object can be computed. The farness of an object i is the f(i, g) to its own class: 

```{r}
vcr.train$farness[1:5]
summary(vcr.train$farness)
```

The "overall farness" of an object is defined as the lowest f(i, g) it has to any class g (including its own):

```{r}
summary(vcr.train$ofarness)
```

Objects with ofarness > cutoff are flagged as "outliers". These can be included in a separate column in the confusion matrix. This confusion matrix can be computed using confmat.vcr, which also returns the accuracy.

To illustrate this we choose a rather low cutoff:

```{r}
confmat.vcr(vcr.train, cutoff = 0.98)
```


With the default cutoff = 0.99 no objects are flagged in this example:

```{r}
confmat.vcr(vcr.train)
```
Note that the accuracy is computed before any objects are flagged, so it does not depend on the cutoff.

The confusion matrix can also be constructed showing class numbers instead of labels. This option can be useful for long level names.

```{r}
confmat.vcr(vcr.train, showClassNumbers = TRUE)
```

A stacked mosaic plot made with the stackedplot() function can be used to visualize the confusion matrix. The outliers, if there are any, appear as grey areas on top.

```{r}
cols <- c("red", "darkgreen", "blue")
stackedplot(vcr.train, classCols = cols, separSize = 1.5,
            minSize = 1, showLegend = TRUE)

stackedplot(vcr.train, classCols = cols, separSize = 1.5,
            minSize = 1, showLegend = TRUE, cutoff = 0.98)
```

The default stacked mosaic plot has no legend:

```{r}
stplot <- stackedplot(vcr.train, classCols = cols, 
                     separSize = 1.5, minSize = 1,
                     main = "QDA on iris data")
stplot
```

We also make the silhouette plot using the silplot() function:

```{r}
# pdf("Iris_QDA_silhouettes.pdf", width=5.0, height=4.6)
silplot(vcr.train, classCols = cols, 
        main = "Silhouette plot of QDA on iris data")      
# dev.off()
```

We now make the class maps based on the vcr object. This can be done using the classmap() function. We make a separate class map for each of the three classes. We see that class 1 is a very tight class (low PAC, no high farness). 
Class 2 is not so tight, and has two points which are predicted as virginica. Class 3 has one point predicted as versicolor.

```{r}
classmap(vcr.train, 1, classCols = cols)
classmap(vcr.train, 2, classCols = cols)
classmap(vcr.train, 3, classCols = cols) # With the default cutoff no farness values stand out:
```



```{r}
# With a lower cutoff:
classmap(vcr.train, 3, classCols = cols, cutoff = 0.98)
# Now one point is to the right of the vertical line.
# It also has a black border, meaning that it is flagged
# as an outlier, in the sense that its farness to _all_
# classes is above 0.98.
```
***

To illustrate the use of new data we create a fake dataset which is a subset of the training data, where not all classes occur, and ynew has NA's.

```{r}
Xnew <- X[c(1:50, 101:150), ]
ynew <- y[c(1:50, 101:150)]
ynew[c(1:10, 51:60)] <- NA
pairs(X, col = as.numeric(y) + 1, pch = 19) # 3 colors
pairs(Xnew, col = as.numeric(ynew) + 1, pch = 19) # only red and blue
```

Now we build the vcr object on the training data.

```{r}
vcr.test <- vcr.da.newdata(Xnew, ynew, vcr.train)
```

Inspect some of the output to confirm that it corresponds with what we would expect:

```{r}
plot(vcr.test$predint, vcr.train$predint[c(1:50, 101:150)]); abline(0, 1)
plot(vcr.test$altint, vcr.train$altint[c(1:50, 101:150)]); abline(0, 1)
plot(vcr.test$PAC, vcr.train$PAC[c(1:50, 101:150)]); abline(0, 1)
vcr.test$farness 
plot(vcr.test$farness, vcr.train$farness[c(1:50, 101:150)]); abline(0, 1)
plot(vcr.test$fig, vcr.train$fig[c(1:50, 101:150), ]); abline(0, 1)
vcr.test$ofarness 
plot(vcr.test$ofarness, vcr.train$ofarness[c(1:50, 101:150)]); abline(0, 1)
```


The confusion matrix for the test data, as for the training data, can be constructed by the confmat.vcr() function. A cutoff of 0.98 flags three outliers in this example.


```{r}
confmat.vcr(vcr.test)
confmat.vcr(vcr.test, cutoff = 0.98)
```

Also the stacked mosaic plot can be constructed on the test data:

```{r}
stplot # to compare with:
stackedplot(vcr.test, classCols = cols, separSize = 1.5, minSize = 1)
```

We now make the silhouette plot on the test data:

```{r}
#pdf("Iris_test_QDA_silhouettes.pdf", width=5.0, height=4.3)
silplot(vcr.test, classCols = cols, 
        main = "Silhouette plot of QDA on iris subset") 
#dev.off()
```

Finally, we construct the class maps for the test data. 
We compare the class map of the training data with that of the test data for each class.

```{r}
classmap(vcr.train, 1, classCols = cols)
classmap(vcr.test, 1, classCols = cols) 
```

```{r, error = TRUE}
classmap(vcr.train, 2, classCols = cols)
classmap(vcr.test, 2, classCols = cols)
```

```{r}
classmap(vcr.train, 3, classCols = cols)
classmap(vcr.test, 3, classCols = cols) 
```

***
# Floral buds data (shown in paper)

We now analyze the floral buds data, which was also used as an illustration in the paper. First load and inspect the data.

```{r}
data(data_floralbuds)
X <- as.matrix(data_floralbuds[, 1:6])
y <- data_floralbuds$y
dim(X) # 550  6
length(y) # 550
table(y)
# branch     bud  scales support 
#     49     363      94      44 

# Pairs plot
cols <- c("saddlebrown", "orange", "olivedrab4", "royalblue3")
pairs(X, gap = 0, col = cols[as.numeric(y)]) # hard to separate visually
```

Now we perform quadratic discriminant analysis:

```{r}
vcr.obj <- vcr.da.train(X, y)
```

Construct the confusion matrix without and with outliers shown:

```{r}
confmat <- confmat.vcr(vcr.obj, showOutliers = FALSE)

confmat.vcr(vcr.obj) 
```

Construct the stacked mosaic plot:

```{r}
stackedplot(vcr.obj, classCols = cols, separSize = 0.6,
            minSize = 1.5,  main = "stacked plot of QDA on floral buds")

# Version in paper:
# pdf("Floralbuds_QDA_stackplot_without_outliers.pdf",
#     width=5, height=4.3)
# stackedplot(vcr.obj, classCols = cols, separSize = 0.6,
#             minSize = 1.5, showOutliers = FALSE,
#             htitle = "given class", vtitle = "predicted class")
# dev.off()
```

Now make the silhouette plot:

```{r}
#pdf("Floralbuds_QDA_silhouettes.pdf", width=5.0, height=4.3)
silplot(vcr.obj, classCols = cols,
        main = "Silhouette plot of QDA on floral bud data")      
#dev.off()
```


The quasi residual plot can be made with the
qresplot() function. We illustate this below by making the quasi residual plot against the sum of the variables.
A correlation test confirms that the images with higher sums are significantly easier to classify:

```{r}
PAC <- vcr.obj$PAC
feat <- rowSums(X); xlab = "rowSums(X)"
# pdf("Floralbuds_QDA_quasi_residual_plot.pdf", width=5, height=4.8)
qresplot(PAC, feat, xlab = xlab, plotErrorBars = TRUE, fac = 2, 
         main = "Floral buds: quasi residual plot")
# dev.off()

cor.test(feat, PAC, method = "spearman") 
```



Construct the class maps, as shown in the paper:

```{r}
labels <- c("branch", "bud", "scale", "support")

# classmap of class "bud"
#
# To identify the points that stand out:
# classmap(vcr.obj, 2, classCols = cols, identify = TRUE)
# Press "Esc" to get out.
#
# pdf("Floralbuds_QDA_classmap_bud.pdf", width=7, height=7)
par(mar = c(3.6, 3.5, 2.4, 3.5))
coords <- classmap(vcr.obj, 2, classCols = cols,
         main = "predictions of buds",
         cex = 1.5, cex.lab = 1.5, cex.axis = 1.5,
         cex.main = 1.5) 
# For marking points:
indstomark <- c(294, 70, 69, 152, 204) # from identify = TRUE above
labs  <- letters[seq_len(5)]
xvals <- coords[indstomark, 1] +
  c(0, 0.10, 0.14, 0.10, 0.08) # visual finetuning
yvals <- coords[indstomark, 2] +
  c(0.04, 0.04, 0, -0.03, +0.04)
text(x = xvals, y = yvals, labels = labs, cex = 1.5)
legend("topleft", fill = cols[1:4], legend = labels, 
       cex = 1, ncol = 1, bg = "white")
# dev.off()
par(oldpar)
```

All class maps:
 
```{r, fig.height = 8, fig.width = 8}

#
# pdf(file = "Floralbuds_all_class_maps.pdf", width = 7, height = 7)
par(mfrow = c(2, 2))
par(mar = c(3.3, 3.2, 2.7, 1.0))
classmap(vcr.obj, 1, classCols = cols,
         main = "predictions of branches")
legend("topright", fill = cols, legend = labels,
       cex = 1, ncol = 1, bg = "white")
#
par(mar = c(3.3, 0.5, 2.7, 0.3))
classmap(vcr.obj, 2, classCols = cols,
         main = "predictions of buds")
labs  <- letters[seq_len(5)]
xvals <- coords[indstomark, 1] +
  c(0, 0.10, 0.14, 0.10, 0.08) # visual finetuning
yvals <- coords[indstomark, 2] +
  c(0.04, 0.04, 0, -0.03, 0.04)
# xvals <- c( 1.75, 1.68, 1.25, 3.25, 4.00)
# yvals <- c(0.045, 0.92, 0.54, 0.97, 0.045)
text(x = xvals, y = yvals, labels = labs, cex = 1.0)
legend("topleft", fill = cols, legend = labels,
       cex = 1, ncol = 1, bg = "white")
#
par(mar = c(3.3, 3.2, 2.7, 1.0))
classmap(vcr.obj, 3, classCols = cols,
         main = "predictions of scales")
legend("left", fill = cols, legend = labels,
       cex = 1, ncol = 1, bg = "white")
# 
par(mar = c(3.3, 0.5, 2.7, 0.3))
classmap(vcr.obj, 4, classCols = cols,
         main = "predictions of supports")
legend("topright", fill = cols, legend = labels,
       cex = 1, ncol = 1, bg = "white")
# dev.off()
par(oldpar)
```
***


# MNIST data

We now analyze the MNIST data, originally from the website of Yann LeCun. As the link on his website is currently down, we use a different source. Note that downloading the data may take a minute or two, depending on the speed of the internet connection.

```{r}

mnist_url <- url("https://wis.kuleuven.be/statdatascience/robust/data/mnist-rdata")
load(mnist_url)

X_train <- mnist$train$x
y_train <- as.factor(mnist$train$y)

head(y_train)
# Levels: 0 1 2 3 4 5 6 7 8 9
dim(X_train) # 60000    28    28
length(y_train) # 60000
```

We now inspect the data by plotting a few images

```{r, fig.height = 1, fig.width = 1}

plotImage = function(tempImage) {
  tdm = reshape2::melt(apply((tempImage), 2, rev))
  p = ggplot(tdm, aes(x = Var2, y = Var1, fill = (value))) +
    geom_raster() +
    guides(color = "none", size = "none", fill = "none") +
    theme(axis.title.x = element_blank(),
          axis.title.y = element_blank(),
          axis.text.x = element_blank(),
          axis.text.y = element_blank(),
          axis.ticks.x = element_blank(),
          axis.ticks.y = element_blank()) +
    scale_fill_gradient(low = "white", high = "black")
  p
}

plotImage(X_train[1, , ])
plotImage(X_train[2, , ])
plotImage(X_train[3, , ])
```

We now unfold the array containing the data to a matrix, 
and inspect some sample images as well as the average image per digit:

```{r, fig.height = 2, fig.width = 5}
# Change the dimensions of X for the sequel:
dim(X_train) <- c(60000, 28 * 28)
dim(X_train) # 60000    784

# Sampled digit images:
set.seed(123)
sampledigits <- list()
for (i in 0:9) {
  digit <- i
  idx <- sample(which(y_train == digit), size = 1)
  tempImage <- matrix(unlist(X_train[idx, ]), 28, 28)
  sampledigits[[i + 1]] <- plotImage(tempImage) 
}
psampledigits <- grid.arrange(grobs = sampledigits, ncol = 5)
# ggsave("MNIST_sampled_images.pdf", plot = psampledigits,
#        width = 10, height = 1)


# Averaged digit images:
meanPlots <- list()
for (j in 0:9) {
  m.out <- colMeans(X_train[which(y_train == j), ])
  dim(m.out) <- c(28, 28)
  meanPlots[[j + 1]] <- plotImage(m.out) 
}
meanplot <- grid.arrange(grobs = meanPlots, ncol = 5)
# ggsave("MNIST_averaged_images.pdf", plot = meanplot,
#        width = 10, height = 1)
```

Before performing discriminant analysis, we reduce the dimension of the data
by PCA.

```{r}
library(svd)
ptm <- proc.time()
svd.out <- svd::propack.svd(X_train, neig = 50)
(proc.time() - ptm)[3]
loadings <- svd.out$v
rm(svd.out)
dataProj <- as.matrix(X_train %*% loadings)
dim(dataProj)
```

Now we perform discriminant analysis, which takes roughly 5 seconds.

```{r}
vcr.train <- vcr.da.train(X = dataProj, y_train)
```

We compute the confusion matrix and make the stacked mosaic plot:

```{r}
confmat.vcr(vcr.train, showOutliers = FALSE)

cols <- c("red3", "darkorange", "gold2", "darkolivegreen3",
         "darkolivegreen4", "cadetblue3", "deepskyblue4", 
         "darkslateblue", "darkorchid3", "deeppink4")

# stacked plot in paper:
# pdf("MNIST_stackplot_with_outliers.pdf", width=5, height=4.3)
stackedplot(vcr.train, classCols = cols, separSize = 0.6,
            minSize = 1.5, htitle = "given class",
            main = "Stacked plot of QDA on MNIST training data", vtitle = "predicted class")
# dev.off()
```



The silhouette plot:

```{r}
# pdf("MNIST_QDA_silhouettes.pdf", width=5.0, height=4.6)
silplot(vcr.train, classCols = cols,
        main = "Silhouette plot of QDA on MNIST training data")      
# dev.off()
```

Now we make the class maps.


```{r}
wnq <- function(string, qwrite=TRUE) { # auxiliary function
  # writes a line without quotes
  if (qwrite) write(noquote(string), file = "", ncolumns = 100)
}

showdigit <- function(digit=digit, i, plotIt = TRUE) {
  idx = which(y_train == digit)[i]
  # wnq(paste("Estimated digit: ", as.numeric(vcr.train$pred[idx]), sep=""))
  tempImage <- matrix(unlist(X_train[idx, ]), 28, 28)
  if (plotIt) {plot(plotImage(tempImage))}
  return(plotImage(tempImage))
}

```


Class map of digit 0, shown in paper:

```{r, fig.height = 4, fig.width = 6}
digit <- 0
#
# To identify outliers:
# classmap(vcr.train, digit+1, classCols = cols, identify = TRUE)
# Press "Esc" to get out.
#
# pdf(paste0("MNIST_classmap_digit", digit, ".pdf"), width = 7, height = 7)
par(mar = c(3.6, 3.5, 2.4, 3.5))
coords <- classmap(vcr.train, digit + 1, classCols = cols,
         main = paste0("predictions of digit ",digit),
         cex = 1.5, cex.lab = 1.5, cex.axis = 1.5, 
         cex.main = 1.5)
indstomark <- c(4000, 3964, 5891, 2485, 822, 
               2280, 2504, 3906, 5869, 1034) # from identify = TRUE
labs  <- letters[1:length(indstomark)]
xvals <- coords[indstomark, 1] +
  c(-0.04, -0.01, 0, -0.11, 0.06,
    0.07, 0.06, 0.10, 0.06, 0.09)
yvals <- coords[indstomark, 2] +
  c(-0.03, -0.03, -0.03, 0.022, -0.025, 
    -0.025, -0.035, -0.025, 0.03, 0.03)
text(x = xvals, y = yvals, labels = labs, cex = 1.5)
legend("left", fill = cols,
       legend = 0:9, cex = 1, ncol = 2, bg = "white")
# dev.off()
par(oldpar)
```

```{r, fig.height = 2, fig.width = 5}
pred <- vcr.train$pred # needed for discussion plots
tempPreds <- (pred[which(y_train == digit)])[indstomark]
discussionPlots <- list()
for (i in 1:length(indstomark)) {
  idx <- indstomark[i]
  tempplot <- showdigit(digit, idx, plotIt = FALSE)
  tempplot <- arrangeGrob(tempplot, 
    bottom = paste0("(", labs[i], ") \"", tempPreds[i], "\""))
  discussionPlots[[i]] = tempplot
}
discussionPlot <- grid.arrange(grobs = discussionPlots, ncol = 5)
# ggsave(paste0("MNIST_discussionplot_digit", digit, ".pdf"),
#        plot = discussionPlot, width = 5,
#        height = (length(indstomark) %/% 5 +
#                    (length(indstomark) %% 5 > 0)))
```


Class map of digit 1, shown in paper:

```{r, fig.height = 4, fig.width = 6}

digit <- 1
# pdf(paste0("MNIST_classmap_digit", digit, ".pdf"), width = 7, height = 7)
par(mar = c(3.6, 3.5, 2.4, 3.5))
classmap(vcr.train, digit + 1, classCols = cols,
         main = paste0("predictions of digit ", digit),
         cex = 1.5, cex.lab = 1.5, cex.axis = 1.5, 
         cex.main = 1.5)
legend("left", fill = cols,
       legend = 0:9, cex = 1, ncol = 2, bg = "white")
# dev.off()
par(oldpar)
```

```{r, fig.height = 13, fig.width = 8}
# indices of the 1s predicted as 2 (takes a while):
#
indstomark <- which(vcr.train$predint[which(y_train == digit)] == 3)
length(indstomark) # 104
labs  <- letters[1:length(indstomark)]
pred <- vcr.train$pred # needed for discussion plots
tempPreds    <- (pred[which(y_train == digit)])[indstomark]
discussionPlots <- list()
for (i in 1:length(indstomark)) {
  idx <- indstomark[i]
  tempplot <- showdigit(digit, idx, FALSE)
  tempplot <- arrangeGrob(tempplot, 
        bottom = paste0("\"", tempPreds[i], "\""))
  discussionPlots[[i]] <- tempplot
}
discussionPlot <- grid.arrange(grobs = discussionPlots, 
                                         ncol = 8)
# ggsave(paste0("MNIST_discussionplot_digit", digit, "predictedAs2b.pdf"),
#        plot = discussionPlot, width = 10,
#        height = (length(indstomark) %/% 10 +
#                    (length(indstomark) %% 10 > 0)))
                 
# The digits 1 predicted as a 2 are mostly ones written with
# a horizontal line at the bottom.
```


Class map of digit 2:

```{r, fig.height = 4, fig.width = 6}
digit <- 2
# To identify outliers:
# classmap(vcr.train, digit + 1, classCols = cols, identify = TRUE)
# Press "Esc" to get out.
#
# pdf(paste0("MNIST_classmap_digit", digit,".pdf"), width = 7, height = 7)
par(mar = c(3.6, 3.5, 2.4, 3.5))
coords <- classmap(vcr.train, digit + 1, classCols = cols,
                  main = paste0("predictions of digit", digit), cex = 1.5, cex.lab = 1.5, cex.axis = 1.5,
                  cex.main = 1.5)
indstomark <- c(3164, 5434, 2319 , 4224, 3682, 
               2642, 4920, 1233, 3741, 3993) # from identify = TRUE
labs  <- letters[1:length(indstomark)]
xvals <- coords[indstomark, 1] +
  c(0, 0.08, 0, 0, 0, 0, 0, 0, 0, 0)
yvals <- coords[indstomark, 2] +
  c(-0.03, -0.03, -0.03, -0.03, -0.03, 
    -0.03, -0.03, -0.03, 0.03, 0.03)  
text(x = xvals, y = yvals, labels = labs, cex = 1.5)
legend("right", fill = cols,
       legend = 0:9, cex = 1, ncol = 2, bg = "white")
# dev.off()
par(oldpar)
```

```{r, fig.height = 2, fig.width = 5}
pred <- vcr.train$pred # needed for discussion plots
tempPreds    <- (pred[which(y_train == digit)])[indstomark]
discussionPlots <- list()
for (i in 1:length(indstomark)) {
  idx <- indstomark[i]
  tempplot <- showdigit(digit, idx, FALSE)
  tempplot <- arrangeGrob(tempplot, 
        bottom = paste0("(", labs[i], ") \"", tempPreds[i], "\""))
  discussionPlots[[i]] <- tempplot
}
discussionPlot <- grid.arrange(grobs = discussionPlots, 
                                         ncol = 5)
# ggsave(paste0("MNIST_discussionplot_digit", digit, ".pdf"),
#        plot = discussionPlot, width = 5,
#        height = (length(indstomark) %/% 5 +
#                    (length(indstomark) %% 5 > 0)))
```

***

Now we analyze the MNIST test data.
First load and inspect the data, and project it onto the PCA subspace extracted from the training data.


```{r, fig.height = 1, fig.width = 1}
X_test <- mnist$test$x
y_test <- as.factor(mnist$test$y)

head(y_test)
#
dim(X_test) # 10000    28    28
length(y_test) # 10000

plotImage(X_test[1, , ])
plotImage(X_test[2, , ])
plotImage(X_test[3, , ])

dim(X_test) <- c(10000, 28 * 28)
dim(X_test) # 10000  784

dataProj_test <- as.matrix(X_test %*% loadings)
```

Now prepare the VCR object:

```{r}
vcr.test <- vcr.da.newdata(Xnew = dataProj_test,
                           ynew = y_test,
                           vcr.da.train.out = vcr.train)
```

Build the confusion matrix and plot a stacked mosaic plot of the classification performance on the test data:

```{r}
confmat.vcr(vcr.test, showOutliers = FALSE, showClassNumbers = TRUE)

# In supplementary material:
# pdf("MNISTtest_stackplot_with_outliers.pdf", width = 5, height = 4.3)
stackedplot(vcr.test, classCols = cols, separSize = 0.6,
            main = "Stacked plot of QDA on MNIST test data",
            minSize = 1.5)
# dev.off()
```

Silhouette plot:



```{r}
#pdf("MNIST_test_QDA_silhouettes.pdf", width = 5.0, height = 4.6)
silplot(vcr.test, classCols = cols,
        main = "Silhouette plot of QDA on MNIST test data")      
#dev.off()
```

Now we can construct the class maps on the test data.
First for digit 0:

```{r, fig.height = 4, fig.width = 6}
showdigit_test <- function(digit = digit, i, plotIt = TRUE) {
  idx = which(y_test == digit)[i]
  # wnq(paste("Estimated digit: ", as.numeric(vcr.test$pred[idx]), sep = ""))
  tempImage <- matrix(unlist(X_test[idx, ]), 28, 28)
  if (plotIt) {plot(plotImage(tempImage))}
  return(plotImage(tempImage))
}

digit <- 0
# classmap(vcr.test, digit+1, classCols = cols, identify = TRUE)
# pdf(paste0("MNISTtest_classmap_digit", digit,".pdf"))
par(mar = c(3.6, 3.5, 2.4, 3.5))
coords <- classmap(vcr.test, digit + 1, classCols = cols,
         main = paste0("predictions of digit ", digit),
         cex = 1.5, cex.lab = 1.5, cex.axis = 1.5, 
         cex.main = 1.5)
indstomark <- c(140, 630, 241, 967, 189,
               377, 78, 943, 64, 354)
labs  <- letters[1:length(indstomark)]
xvals <- coords[indstomark, 1] +
  c(0.08, 0.07, -0.07, 0.06, 0,
    0.04, 0.05, 0.09, -0.04, 0.09)
yvals <- coords[indstomark, 2] +
  c(-0.025, -0.03, -0.024, -0.025, -0.03, 
    -0.03, -0.03, 0.022, 0.035, 0.03)
text(x = xvals, y = yvals, labels = labs, cex = 1.5)
legend("left", fill = cols,
       legend = 0:9, cex = 1, ncol = 2, bg = "white")
# dev.off()
par(oldpar)
```

```{r, fig.height = 2, fig.width = 5}
pred <- vcr.test$pred # needed for discussion plots
tempPreds <- (pred[which(y_test == digit)])[indstomark]
discussionPlots <- list()
for (i in 1:length(indstomark)) {
  idx <- indstomark[i]
  tempplot <- showdigit_test(digit, idx, FALSE)
  tempplot <- arrangeGrob(tempplot, 
      bottom = paste0("(", labs[i], ") \"", tempPreds[i], "\""))
  discussionPlots[[i]] <- tempplot
}
discussionPlot <- grid.arrange(grobs = discussionPlots, 
                                         ncol = 5)
# ggsave(paste0("MNISTtest_discussionplot_digit", digit, ".pdf"),
#        plot = discussionPlot, width = 5,
#        height = (length(indstomark) %/% 5 +
#                    (length(indstomark) %% 5 > 0)))
```

Now for digit 3:

```{r, fig.height = 4, fig.width = 6}
digit <- 3
# classmap(vcr.test, digit + 1, classCols = cols, identify = TRUE)
# pdf(paste0("MNISTtest_classmap_digit", digit, ".pdf"))
par(mar = c(3.6, 3.5, 2.4, 3.5))
coords <- classmap(vcr.test, digit + 1, classCols = cols,
         main = paste0("predictions of digit ", digit),
         cex = 1.5, cex.lab = 1.5, cex.axis = 1.5, 
         cex.main = 1.5)
indstomark <- c(883, 659, 262, 60, 310,
               832, 223, 784, 835, 289)
labs  <- letters[1:length(indstomark)]
xvals <- coords[indstomark, 1] +
  c(-0.01, 0.08, -0.10, 0.06, 0.07, 
    0.06, 0.03, 0.11, 0.02, 0.06)
yvals <- coords[indstomark, 2] +
  c(0.035, 0.033, -0.017, -0.022, -0.025, 
    -0.025, -0.033, -0.022, 0.035, 0.038)
text(x = xvals, y = yvals, labels = labs, cex = 1.5)
legend("right", fill = cols,
       legend = 0:9, cex = 1, ncol = 2, bg = "white")
# dev.off()
par(oldpar)
```

```{r, fig.height = 2, fig.width = 5}
pred <- vcr.test$pred # needed for discussion plots
tempPreds    <- (pred[which(y_test == digit)])[indstomark]
discussionPlots <- list()
for (i in 1:length(indstomark)) {
  idx <- indstomark[i]
  tempplot <- showdigit_test(digit, idx, FALSE)
  tempplot <- arrangeGrob(tempplot, 
    bottom = paste0("(", labs[i], ") \"", tempPreds[i], "\""))
  discussionPlots[[i]] <- tempplot
}
discussionPlot <- grid.arrange(grobs = discussionPlots, 
                                         ncol = 5)
# ggsave(paste0("MNISTtest_discussionplot_digit", digit, ".pdf"),
#        plot = discussionPlot, width = 5,
#        height = (length(indstomark) %/% 5 +
#                    (length(indstomark) %% 5 > 0)))
```
***

---
title: "Rpart_examples"
author: "Raymaekers, J. and Rousseeuw, P.J."
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Rpart_examples}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, echo = FALSE}
knitr::opts_chunk$set(
 fig.width  = 5 ,
 fig.height = 3.5,
 fig.align  = 'center'
)
oldpar <- list(mar = par()$mar, mfrow = par()$mfrow)
```

# Introduction

This vignette visualizes classification results from rpart (CART), using tools from the package. The displays in this vignette are discussed in section 4 of Raymaekers and Rousseeuw (2021) (for the training data), and in section A.4 of the Supplementary Material (for the test data).

```{r}
library(rpart)
library(classmap)
```


# Titanic training data 

We first load and inspect the data.

```{r}
data("data_titanic")

traindata <- data_titanic[which(data_titanic$dataType == "train"), -13]

dim(traindata) 
colnames(traindata)
# SibSp: number of siblings + spouses aboard
# Parch: number of parents + children aboard

str(traindata)
table(traindata$y)
```

Now we want to predict the response y by CART (rpart).
We set the seed as rpart is not deterministic.
We also inspect the resulting classification tree.

```{r}
set.seed(123) 
rpart.out <- rpart(y ~ Pclass + Sex + SibSp + 
                    Parch + Fare + Embarked,
                  data = traindata, method = 'class',
                  model = TRUE)

## plot the tree:
# pdf(file = "titanic_train_tree.pdf", width = 6, height = 6)
rpart.plot::rpart.plot(rpart.out, box.palette = "RdBu")
# dev.off()
```

We now inspect the variable importance, which will be used to calculate the variable weights in the farness computation. 

```{r}
rpart.out$variable.importance
```

Now we declare the types of the variables. This is used in the calculation of the Daisy distance, which in turn is needed for the farness computation. There are 5 nominal columns and one ordinal. The variables not listed here are interval-scaled by default.
```{r}
mytype <- list(nominal = c("Name", "Sex", "Ticket", "Cabin", "Embarked"), ordratio = c("Pclass"))
```

Now we prepare for visualization and inspect the elements of the output.

```{r}
x_train <- traindata[, -12]
y_train <- traindata[,  12]
vcrtrain <- vcr.rpart.train(x_train, y_train, rpart.out, mytype)
names(vcrtrain)

vcrtrain$predint[1:10] # prediction as integer
vcrtrain$pred[1:10]    # prediction as label
vcrtrain$altint[1:10]  # alternative label as integer
vcrtrain$altlab[1:10]  # alternative label

# Probability of Alternative Class (PAC) of each object:
vcrtrain$PAC[1:3] 
#
summary(vcrtrain$PAC)

# f(i, g) is the distance from case i to class g:
vcrtrain$fig[1:3, ] # for the first 3 objects:

# The farness of an object i is the f(i, g) to its own class: 
vcrtrain$farness[1:3]
#
summary(vcrtrain$farness)

# The "overall farness" of an object is defined as the 
# lowest f(i, g) it has to any class g (including its own):
summary(vcrtrain$ofarness)

sum(vcrtrain$ofarness > 0.99, na.rm = TRUE) 
# No farness is considered outlying in these data.

confmat.vcr(vcrtrain) 

cols <- c("firebrick", "blue")
```

We first make the stacked plot, visualizing the confusion matrix:

```{r}
stackedplot(vcrtrain, classCols = cols,
            main = "Stacked plot of rpart on Titanic training data")
```

Now we consider the silhouette plot

```{r}
silplot(vcrtrain, classCols = cols, 
        main = "Silhouettes of rpart on Titanic training data")
# silplot.out <- silplot(vcrtrain, classCols = cols)
# ggsave("titanic_train_silhouettes.pdf", silplot.out,
#        width = 5, height = 5)
```


Now we construct the quasi residual plot of PAC vs. age for males only. The age variable is not in the model, but including it did not improve the classification. We see that the very young male passengers are often misclassified.
```{r}
hist(x_train$Age)

# Quasi residual plot versus age, for males only:

# pdf("titanic_qrp_versus_age_males.pdf", width = 4.8, height = 4.8)
PAC <- vcrtrain$PAC[which(x_train$Sex == "male")]
feat <- x_train$Age[which(x_train$Sex == "male")]
qresplot(PAC, feat, xlab = "Age (years)", opacity = 0.5,
         main = "quasi residual plot for male passengers",
         plotLoess = TRUE)
text(x = 14, y = 0.60, "loess curve", col = "red", cex = 1)
# dev.off()
```

Now we construct the class maps. First of the casualties.

```{r}
classmap(vcrtrain, "casualty", classCols = cols)
# classmap(vcrtrain, "casualty", classCols = cols, identify = TRUE)

# blue points top right: cases "a" and "b" in the paper
x_train[which(y_train == "casualty")[119], ]
# Woman in 1st class, should have survived.
#
x_train[which(y_train == "casualty")[192], ]

# Similar, but child.

# red point most to the right: case "c" in the paper
x_train[which(y_train == "casualty")[268], ]

```

Now the class map of the survivors.

```{r}
classmap(vcrtrain, "survived", classCols = cols)
# classmap(vcrtrain, "survived", classCols = cols, identify = TRUE)

# red point with highest farness among highest PAC: case "d" in the paper
x_train[which(y_train == "survived")[c(14)], ] 

# near-coinciding points with highest farness: cases "e" and "f" in the paper
#
x_train[which(y_train == "survived")[265], ]

x_train[which(y_train == "survived")[287], ] 

# man --> predicted as not survived. also paid the highest 
# fare in the data and has same ticket number as passenger 
# 265. Also embarked at the same place. It turns out that
# Gustave Lesueur was the man servant of the banker Thomas Cardeza:
  
# https://www.encyclopedia-titanica.org/titanic-survivor/thomas-cardeza.html
# https://www.encyclopedia-titanica.org/titanic-survivor/gustave-lesueur.html

# blue point bottom right: case "g" in the paper
x_train[which(y_train == "survived")[90], ]

# # woman  + first class so predicted in survivors.
# # paid highest fare in whole dataset 
```

# Titanic test data 
In addition to the training data, we consider the titanic test data.
First we load the data and inspect.


```{r}
testdata <- data_titanic[which(data_titanic$dataType == "test"), -13]

dim(testdata)
x_test <- testdata[, -12]
y_test <- testdata[, 12]
table(y_test)
```

Now we prepare for visualization:

```{r}
vcrtest <- vcr.rpart.newdata(x_test, y_test, vcrtrain)

confmat.vcr(vcrtest)
cols <- c("firebrick", "blue")
``` 

Now we can visualize, starting with the stacked plot:
```{r}
stackedplot(vcrtest, classCols = cols,
            main = "Stacked plot of Titanic test data")
```

Now we construct the silhouette plot:

```{r}
silplot(vcrtest, classCols = cols, 
        main = "Silhouettes of rpart on Titanic test data")
```

Finally, we make the class maps. First of the casualties, in which the misclassifications are all female (since all males 
are predicted as casualty by this model).

```{r}
classmap(vcrtest, "casualty", classCols = cols) 
```

Now the class map of survivors:

```{r}
classmap(vcrtest, "survived", classCols = cols) 
```

***

---
title: "Rpart_examples"
author: "Raymaekers, J. and Rousseeuw, P.J."
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Rpart_examples}
  %\VignetteEngine{knitr::knitr}
  %\VignetteEncoding{UTF-8}
---

```{r, echo = FALSE}
knitr::opts_chunk$set(
 fig.width  = 5 ,
 fig.height = 3.5,
 fig.align  = 'center'
)
oldpar <- list(mar = par()$mar, mfrow = par()$mfrow)
```

# Introduction

This vignette visualizes classification results from rpart (CART), using tools from the package. The displays in this vignette are discussed in section 4 of Raymaekers and Rousseeuw (2021) (for the training data), and in section A.4 of the Supplementary Material (for the test data).

```{r}
library(rpart)
library(classmap)
```


# Titanic training data 

We first load and inspect the data.

```{r}
data("data_titanic")

traindata <- data_titanic[which(data_titanic$dataType == "train"), -13]

dim(traindata) 
colnames(traindata)
# SibSp: number of siblings + spouses aboard
# Parch: number of parents + children aboard

str(traindata)
table(traindata$y)
```

Now we want to predict the response y by CART (rpart).
We set the seed as rpart is not deterministic.
We also inspect the resulting classification tree.

```{r}
set.seed(123) 
rpart.out <- rpart(y ~ Pclass + Sex + SibSp + 
                    Parch + Fare + Embarked,
                  data = traindata, method = 'class',
                  model = TRUE)

## plot the tree:
# pdf(file = "titanic_train_tree.pdf", width = 6, height = 6)
rpart.plot::rpart.plot(rpart.out, box.palette = "RdBu")
# dev.off()
```

We now inspect the variable importance, which will be used to calculate the variable weights in the farness computation. 

```{r}
rpart.out$variable.importance
```

Now we declare the types of the variables. This is used in the calculation of the Daisy distance, which in turn is needed for the farness computation. There are 5 nominal columns and one ordinal. The variables not listed here are interval-scaled by default.
```{r}
mytype <- list(nominal = c("Name", "Sex", "Ticket", "Cabin", "Embarked"), ordratio = c("Pclass"))
```

Now we prepare for visualization and inspect the elements of the output.

```{r}
x_train <- traindata[, -12]
y_train <- traindata[,  12]
vcrtrain <- vcr.rpart.train(x_train, y_train, rpart.out, mytype)
names(vcrtrain)

vcrtrain$predint[1:10] # prediction as integer
vcrtrain$pred[1:10]    # prediction as label
vcrtrain$altint[1:10]  # alternative label as integer
vcrtrain$altlab[1:10]  # alternative label

# Probability of Alternative Class (PAC) of each object:
vcrtrain$PAC[1:3] 
#
summary(vcrtrain$PAC)

# f(i, g) is the distance from case i to class g:
vcrtrain$fig[1:3, ] # for the first 3 objects:

# The farness of an object i is the f(i, g) to its own class: 
vcrtrain$farness[1:3]
#
summary(vcrtrain$farness)

# The "overall farness" of an object is defined as the 
# lowest f(i, g) it has to any class g (including its own):
summary(vcrtrain$ofarness)

sum(vcrtrain$ofarness > 0.99, na.rm = TRUE) 
# No farness is considered outlying in these data.

confmat.vcr(vcrtrain) 

cols <- c("firebrick", "blue")
```

We first make the stacked plot, visualizing the confusion matrix:

```{r}
stackedplot(vcrtrain, classCols = cols,
            main = "Stacked plot of rpart on Titanic training data")
```

Now we consider the silhouette plot

```{r}
silplot(vcrtrain, classCols = cols, 
        main = "Silhouettes of rpart on Titanic training data")
# silplot.out <- silplot(vcrtrain, classCols = cols)
# ggsave("titanic_train_silhouettes.pdf", silplot.out,
#        width = 5, height = 5)
```


Now we construct the quasi residual plot of PAC vs. age for males only. The age variable is not in the model, but including it did not improve the classification. We see that the very young male passengers are often misclassified.
```{r}
hist(x_train$Age)

# Quasi residual plot versus age, for males only:

# pdf("titanic_qrp_versus_age_males.pdf", width = 4.8, height = 4.8)
PAC <- vcrtrain$PAC[which(x_train$Sex == "male")]
feat <- x_train$Age[which(x_train$Sex == "male")]
qresplot(PAC, feat, xlab = "Age (years)", opacity = 0.5,
         main = "quasi residual plot for male passengers",
         plotLoess = TRUE)
text(x = 14, y = 0.60, "loess curve", col = "red", cex = 1)
# dev.off()
```

Now we construct the class maps. First of the casualties.

```{r}
classmap(vcrtrain, "casualty", classCols = cols)
# classmap(vcrtrain, "casualty", classCols = cols, identify = TRUE)

# blue points top right: cases "a" and "b" in the paper
x_train[which(y_train == "casualty")[119], ]
# Woman in 1st class, should have survived.
#
x_train[which(y_train == "casualty")[192], ]

# Similar, but child.

# red point most to the right: case "c" in the paper
x_train[which(y_train == "casualty")[268], ]

```

Now the class map of the survivors.

```{r}
classmap(vcrtrain, "survived", classCols = cols)
# classmap(vcrtrain, "survived", classCols = cols, identify = TRUE)

# red point with highest farness among highest PAC: case "d" in the paper
x_train[which(y_train == "survived")[c(14)], ] 

# near-coinciding points with highest farness: cases "e" and "f" in the paper
#
x_train[which(y_train == "survived")[265], ]

x_train[which(y_train == "survived")[287], ] 

# man --> predicted as not survived. also paid the highest 
# fare in the data and has same ticket number as passenger 
# 265. Also embarked at the same place. It turns out that
# Gustave Lesueur was the man servant of the banker Thomas Cardeza:
  
# https://www.encyclopedia-titanica.org/titanic-survivor/thomas-cardeza.html
# https://www.encyclopedia-titanica.org/titanic-survivor/gustave-lesueur.html

# blue point bottom right: case "g" in the paper
x_train[which(y_train == "survived")[90], ]

# # woman  + first class so predicted in survivors.
# # paid highest fare in whole dataset 
```

# Titanic test data 
In addition to the training data, we consider the titanic test data.
First we load the data and inspect.


```{r}
testdata <- data_titanic[which(data_titanic$dataType == "test"), -13]

dim(testdata)
x_test <- testdata[, -12]
y_test <- testdata[, 12]
table(y_test)
```

Now we prepare for visualization:

```{r}
vcrtest <- vcr.rpart.newdata(x_test, y_test, vcrtrain)

confmat.vcr(vcrtest)
cols <- c("firebrick", "blue")
``` 

Now we can visualize, starting with the stacked plot:
```{r}
stackedplot(vcrtest, classCols = cols,
            main = "Stacked plot of Titanic test data")
```

Now we construct the silhouette plot:

```{r}
silplot(vcrtest, classCols = cols, 
        main = "Silhouettes of rpart on Titanic test data")
```

Finally, we make the class maps. First of the casualties, in which the misclassifications are all female (since all males 
are predicted as casualty by this model).

```{r}
classmap(vcrtest, "casualty", classCols = cols) 
```

Now the class map of survivors:

```{r}
classmap(vcrtest, "survived", classCols = cols) 
```

***

---
title: "bixplot_examples"
author: "Rousseeuw, P.J."
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{bixplot_examples}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
editor_options: 
  markdown: 
    wrap: 72
---

```{r, echo = FALSE}
knitr::opts_chunk$set(
  fig.width  = 7,
  fig.height = 4,
  fig.align  = 'center'
)
oldpar <- list(mar = par()$mar, mfrow = par()$mfrow)
```

# Introduction

This vignette illustrates the `bixplot` function, which visualizes a
univariate dataset by combining an estimated density body, a
box-and-whisker summary, and a rug of individual data values. For
variables with more than one mode, the function automatically detects
clusters and displays each mode separately. The examples below start
with simple univariate comparisons and continue with more elaborate
multivariable layouts and rug coloring options.

```{r}
library(vioplot)
library(robustHD)
library(classmap)
```

------------------------------------------------------------------------

# Unimodal, bimodal and multimodal data

We start with three simulated variables: one unimodal, one bimodal and
one trimodal. This illustrates the core difference between a violin plot
and a bixplot.

```{r}
set.seed(1)
dat1 <- rnorm(120, 0, 2.5)
dat2 <- c(rnorm(80, -3, 1), rnorm(40, 3, 1))
dat3 <- c(rnorm(25, -4, 0.8), rnorm(50, 0, 0.8), rnorm(40, 4, 0.8))
xlist <- list(Unimodal = dat1, Bimodal = dat2, Multimodal = dat3)
```

The default call already gives an informative result:

```{r}
bixout <- bixplot(xlist, main = "bixplot")
```

The returned list `bixout` contains the cluster structure and
five-number summaries for each variable:

```{r}
bixout
```

For comparison, below we display the violin plot and the bixplot side by
side. The bixplot separates the modes of the bimodal and trimodal
variables and draws a dedicated box for each cluster.

```{r, fig.width = 8, fig.height = 4}
ylim <- c(-8.5, 9)
par(las = 1, mfrow = c(1, 2))
par(mar = c(2.1, 2.2, 1.7, 2))
viocol <- adjustcolor("chocolate3", alpha.f = 0.5)
vioplot::vioplot(xlist, ylim = ylim, main = "", col = viocol)
title(main = "violin plot", line = 0.5, cex.main = 1)
par(mar = c(2.1, 2.2, 1.7, 0.2))
bixplot(xlist, ylim = ylim, main = "")
title(main = "bixplot", line = 0.5, cex.main = 1)
par(oldpar)
```

------------------------------------------------------------------------

# Latency data and penguin bill length

Next we apply `bixplot` to two real datasets: a latency dataset with
three variables and the penguin bill length variable grouped by island.

```{r, fig.width = 10, fig.height = 7}
data("data_latenc")
dim(data_latenc) # 40 rows, 3 columns

par(las = 1, mfrow = c(2, 2))
mar1 <- c(2.1, 2.4, 2.3, 2)
mar2 <- c(2.1, 2.2, 2.3, 0.2)
viocol <- adjustcolor("chocolate3", alpha.f = 0.5)

par(mar = mar1)
vioplot::vioplot(data_latenc, main = "", col = viocol)
title(main = "violin plot", cex.main = 1.2, line = 0.6)

par(mar = mar2)
mymodeCol <- c("cadetblue3", "hotpink2")
bixplot(data_latenc, main = "", cutmin = 0, cutmax = 300,
        ylim = c(0, 300), modeCol = mymodeCol)
title(main = "bixplot", cex.main = 1.2, line = 0.6)

# Score the islands in a meaningful order
islandscore <- rep(NA, length(penguins$island))
islandscore[penguins$island == "Torgersen"] <- 1
islandscore[penguins$island == "Biscoe"]    <- 2
islandscore[penguins$island == "Dream"]     <- 3
ylim <- c(29, 62)

par(mar = mar1)
vioplot::vioplot(bill_len ~ reorder(island, islandscore, mean),
                 data = penguins, main = "",
                 col = viocol, xlab = "", ylab = "",
                 ylim = ylim, cex.axis = 1)
title(main = "violin plot", cex.main = 1.2, line = 0.6)

par(mar = mar2)
bixplot(bill_len ~ reorder(island, islandscore, mean),
        data = penguins, main = "", ylim = ylim,
        bodyCol = "gray40", bodyOpaque = 0.3)
title(main = "bixplot", cex.main = 1.2, line = 0.6)
par(oldpar)
```

For the latency variables, the bixplot correctly identifies a bimodal
structure that is obscured by the smooth envelope of the violin plot.
For the penguin bill lengths, the Biscoe island group is clearly
bimodal, reflecting the presence of two penguin species.

------------------------------------------------------------------------

# Iris data

The standardized iris dataset has four variables. The bixplot reveals
that petal length and petal width have more than one mode (cluster).

```{r, fig.width = 10, fig.height = 5}
par(mfrow = c(1, 2))
par(mar = c(4, 2, 2, 0.1))

diris <- data.frame(scale(iris[, 1:4]))
colnames(diris) <- c("Sepal.L", "Sepal.W", "Petal.L", "Petal.W")

mymodeCol <- c("cadetblue3", "hotpink2",
               "cadetblue3", "hotpink2",
               "lawngreen")
bixplot(diris, main = "", cut = 3, col = "gray75",
        bodyOpaque = 0.6, rugW = 0.16,
        rugoutCol = "red", curveLwd = 0.5,
        modeCol = mymodeCol,
        ylim = c(-3.6, 4.2), yaxs = "i",
        xlab = "standardized variables")
title(main = "bixplot display of iris data", cex.main = 1, line = 0.7)
```

We can also overlay marginal bixplots on top of a scatter plot, to show
the marginal distributions of the axes variables. Here we add horizontal
and vertical half-bixplots to a scatter plot of petal length versus
petal width.

```{r, fig.width = 5, fig.height = 5}
par(mar = c(4, 4, 2, 0.2))
x <- diris$Petal.L
y <- diris$Petal.W
xlim <- c(-2.3, 2.4)
ylim <- c(-2.2, 2.4)
xyratio <- (xlim[2] - xlim[1]) / (ylim[2] - ylim[1])

plot(x, y, xlim = xlim, ylim = ylim, pch = 16,
     xlab = "", ylab = "", xaxs = "i", yaxs = "i")
title(xlab = "Petal.L", line = 2)
title(ylab = "Petal.W", line = 2)
title(main = "petal length versus petal width", cex.main = 1, line = 0.7)

bixplot(x, add = TRUE, horizontal = TRUE,
        at = ylim[1] + 0.015, cutmin = xlim[1],
        boxwex = 0.9, curveLwd = 0.5,
        border = "black", side = "second",
        bodyOpaque = 0.6)
bixplot(y, add = TRUE, horizontal = FALSE,
        at = xlim[1] + 0.015, cutmin = ylim[1],
        boxwex = xyratio * 0.9,
        modeCol = c("cadetblue3", "hotpink2", "lawngreen"),
        curveLwd = 0.5, side = "second",
        bodyOpaque = 0.6)
par(oldpar)
```

The argument `xyratio` scales the vertical bixplot so that its density
body has the same visual height as the horizontal one.

------------------------------------------------------------------------

# Body sizing options for multimodal variables

The `bodysize` argument controls how the density bodies of individual
modes within a variable are sized relative to each other. The three
options are illustrated below on the (bimodal) petal length variable.

```{r, fig.width = 6, fig.height = 2.4}
par(mfrow = c(1, 3))
par(mar = c(2.5, 2, 2, 1))

for (bs in c("width_is_constant", "area_is_constant", "area_from_count")) {
  bixplot(diris[, 3], main = "", cut = 3, col = "gray75",
          bodyOpaque = 0.5, bodysize = bs, curveLwd = 0.5,
          ylim = c(-2.2, 2.4), yaxs = "i", names = "Petal.L")
  title(main = switch(bs,
                      width_is_constant = "equal width",
                      area_is_constant  = "equal area",
                      area_from_count   = "area from count"),
        cex.main = 1, line = 0.7)
}
par(oldpar)
```

With `"area_from_count"` (the default), the larger cluster receives a
proportionally higher area, making cluster sizes immediately apparent.

------------------------------------------------------------------------

# Bill length by island and sex

Here we use the formula interface to plot bill length split by the
interaction of sex and island. We compare a standard multi-group bixplot
with the `side = "both"` layout, which pairs male and female
distributions on opposing sides of a shared axis.

```{r, fig.width = 12, fig.height = 6}
par(las = 1, mfrow = c(1, 2))

islscore <- rep(NA, length(penguins$island))
islscore[penguins$island == "Torgersen"] <- 1
islscore[penguins$island == "Biscoe"]    <- 2
islscore[penguins$island == "Dream"]     <- 3
ylim    <- c(27, 64)
mynames <- c("Torger.F", "Torger.M", "Biscoe.F",
             "Biscoe.M", "Dream.F",  "Dream.M")
mycol      <- c("slateblue2", "orange")
mymodeCol  <- c("darkorchid1", "slateblue3",
                "goldenrod1",  "darkorange2")

par(mar = c(2.9, 2, 0.8, 1))
bixplot(bill_len ~ sex + reorder(island, islscore, mean),
        data = penguins, names = mynames, modeCol = mymodeCol,
        bodyOpaque = 0.6, ylim = ylim, bodyW = 0.9,
        col = mycol, main = "", rugCol = "black", las = 1)
legend("topleft", legend = c("female", "male"),
       fill = c("slateblue3", "orange"), cex = 1.5)

par(mar = c(2.9, 3, 0.8, 0.1))
bixplot(bill_len ~ sex + reorder(island, islscore, mean),
        data = penguins, main = "", rugCol = "black",
        bodyOpaque = 0.6, ylim = ylim, bodyW = 0.9,
        col = mycol, stickCol = "gray10", stickLwd = 1,
        side = "both", modeCol = mymodeCol, las = 1,
        names = c("Torgersen", "Biscoe", "Dream"))
legend("topleft", legend = c("female", "male"),
       fill = mycol, cex = 1.5)
par(oldpar)
```

With `side = "both"` the number of axis labels is halved and the
male/female pair for each island shares one axis, making comparisons
within an island more direct. Here, male penguins tend to have larger
bill lengths than females.

------------------------------------------------------------------------

# Rug colored by an external numeric variable

The `rugNumeric` argument colors each rug tick mark according to a
continuous covariate via a smooth color palette. Here the rug is colored
by body mass, and a color bar legend is added automatically.

```{r, fig.width = 5, fig.height = 5}
par(las = 1, mfrow = c(1, 1))
par(mar = c(3.2, 3.2, 2, 2.6))

scpenguins <- scale(penguins[, c(4, 3, 5)])
varnames   <- c("bill_depth", "bill_length", "flipper_length")

bixplot(scpenguins, rugNumeric = penguins$body_mass,
        main = "", boxW = 0.30, rugW = 0.24,
        names = varnames, boxLwd = 1.2, boxOpaque = 1,
        curveLwd = 2, colorbarW = 0.16, cex.colorbar = 0.8,
        bodysize = "width_is_constant",
        bodyCol = "grey90", modeCol = "grey90",
        xlab = "feature", ylab = "standardized value")
title(main = "penguins with rug color by body_mass       ",
      line = 0.6)
par(oldpar)
```

We see that heavier penguins (green) tend to have a larger flipper
length, revealing a positive association between these two variables.

------------------------------------------------------------------------

# Rug colored by a factor variable

The `rugFactor` argument colors each rug tick mark by a factor level.
Below we display the same three penguin measurements with rug color
indicating species, this time using a horizontal layout.

```{r, fig.width = 5, fig.height = 5}
par(mfrow = c(1, 1))
par(mar = c(3.2, 3.8, 2, 0.2))

rugFactorColors <- c("red", "blue", "forestgreen")

bixplot(scpenguins, rugFactor = penguins$species,
        main = "", boxW = 0.30, rugW = 0.24,
        ylim = c(-3, 4), names = varnames,
        boxLwd = 1.2, boxOpaque = 1, curveLwd = 2,
        horizontal = TRUE, bodysize = "width_is_constant",
        bodyCol = "grey90", modeCol = "grey90",
        xlab = "standardized value", ylab = "feature",
        rugFactorColors = rugFactorColors, las = 0)
title(main = "penguins with rug color by species", line = 0.6)
legend(x = 2.36, y = 2.8,
       legend = c("Adelie", "Chinstrap", "Gentoo"),
       fill = rugFactorColors, cex = 0.88)
par(oldpar)
```

The species labels in the rug indicate that Gentoo penguins tend to have
smaller `bill_depth` and longer `flipper_length`.

------------------------------------------------------------------------

# Top Gear car data

We now illustrate `bixplot` on the Top Gear dataset from the `robustHD`
package, which contains performance and price data for about three
hundred cars.

```{r}
data(TopGear, package = "robustHD")
scars        <- TopGear[, c(14, 12, 5, 7, 8)]
scars[, 3]   <- log(scars[, 3])
colnames(scars)[3] <- "log(Price)"
scars[, 1:4] <- scale(scars[, 1:4])
```

An initial bixplot reveals a severe outlier in the Weight variable:

```{r}
bixplot(scars[, 1:4])
```

```{r}
which.min(TopGear$Weight) # row 199
TopGear[199, c(1, 2, 14, 12, 5, 7)]
```

The Peugeot 107 has a recorded weight of 210, which is clearly a data
error. We set it to `NA` and replot:

```{r}
scars[199, 1] <- NA
bixplot(scars[, 1:4])
```

There is also an outlier in TopSpeed: the value 50 of the Renault Twizy,
that is a tiny city car. This value is correct, so it is retained.

```{r}
which.min(TopGear$TopSpeed) # row 220
TopGear[220, c(1, 2, 14, 12, 5, 7)]
```

We now produce a final two-panel figure: a bixplot of the four
standardized variables alongside a scatter plot of TopSpeed versus
Displacement with marginal bixplots added on each axis.

```{r, fig.width = 11, fig.height = 5.5}
par(las = 1, mfrow = c(1, 2))
par(mar = c(4, 2, 2, 0.1))

bixplot(scars[, 1:4], ylim = c(-3.8, 5.2), main = "",
        col = "darkgoldenrod2", yaxs = "i",
        modeCol = c("cadetblue3", "hotpink2"),
        bodyOpaque = 0.6, las = 1)
title(main = "standardized Top Gear variables",
      cex.main = 1, line = 0.7)

par(mar = c(4, 4, 2, 0.1))
x <- jitter(scars[, 4])
y <- jitter(scars[, 2])
mycol <- rep(NA, nrow(scars))
mycol[scars[, 5] == "Front"] <- "red"
mycol[scars[, 5] == "Rear"]  <- "forestgreen"
mycol[scars[, 5] == "4WD"]   <- "orange"

plot(x, y, xlim = c(-1.999, 3.999), ylim = c(-2.999, 4.5),
     xaxs = "i", yaxs = "i", xlab = "", ylab = "",
     pch = 16, cex = 1, col = mycol, las = 1)
title(xlab = "Displacement", line = 2)
title(ylab = "TopSpeed",     line = 2)
title(main = "TopSpeed versus Displacement",
      cex.main = 1, line = 0.7)

bixplot(scars[, 4], add = TRUE, at = -2.98, horizontal = TRUE,
        side = "second", boxwex = 2, bodyOpaque = 0.6)
bixplot(scars[, 2], add = TRUE, at = -1.985, horizontal = FALSE,
        side = "second", boxwex = 1.4, bodyOpaque = 0.6)

legend(x = -1.5, y = 4, title = "DriveWheel",
       legend = c("Front", "Rear", "4WD"),
       fill = c("red", "forestgreen", "orange"), cex = 1)
par(oldpar)
```

Rear-wheel-drive cars (green) span the full range of displacement and
are concentrated at the higher end of `TopSpeed`, whereas
front-wheel-drive cars (red) have rather low values of both `TopSpeed`
and `Displacement`.

------------------------------------------------------------------------

# Tooth growth and iris data with `side = "both"`

The `side = "both"` option is particularly useful for two-group
comparisons, as it places paired distributions on opposite sides of a
shared axis.

```{r, fig.width = 9.2, fig.height = 4.6}
par(las = 0, mfrow = c(1, 2))
par(mar = c(3.5, 2, 2, 1))

bixplot(len ~ supp * dose,
        data = ToothGrowth, side = "both",
        col = c("orange", "slateblue3"),
        main = "", xlab = "Vitamin C dose (mg)",
        ylab = "tooth length", bodyOpaque = 0.7,
        ylim = c(-1, 44), yaxs = "i", las = 0,
        stickCol = "black", stickLwd = 1)
title(main = "guinea pigs' tooth growth by supplement type",
      cex.main = 1, line = 0.7)
legend("topleft", title = "supplement",
       legend = c("OJ", "AA"),
       fill = c("orange", "slateblue3"), cex = 1)

par(mar = c(3.5, 2.8, 2, 0.1))
diris <- data.frame(scale(iris[, 1:4]))
colnames(diris) <- c("Sepal.L", "Sepal.W", "Petal.L", "Petal.W")

bixplot(diris, side = "both", main = "",
        bodysize = "width_is_constant",
        col = c("orange", "slateblue3"),
        modeCol = c("cadetblue3", "hotpink2",
                    "lawngreen", "gray60", "cyan3"),
        stickCol = "red", stickLwd = 1,
        bodyOpaque = 0.7, las = 0,
        xlab = "standardized measurements")
title(main = "iris data by length and width",
      cex.main = 1, line = 0.7)
legend("top", legend = c("length", "width"),
       fill = c("orange", "slateblue3"), cex = 1)
par(oldpar)
```

In the tooth growth data we see that the orange juice (OJ) supplement
yields higher tooth growth than ascorbic acid (AA), except at the
highest dose, where the medians of OJ and AA align and AA has a larger
spread than OJ. In the right panel, the iris data are now shown on two
vertical axes instead of four.

------------------------------------------------------------------------

# Iris data with rug coloring

Finally, we revisit the iris data with rug colors tied to an external
variable, first a continuous one (sepal length) and then a factor
(species).

```{r, fig.width = 5.5, fig.height = 5}
par(las = 1, mfrow = c(1, 1))
par(mar = c(2.2, 2.2, 2, 2.6))

bixplot(diris, main = "",
        rugNumeric = iris$Sepal.Length, colorbarW = 0.15,
        bodyCol = "grey80", modeCol = "grey80", las = 1)
title(main = "iris data with rug color by sepal length      ",
      line = 0.6)
par(oldpar)
```

Flowers with high sepal length (green) tend to have high petal length
and petal width. There is no such relation between sepal length and
sepal width.

```{r, fig.width = 5, fig.height = 5}
par(las = 1, mfrow = c(1, 1))
par(mar = c(2.2, 2.2, 2, 0.2))

rugFactorColors <- c("red", "blue", "forestgreen")
bixplot(diris, main = "", rugFactor = iris$Species,
        bodyCol = "grey80", modeCol = "grey80",
        rugFactorColors = rugFactorColors, las = 1)
title(main = "iris data with rug color by species", line = 0.6)
legend("topright",
       legend = c("setosa", "versicolor", "virginica"),
       fill = rugFactorColors, cex = 1)
par(oldpar)
```

The three species nicely match the clusters in `Petal.W` and `Petal.L`.
The colors inside `Sepal.L` are roughly similar, whereas `Sepal.W` again
behaves differently.

------------------------------------------------------------------------

# Titanic data

As a final example we use the Titanic dataset from the `classmap`
package. We compare the standardized age and log-fare distributions for
casualties and survivors using the `side = "both"` layout, and then
color the rug by cabin class.

```{r}
data(data_titanic, package = "classmap")
titanic         <- data_titanic[1:100, ]
titanic$logFare <- log(titanic$Fare + 1)
titanic$Pclass  <- as.factor(titanic$Pclass)
titanic[, c(5, 14)] <- scale(titanic[, c(5, 14)])

xt <- list(titanic[titanic$y == "casualty",  5],
           titanic[titanic$y == "survived",  5],
           titanic[titanic$y == "casualty",  14],
           titanic[titanic$y == "survived",  14])
names(xt) <- c("standardized Age.C",
               "standardized Age.S",
               "standardized log(Fare).C",
               "standardized log(Fare).S")
```

```{r, fig.width = 9.2, fig.height = 4.6}
par(las = 1, mfrow = c(1, 2))
par(mar = c(2.4, 2, 2, 1))

mycol <- c("coral2", "cadetblue3")

bixplot(xt, side = "both", main = "", col = mycol,
        stickLwd = 1, stickCol = "purple",
        boxW = 0.18, rugW = 0.10, las = 1)
title(main = "Titanic data by survival", line = 0.6)
legend("top", legend = c("casualty", "survived"),
       fill = mycol, cex = 0.9)

par(mar = c(2.4, 3, 2, 0.1))
bixplot(titanic[, c(5, 14)], main = "",
        rugFactor = titanic$Pclass, boxW = c(0.22, 0.18),
        names = c("standardized Age",
                  "standardized log(Fare)"), las = 1)
title(main = "Titanic data by cabin class", line = 0.6)
legend("top", title = "cabin class",
       legend = c("1", "2", "3"),
       fill = c("red", "blue", "forestgreen"), cex = 0.9)
par(oldpar)
```

In the left panel we see that the median of `Age` and especially its
mode are higher for the survivors, with a smaller effect of `Fare`. On
the right we see that both `Age` and `Fare` have an effect on cabin
class.

------------------------------------------------------------------------

---
title: "predsplot_examples"
author: "Rousseeuw, P.J."
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{predsplot_examples}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, echo = FALSE}
knitr::opts_chunk$set(
 fig.width  = 5 ,
 fig.height = 3.5,
 fig.align  = 'center'
)
oldpar <- list(mar = par()$mar, mfrow = par()$mfrow)
```


```{r}
library(classmap)
```

# TopGear data


We start by loading the data, and selecting the relevant features.
As pre-processing, we set values of accel==0 to missing because it is impossible.
We also remove the incorrect weight of the Peugeot 107.
```{r}
data(TopGear, package = "robustHD")
cars = TopGear; rm(TopGear)
dim(cars) # [1] 297  32
rownames(cars) = paste(cars[,1],cars[,2])
# Now the rownames are make and model of the cars.
colnames(cars)
colnames(cars)[4:17] = c("fuel","price","cyl","displ","drive",
                         "hp","torque","accel","topspeed",
                         "MPG","weight","length","width",
                         "height")
summary(cars$accel) # zero minimum is impossible
rownames(cars)[which(cars$accel == 0)]
# [1] "Citroen C5 Tourer" "Ford Mondeo"   "Lotus Elise"
# [4] "Renault Twizy"     "Ssangyong Rodius"
cars$accel[cars$accel == 0] = NA
summary(cars$accel) # OK

cars$weight[order(cars$weight)[1:4]]
rownames(cars)[order(cars$weight)[1:4]]
cars$weight[cars$weight == 210] = NA
summary(cars$weight) # OK
# save(cars, file = "Topgear.RData")
# load("Topgear.RData")
```

## Numerical example in the introduction


```{r}
summary(cars$topspeed) # in mph
summary(cars$displ)    # in cc
summary(cars$length)   # in mm
cars$length = cars$length/1000 # in meters

fit = lm(hp ~ topspeed + length + displ, data = cars)
summary(fit)

predsplot(fit, main = "Top Gear data")

fact = 0.51
width = fact*10; height = fact*8
maxnchar = 6
main = "Top Gear data"
pdf(file = "topgear_numerical_predsplot.pdf",
    width = width, height = height)
predsplot(fit, main=main, maxchar.level = maxnchar)
dev.off()


predscor(fit, sort.by.stdev = FALSE) 
# fact = 0.6
# width = fact*10; height = fact*8
# pdf(file = "topgear_numerical_predscor.pdf", 
#     width = height, height = height)
# predscor(fit, sort.by.stdev = FALSE)
# dev.off()

```

## Figure 2 and TopGear figures in Supplementary Material:


```{r}
cars$GPM = 1/cars$MPG

fit4 = lm(GPM ~ accel + drive + weight + fuel, data = cars)
summary(fit4)

## Figure 2:
## 
fact = 0.52
width = fact*10; height = fact*8
maxnchar = 6
main = "Top Gear data"
# pdf(file = "topgear_4_predsplot.pdf",
#     width = width, height = height)
predsplot(fit4, main=main, maxchar.level = maxnchar)
# dev.off()
```


## Three figures in the Supplementary Material:

```{r}
## 
# fact = 0.6
# width = fact*10; height = fact*8
# pdf(file = "topgear_4_predscor.pdf", 
#     width = height, height = height)
predscor(fit4, sort.by.stdev = FALSE)
# dev.off()

car = "Bentley Continental"
# pdf(file = "topgear_4_predsplot_1_dens.pdf",
#     width = width, height = height)
predsplot(fit4, main = main, casetoshow=car, 
          displaytype = "density", 
          maxchar.level = maxnchar, 
          xlab = paste0("prediction for ",car))
# dev.off()

car = "Kia Rio"
# pdf(file = "topgear_4_predsplot_2_stairs.pdf",
#     width = width, height = height)
predsplot(fit4, main = main, casetoshow=car, 
          staircase = TRUE, maxchar.level = maxnchar, 
          xlab = paste0("prediction for ",car))
# dev.off()
```

## Test combination of expressions and types

We add a logical variable (AlarmSystem) as well as a character variable (SatNav)

```{r}
summary(cars$AlarmSystem) 
cars$alarm = (cars$AlarmSystem == "standard")
summary(cars$alarm)

summary(cars$SatNav)
cars$navig = "y"
cars$navig[cars$SatNav == "no"] = "n"
summary(cars$navig)

table(cars$navig)
class(cars$navig) 
str(cars)

fitcombin = lm(1/MPG ~ accel + log(weight) + accel:torque + alarm + navig, data = cars)
summary(fitcombin)

fact = 0.45
width = fact*10; height = fact*8
main = "Example with transformation, interaction, logical, character"
# pdf(file = "topgear_logical+char_predsplot.pdf", width = width, height = height)
predsplot(fitcombin, main=main)
# dev.off()  

fact = 0.6
width = fact*10; height = fact*8
# pdf(file = "topgear_logical+char_predscor.pdf",
#     width = height, height = height)
predscor(fitcombin)
# dev.off()

```

# Titanic data


We start by loading the data, turning the pclass and sex variables into factors.

```{r}
data(data_titanic, package = "classmap")
titanic = data_titanic; rm(data_titanic)
dim(titanic) # 1309  13
# A data frame with 1309 observations on the following variables.
#   passengerId: a unique identifier for each passenger.
#   pclass: travel class of the passenger.
#   name: name of the passenger.
#   sex: sex of the passenger.
#   age: age of the passenger.
#   sibsp: number of siblings and spouses traveling with the passenger.
#   parch: number of parents and children traveling with the passenger.
#   ticket: ticket number of the passenger.
#   fare: fare paid for the ticket.
#   Cabin: cabin number of the passenger.
#   embarked: Port of embarkation. Takes the values
#             C (Cherbourg), Q (Queenstown) and
#             S (Southampton).
#   y: factor indicating casualty or survivor.
#   dataType: vector taking the values “train” or “test”
#             indicating whether the observation belongs
#             to training or test data.
#
colnames(titanic) = c("passengerId","pclass","name","sex",
                   "age", "sibsp", "parch", "ticket",
                   "fare", "cabin", "embarked", "survival",
                   "dataType")
str(titanic)

unique(titanic$pclass) # 3 1 2
titanic$pclass = factor(titanic$pclass, levels = unique(titanic$pclass))
#

unique(titanic$sex) 
titanic$sex = factor(titanic$sex, levels = unique(titanic$sex) )
titanic$sex = factor(titanic$sex, labels = c("M","F"))
head(titanic)
# save(titanic, file="Titanic.RData")
# load("Titanic.RData")
```

Now fit a logistic regression model.

```{r}
fit5 <- glm(survival ~ sex + age + sibsp + parch + pclass,
            family=binomial, data = titanic)
summary(fit5)
```

## Figure 3:

```{r} 
fact = 0.5
width = fact*10; height = fact*8
main = "Titanic data"
# pdf(file = "titanic_5_predsplot_dens.pdf",
#     width = width, height = height)
predsplot(fit5, main=main, displaytype = "density")

# The glm fit used the logit link function.
# dev.off()

# With other options for density estimation:
predsplot(fit5, main=main, displaytype = "density",
          bw = "SJ", adjust = 0.5)
```


## Three figures in Supplementary Material:

```{r} 
# pdf(file = "titanic_5_predsplot_1.pdf",
#     width = width, height = height)
predsplot(fit5, main = main, casetoshow=1)
# dev.off()

# pdf(file = "titanic_5_predsplot_2_stairs.pdf",
#     width = width, height = height)
predsplot(fit5, main = main, casetoshow=2, staircase = TRUE)
# dev.off()

# fact = 0.6
# width = fact*10; height = fact*8
# pdf(file = "titanic_5_predscor.pdf", 
#     width = height, height = height)
predscor(fit5, sort.by.stdev = FALSE)
# dev.off()
```

# German credit data


```{r}
data(german.credit, package = "fairml")
credit = german.credit; rm(german.credit)
dim(credit) # 1000 21
str(credit)
#
# Shorten variable names for plotting:
colnames(credit) <- c("checking","months","history","purpose","amount","savings","employment","rate","guarantors","residence","property","age","inst_plans","housing","nloans","job","nclients","phone","foreign","credit","sex")
#
# Give factors sex and purpose shorter labels for plotting:
credit$sex = factor(as.numeric(credit$sex), labels = c("F","M"))
levels(credit$purpose)
# [10] "retrainin"
levels(credit$purpose) = c("business","n.car","u.car","appliances","education","furniture","other","TV","repairs","retraining")
# save(credit, file = "German_credit.RData")
# load("German_credit.RData")
```

Now fit a logistic regression model

```{r}
fit7 <- glm(credit ~ months + purpose + amount + rate + age +
            nclients + sex, family=binomial, data = credit)
```

 
## Figure in Supplementary Material:

```{r} 
main = "German credit data"
# fact = 0.48
# width = fact*10; height = fact*8
# pdf(file = "germancredit_7_predsplot.pdf",
#     width = width, height = height)
predsplot(fit7, main = main)       
# dev.off()
```


## Figure 4:
```{r} 
# fact = 0.48
# width = fact*10; height = fact*8
# pdf(file = "germancredit_7_predsplot_1.pdf",
#     width = width, height = height)
predsplot(fit7, main = main, casetoshow=1, 
          displaytype = "density")
# dev.off()
```

## Figure 5:
```{r} 
# pdf(file = "germancredit_7_predsplot_2_stairs.pdf",
#     width = width, height = height)
predsplot(fit7, main = main, casetoshow=2, staircase = TRUE) 
# dev.off()
```

## Figure 6:

```{r}
# fact = 0.6
# width = fact*10; height = fact*8
# pdf(file = "germancredit_7_predscor_equalsizes.pdf", 
#     width = height, height = height)
predscor(fit7, sort.by.stdev = FALSE, cell.length = "equal")
# dev.off()
#
# pdf(file = "germancredit_7_predscor.pdf", 
#     width = height, height = height)
predscor(fit7, sort.by.stdev = FALSE)
# dev.off()
```

## Two figures in Supplementary Material:
```{r} 
## Prediction for a new case:
newc = list("u.car", 36, 2, 6000, 55, "F", 1)
names(newc) = c("purpose","months","rate","amount",
                "age","sex","nclients")
# fact = 0.48
# width = fact*10; height = fact*8
# pdf(file = "germancredit_7_predsplot_new_stairs.pdf",
#     width = width, height = height)
predsplot(fit7, main = main, casetoshow = newc, 
          staircase = TRUE)
# dev.off()

## Figure with profile:
# pdf(file = "germancredit_7_predsplot_1_dens_profile.pdf",
#     width = width, height = height)
predsplot(fit7, main = main, casetoshow=1, 
               displaytype = "density", profile = TRUE)
# dev.off()

```

# Artificial example to illustrate high correlation:


```{r}
credit$x1 = credit$months + credit$nclients
credit$x2 = credit$months - credit$nclients
cor(credit$x1,credit$x2) # 0.998199
# But minus that for prediction terms!

fitart <- glm(credit ~ x1 + x2 + purpose + amount + rate + age +
           sex, family=binomial, data = credit)

## Figure 7:
## 
# fact = 0.48
# width = fact*10; height = fact*8
# main = "German credit data with artificial variables x1 and x2"
# pdf(file = "germancredit_7_artif_predsplot.pdf", 
#     width = width, height = height)
predsplot(fitart, main = main)
# dev.off()

## Figure 8:
## 
# fact = 0.6
# width = fact*10; height = fact*8
# pdf(file = "germancredit_7_artif_predscor.pdf", 
#     width = height, height = height)
predscor(fitart, sort.by.stdev = FALSE)
# dev.off()

```

---
title: "Neural_net_examples"
author: "Raymaekers, J. and Rousseeuw, P.J."
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Neural_net_examples}
  %\VignetteEngine{knitr::knitr}
  %\VignetteEncoding{UTF-8}
---

```{r, echo = FALSE}
knitr::opts_chunk$set(
 fig.width  = 5 ,
 fig.height = 3.5,
 fig.align  = 'center'
)
oldpar <- list(mar = par()$mar, mfrow = par()$mfrow)
```


# Introduction


This vignette visualizes classification results of neural nets using tools from the classmap package. We will fit neural nets by the nnet package, which allows the training of simple neural networks with a single hidden layer.


```{r}
library(nnet)
library(classmap)
```

# Iris data
## Training data
As a first small example, we consider the Iris data.
We start by training a neural network with one hidden layer
containing 3 neurons on the full iris data.

```{r}
set.seed(123)
nn.iris <- nnet(Species ~ ., data = iris, size = 3)
names(nn.iris)
head(nn.iris$fitted.values) # matrix of posterior probabilities.
nn.iris$wts # the "weights" (coefficients, some are negative).
summary(nn.iris) 
```
In the above output, the coefficients are listed with their arrows:
i1,...,i4 are the 4 _i_nput variables.
h1, h2, h3 are the 3 _h_idden nodes.
o1, o2, o3 are the _o_utput nodes, one for each class
            (before the softmax function is applied).
b->h1, b->h2, b->h3 are the intercepts to the hidden layer.
b->o1, b->o2, b->o3 are the intercepts to the output layer.

We can use this to reconstruct the model from the coefficients.
We first extract the coefficients from each layer:

```{r}
train_x <- as.matrix(iris[, -5]) # original input: 4 variables.
# extract "weights" (coefficients) from each layer:
bias1  <- nn.iris$wts[c(1, 6, 11)] # intercepts to hidden layer
betas1 <- cbind(nn.iris$wts[2:5], nn.iris$wts[7:10], 
                nn.iris$wts[12:15]) # slopes to hidden layer
bias2  <- nn.iris$wts[c(16, 20, 24)] # intercepts to output layer
betas2 <- cbind(nn.iris$wts[17:19], nn.iris$wts[21:23], 
                nn.iris$wts[25:27]) # slopes to output layer.
```

Now we use the weights and the sigmoid function (for activation in the hidden layer) to reconstruct the
representations of the data in the hidden layer (H), as well as in the final layer before softmax (X). X is the set of G-dimensional points v_i given by (6) in the paper. This X will be used for the farness in the class maps.

```{r}
sigmoid <- function(x) 1 / (1 + exp(-x)) 

H <- t(t(train_x %*% betas1) + bias1) # = hidden layer
X <- t(t(sigmoid(H) %*% betas2) + bias2) # = layer before softmax.
pairs(X, col = unclass(iris$Species)) # G=3 classes ==> 3 dimensions.
```

To obtain the posterior probabilities, we can apply the 
softmax function to the last layer. We verify that
this yields the same result as the probabilities returned by 
nnet().

```{r}
softmax <- function(x) exp(x) / sum(exp(x)) 
outprobs <- t(apply(X, 1, softmax))
trainprobs <- nn.iris$fitted.values # this is after softmax
range(trainprobs - outprobs) # near 0, so we have a match.
```

Now we prepare the output for visualizations.

```{r}
y <- iris[, 5]
vcrtrain <- vcr.neural.train(X, y, trainprobs)

names(vcrtrain)
vcrtrain$predint[c(1:10, 51:60, 101:110)] # prediction as integer
vcrtrain$pred[c(1:10, 51:60, 101:110)]    # prediction as label
vcrtrain$altint[c(1:10, 51:60, 101:110)]  # alternative label as integer
vcrtrain$altlab[c(1:10, 51:60, 101:110)]  # alternative label

# Probability of Alternative Class (PAC) of each object:
vcrtrain$PAC[1:3] 
#
summary(vcrtrain$PAC)

# f(i, g) is the distance from case i to class g:
vcrtrain$fig[1:5, ] # for the first 5 objects:

# The farness of an object i is the f(i, g) to its own class: 
vcrtrain$farness[1:5]
#
summary(vcrtrain$farness)

# The "overall farness" of an object is defined as the 
# lowest f(i, g) it has to any class g (including its own):
summary(vcrtrain$ofarness)

confmat.vcr(vcrtrain)
```


The vcrtrain object can now be used to create the visualizations.

```{r}
cols <- c("red", "darkgreen", "blue")

# stacked mosaic plot:
stplot <- stackedplot(vcrtrain, classCols = cols,
                     main = "Stacked plot of nnet on iris data") 
stplot

# silhouette plot:
silplot(vcrtrain, classCols = cols, main =
          "Silhouette plot of nnet on iris data") 

# class maps:
classmap(vcrtrain, "setosa", classCols = cols)
# Very tight class (low PAC, no high farness).
classmap(vcrtrain, "versicolor", classCols = cols)
# Not so tight, one point is predicted as virginica.
classmap(vcrtrain, "virginica", classCols = cols)
# Also tight.
```


## New data


To illustrate the use of new data we create a
"fake" dataset which is a subset of the training data,
where not all classes occur, and ynew has NA's.

```{r}
test_x <- train_x[1:100, ]
ynew <- y[1:100]
ynew[c(1:10, 51:60)] <- NA
pairs(train_x, col = as.numeric(y) + 1, pch = 19) # 3 colors
pairs(test_x, col = as.numeric(ynew) + 1, pch = 19) # only red and green
```

We now calculate the predictions of this test data, and reconstruct the representation of the test data in the hidden and final layer.

```{r}
predprobs <- predict(nn.iris, test_x, type = "raw")
range(predprobs - trainprobs[1:100, ]) # perfect match

# Reconstruct this prediction:
Hnew <- t(t(test_x %*% betas1) + bias1) # hidden layer
Xnew <- t(t(sigmoid(Hnew) %*% betas2) + bias2) # layer before softmax
probsnew <- t(apply(Xnew, 1, softmax))
range(probsnew - predprobs) # ~0, so we have a match
```

Now prepare for visualization:

```{r}
vcrtest <- vcr.neural.newdata(Xnew, ynew, probsnew, vcrtrain)

plot(vcrtest$predint, vcrtrain$predint[1:100]); abline(0, 1) # identical, OK
plot(vcrtest$altint, vcrtrain$altint[1:100]); abline(0, 1) # identical when not NA, OK
plot(vcrtest$PAC, vcrtrain$PAC[1:100]); abline(0, 1) # OK
vcrtest$farness # length 100, with NA's where ynew is NA
plot(vcrtest$farness, vcrtrain$farness[1:100]); abline(0, 1) # identical where not NA
plot(vcrtest$fig, vcrtrain$fig[1:100, ]); abline(0, 1) # same, OK
vcrtest$ofarness # 100, withOUT NA's: ofarness even exists for cases with missing ynew, for which farness cannot exist.
plot(vcrtest$ofarness, vcrtrain$ofarness[1:100]); abline(0, 1) # same, OK

confmat.vcr(vcrtest) # as expected:
```

Now we can visualize the classification on the test data,
and compare the resulting plots with their counterparts on the 
training data.

```{r}
stplot # to compare with:
stackedplot(vcrtest, classCols = cols, separSize = 1.5,
            minSize = 1, main = "Stacked plot of nnet on subset of iris data")
silplot(vcrtest, classCols = cols, main = "Silhouette plot of nnet on subset of iris data")

classmap(vcrtrain, 1, classCols = cols)
classmap(vcrtest, 1, classCols = cols) # same, but fewer points
#
classmap(vcrtrain, 2, classCols = cols)
classmap(vcrtest, 2, classCols = cols) # same, but fewer points
#
# classmap(vcrtrain, 3, classCols = cols)
# classmap(vcrtest, 3, classCols = cols)
# # Class number 3 with label virginica has no objects to visualize.
``` 

# floral buds data: 

As a second example, we consider the floral buds data.

```{r}
data("data_floralbuds")
```

We start by training a neural network with one hidden layer consisting of 2 neurons.
```{r}
set.seed(123)
nn.buds <- nnet(y ~ ., data = data_floralbuds, size = 2)
names(nn.buds)
head(nn.buds$fitted.values) # matrix of posterior probabilities of each class
nn.buds$wts # the "weights" (coefficients, some negative)
summary(nn.buds) # A hidden layer with 2 neurons:
```


In the output above, the coefficients are listed with their arrows:
i1, ..., i6 are the 6 input variables
h1, h2 are the 2 hidden nodes
o1, .., o4 are the output nodes, one for each class (before softmax)
b->h1 and b->h2 are both intercepts to the hidden layer
b_o1, ..., b->o4 are the intercepts to the output layer

We first extract these weights from the output of nnet():
```{r}
train_x <- as.matrix(data_floralbuds[, -7])
# extract weights for each layer:
bias1  <- nn.buds$wts[c(1, 8)]
betas1 <- cbind(nn.buds$wts[2:7], nn.buds$wts[9:14])
bias2  <- nn.buds$wts[c(15, 18, 21, 24)]
betas2 <- cbind(nn.buds$wts[16:17], nn.buds$wts[19:20], 
                nn.buds$wts[22:23], nn.buds$wts[25:26])
```


We now use these weights and the sigmoid activation function
to reconstruct the data in the hidden and final layer (before applying the softmax function). We verify that, after applying softmax, we obtain the same posterior probabilities as those in the output of nnet().

```{r}
H <- t(t(train_x %*% betas1) + bias1) # hidden layer
X <- t(t(sigmoid(H) %*% betas2) + bias2) # layer before softmax
outprobs <- t(apply(X, 1, softmax))
trainprobs <- nn.buds$fitted.values # posterior probabilities
range(trainprobs - outprobs) # ~0
# OK, we have reconstructed the posterior probabilities
pairs(X, col = data_floralbuds$y)
```

Now prepare for visualization:

```{r}
y <- data_floralbuds$y
vcrtrain <- vcr.neural.train(X, y, trainprobs)
cols <- c("saddlebrown", "orange", "olivedrab4", "royalblue3")
``` 


```{r}
stackedplot(vcrtrain, classCols = cols, main =
              "Stacked plot of nnet on floral buds data")

# Silhouette plot:
silplot(vcrtrain, classCols = cols, 
        main = "Silhouette plot of nnet on floral buds data")



# Quasi residual plot:

PAC <- vcrtrain$PAC
feat <- rowSums(train_x); xlab <- "rowSums(X)"
# pdf("Floralbuds_quasi_residual_plot.pdf", width = 5, height = 4.8)
qresplot(PAC, feat, xlab = xlab, plotErrorBars = TRUE, fac = 2, main = "Floral buds: quasi residual plot")
# images with higher sum are easier to classify
# dev.off()
cor.test(feat, PAC, method = "spearman") 
# rho = -0.238, p-value < 2e-8 
# ==> decreasing trend is significant



# Class maps:

classmap(vcrtrain, "branch", classCols = cols)

classmap(vcrtrain, "bud", classCols = cols)

classmap(vcrtrain, "scales", classCols = cols)

classmap(vcrtrain, "support", classCols = cols)
```
***

---
title: "K_nearest_neighbors_examples"
author: "Raymaekers, J. and Rousseeuw, P.J."
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{K_nearest_neighbors_examples}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---
  
```{r, echo = FALSE}
knitr::opts_chunk$set(
fig.width  = 5 ,
fig.height = 3.5,
fig.align  = 'center'
)
oldpar <- list(mar = par()$mar, mfrow = par()$mfrow)
```

# Introduction

This vignette visualizes classification results from k-nearest neighbors, using tools from the package.

```{r}
library("classmap")
```

# Iris data

As a first small example, we consider the Iris data.
We first load the data and inspect it. We also scale the data, to make the results of the k-nearest neighbor classifier scale-invariant.


```{r}
data(iris)
X <- iris[, 1:4]
y <- iris[, 5]
is.factor(y) 
table(y) 

X <- scale(X) # if you want to scale, best to do it yourself.
pairs(X, col = as.numeric(y) + 1, pch = 19)
dis <- dist(X, method = "euclidean")
```

The VCR object for the k-nearest neighbors classifier can be constructed based on a distance matrix or the original data. We use k = 5 here:

```{r}
vcr.train <- vcr.knn.train(dis, y, k = 5)

vcr.train <- vcr.knn.train(X, y, k = 5) # Gives the same result.
```

Let's inspect some of the elements in the output of the vcr.train function;
We first look at the prediction as integer, prediction as label, the 
alternative label as integer and the alternative label.

```{r}
names(vcr.train)
vcr.train$predint 
vcr.train$pred[c(1:10, 51:60, 101:110)]
vcr.train$altint  
```

The NA's stem from k-neighborhoods in which all members are from the given class. In such cases there is no alternative class.

```{r}
vcr.train$altlab  # has NA's for the same cases:
```

The output also contains the value of k, which was an input value, as well as the "true" value of k used for each instance, which can differ from the input value of k in case of ties:

```{r}
vcr.train$k 
vcr.train$ktrues[1:10]
```


We can also extract the probability of Alternative Class (PAC) of each object, which we will use later to construct the class map:

```{r}
vcr.train$PAC # length 150, all between 0 and 1, no NA's
```

The second ingredient of the class map is the farness, which is computed from the fig values defined as f(i, g)= the distance from case i to class g:
```{r}
vcr.train$fig[1:5, ]
```

The farness of an object i is the f(i, g) to its own class:

```{r}
vcr.train$farness[1:5]
summary(vcr.train$farness)
```

The "overall farness" of an object is defined as the  lowest f(i, g) it has to any class g (including its own). This always exists.

```{r}
summary(vcr.train$ofarness)
```

Using confmat.vcr(), we can construct the usual confusion matrix and obtain the accuracy of the classification:

```{r}
confmat.vcr(vcr.train, showOutliers = FALSE)
```

By using showOutliers = TRUE and setting a cutoff, objects with ofarness > cutoff are flagged as outliers and shown in a separate column of the confusion matrix:

```{r}
confmat.vcr(vcr.train, cutoff = 0.98)
```

With the default cutoff = 0.99 only one object is flagged in this example:

```{r}
confmat.vcr(vcr.train)
```

Note that the accuracy is computed before any objects are flagged, so it does not depend on the cutoff. Finally, we can also show the class numbers instead of the labels in the confusion matrix. This option can be useful for long level
names.

```{r}
confmat.vcr(vcr.train, showClassNumbers = TRUE)
```

The stacked mosaic plot is a graphical representation of the confusion matrix, and can be made using the stackedplot() function. The outliers can optionally be shown as separate gray areas on top of each rectangle:

```{r}
par(mfrow = c(1, 1))
cols <- c("red", "darkgreen", "blue")
stackedplot(vcr.train, classCols = cols, separSize = 1.5,
            minSize = 1, showOutliers = FALSE, showLegend = TRUE,
            main = "Stacked plot of kNN on iris data")

stackedplot(vcr.train, classCols = cols, separSize = 1.5,
            minSize = 1, showLegend = TRUE,
            main = "Stacked plot of kNN on iris data")
par(oldpar)
```
By default, no legend is added to the stacked mosaic plot, because we see the colors at the bottom of each given class anyway:

```{r}
stplot <- stackedplot(vcr.train, classCols = cols, 
                      separSize = 1.5, minSize = 1,
                      main = "Stacked plot of kNN on iris data")
stplot 
```

We now construct the silhouette plot:
```{r}
# pdf("Iris_kNN_silhouettes.pdf", width=5.0, height=4.6)
silplot(vcr.train, classCols = cols, 
        main = "Silhouette plot of kNN on iris data")      
# dev.off()
```

Now we turn to the construction of the class maps.

For class 1, only one point has PAC > 0, around 0.4.
All the others are classified perfectly:
```{r}
classmap(vcr.train, 1, classCols = cols)
```

Class 2 shows only 4 points with PAC > 0.5:
```{r}
classmap(vcr.train, 2, classCols = cols)
```

Finally, class 3 also has 4 points with PAC > 0.5:
```{r}
classmap(vcr.train, 3, classCols = cols)
```

***

To illustrate the use of new data we create a `fake' dataset which is a subset of training data, where not all classes occur, and ynew has NA's.
First create and inspect this test data:
```{r}
Xnew <- X[c(1:50, 101:150),]
ynew <- y[c(1:50, 101:150)]
ynew[c(1:10, 51:60)] <- NA
pairs(X, col = as.numeric(y) + 1, pch = 19) # 3 colors
pairs(Xnew, col = as.numeric(ynew) + 1, pch = 19) # only red and blue
```

Now build the VCR object on the new data, using the output vcr.train which was built on the training data.

```{r}
vcr.test <- vcr.knn.newdata(Xnew, ynew, vcr.train, LOO = TRUE)
```


We perform a few sanity checks by comparing the vcr object of the training data with that of the test data. Note that the following only match when LOO = TRUE since we use "leave-one-out" in training! The default is LOO = FALSE.

```{r}
plot(vcr.test$predint, vcr.train$predint[c(1:50, 101:150)]); abline(0, 1)
plot(vcr.test$altint, vcr.train$altint[c(1:50, 101:150)]); abline(0, 1)
plot(vcr.test$PAC, vcr.train$PAC[c(1:50, 101:150)]); abline(0, 1) 
vcr.test$farness 
plot(vcr.test$farness, vcr.train$farness[c(1:50, 101:150)]); abline(0, 1)
plot(vcr.test$fig, vcr.train$fig[c(1:50, 101:150), ]); abline(0, 1)
vcr.test$ofarness  # This does exist for every case, even if is given label is NA:
plot(vcr.test$ofarness, vcr.train$ofarness[c(1:50, 101:150)]); abline(0, 1)
```

We now construct and inspect the confusion matrix and the stacked mosaic plot. We plot the mosaic plot on the training data again, for comparison:

```{r}
confmat.vcr(vcr.test) 

stplot 
stackedplot(vcr.test, classCols = cols, separSize = 1.5, minSize = 1, main = "Stacked plot of kNN on iris subset") 
```


And we also make the silhouette plot on the test data:
```{r}
# pdf("Iris_test_kNN_silhouettes.pdf", width=5.0, height=4.6)
silplot(vcr.test, classCols = cols, 
        main = "Silhouette plot of kNN on iris subset")      
# dev.off()
```


For each class, we now make the class map on the test data,
and compare with the class map on the training data.
First for class 1:
```{r}
classmap(vcr.train, 1, classCols = cols)
classmap(vcr.test, 1, classCols = cols) 
```

Now for class 2. This throws an error on the test data, as there are no observations with the label versicolor in the test data.
```{r, error = TRUE}
classmap(vcr.train, 2, classCols = cols)
classmap(vcr.test, 2, classCols = cols)
# Class number 2 with label versicolor has no objects to visualize.
```

Finally for class 3. It looks the same as the class map for the training data, but there are fewer points:

```{r}
classmap(vcr.train, 3, classCols = cols)
classmap(vcr.test, 3, classCols = cols) # same, but fewer points
```
***

# Spam data

In the example we analyze the spam data. It can be obtained via the kernlab package. It contains 4601 emails, 57 variables and a categorical variable indicating whether an email is nonspam or spam:

```{r}
library(kernlab)
#?kernlab::spam 
data(spam)
```

We now create a matrix of predictors and a response vector. Inspection shows that the predictors contain 394 duplicate rows:

```{r}
y <- spam$type
table(y)
X <- spam[, -58] 
sum(duplicated(X))
```


Now we can construct the vcr object based on the k-nearest neighbor classifier. All knn computations will be done on scale(X), but we  keep the original data X for interpreting the results.

```{r}
vcr.obj <- vcr.knn.train(scale(X), y, k = 5) 
names(vcr.obj) 
vcr.obj$predint[1:100]; length(vcr.obj$predint) 
vcr.obj$altint[1:100]
vcr.obj$k 
summary(vcr.obj$ktrues) 
```

We now inspect the confusion matrices, with and without the outliers in a separate column:
```{r}
confmat.vcr(vcr.obj, showOutliers = FALSE)

confmat.vcr(vcr.obj)
```

Now construct the stacked mosaic plot. The nonspam emails get the blue color, whereas the spam class is shown in red:

```{r}
cols <- c("deepskyblue2", "red") 
# nonspam is blue, spam is red

stackedplot(vcr.obj, separSize = 1.5, minSize = 1,
            classCols = cols, showOutliers = FALSE,
            main = "Spam data")

stackedplot(vcr.obj, separSize = 1.5, minSize = 2,
            classCols = cols, main = "Spam data")
```

We now make the silhouette plot:

```{r}
# pdf("Spam_kNN_silhouettes.pdf", width=5.0, height=4.6)
silplot(vcr.obj, classCols = cols, 
        main = "Silhouette plot of kNN on spam data")      
# dev.off()
```

Now we build the class maps. First for the nonspam messages:

```{r, fig.height = 4, fig.width = 6}
# To identify the points that stand out:
# classmap(vcr.obj, 1, classCols = cols, identify = T)
# Press "Esc" to get out.
#
# Class map in paper:
# pdf("Spamdata_classmap_ham.pdf")
par(mar = c(3.5, 4.3, 2.0, 0.2))
coords <- classmap(vcr.obj, 1, classCols = cols,
                   main = "predictions of non-spam mails",
                   cex = 1.5, cex.lab = 1.5, cex.axis = 1.5,
                   cex.main = 1.5, maxfactor = 1.03)
# From identify = TRUE above we can mark points:
indstomark <- c(202, 1434, 1596, 2651, 1576, 1804)
labs  <- letters[1:length(indstomark)]
xvals <- coords[indstomark, 1] +
  c(0, 0.125, 0.125, 0.125, 0.125,0.0625) # visual finetuning
yvals <- coords[indstomark, 2] +  
  c(-0.03, 0, 0, 0, 0, 0.03)
text(x = xvals, y = yvals, labels = labs, cex = 1.5)
legend("topleft", fill = cols,
       legend = c("ham", "spam"), cex = 1,
       ncol = 1, bg = "white")
# dev.off()
par(oldpar)
# To interpret the marked points:
# markInds = which(y == "nonspam")[indstomark]
# X[markInds, ] 
```

Now for the spam messages:

```{r, fig.height = 4, fig.width = 6}
#
# To identify the points that stand out:
# classmap(vcr.obj, 2, classCols = cols, identify = TRUE)
# Press "Esc" to get out.
#
# Class map in paper:
# pdf("Spamdata_classmap_spam.pdf")
par(mar = c(3.5, 4.3, 2.0, 0.2))
coords <- classmap(vcr.obj, 2, classCols = cols,
                   main = "predictions of spam mails",
                   cex = 1.5, cex.lab = 1.5, cex.axis = 1.5,
                   cex.main = 1.5, maxfactor = 1.03)
indstomark <- c(1306, 1294, 1754, 177)
labs  <- letters[6 + (1:length(indstomark))]
xvals <- coords[indstomark, 1] + c(0.1, 0.1, 0.1, 0.1)
yvals <- coords[indstomark, 2] + c(-0.03, 0, 0, 0.03)
text(x = xvals, y = yvals, labels = labs, cex = 1.5)
legend("topleft", fill = cols,
       legend = c("ham", "spam"), cex = 1,
       ncol = 1, bg = "white")
# dev.off()
par(oldpar)
```
***

---
title: "Neural_net_examples"
author: "Raymaekers, J. and Rousseeuw, P.J."
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Neural_net_examples}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, echo = FALSE}
knitr::opts_chunk$set(
 fig.width  = 5 ,
 fig.height = 3.5,
 fig.align  = 'center'
)
oldpar <- list(mar = par()$mar, mfrow = par()$mfrow)
```


# Introduction


This vignette visualizes classification results of neural nets using tools from the classmap package. We will fit neural nets by the nnet package, which allows the training of simple neural networks with a single hidden layer.


```{r}
library(nnet)
library(classmap)
```

# Iris data
## Training data
As a first small example, we consider the Iris data.
We start by training a neural network with one hidden layer
containing 3 neurons on the full iris data.

```{r}
set.seed(123)
nn.iris <- nnet(Species ~ ., data = iris, size = 3)
names(nn.iris)
head(nn.iris$fitted.values) # matrix of posterior probabilities.
nn.iris$wts # the "weights" (coefficients, some are negative).
summary(nn.iris) 
```
In the above output, the coefficients are listed with their arrows:
i1,...,i4 are the 4 _i_nput variables.
h1, h2, h3 are the 3 _h_idden nodes.
o1, o2, o3 are the _o_utput nodes, one for each class
            (before the softmax function is applied).
b->h1, b->h2, b->h3 are the intercepts to the hidden layer.
b->o1, b->o2, b->o3 are the intercepts to the output layer.

We can use this to reconstruct the model from the coefficients.
We first extract the coefficients from each layer:

```{r}
train_x <- as.matrix(iris[, -5]) # original input: 4 variables.
# extract "weights" (coefficients) from each layer:
bias1  <- nn.iris$wts[c(1, 6, 11)] # intercepts to hidden layer
betas1 <- cbind(nn.iris$wts[2:5], nn.iris$wts[7:10], 
                nn.iris$wts[12:15]) # slopes to hidden layer
bias2  <- nn.iris$wts[c(16, 20, 24)] # intercepts to output layer
betas2 <- cbind(nn.iris$wts[17:19], nn.iris$wts[21:23], 
                nn.iris$wts[25:27]) # slopes to output layer.
```

Now we use the weights and the sigmoid function (for activation in the hidden layer) to reconstruct the
representations of the data in the hidden layer (H), as well as in the final layer before softmax (X). X is the set of G-dimensional points v_i given by (6) in the paper. This X will be used for the farness in the class maps.

```{r}
sigmoid <- function(x) 1 / (1 + exp(-x)) 

H <- t(t(train_x %*% betas1) + bias1) # = hidden layer
X <- t(t(sigmoid(H) %*% betas2) + bias2) # = layer before softmax.
pairs(X, col = unclass(iris$Species)) # G=3 classes ==> 3 dimensions.
```

To obtain the posterior probabilities, we can apply the 
softmax function to the last layer. We verify that
this yields the same result as the probabilities returned by 
nnet().

```{r}
softmax <- function(x) exp(x) / sum(exp(x)) 
outprobs <- t(apply(X, 1, softmax))
trainprobs <- nn.iris$fitted.values # this is after softmax
range(trainprobs - outprobs) # near 0, so we have a match.
```

Now we prepare the output for visualizations.

```{r}
y <- iris[, 5]
vcrtrain <- vcr.neural.train(X, y, trainprobs)

names(vcrtrain)
vcrtrain$predint[c(1:10, 51:60, 101:110)] # prediction as integer
vcrtrain$pred[c(1:10, 51:60, 101:110)]    # prediction as label
vcrtrain$altint[c(1:10, 51:60, 101:110)]  # alternative label as integer
vcrtrain$altlab[c(1:10, 51:60, 101:110)]  # alternative label

# Probability of Alternative Class (PAC) of each object:
vcrtrain$PAC[1:3] 
#
summary(vcrtrain$PAC)

# f(i, g) is the distance from case i to class g:
vcrtrain$fig[1:5, ] # for the first 5 objects:

# The farness of an object i is the f(i, g) to its own class: 
vcrtrain$farness[1:5]
#
summary(vcrtrain$farness)

# The "overall farness" of an object is defined as the 
# lowest f(i, g) it has to any class g (including its own):
summary(vcrtrain$ofarness)

confmat.vcr(vcrtrain)
```


The vcrtrain object can now be used to create the visualizations.

```{r}
cols <- c("red", "darkgreen", "blue")

# stacked mosaic plot:
stplot <- stackedplot(vcrtrain, classCols = cols,
                     main = "Stacked plot of nnet on iris data") 
stplot

# silhouette plot:
silplot(vcrtrain, classCols = cols, main =
          "Silhouette plot of nnet on iris data") 

# class maps:
classmap(vcrtrain, "setosa", classCols = cols)
# Very tight class (low PAC, no high farness).
classmap(vcrtrain, "versicolor", classCols = cols)
# Not so tight, one point is predicted as virginica.
classmap(vcrtrain, "virginica", classCols = cols)
# Also tight.
```


## New data


To illustrate the use of new data we create a
"fake" dataset which is a subset of the training data,
where not all classes occur, and ynew has NA's.

```{r}
test_x <- train_x[1:100, ]
ynew <- y[1:100]
ynew[c(1:10, 51:60)] <- NA
pairs(train_x, col = as.numeric(y) + 1, pch = 19) # 3 colors
pairs(test_x, col = as.numeric(ynew) + 1, pch = 19) # only red and green
```

We now calculate the predictions of this test data, and reconstruct the representation of the test data in the hidden and final layer.

```{r}
predprobs <- predict(nn.iris, test_x, type = "raw")
range(predprobs - trainprobs[1:100, ]) # perfect match

# Reconstruct this prediction:
Hnew <- t(t(test_x %*% betas1) + bias1) # hidden layer
Xnew <- t(t(sigmoid(Hnew) %*% betas2) + bias2) # layer before softmax
probsnew <- t(apply(Xnew, 1, softmax))
range(probsnew - predprobs) # ~0, so we have a match
```

Now prepare for visualization:

```{r}
vcrtest <- vcr.neural.newdata(Xnew, ynew, probsnew, vcrtrain)

plot(vcrtest$predint, vcrtrain$predint[1:100]); abline(0, 1) # identical, OK
plot(vcrtest$altint, vcrtrain$altint[1:100]); abline(0, 1) # identical when not NA, OK
plot(vcrtest$PAC, vcrtrain$PAC[1:100]); abline(0, 1) # OK
vcrtest$farness # length 100, with NA's where ynew is NA
plot(vcrtest$farness, vcrtrain$farness[1:100]); abline(0, 1) # identical where not NA
plot(vcrtest$fig, vcrtrain$fig[1:100, ]); abline(0, 1) # same, OK
vcrtest$ofarness # 100, withOUT NA's: ofarness even exists for cases with missing ynew, for which farness cannot exist.
plot(vcrtest$ofarness, vcrtrain$ofarness[1:100]); abline(0, 1) # same, OK

confmat.vcr(vcrtest) # as expected:
```

Now we can visualize the classification on the test data,
and compare the resulting plots with their counterparts on the 
training data.

```{r}
stplot # to compare with:
stackedplot(vcrtest, classCols = cols, separSize = 1.5,
            minSize = 1, main = "Stacked plot of nnet on subset of iris data")
silplot(vcrtest, classCols = cols, main = "Silhouette plot of nnet on subset of iris data")

classmap(vcrtrain, 1, classCols = cols)
classmap(vcrtest, 1, classCols = cols) # same, but fewer points
#
classmap(vcrtrain, 2, classCols = cols)
classmap(vcrtest, 2, classCols = cols) # same, but fewer points
#
# classmap(vcrtrain, 3, classCols = cols)
# classmap(vcrtest, 3, classCols = cols)
# # Class number 3 with label virginica has no objects to visualize.
``` 

# floral buds data: 

As a second example, we consider the floral buds data.

```{r}
data("data_floralbuds")
```

We start by training a neural network with one hidden layer consisting of 2 neurons.
```{r}
set.seed(123)
nn.buds <- nnet(y ~ ., data = data_floralbuds, size = 2)
names(nn.buds)
head(nn.buds$fitted.values) # matrix of posterior probabilities of each class
nn.buds$wts # the "weights" (coefficients, some negative)
summary(nn.buds) # A hidden layer with 2 neurons:
```


In the output above, the coefficients are listed with their arrows:
i1, ..., i6 are the 6 input variables
h1, h2 are the 2 hidden nodes
o1, .., o4 are the output nodes, one for each class (before softmax)
b->h1 and b->h2 are both intercepts to the hidden layer
b_o1, ..., b->o4 are the intercepts to the output layer

We first extract these weights from the output of nnet():
```{r}
train_x <- as.matrix(data_floralbuds[, -7])
# extract weights for each layer:
bias1  <- nn.buds$wts[c(1, 8)]
betas1 <- cbind(nn.buds$wts[2:7], nn.buds$wts[9:14])
bias2  <- nn.buds$wts[c(15, 18, 21, 24)]
betas2 <- cbind(nn.buds$wts[16:17], nn.buds$wts[19:20], 
                nn.buds$wts[22:23], nn.buds$wts[25:26])
```


We now use these weights and the sigmoid activation function
to reconstruct the data in the hidden and final layer (before applying the softmax function). We verify that, after applying softmax, we obtain the same posterior probabilities as those in the output of nnet().

```{r}
H <- t(t(train_x %*% betas1) + bias1) # hidden layer
X <- t(t(sigmoid(H) %*% betas2) + bias2) # layer before softmax
outprobs <- t(apply(X, 1, softmax))
trainprobs <- nn.buds$fitted.values # posterior probabilities
range(trainprobs - outprobs) # ~0
# OK, we have reconstructed the posterior probabilities
pairs(X, col = data_floralbuds$y)
```

Now prepare for visualization:

```{r}
y <- data_floralbuds$y
vcrtrain <- vcr.neural.train(X, y, trainprobs)
cols <- c("saddlebrown", "orange", "olivedrab4", "royalblue3")
``` 


```{r}
stackedplot(vcrtrain, classCols = cols, main =
              "Stacked plot of nnet on floral buds data")

# Silhouette plot:
silplot(vcrtrain, classCols = cols, 
        main = "Silhouette plot of nnet on floral buds data")



# Quasi residual plot:

PAC <- vcrtrain$PAC
feat <- rowSums(train_x); xlab <- "rowSums(X)"
# pdf("Floralbuds_quasi_residual_plot.pdf", width = 5, height = 4.8)
qresplot(PAC, feat, xlab = xlab, plotErrorBars = TRUE, fac = 2, main = "Floral buds: quasi residual plot")
# images with higher sum are easier to classify
# dev.off()
cor.test(feat, PAC, method = "spearman") 
# rho = -0.238, p-value < 2e-8 
# ==> decreasing trend is significant



# Class maps:

classmap(vcrtrain, "branch", classCols = cols)

classmap(vcrtrain, "bud", classCols = cols)

classmap(vcrtrain, "scales", classCols = cols)

classmap(vcrtrain, "support", classCols = cols)
```
***

\name{vcr.svm.newdata}
\alias{vcr.svm.newdata}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{
Prepare for visualization of a support vector machine classification on new data.
}
\description{
Carries out a support vector machine classification of new data using the output of \code{\link{vcr.svm.train}} on the training data, and computes the quantities needed for its visualization.
}
\usage{
vcr.svm.newdata(Xnew, ynew = NULL, vcr.svm.train.out)
}

\arguments{

\item{Xnew}{data matrix of the new data, with the same number of columns as in the training data. Missing values in \code{Xnew} are not allowed.}
\item{ynew}{factor with class membership of each new case. Can be \code{NA} for some or all cases. If \code{NULL}, is assumed to be \code{NA} everywhere.}
\item{vcr.svm.train.out}{ output of \code{\link{vcr.svm.train}} on the training data.
}
}

\value{
A list with components: \cr
\item{yintnew}{number of the given class of each case. Can contain \code{NA}'s.}
\item{ynew}{given class label of each case. Can contain \code{NA}'s.}
\item{levels}{levels of the response, from \code{vcr.svm.train.out}.}
\item{predint}{predicted class number of each case. Always exists.}
\item{pred}{predicted label of each case.}
\item{altint}{number of the alternative class. Among the classes different from the given class, it is the one with the highest posterior probability. Is \code{NA} for cases whose \code{ynew} is missing.}
\item{altlab}{alternative label if yintnew was given, else \code{NA}.}
\item{PAC}{probability of the alternative class. Is \code{NA} for cases whose \code{ynew} is missing.}
\item{fig}{distance of each case \eqn{i} from each class \eqn{g}. Always exists.}
\item{farness}{farness of each case from its given class. Is \code{NA} for cases whose \code{ynew} is missing.}
\item{ofarness}{for each case \eqn{i}, its lowest \code{fig[i,g]} to any class \eqn{g}. Always exists.}
}

\references{
Raymaekers J., Rousseeuw P.J., Hubert M. (2021). Class maps for visualizing classification results. \emph{Technometrics}, 64(2), 151–165. \doi{10.1080/00401706.2021.1927849}
}

\author{
 Raymaekers J., Rousseeuw P.J.
}

\seealso{
\code{\link{vcr.svm.train}}, \code{\link{classmap}}, \code{\link{silplot}}, \code{\link{stackedplot}}, \code{\link[e1071:svm]{e1071::svm}}
}

\examples{

library(e1071)
set.seed(1); X <- matrix(rnorm(200 * 2), ncol = 2)
X[1:100, ] <- X[1:100, ] + 2
X[101:150, ] <- X[101:150, ] - 2
y <- as.factor(c(rep("blue", 150), rep("red", 50)))
# We now fit an SVM with radial basis kernel to the data:
set.seed(1) # to make the result of svm() reproducible.
svmfit <- svm(y~., data = data.frame(X = X, y = y),
scale = FALSE, kernel = "radial", cost = 10,
gamma = 1, probability = TRUE)
vcr.train <- vcr.svm.train(X, y, svfit = svmfit)
# As "new" data we take a subset of the training data:
inds <- c(1:25, 101:125, 151:175)
vcr.test <- vcr.svm.newdata(X[inds, ], y[inds], vcr.train)
plot(vcr.test$PAC, vcr.train$PAC[inds]); abline(0, 1) # match
plot(vcr.test$farness, vcr.train$farness[inds]); abline(0, 1)
confmat.vcr(vcr.test)
cols <- c("deepskyblue3", "red")
stackedplot(vcr.test, classCols = cols)
classmap(vcr.train, "blue", classCols = cols) # for comparison
classmap(vcr.test, "blue", classCols = cols)
classmap(vcr.train, "red", classCols = cols) # for comparison
classmap(vcr.test, "red", classCols = cols)


# For more examples, we refer to the vignette:
\dontrun{
vignette("Support_vector_machine_examples")
}
}

\name{vcr.knn.newdata}
\alias{vcr.knn.newdata}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{
Carry out a k-nearest neighbor classification on new data, and prepare to visualize its results.
}
\description{
Predicts class labels for new data by k nearest neighbors, using the output of \code{\link{vcr.knn.train}} on the training data. For cases in the new data whose given label \code{ynew} is not \code{NA}, additional output is produced for constructing graphical displays such as the \code{\link{classmap}}.
}
\usage{
vcr.knn.newdata(Xnew, ynew = NULL, vcr.knn.train.out, LOO = FALSE)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{Xnew}{If the training data was a matrix of coordinates, \code{Xnew} must be such a matrix with the same number of columns. If the training data was a set of dissimilarities, \code{Xnew} must be a rectangular matrix of dissimilarities, with each row containing the dissmilarities of a new case to all training cases. Missing values are not allowed.}
\item{ynew}{factor with class membership of each new case. Can be \code{NA} for some or all cases. If \code{NULL}, is assumed to be \code{NA} everywhere.}
\item{vcr.knn.train.out}{output of \code{\link{vcr.knn.train}} on the training data.}
\item{LOO}{leave one out. Only used when testing this function on a subset of the training data. Default is \code{LOO=FALSE}.}
}

\value{
A list with components: \cr

\item{yintnew}{number of the given class of each case. Can contain \code{NA}'s.}
\item{ynew}{given class label of each case. Can contain \code{NA}'s.}
\item{levels}{levels of the response, from \code{vcr.knn.train.out}.}
\item{predint}{predicted class number of each case. Always exists.}
\item{pred}{predicted label of each case.}
\item{altint}{number of the alternative class. Among the classes different from the given class, it is the one with the highest posterior probability. Is \code{NA} for cases whose \code{ynew} is missing.}
\item{altlab}{label of the alternative class. Is \code{NA} for cases whose \code{ynew} is missing.}
\item{PAC}{probability of the alternative class. Is \code{NA} for cases whose \code{ynew} is missing.}
\item{fig}{distance of each case \eqn{i} from each class \eqn{g}. Always exists.}
\item{farness}{farness of each case from its given class. Is \code{NA} for cases whose \code{ynew} is missing.}
\item{ofarness}{for each case \code{i}, its lowest  \code{fig[i,g]} to any class \code{g}. Always exists.}
\item{k}{the requested number of nearest neighbors, from \code{vcr.knn.train.out}.}
\item{ktrues}{for each case this contains the actual number of elements in its neighborhood. This can be higher than \code{k} due to ties.}
\item{counts}{a matrix with 3 columns, each row representing a case. For the neighborhood of each case it says how many members it has from the given class, the predicted class, and the alternative class. The first and third entry is \code{NA} for cases whose \code{ynew} is missing.}

}

\references{
Raymaekers J., Rousseeuw P.J., Hubert M. (2021). Class maps for visualizing classification results. \emph{Technometrics}, 64(2), 151–165. \doi{10.1080/00401706.2021.1927849}
}

\author{
 Raymaekers J., Rousseeuw P.J.
}

\seealso{
\code{\link{vcr.knn.train}}, \code{\link{classmap}}, \code{\link{silplot}}, \code{\link{stackedplot}}
}

\examples{
data("data_floralbuds")
X <- data_floralbuds[, 1:6]; y <- data_floralbuds[, 7]
set.seed(12345); trainset <- sample(1:550, 275)
vcr.train <- vcr.knn.train(X[trainset, ], y[trainset], k = 5)
vcr.test <- vcr.knn.newdata(X[-trainset, ], y[-trainset], vcr.train)
confmat.vcr(vcr.train) # for comparison
confmat.vcr(vcr.test)
cols <- c("saddlebrown", "orange", "olivedrab4", "royalblue3")
stackedplot(vcr.train, classCols = cols) # for comparison
stackedplot(vcr.test, classCols = cols)
classmap(vcr.train, "bud", classCols = cols) # for comparison
classmap(vcr.test, "bud", classCols = cols)

# For more examples, we refer to the vignette:
\dontrun{
vignette("K_nearest_neighbors_examples")
}
}

\name{vcr.neural.newdata}
\alias{vcr.neural.newdata}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{
Prepare for visualization of a neural network classification on new data.
}
\description{
  Prepares graphical display of new data fitted by a neural
  net that was modeled on the training data, using the output
  of \code{\link{vcr.neural.train}} on the training data.
}
\usage{
vcr.neural.newdata(Xnew, ynew = NULL, probs,
                   vcr.neural.train.out)
}

\arguments{
\item{Xnew}{data matrix of the new data, with the same number of columns as in the training data. Missing values in \code{Xnew} are not allowed.}
\item{ynew}{factor with class membership of each new case. Can be \code{NA} for some or all cases. If \code{NULL}, is assumed to be \code{NA} everywhere.}
\item{probs}{posterior probabilities obtained by running the neural net on the new data.}
\item{vcr.neural.train.out}{ output of \code{\link{vcr.neural.train}} on the training data.}
}

\value{
A list with components: \cr
\item{yintnew}{number of the given class of each case. Can contain \code{NA}'s.}
\item{ynew}{given class label of each case. Can contain \code{NA}'s.}
\item{levels}{levels of the response, from \code{vcr.svm.train.out}.}
\item{predint}{predicted class number of each case. Always exists.}
\item{pred}{predicted label of each case.}
\item{altint}{number of the alternative class. Among the classes different from the given class, it is the one with the highest posterior probability. Is \code{NA} for cases whose \code{ynew} is missing.}
\item{altlab}{alternative label if yintnew was given, else \code{NA}.}
\item{PAC}{probability of the alternative class. Is \code{NA} for cases whose \code{ynew} is missing.}
\item{fig}{distance of each case \eqn{i} from each class \eqn{g}. Always exists.}
\item{farness}{farness of each case from its given class. Is \code{NA} for cases whose \code{ynew} is missing.}
\item{ofarness}{for each case \eqn{i}, its lowest \code{fig[i,g]} to any class \eqn{g}. Always exists.}
}

\references{
Raymaekers J., Rousseeuw P.J.(2021). Silhouettes and quasi residual plots for neural nets and tree-based classifiers. \emph{Journal of Computational and Graphical Statistics}, 31(4), 1332–1343. \doi{10.1080/10618600.2022.2050249}
}

\author{
 Raymaekers J., Rousseeuw P.J.
}

\seealso{
\code{\link{vcr.neural.train}}, \code{\link{classmap}}, \code{\link{silplot}}, \code{\link{stackedplot}}
}

\examples{
# For examples, we refer to the vignette:
\dontrun{
vignette("Neural_net_examples")
}
}

\name{classmap}
\alias{classmap}
\title{
Draw the class map to visualize classification results.
}
\description{
Draw the class map to visualize classification results, based on the output of one of the
\code{vcr.*.*} functions in this package. The vertical axis of the class map shows each case's \code{PAC}, the conditional probability that it belongs to an alternative class. The \code{farness} on the horizontal axis is the probability of a member of the given class being at most as far from the class as the case itself.
}
\usage{
classmap(vcrout, whichclass, classLabels = NULL, classCols = NULL,
         main = NULL, cutoff = 0.99, plotcutoff = TRUE,
         identify = FALSE, cex = 1, cex.main = 1.2, cex.lab = NULL,
         cex.axis = NULL, opacity = 1,
         squareplot = TRUE,  maxprob = NULL, maxfactor = NULL)
}

\arguments{
 \item{vcrout}{output of \code{vcr.*.train} or \code{vcr.*.newdata}. Required.}

\item{whichclass}{the number or level of the class to be displayed. Required.}
 \item{classLabels}{the labels (levels) of the classes. If \code{NULL}, they are taken from \code{vcrout}.}
 \item{classCols}{a list of colors for the class labels. There should be at least as many as there are levels. If \code{NULL} the \code{classCols} are taken as 2, 3, 4, ...}
 \item{main}{title for the plot.}
  \item{cutoff}{cases with overall farness \code{vcrout$ofarness} > \code{cutoff} are flagged as outliers.}
   \item{plotcutoff}{If true, plots the cutoff on the farness values as a vertical line.}
 \item{identify}{if \code{TRUE}, left-click on a point to get its number, then ESC to exit.}
 \item{cex}{passed on to \code{\link[graphics:plot]{graphics::plot}}.}
 \item{cex.main}{same, for title.}
 \item{cex.lab}{same, for labels on horizontal and vertical axes.}
 \item{cex.axis}{same, for axes.}
 \item{opacity}{determines opacity of plotted dots. Value between 0 and 1, where 0 is transparent and 1 is opaque.}
 \item{squareplot}{If \code{TRUE}, makes the axes of the plot equally long.}
 \item{maxprob}{draws the farness axis at least upto probability maxprob. If \code{NULL}, the limits are obtained automatically.}
 \item{maxfactor}{if not \code{NULL}, a number slightly higher than 1 to increase the space at the right hand side of the plot, to make room for marking points.}
}

\value{
Executing the function plots the class map and returns
\item{coordinates}{ a matrix with 2 columns containing the coordinates of the plotted points. The first coordinate is the quantile of the farness probability. This makes it easier to add text next to interesting points. If \code{identify = T}, the attribute \code{ids} of \code{coordinates} contains the row numbers of the identified points in the matrix \code{coordinates}.}
}

\references{
Raymaekers J., Rousseeuw P.J., Hubert M. (2021). Class maps for visualizing classification results. \emph{Technometrics}, 64(2), 151–165. \doi{10.1080/00401706.2021.1927849}

Raymaekers J., Rousseeuw P.J.(2021). Silhouettes and quasi residual plots for neural nets and tree-based classifiers. \emph{Journal of Computational and Graphical Statistics}, 31(4), 1332–1343. \doi{10.1080/10618600.2022.2050249}
}

\author{
 Raymaekers J., Rousseeuw P.J.
}

\seealso{
\code{\link{vcr.da.train}}, \code{\link{vcr.da.newdata}},\cr \code{\link{vcr.knn.train}}, \code{\link{vcr.knn.newdata}},\cr \code{\link{vcr.svm.train}}, \code{\link{vcr.svm.newdata}},\cr \code{\link{vcr.rpart.train}}, \code{\link{vcr.rpart.newdata}},\cr \code{\link{vcr.forest.train}}, \code{\link{vcr.forest.newdata}},\cr \code{\link{vcr.neural.train}}, \code{\link{vcr.neural.newdata}}
}

\examples{
vcrout <- vcr.da.train(iris[, 1:4], iris[, 5])
classmap(vcrout, "setosa", classCols = 2:4) # tight class
classmap(vcrout, "versicolor", classCols = 2:4) # less tight
# The cases misclassified as virginica are shown in blue.
classmap(vcrout, "virginica", classCols = 2:4)
# The case misclassified as versicolor is shown in green.

# For more examples, we refer to the vignettes:
\dontrun{
vignette("Discriminant_analysis_examples")
vignette("K_nearest_neighbors_examples")
vignette("Support_vector_machine_examples")
vignette("Rpart_examples")
vignette("Random_forest_examples")
vignette("Neural_net_examples")
}
}

\name{makeKernel}
\alias{makeKernel}
\title{
Compute kernel matrix
}
\description{
Computes kernel value or kernel matrix, where the kernel type is extracted from an svm trained by \code{\link[e1071:svm]{e1071::svm}}.
}
\usage{
makeKernel(X1, X2 = NULL, svfit)
}

\arguments{

\item{X1}{first matrix (or vector) of coordinates.}
\item{X2}{if not \code{NULL}, second data matrix or vector. If NULL, \code{X2} is assumed equal to \code{X1}.}
\item{svfit}{output from \code{\link[e1071:svm]{e1071::svm}}}.
}


\value{the kernel matrix, of dimensions \code{nrow(X1)} by \code{nrow(X2)}. When both \code{X1} and \code{X2} are vectors, the result is a single number.
}

\references{
Raymaekers J., Rousseeuw P.J., Hubert M. (2021). Class maps for visualizing classification results. \emph{Technometrics}, 64(2), 151–165. \doi{10.1080/00401706.2021.1927849}
}

\author{
 Raymaekers J., Rousseeuw P.J.
}

\seealso{
\code{\link{makeFV}}
}

\examples{

library(e1071)
set.seed(1); X <- matrix(rnorm(200 * 2), ncol = 2)
X[1:100, ] <- X[1:100, ] + 2
X[101:150, ] <- X[101:150, ] - 2
y <- as.factor(c(rep("blue", 150), rep("red", 50))) # two classes
# We now fit an SVM with radial basis kernel to the data:
set.seed(1) # to make the result of svm() reproducible.
svmfit <- svm(y~.,  data = data.frame(X = X, y = y),  scale = FALSE,
             kernel = "radial", cost = 10, gamma = 1, probability = TRUE)
Kxx <- makeKernel(X, svfit = svmfit)
# The result is a square kernel matrix:
dim(Kxx) # 200 200
Kxx[1:5, 1:5]

# For more examples, we refer to the vignette:
\dontrun{
vignette("Support_vector_machine_examples")
}
}

\name{data_instagram}
\alias{data_instagram}
\docType{data}
\title{
Instagram data
}
\description{

This dataset contains information on fake (spam) accounts on Instagram.
The original source is https://www.kaggle.com/free4ever1/instagram-fake-spammer-genuine-accounts by Bardiya Bakhshandeh.

The data contains information on 696 Instagram accounts.
For each account, 11 variables were recorded describing its characteristics.
The goal is to detect fake instagram accounts, which are used for spamming.
}

\usage{data("data_instagram")}

\format{
  A data frame with 696 observations on the following variables.
  \describe{
  \item{profile.pic}{binary, indicates whether profile has picture.}
\item{nums.length.username}{ratio of number of numerical chars in username to its length.}
\item{fullname.words}{number of words in full name.}
\item{nums.length.fullname}{ratio of number of numerical characters in full name to its length.}
\item{name..username}{binary, indicates whether the name and username of the profile
are the same.}
\item{description.length}{length of the description/biography of the profile (in number of characters).}
\item{external.URL}{binary, indicates whether profile has external url.}
\item{private}{binary, indicates whether profile is private or not.}
\item{X.posts}{number of posts made by profile.}
\item{X.followers}{number of followers.}
\item{X.follows}{numbers of follows.}
\item{y}{whether profile is fake or not.}
  \item{dataType}{vector taking the values ``train'' or ``test'' indicating
  whether the observation belongs to the training or the test data.}
  }
}


\source{
https://www.kaggle.com/free4ever1/instagram-fake-spammer-genuine-accounts
}

\examples{
data(data_instagram)
str(data_instagram)

# The data are used in:
\dontrun{
vignette("Random_forest_examples")
}
}

\name{data_floralbuds}
\alias{data_floralbuds}
\docType{data}
\title{
Floral buds data
}
\description{
This data on floral pear bud detection was first described by Wouters et al.
The goal is to classify the instances into buds, branches, scales and support.
The numeric vectors resulted from a multispectral vision sensor and describe the scanned images.
}
\usage{data("data_floralbuds")}

\format{
  A data frame with 550 observations on the following 7 variables.
  \describe{
    \item{\code{X1}}{numeric vector}
    \item{\code{X2}}{numeric vector}
    \item{\code{X3}}{numeric vector}
    \item{\code{X4}}{numeric vector}
    \item{\code{X5}}{numeric vector}
    \item{\code{X6}}{numeric vector}
    \item{\code{y}}{a factor with levels \code{branch} \code{bud} \code{scales} \code{support}}
  }
}


\source{
Wouters, N., De Ketelaere, B., Deckers, T. De Baerdemaeker, J., Saeys, W. (2015).
Multispectral detection of floral buds for automated thinning of pear.
\emph{Comput. Electron. Agric.} 113, C, 93–103. <doi:10.1016/j.compag.2015.01.015>
}

\examples{
data("data_floralbuds")
str(data_floralbuds)
summary(data_floralbuds)

# The data are used in:
\dontrun{
vignette("Discriminant_analysis_examples")
vignette("Neural_net_examples")
}
}

\name{pamc1d}
\alias{pamc1d}

\title{
Constrained k-medoids clustering for univariate data
}
\description{
Performs constrained partitioning around medoids (PAM) clustering on
univariate data. This is a modification of the constrained k-means
algorithm by Bradley, Bennett and Demiriz (2000), adapted to use
medians instead of means as cluster centers, and to enforce a minimum
cluster size constraint on unique values rather than all cases. This
prevents tied observations from being split across separate clusters.
}
\usage{
pamc1d(y, k, minsize = 4, countwhat = "unique",
       stand = TRUE, maxit = 100, verbose = TRUE)
}
\arguments{
\item{y}{a numeric vector of univariate data to be clustered.
Missing values (\code{NA}) are removed before clustering.}
\item{k}{integer specifying the desired number of clusters.}
\item{minsize}{integer giving the minimum number of members required
in each cluster. When \code{countwhat = "unique"} (the default),
this refers to unique values; when \code{countwhat = "any"} it
refers to all cases including duplicates. Defaults to \code{4}.}
\item{countwhat}{character string specifying whether the minimum
cluster size constraint applies to \code{"unique"} values (the
default) or to \code{"any"} cases including duplicates.}
\item{stand}{logical. If \code{TRUE} (the default), the data are
standardized by subtracting the mean and dividing by the standard
deviation before clustering. The returned cluster centers are on
the standardized scale.}
\item{maxit}{integer giving the maximum number of iterations in the
reassignment loop. Defaults to \code{100}.}
\item{verbose}{logical. If \code{TRUE} (the default), intermediate
results including the clustering vector, cluster table, and
objective function value are printed at each iteration.}
}

\details{
The algorithm starts from an unconstrained PAM solution obtained via
\code{\link[cluster]{pam}}. If any cluster fails to meet the
\code{minsize} constraint, a constrained reassignment loop is
entered. At each iteration, a transportation linear program is solved
(via \code{\link[lpSolve]{lp.transport}}) to reassign cases to
clusters in a way that minimizes the total L1 distance to the cluster
medians while satisfying the size constraints. Cluster medians are
then recomputed, and the procedure repeats until no assignment
changes or \code{maxit} iterations are reached.

When \code{countwhat = "unique"}, only the unique values are passed
to the linear program, and duplicate cases follow the cluster
assignment of the unique value they match. This ensures that tied
observations are never separated into different clusters.

The objective function is the mean absolute deviation of all cases
from their assigned cluster median.
}

\value{
A list with the following components:
\item{iter}{integer, the number of iterations performed.}
\item{converged}{logical, \code{TRUE} if the algorithm converged
(no assignment changes in the final iteration), \code{FALSE} if
\code{maxit} was reached before convergence.}
\item{clustering}{integer vector of length \code{n} (after removing
\code{NA}s) giving the cluster assignment of each observation.
Observations are sorted by value.}
\item{obj}{numeric, the value of the objective function (mean
absolute deviation from cluster medians) at the final iteration.}
\item{centers}{numeric vector of length \code{k} with the median of
each cluster on the (possibly standardized) scale.}
\item{clustable}{a table giving the number of members in each
cluster.}
}

\references{
Bradley P.S., Bennett K.P., Demiriz A. (2000). Constrained k-means
clustering. \emph{Microsoft Research Technical Report},
MSR-TR-2000-65.
}

\author{
Rousseeuw P.J.
}

\seealso{
\code{\link[cluster]{pam}}, \code{\link[lpSolve]{lp.transport}}
}

\examples{
set.seed(42)
y <- c(rnorm(30, mean = 0), rnorm(30, mean = 5), rnorm(30, mean = 10))

# Basic usage with k = 3 clusters and minimum cluster size of 4
result <- pamc1d(y, k = 3, minsize = 4, verbose = FALSE)

# Inspect the results
result$converged
result$clustable
result$centers

# Apply the constraint to all cases (not just unique values)
result2 <- pamc1d(y, k = 3, minsize = 4, countwhat = "any",
                  verbose = FALSE)

# Example with tied values
y_tied <- c(rep(1, 10), rep(3, 10), rep(5, 10),
            rep(7, 10), rep(9, 10))
result3 <- pamc1d(y_tied, k = 2, minsize = 2,
                  countwhat = "unique", verbose = TRUE)
result3$clustable
}

\name{bixplot}
\alias{bixplot}

\title{
Boxplot version suited for bimodal and multimodal data, combining density, box,
and rug elements with automatic cluster detection
}
\description{
Draws a bixplot for one or more numeric variables. A bixplot extends
the violin plot and boxplot by automatically testing each variable for
unimodality (via Hartigan's dip test) and, when multimodality is
detected, fitting a constrained k-medoids clustering to identify and
separately display the modes. Each variable is rendered as a filled
density body, a box-and-whisker summary, and a rug of individual
data values. The rug can optionally be colored by an external numeric
or factor variable. Input can be provided as vectors, a data frame, a
matrix, a list of vectors, or a formula.
}
\usage{
bixplot(...,
        names = NA,
        add = FALSE,
        at = NULL,
        horizontal = FALSE,
        col = "gray60", bodyCol = NULL,
        bodyOpaque = 0.5,
        bodyW = 0.80,
        bodysize = "area_from_count",
        modeCol = c("cadetblue3", "hotpink2",
                    "lawngreen", "orange", "cyan3",
                    "coral2", "gray60", "darkgoldenrod2",
                    "slateblue2", "brown2", "khaki3"),
        curveCol = "black",
        curveLwd = 1,
        border = "black", boxCol = NULL,
        boxOpaque = 0.5,
        boxW = 0.32,
        boxLwd = 2,
        makewhiskers = TRUE,
        innerwhiskers = TRUE,
        rugCol = "black",
        rugoutCol = par("fg"),
        rugNumeric = NULL,
        rugNumericColors = c("blue", "green"),
        colorbarW = 0.12,
        rugFactor = NULL,
        rugFactorColors = c("red", "blue", "forestgreen"),
        rugLwd = par("lwd"),
        rugoutLwd = 2 * rugLwd,
        rugOpaque = 0.4,
        jittering = TRUE,
        jitteramount = NULL,
        rugW = 0.12,
        stickCol = "black",
        stickLwd = 2,
        boxwex = 1,
        width = NULL,
        side = "no",
        tick = TRUE,
        diplevel = 0.01,
        minN = 15,
        kmax = NULL,
        bigN = 500,
        clusMinN = 3,
        bw = "SJ-dpi",
        kernel = "gaussian",
        cut = 3,
        cutmin = -Inf,
        cutmax = Inf,
        xlim = NULL,
        ylim = NULL,
        cex.axis = 1,
        main = "bixplot",
        cex.main = 1, line.main = 1,
        cex.colorbar = 1,
        xlab = "",
        ylab = "",
        cex.lab = 1, line.lab = 2.2,
        xaxs = "r", yaxs = "r", las = 1,
        ann = !add,
        plot = TRUE,
        log = NULL)
}

\arguments{
\item{...}{one or more numeric vectors to plot, or a single data
frame or matrix whose columns are plotted, or a single list of
numeric vectors (which may have different lengths), or a formula
such as \code{y ~ grp} where \code{y} is a numeric vector to be
split into groups by the grouping variable \code{grp} (usually a
factor). Note that \code{~ g1 + g2} is equivalent to
\code{~ g1:g2}. When a formula is used it may be necessary to also
supply a \code{data =} argument. Missing values are silently
removed.}
\item{names}{character vector of group labels printed next to each
variable. If \code{NA} (the default), labels are extracted from
the input data; if the data carry no names the integers
\code{1, 2, 3, \ldots} are used. [Argument from
\code{\link[graphics]{boxplot}}.]}
\item{add}{logical. If \code{TRUE}, the bixplot is added to the
current plot without drawing new axes. Defaults to \code{FALSE}.
[Argument from \code{boxplot}.]}
\item{at}{numeric vector specifying the axis positions at which the
bixplots are drawn, particularly useful when \code{add = TRUE}.
Defaults to \code{1:p} where \code{p} is the number of variables.
[Argument from \code{boxplot}.]}
\item{horizontal}{logical. If \code{TRUE}, bixplots are drawn
horizontally. Defaults to \code{FALSE} (vertical). [Argument from
\code{boxplot}.]}
\item{col}{color(s) for the density body of variables deemed
unimodal. Can be a single color or a vector recycled to length
\code{p}. When \code{NA}, the body is not filled. [Argument from
\code{boxplot}.]}
\item{bodyCol}{if not \code{NULL}, overrides \code{col} for the
body fill color.}
\item{bodyOpaque}{numeric between 0 (transparent) and 1 (opaque)
controlling the opacity of the filled density body. Defaults to
\code{0.5}.}
\item{bodyW}{numeric vector (recycled to length \code{p}) giving the
width of the widest density body for each variable. When two or
more modes are shown for a variable, this is the width of the
widest mode body. Set to zero to suppress the body entirely.
Defaults to \code{0.80}.}
\item{bodysize}{character string determining how the density bodies
of individual modes within the same variable are sized relative to
each other. One of \code{"area_from_count"} (the default; body
area proportional to cluster membership count),
\code{"area_is_constant"} (all modes have equal area), or
\code{"width_is_constant"} (all modes have equal width). In all
cases the widest body has width \code{bodyW[j]}, possibly scaled
by \code{width}.}
\item{modeCol}{color(s) of the density bodies for variables with
more than one detected mode (cluster). Cycled across all modes of
all multimodal variables. If \code{NULL}, each mode inherits the
\code{col} color of its variable.}
\item{curveCol}{color(s) of the density curve boundary drawn around
each body. Can be a single color or a vector recycled to length
\code{p}. Set to \code{NA} to suppress the curve. Defaults to
\code{"black"}.}
\item{curveLwd}{single number giving the line width of the density
curve.}
\item{border}{color(s) of the box and whiskers. Can be a single
color or a vector recycled to length \code{p}. Set to \code{NA}
to suppress the box. [Argument from \code{boxplot}.]}
\item{boxCol}{if not \code{NULL}, overrides \code{border} for the
box color.}
\item{boxOpaque}{numeric between 0 and 1 controlling the opacity of
the lines making up the boxplot. Defaults to \code{0.5}.}
\item{boxW}{numeric vector (recycled to length \code{p}) giving the
width of the interquartile box for each variable. Set to zero to
suppress the box. Defaults to \code{0.32}.}
\item{boxLwd}{single number giving the line width of the box and
whiskers.}
\item{makewhiskers}{logical. If \code{TRUE} (the default), whiskers
are drawn extending from the box to the most extreme non-outlying
values.}
\item{innerwhiskers}{logical. Relevant only when a variable has more
than one mode. If \code{TRUE}, whiskers are also drawn between
adjacent modes. If \code{FALSE}, inner whiskers are
omitted because their interpretation is ambiguous when mode
densities overlap.}
\item{rugCol}{color(s) of the rug tick marks for each variable.
Can be a single color or a vector recycled to length \code{p}.
Set to \code{NA} to suppress the rug. Defaults to
\code{"black"}.}
\item{rugoutCol}{color(s) of the portion of rug lines that extends
outside the density body. If \code{NULL} or \code{NA}, the same
color as \code{rugCol} is used throughout. Defaults to
\code{par("fg")}.}
\item{rugNumeric}{optional numeric vector of the same length as each
\code{y} variable (requires all variables to have the same length)
used to color the rug lines via a continuous color palette. Cannot
be combined with \code{rugFactor}.}
\item{rugNumericColors}{vector of two or three colors passed to
\code{\link[grDevices]{colorRampPalette}} to construct the palette
for \code{rugNumeric}. Defaults to \code{c("blue", "green")}.}
\item{colorbarW}{width of the color bar legend for \code{rugNumeric}
relative to the main plot width. Only has an effect when
\code{rugNumeric} is specified. Defaults to \code{0.12}.}
\item{rugFactor}{optional factor variable of the same length as each
\code{y} variable used to color the rug lines by factor level.
Cannot be combined with \code{rugNumeric}.}
\item{rugFactorColors}{character vector of colors for the levels of
\code{rugFactor}, recycled as needed. Defaults to
\code{c("red", "blue", "forestgreen")}.}
\item{rugLwd}{single number giving the line width of the rug marks.
Defaults to \code{par("lwd")}.}
\item{rugoutLwd}{single number giving the line width of the part of
rug lines outside the density body. Defaults to
\code{2 * rugLwd} to make isolated points visually prominent.}
\item{rugOpaque}{numeric between 0 and 1 controlling the opacity of
the rug marks. Defaults to \code{0.4}.}
\item{jittering}{logical. If \code{TRUE} (the default), rug tick
positions are jittered via \code{\link[base]{jitter}} to make tied
values distinguishable.}
\item{jitteramount}{amount of jittering passed to
\code{\link[base]{jitter}}. If \code{NULL}, the default amount is
used.}
\item{rugW}{numeric vector (recycled to length \code{p}) giving the
width of the rug (i.e. the length of each tick mark). Set to zero
to suppress the rug. Defaults to \code{0.12}.}
\item{stickCol}{color(s) of the vertical or horizontal "stick" drawn
when \code{side = "both"}, which separates the two half-bixplots
sharing an axis. Set to \code{NA} to suppress the stick. Can be a
single color or a vector recycled to length \code{p}. Defaults to
\code{"black"}.}
\item{stickLwd}{single number giving the line width of the stick.
Set to zero to suppress the stick. Defaults to \code{2}.}
\item{boxwex}{scale factor applied uniformly to the widths of the
body, box and rug across all bixplots. [Argument from
\code{boxplot}.] Defaults to \code{1}.}
\item{width}{optional numeric vector of length \code{p} giving the
relative widths of the bixplots. Entries must be strictly
positive; they are divided by their maximum so that the widest
bixplot has relative width 1. The resulting ratios multiply the
body, box and rug widths. If \code{NULL} (the default), all
bixplots have equal width. [Argument from \code{boxplot}.]}
\item{side}{character string specifying which side of the variable
axis the body, box and rug are drawn on. One of \code{"no"} (the
default; each bixplot is symmetric about its axis), \code{"first"}
or \code{"second"} (all half-bixplots on that side), or
\code{"both"} (adjacent variables are plotted on alternate sides
of shared axes, halving the number of axes needed). [Argument from
\code{\link[beanplot]{beanplot}}.]}
\item{tick}{logical indicating whether tick marks are drawn on the
group-label axis. Only has an effect when \code{add = FALSE}.
Defaults to \code{TRUE}.}
\item{diplevel}{significance level for Hartigan's dip test for
unimodality. A cluster search is only performed for variables
whose dip test p-value is at most \code{diplevel}. Defaults to
\code{0.01}. Increasing this value may yield more clusters;
decreasing it fewer.}
\item{minN}{minimum number of observations required per potential
cluster. Defaults to \code{15}. The maximum number of clusters
searched is bounded by \code{floor(n / minN)}. If
\code{n < 2 * minN}, clustering is not attempted.}
\item{kmax}{maximum number of clusters to consider. Internally
capped at 5. If \code{NULL} (the default), it is set to
\code{min(floor(n / minN), 5)}. Setting \code{kmax = 1} treats
all variables as unimodal and the display resembles a violin
plot.}
\item{bigN}{when a variable has more than \code{bigN} non-missing
values, a sample of \code{bigN} observations (always including
the minimum and maximum) is drawn without replacement before
computing densities and clusters, to reduce computation time.
Defaults to \code{500}; values below 300 are silently raised to
300.}
\item{clusMinN}{minimum number of unique values that each cluster
must contain. The constrained clustering enforces this bound.
Defaults to \code{3}.}
\item{bw}{bandwidth for \code{\link[stats]{density}}, used to
construct the density body. Can be a numeric value or a character
string naming a bandwidth selector (see \code{?density}).
Defaults to \code{"SJ-dpi"}.}
\item{kernel}{kernel for \code{\link[stats]{density}}. Defaults to
\code{"gaussian"}.}
\item{cut}{the density is computed from
\code{min(y) - cut * bw} to \code{max(y) + cut * bw}
where \code{bw} is the numeric bandwidth. Defaults to \code{3}.
[Argument from \code{beanplot}.]}
\item{cutmin}{if finite, the density of every variable and mode is
truncated to begin no lower than \code{cutmin}. Defaults to
\code{-Inf} (no effect). [Argument from \code{beanplot}.]}
\item{cutmax}{if finite, the density of every variable and mode is
truncated to end no higher than \code{cutmax}. Defaults to
\code{Inf} (no effect). [Argument from \code{beanplot}.]}
\item{xlim}{numeric vector of length 2 giving the limits of the
group axis (whether \code{horizontal} is \code{TRUE} or
\code{FALSE}). If \code{NULL}, limits are set automatically.
[Convention from \code{boxplot}.]}
\item{ylim}{numeric vector of length 2 giving the limits of the
numeric value axis (whether \code{horizontal} is \code{TRUE} or
\code{FALSE}). If \code{NULL}, limits are set automatically.
[Convention from \code{boxplot}.]}
\item{cex.axis}{character expansion factor for axis tick labels.}
\item{main}{title of the plot. Defaults to \code{"bixplot"}.}
\item{cex.main}{character expansion factor for the title.}
\item{line.main}{margin line for the title (passed to
\code{\link[graphics]{title}}).}
\item{cex.colorbar}{character expansion factor for the color bar
axis labels when \code{rugNumeric} is used.}
\item{xlab}{label for the horizontal axis. Defaults to \code{""}.}
\item{ylab}{label for the vertical axis. Defaults to \code{""}.}
\item{cex.lab}{character expansion factor for axis labels.}
\item{line.lab}{margin line for axis labels. Defaults to
\code{2.2}.}
\item{xaxs}{axis interval calculation style for the x-axis
(see \code{\link[graphics]{par}}). Defaults to \code{"r"}.}
\item{yaxs}{axis interval calculation style for the y-axis.
Defaults to \code{"r"}.}
\item{las}{orientation of axis tick labels
(see \code{\link[graphics]{par}}). Defaults to \code{1}
(always horizontal).}
\item{ann}{logical indicating whether the plot should be annotated
with \code{xlab}, \code{ylab} and \code{main}. Defaults to
\code{!add}. [Argument from \code{boxplot}.]}
\item{plot}{if \code{TRUE} (the default), the summary list is
returned invisibly. If \code{FALSE}, it is also printed to the
console. [Argument from \code{boxplot}.]}
\item{log}{this argument from \code{boxplot} must be \code{NULL}.
If a non-\code{NULL} value is supplied, a warning is issued and
the argument is ignored, because a log transformation can change
the number and position of modes. Apply the log transform
explicitly to the data before calling \code{bixplot} if
desired.}
}

\details{
For each variable, \code{bixplot} proceeds as follows. Hartigan's
dip test is applied to test for unimodality; if the p-value exceeds
\code{diplevel}, or if the sample is too small (\code{n < 2 * minN}),
the variable is treated as unimodal (\code{k = 1}). Otherwise,
constrained k-medoids clustering (via \code{\link{pamc1d}}) is fitted
for \code{k = 2, \ldots, kmax} clusters, and the best \code{k} is
selected by the highest mean silhouette width (computed via
\code{\link[cluster]{silhouette}}). If no \code{k > 1} yields a
positive mean silhouette width, the variable is treated as unimodal.

A single global bandwidth is computed once per variable and reused
for the density of every mode, ensuring that density bodies are
comparable across modes. Mode sizes are scaled according to
\code{bodysize}.

When \code{side = "both"}, adjacent variables are paired and plotted
as half-bixplots on opposite sides of shared axes, and axis labels
are combined automatically from the names of each pair.

The arguments \code{rugNumeric} and \code{rugFactor} require all
variables to have the same number of observations. They cannot be
specified simultaneously.
}

\value{
A list returned invisibly when \code{plot = TRUE} (or visibly when
\code{plot = FALSE}), containing:

\item{call}{the matched call.}
\item{p}{the number of variables plotted.}
\item{<name_1>, <name_2>, \ldots}{one list entry per variable,
named after the variable. For a unimodal variable the list
contains \code{values} (the sorted non-missing observations) and
\code{fivenumbersummary} (the five-number summary from
\code{\link[stats]{fivenum}} used to draw the box). For a
multimodal variable the list additionally contains
\code{clustering} (an integer vector of cluster assignments) and
one sub-list per cluster named \code{cluster_1},
\code{cluster_2}, \ldots, each containing \code{members} and
\code{fivenumbersummary} for that cluster.}
}


\references{
Montalcini, C., Rousseeuw, P.J. (2025). The bixplot: A variation on the boxplot suited for bimodal data, \doi{10.48550/arXiv.2510.09276} (open access).
}

\author{
 P.J. Rousseeuw
}

\seealso{
\code{\link{pamc1d}},
\code{\link[cluster]{pam}},
\code{\link[cluster]{silhouette}},
\code{\link[diptest]{dip.test}},
\code{\link[stats]{density}},
\code{\link[graphics]{boxplot}}
}

\examples{
set.seed(1)
# A unimodal and a clearly bimodal variable
x1 <- rnorm(100)
x2 <- c(rnorm(60, mean = -3), rnorm(60, mean = 3))
bixplot(x1, x2, names = c("unimodal", "bimodal"),
        main = "Basic bixplot example")

# Formula interface, coloring rug by a factor
n <- 150
grp <- factor(rep(c("A", "B", "C"), each = 50))
y   <- c(rnorm(50, 0), rnorm(50, 4), rnorm(50, 8))
bixplot(y ~ grp, main = "Formula interface")

# Horizontal layout with an external numeric rug variable
set.seed(42)
vals <- c(rnorm(80, mean = -2), rnorm(80, mean = 2))
covariate <- runif(160)
bixplot(vals, horizontal = TRUE,
        rugNumeric = covariate,
        rugNumericColors = c("purple", "yellow"),
        main = "Horizontal bixplot with numeric rug")

# Side-by-side ("both") half-bixplots
bixplot(x1, x2, side = "both",
        names = c("unimodal vs bimodal"),
        main = "side = both")

# For more examples, we refer to the vignette:
\dontrun{
vignette("bixplot_examples")
}
}

\name{data_titanic}
\alias{data_titanic}
\docType{data}
\title{
Titanic data
}
\description{

This dataset contains information on 1309 passengers of the RMS Titanic.
The goal is to predict survival based on 11 characteristics such as the travel class, age and sex of the passengers.

The original data source is
https://www.kaggle.com/c/titanic/data

The data is split up in a training data consisting of 891 observations and a test data of 418 observations. The response in the test set was obtained by combining information from other data files, and has been verified by submitting it as a `prediction' to kaggle and getting perfect marks.
}

\usage{data("data_titanic")}

\format{
  A data frame with 1309 observations on the following variables.
  \describe{
  \item{PassengerId}{a unique identified for each passenger.}
\item{Pclass}{travel class of the passenger.}
\item{Name}{name of the passenger.}
\item{Sex}{sex of the passenger.}
\item{Age}{age of the passenger.}
\item{SibSp}{number of siblings and spouses traveling with the passenger.}
\item{Parch}{number of parents and children traveling with the passenger.}
\item{Ticket}{Ticket number of the passenger.}
\item{Fare}{fare paid for the ticket.}
\item{Cabin}{cabin number of the passenger.}
\item{Embarked}{Port of embarkation. Takes the values C (Cherbourg), Q (Queenstown) and S (Southampton).}
\item{y}{factor indicating casualty or survivor.}
  \item{dataType}{vector taking the values ``train'' or ``test'' indicating whether the observation belongs to the training or the test data.}
  }
}


\source{
https://www.kaggle.com/c/titanic/data
}

\examples{
data("data_titanic")
traindata <- data_titanic[which(data_titanic$dataType == "train"), -13]
testdata <- data_titanic[which(data_titanic$dataType == "test"), -13]
str(traindata)
table(traindata$y)

# The data are used in:
\dontrun{
vignette("Rpart_examples")
}
}

\name{predscor}
\alias{predscor}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{
Draws a predictions correlation plot, which visualizes the correlations between the prediction terms in a regression fit.
}
\description{
Computes the correlations between the prediction terms in a regression fit, and displays them graphically in a way that takes the standard deviations of the prediction terms into account. The input variables of the regression can be numerical, categorical, logical and character, and the regression model can be linear or generalized linear. The regression formula in \code{\link[stats:lm]{lm}} or \code{\link[stats:glm]{glm}} may contain transformations and interactions.
}
\usage{
predscor(fit, maxnpreds = 8, sort.by.stdev = TRUE, adj.order = FALSE,
cell.length = "stdev", plot.abs.cor = FALSE, palette = NULL,
diagonalcolor = "black")
}
%- maybe also 'usage' for other objects documented here.
\arguments{
 \item{fit}{an output object of \code{\link[stats:lm]{lm}} or \code{\link[stats:glm]{glm}}.}
 \item{maxnpreds}{the maximal number of prediction terms
 to plot. When there are more prediction
 terms than this, those with smallest
 standard deviations are combined.}
 \item{sort.by.stdev}{if \code{TRUE}, sorts the prediction terms by
 decreasing standard deviation.}
 \item{adj.order}{if \code{TRUE}, this modifies the order
 of the prediction terms in an attempt
 to bring highly correlated
 prediction terms close to each other
 in the \code{\link{predscor}} display. If
 \code{sort.by.stdev} is \code{TRUE}, this happens
 after that sorting.}
 \item{cell.length}{if "\code{stdev}", the sides of the square
 cells on the diagonal of the correlation
 matrix are proportional to the standard
 deviation of their prediction term.
 If "\code{sqrt}" they are proportional to the
 square root of the standard deviation.
 If "\code{equal}" all sides are the same.}
 \item{plot.abs.cor}{if \code{FALSE}, the default, positive and
 negative correlations are shown in
 different colors, typically red and
 blue. If \code{TRUE} the absolute values
 of the correlations are shown.}
 \item{palette}{a vector with colors to display
 correlations ranging from -1 to 1.
 If \code{NULL}, the default palette shows
 positive correlations in red,
 negative correlations in blue, and
 uses white for correlation zero.}
 \item{diagonalcolor}{color of the cells on the diagonal
 of the correlation matrix. The
 default is "black".}

}

\value{
A list containing
  \item{cormat}{the correlation matrix of the prediction terms.}
  \item{predterms}{matrix of cases by prediction terms.}
  \item{predsummary}{data frame with the standard deviation of each prediction term and the total linear prediction.}
}
\references{
Rousseeuw, P.J. (2025). Explainable Linear and Generalized Linear Models by the Predictions Plot
\url{https://arxiv.org/abs/2412.16980v2} (open access).
}
\author{
 Rousseeuw, P.J.
}



\seealso{
\code{\link{predsplot}}
}
\examples{
data(data_titanic)
attach(data_titanic)
Pclass = factor(Pclass, unique(Pclass))
Sex = factor(Sex, labels = c("F","M"))
fit <- glm(y ~ Sex + Age + SibSp + Parch + Pclass, family=binomial)
predscor(fit)

# For more examples, we refer to the vignette:
\dontrun{
vignette("predsplot_examples")
}
}

\name{confmat.vcr}
\alias{confmat.vcr}
\title{
Build a confusion matrix from the output of a function \code{vcr.*.*}.
}

\description{
Build a confusion matrix from the output of a function \code{vcr.*.*}.
Optionally, a separate column for outliers can be added to the confusion matrix.
}

\usage{
confmat.vcr(vcrout, cutoff = 0.99, showClassNumbers = FALSE,
            showOutliers = TRUE, silent = FALSE)
}

\arguments{
 \item{vcrout}{output of \code{vcr.*.train} or \code{vcr.*.newdata}.}
 \item{cutoff}{cases with overall farness \code{vcrout$ofarness} > \code{cutoff} are flagged as outliers.}
 \item{showClassNumbers}{if \code{TRUE}, the row and column names are the number of each level instead of the level itself. Useful for long level names.}
 \item{showOutliers}{if \code{TRUE} and some points were flagged as outliers, it adds an extra column on the right of the confusion matrix for these outliers, with label "outl".}
 \item{silent}{if \code{FALSE}, the confusion matrix and accuracy are shown on the screen.}
}

\value{
A confusion matrix
}

\references{
Raymaekers J., Rousseeuw P.J., Hubert M. (2021). Class maps for visualizing classification results. \emph{Technometrics}, 64(2), 151–165. \doi{10.1080/00401706.2021.1927849}
}

\author{
 Raymaekers J., Rousseeuw P.J.
}

\seealso{
\code{\link{vcr.da.train}}, \code{\link{vcr.da.newdata}},\cr \code{\link{vcr.knn.train}}, \code{\link{vcr.knn.newdata}},\cr \code{\link{vcr.svm.train}}, \code{\link{vcr.svm.newdata}},\cr \code{\link{vcr.rpart.train}}, \code{\link{vcr.rpart.newdata}},\cr \code{\link{vcr.forest.train}}, \code{\link{vcr.forest.newdata}},\cr \code{\link{vcr.neural.train}}, \code{\link{vcr.neural.newdata}}
}

\examples{

vcrout <- vcr.knn.train(scale(iris[, 1:4]), iris[, 5], k = 5)
# The usual confusion matrix:
confmat.vcr(vcrout, showOutliers = FALSE)

# Cases with ofarness > cutoff are flagged as outliers:
confmat.vcr(vcrout, cutoff = 0.98)

# With the default cutoff = 0.99 only one case is flagged here:
confmat.vcr(vcrout)
# Note that the accuracy is computed before any cases
# are flagged, so it does not depend on the cutoff.

confmat.vcr(vcrout, showClassNumbers = TRUE)
# Shows class numbers instead of labels. This option can
# be useful for long level names.

# For more examples, we refer to the vignettes:
\dontrun{
vignette("Discriminant_analysis_examples")
vignette("K_nearest_neighbors_examples")
vignette("Support_vector_machine_examples")
vignette("Rpart_examples")
vignette("Random_forest_examples")
vignette("Neural_net_examples")
}
}

\name{predsplot}
\alias{predsplot}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{
Make a predictions plot
}
\description{
Plots the prediction terms of a regression, together with the total prediction.  The input variables of the regression can be numerical, categorical, logical and character, and are visualized by histograms, densities, or bar plots. The regression model can be linear or generalized linear. The regression formula in lm() or glm() may contain transformations and interactions.  The plot shows all cases in the data, but can also be made for a single case, either in-sample or out-of-sample, to explain its prediction.
}
\usage{

predsplot(fit, maxnpreds = 8, sort.by.stdev = TRUE, displaytype
                   = "histogram", totalpred.type = "response",
                   trunc.totalpred = TRUE, casetoshow = NULL, staircase =
                   FALSE, verbose = TRUE, nfact = 8, main = NULL,
                   cex.main = 1, xlab = NULL, ylab = NULL, cex.labs = 1,
                   cex.varnames = 1, cex.values = 1, cex.levelnames = 1,
                   maxchar.level = 5, plotcentercept = FALSE, vlinewidth
                   = 0.25, hlinewidth = 0.25, drawborder = TRUE,
                   borderwidth = 1, densitycolors = NULL, predupcolor =
                   "red", preddowncolor = "blue", predpointsize = 3,
                   draw.segments = TRUE, predsegmentwidth = 0.8, profile
                   = FALSE, bw = "nrd0", adjust = 1)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
 \item{fit}{an output object of \code{\link[stats:lm]{lm}} or \code{\link[stats:glm]{glm}}.}
 \item{maxnpreds}{the maximal number of prediction terms
to plot. When there are more prediction
terms than this, those with smallest
standard deviations are combined.}
 \item{sort.by.stdev}{if \code{TRUE}, sorts the prediction terms by
decreasing standard deviation.}
 \item{displaytype}{when \code{"histogram"}, the distributions
of the numerical prediction terms and
the total prediction are displayed as
histograms. When \code{"density"}, the
default density estimate of R is
plotted instead.}
 \item{totalpred.type}{when fit is a \code{glm} object, option
\code{"response"} plots labels of the total
prediction that are obtained by the
inverse link function. Option
\code{"linear"} plots labels of the total
linear prediction. When fit is an
\code{lm} object, argument \code{totalpred.type}
is ignored.}
 \item{trunc.totalpred}{if \code{TRUE}, the default, the range of
the total prediction is truncated
so ymin and ymax are determined by
the individual prediction terms. \code{FALSE}
may make the prediction terms look small
in the plot.}
 \item{casetoshow}{if not \code{NULL}, the particular case
to be displayed. This can be a case
number or row name of a case in the
dataset, or a list or vector with
input values of a new case.}
 \item{staircase}{if \code{TRUE} and casetoshow is not
\code{NULL}, the prediction for the case
is shown in staircase style.}
 \item{verbose}{if \code{TRUE}, some intermediate results are
shown on the console.}
 \item{nfact}{if a numeric input variable has at most
\code{nfact} unique values, it will be
displayed as if it were a factor.
This may be useful since R considers
a binary variable to be numeric.}
 \item{main}{main title of the plot.}
 \item{cex.main}{its size.}
 \item{xlab}{horizontal axis label.}
 \item{ylab}{vertical axis label.}
 \item{cex.labs}{size of the axis labels.}
 \item{cex.varnames}{size of the variable names. It may help to shorten the names in the data frame before running the regression.}
 \item{cex.values}{size of the values of the numeric variables.}
 \item{cex.levelnames}{size of the level names of non-numeric variables.}
 \item{maxchar.level}{only the first maxchar.level
characters of the non-numeric levels
are displayed.}
\item{plotcentercept}{logical, whether to list the centercept in the plot.}
 \item{vlinewidth}{width of the vertical lines.}
 \item{hlinewidth}{width of the horizontal line.}
 \item{drawborder}{if \code{TRUE}, draws a box outside the entire
plot.}
\item{borderwidth}{width of the border. Defaults to 1.}
 \item{densitycolors}{a vector with 4 colors. The first is
for numeric input variables pointing up,
the second for numeric input variables
pointing down, the third for
prediction terms without orientation
and the total prediction, and the
fourth for factors. If \code{NULL}, the
default colors are used.}
 \item{predupcolor}{the color of a positive prediction
for casetoshow.}
 \item{preddowncolor}{the color of a negative prediction
for casetoshow.}
 \item{predpointsize}{the size of the points displaying the
predictions for casetoshow.}
 \item{draw.segments}{if \code{TRUE}, also plots a line segment
from the center of each prediction term
to the prediction term for casetoshow,
in the same color as the prediction.}
 \item{predsegmentwidth}{the width of that line segment.}
 \item{profile}{when casetoshow is not \code{NULL} and staircase
is \code{FALSE}, this plots the profile of the
case with a feint grey line.}
 \item{bw}{the bandwidth of the density estimation,
only used when \code{displaytype = "density"}.
This is the argument \code{'bw'} of the function
\code{\link[stats:density]{density}}.}
 \item{adjust}{multiplier of the bandwidth of the density
estimation, only used when \code{displaytype =
"density"}. This is the argument \code{'adjust'}
of the function \code{\link[stats:density]{density}}.}
}

\value{
A list with items
 \item{p}{the predictions plot, which is a ggplot2
 object.}
 \item{totpred}{vector with the total linear
 prediction of all cases for which
 it can be computed.}
 \item{centercept }{the centercept, which is the total
 linear prediction when all prediction
 terms have their average value.}
 \item{predterms}{matrix of cases by prediction terms.}
 \item{predsummary}{data frame with the standard deviation of
 each prediction term and the total linear
 prediction.}
 \item{casetotpred}{a number, the total linear prediction
 for casetoshow if one is given.}
 \item{casepredterms}{a vector with the values of the
 prediction terms for casetoshow.}
 \item{casesummary}{data frame which shows all prediction
 terms for casetoshow together with
 the centercept, total linear
 prediction, and for a glm fit also
 the total prediction in response units.}
}
\references{
Rousseeuw, P.J. (2026). Explainable Linear and Generalized Linear Models by the Predictions Plot. The American Statistician, 80, 157-163, \doi{10.1080/00031305.2025.2539235} (open access)
}
\author{
 Rousseeuw, P.J.
}

\seealso{
\code{\link{predscor}}
}
\examples{
data(data_titanic)
attach(data_titanic)
survival = y
Pclass = factor(Pclass, unique(Pclass))
Sex = factor(Sex, labels = c("F","M"))
fit <- glm(survival ~ Sex + Age + SibSp + Parch + Pclass, family=binomial)
predsplot(fit, main = "Titanic data", displaytype = "density")

# For more examples, we refer to the vignette:
\dontrun{
vignette("predsplot_examples")
}
}

\name{vcr.rpart.newdata}
\alias{vcr.rpart.newdata}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{
Prepare for visualization of an rpart classification on new data.
}
\description{
Produces output for the purpose of constructing graphical displays such as the \code{\link{classmap}} on new data. Requires the output of
\code{\link{vcr.rpart.train}} as an argument.
}

\usage{
vcr.rpart.newdata(Xnew, ynew = NULL, vcr.rpart.train.out,
                  LOO = FALSE)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{Xnew}{data matrix of the new data, with the same
  number of columns \code{d} as in the training data. Missing values are not allowed.
}
\item{ynew}{factor with class membership of each new case. Can be \code{NA} for some or all cases. If \code{NULL}, is assumed to be \code{NA} everywhere.}
\item{vcr.rpart.train.out}{output of \code{\link{vcr.rpart.train}} on the training data.}
\item{LOO}{leave one out. Only used when testing this function on a subset of the training data. Default is \code{LOO=FALSE}.}
}


\value{
A list with components: \cr
 \item{yintnew}{number of the given class of each case. Can contain \code{NA}'s.}
\item{ynew}{given class label of each case. Can contain \code{NA}'s.}
\item{levels}{levels of the response, from \code{vcr.rpart.train.out}.}
\item{predint}{predicted class number of each case. Always exists.}
\item{pred}{predicted label of each case.}
\item{altint}{number of the alternative class. Among the classes different from the given class, it is the one with the highest posterior probability. Is \code{NA} for cases whose \code{ynew} is missing.}
\item{altlab}{alternative label if yintnew was given, else \code{NA}.}
\item{PAC}{probability of the alternative class. Is \code{NA} for cases whose \code{ynew} is missing.}
\item{fig}{distance of each case \eqn{i} from each class \eqn{g}. Always exists.}
\item{farness}{farness of each case from its given class. Is \code{NA} for cases whose \code{ynew} is missing.}
\item{ofarness}{for each case \eqn{i}, its lowest \code{fig[i,g]} to any class \eqn{g}. Always exists.}
}
\references{
Raymaekers J., Rousseeuw P.J.(2021). Silhouettes and quasi residual plots for neural nets and tree-based classifiers. \emph{Journal of Computational and Graphical Statistics}, 31(4), 1332–1343. \doi{10.1080/10618600.2022.2050249}
}

\author{
 Raymaekers J., Rousseeuw P.J.
}

\seealso{
\code{\link{vcr.rpart.train}}, \code{\link{classmap}}, \code{\link{silplot}}, \code{\link{stackedplot}}
}

\examples{
library(rpart)
data("data_titanic")
traindata <- data_titanic[which(data_titanic$dataType == "train"), -13]
str(traindata); table(traindata$y)
set.seed(123) # rpart is not deterministic
rpart.out <- rpart(y ~ Pclass + Sex + SibSp +
                    Parch + Fare + Embarked,
                  data = traindata, method = 'class', model = TRUE)
y_train <- traindata[, 12]
x_train <- traindata[, -12]
mytype <- list(nominal = c("Name", "Sex", "Ticket", "Cabin", "Embarked"), ordratio = c("Pclass"))
# These are 5 nominal columns, and one ordinal.
# The variables not listed are by default interval-scaled.
vcrtrain <- vcr.rpart.train(x_train, y_train, rpart.out, mytype)
testdata <- data_titanic[which(data_titanic$dataType == "test"), -13]
dim(testdata)
x_test <- testdata[, -12]
y_test <- testdata[, 12]
vcrtest <- vcr.rpart.newdata(x_test, y_test, vcrtrain)
confmat.vcr(vcrtest)
silplot(vcrtest, classCols = c(2, 4))
classmap(vcrtest, "casualty", classCols = c(2, 4))
classmap(vcrtest, "survived", classCols = c(2, 4))

# For more examples, we refer to the vignette:
\dontrun{
vignette("Rpart_examples")
}
}

\name{stackedplot}
\alias{stackedplot}
\title{
Make a vertically stacked mosaic plot of class predictions.
}
\description{
Make a vertically stacked mosaic plot of class predictions from the output of
\code{vcr.*.train} or \code{vcr.*.newdata}. Optionally,
the outliers for each class can be shown as a gray rectangle at the top.
}
\usage{
stackedplot(vcrout, cutoff = 0.99, classCols = NULL,
classLabels = NULL, separSize=1, minSize=1.5,
showOutliers = TRUE, showLegend = FALSE, main = NULL,
htitle = NULL, vtitle = NULL)
}

\arguments{
 \item{vcrout}{output of \code{vcr.*.train} or \code{vcr.*.newdata}.}
 \item{cutoff}{cases with overall farness \code{vcrout$ofarness} > \code{cutoff} are flagged as outliers.}
\item{classCols}{user-specified colors for the classes. If \code{NULL} a default palette is used.}
 \item{classLabels}{names of given labels. If \code{NULL} they are taken from \code{vcrout}.}
 \item{separSize}{how much white between rectangles.}
 \item{minSize}{rectangles describing less than \code{minSize} percent of the data, are shown as \code{minSize} percent.}
 \item{showOutliers}{if \code{TRUE}, shows a separate class in gray with the outliers, always at the top.}
 \item{showLegend}{if \code{TRUE}, a legend is shown to the right of the plot. Default \code{FALSE}, since the legend is not necessary as the colors are already visible in the bottom part of each stack.}
 \item{main}{title for the plot.}
 \item{htitle}{title for horizontal axis (given labels). If  \code{NULL}, a default title is shown.}
 \item{vtitle}{title for vertical axis (predicted labels). If  \code{NULL}, a default title is shown.}
}

\value{
A ggplot object.
}

\references{
Raymaekers J., Rousseeuw P.J., Hubert M. (2021). Class maps for visualizing classification results. \emph{Technometrics}, 64(2), 151–165. \doi{10.1080/00401706.2021.1927849}
}

\author{
 Raymaekers J., Rousseeuw P.J.
}

\seealso{
\code{\link{vcr.da.train}}, \code{\link{vcr.da.newdata}},\cr \code{\link{vcr.knn.train}}, \code{\link{vcr.knn.newdata}},\cr \code{\link{vcr.svm.train}}, \code{\link{vcr.svm.newdata}},\cr \code{\link{vcr.rpart.train}}, \code{\link{vcr.rpart.newdata}},\cr \code{\link{vcr.forest.train}}, \code{\link{vcr.forest.newdata}},\cr \code{\link{vcr.neural.train}}, \code{\link{vcr.neural.newdata}}
}

\examples{

data("data_floralbuds")
X <- data_floralbuds[, 1:6]; y <- data_floralbuds[, 7]
vcrout <- vcr.da.train(X, y)
cols <- c("saddlebrown", "orange", "olivedrab4", "royalblue3")
stackedplot(vcrout, classCols = cols, showLegend = TRUE)

# The legend is not really needed, since we can read the
# color of a class from the bottom of its vertical bar:
stackedplot(vcrout, classCols = cols, main = "Stacked plot of QDA on foral buds data")

# If we do not wish to show outliers:
stackedplot(vcrout, classCols = cols, showOutliers = FALSE)

# For more examples, we refer to the vignettes:
\dontrun{
vignette("Discriminant_analysis_examples")
vignette("K_nearest_neighbors_examples")
vignette("Support_vector_machine_examples")
vignette("Rpart_examples")
vignette("Random_forest_examples")
vignette("Neural_net_examples")
}
}

\name{vcr.rpart.train}
\alias{vcr.rpart.train}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{
Prepare for visualization of an rpart classification on training data.
}
\description{
Produces output for the purpose of constructing graphical displays such as the \code{\link{classmap}}. The user first needs to train a
classification tree on the data by \code{\link[rpart:rpart]{rpart::rpart}}.
This then serves as an argument to  \code{\link{vcr.rpart.train}}.
}

\usage{
vcr.rpart.train(X, y, trainfit, type = list(),
                k = 5, stand = TRUE)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{X}{A rectangular matrix or data frame, where the
             columns (variables) may be of mixed type and
             may contain \code{NA}'s.
}
\item{y}{factor with the given class labels.
             It is crucial that \code{X} and \code{y} are exactly the same
             as in the call to \code{\link[rpart:rpart]{rpart::rpart}}.
             \code{y} is allowed to contain \code{NA}'s.}
\item{k}{the number of nearest neighbors used in the
             farness computation.}
\item{trainfit}{the output of an \code{\link[rpart:rpart]{rpart::rpart}} training cycle.}
\item{type}{list for specifying some (or all) of the types of the
    variables (columns) in \code{X}, used for computing the dissimilarity matrix, as in \code{\link[cluster:daisy]{cluster::daisy}}.  The list may contain the following components: \code{"ordratio"} (ratio scaled variables to be treated as ordinal variables), \code{"logratio"} (ratio scaled variables that
    must be logarithmically transformed), \code{"asymm"} (asymmetric
    binary) and \code{"symm"} (symmetric binary variables).  Each
    component's value is a vector, containing the names or the numbers
    of the corresponding columns of \code{X}.
    Variables not mentioned in the \code{type} list are interpreted as
    usual (see argument \code{X}).}
\item{stand}{whether or not to standardize numerical (interval scaled) variables by their range as in the original \code{\link[cluster:daisy]{cluster::daisy}} code for the farness computation. Defaults to \code{TRUE}.}
}


\value{
A list with components: \cr
  \item{X}{The input data \code{X}. Keep??}
  \item{yint}{number of the given class of each case. Can contain \code{NA}'s.}
  \item{y}{given class label of each case. Can contain \code{NA}'s.}
  \item{levels}{levels of \code{y}}
  \item{predint}{predicted class number of each case. For each case this is the class with the highest posterior probability. Always exists.}
  \item{pred}{predicted label of each case.}
  \item{altint}{number of the alternative class. Among the classes different from the given class, it is the one with the highest posterior probability. Is \code{NA} for cases whose \code{y} is missing.}
  \item{altlab}{label of the alternative class. Is \code{NA} for cases whose \code{y} is missing.}
  \item{PAC}{probability of the alternative class. Is \code{NA} for cases whose \code{y} is missing.}
  \item{figparams}{parameters for computing \code{fig}, can be used for new data.}
  \item{fig}{distance of each case \eqn{i} from each class \eqn{g}. Always exists.}
  \item{farness}{farness of each case from its given class. Is \code{NA} for cases whose \code{y} is missing.}
  \item{ofarness}{for each case \code{i}, its lowest  \code{fig[i,g]} to any class \code{g}. Always exists.}
 \item{trainfit}{the trainfit used to build the VCR object.}
}
\references{
Raymaekers J., Rousseeuw P.J.(2021). Silhouettes and quasi residual plots for neural nets and tree-based classifiers. \emph{Journal of Computational and Graphical Statistics}, 31(4), 1332–1343. \doi{10.1080/10618600.2022.2050249}
}

\author{
 Raymaekers J., Rousseeuw P.J.
}

\seealso{
\code{\link{vcr.rpart.newdata}}, \code{\link{classmap}}, \code{\link{silplot}}, \code{\link{stackedplot}}
}

\examples{
library(rpart)
data("data_titanic")
traindata <- data_titanic[which(data_titanic$dataType == "train"), -13]
str(traindata); table(traindata$y)
set.seed(123) # rpart is not deterministic
rpart.out <- rpart(y ~ Pclass + Sex + SibSp +
                    Parch + Fare + Embarked,
                  data = traindata, method = 'class', model = TRUE)
y_train <- traindata[, 12]
x_train <- traindata[, -12]
mytype <- list(nominal = c("Name", "Sex", "Ticket", "Cabin", "Embarked"), ordratio = c("Pclass"))
# These are 5 nominal columns, and one ordinal.
# The variables not listed are by default interval-scaled.
vcrtrain <- vcr.rpart.train(x_train, y_train, rpart.out, mytype)
confmat.vcr(vcrtrain)
silplot(vcrtrain, classCols = c(2, 4))
classmap(vcrtrain, "casualty", classCols = c(2, 4))
classmap(vcrtrain, "survived", classCols = c(2, 4))

# For more examples, we refer to the vignette:
\dontrun{
vignette("Rpart_examples")
}
}

\name{silplot}
\alias{silplot}
\title{
Draw the silhouette plot of a classification
}
\description{
Draw the silhouette plot to visualize classification results, based on the output of one of the \code{vcr.*.*} functions in this package. The horizontal axis of the silhouette plot shows each case's \code{s(i)}.
}
\usage{
silplot(vcrout, classLabels = NULL, classCols = NULL,
        showLegend = TRUE, showClassNumbers = FALSE,
        showCases = FALSE, drawLineAtAverage = FALSE,
        topdown = TRUE, main = NULL, summary = TRUE)
}

\arguments{
 \item{vcrout}{output of \code{vcr.*.train} or \code{vcr.*.newdata}. Required.}
 \item{classLabels}{the labels (levels) of the classes. If \code{NULL}, they are taken from \code{vcrout}.}
 \item{classCols}{a list of colors for the classes. There should be at least as many as there are levels. If \code{NULL} a default palette is used.}

\item{showLegend}{if \code{TRUE}, a legend is shown to the right of the plot.}
\item{showClassNumbers}{if \code{TRUE}, the legend will show the class numbers
 instead of the class labels.}
\item{showCases}{if \code{TRUE}, the plot shows the numbers of the cases.
 They are only readable when the number of cases
 is relatively small.}
\item{topdown}{if \code{TRUE} (the default), the silhouettes are
 plotted from top to bottom. Otherwise they
 are plotted from left to right.}
\item{drawLineAtAverage}{if \code{TRUE}, drwas a line at the average value
 of the \code{s(i)}.}
\item{main }{title for the plot. If \code{NULL}, a default title is used.}
\item{summary}{if \code{TRUE}, puts a summary table on the screen
 with for each class its number, label, number of class members,
 and the average of its \code{s(i)}.}

}

\value{
A ggplot object containing the silhouette plot.
}

\references{
Raymaekers J., Rousseeuw P.J.(2021). Silhouettes and quasi residual plots for neural nets and tree-based classifiers. \emph{Journal of Computational and Graphical Statistics}, 31(4), 1332–1343. \doi{10.1080/10618600.2022.2050249}
}

\author{
 Raymaekers J., Rousseeuw P.J.
}

\seealso{
\code{\link{vcr.da.train}}, \code{\link{vcr.da.newdata}},\cr \code{\link{vcr.knn.train}}, \code{\link{vcr.knn.newdata}},\cr \code{\link{vcr.svm.train}}, \code{\link{vcr.svm.newdata}},\cr \code{\link{vcr.rpart.train}}, \code{\link{vcr.rpart.newdata}},\cr \code{\link{vcr.forest.train}}, \code{\link{vcr.forest.newdata}},\cr \code{\link{vcr.neural.train}}, \code{\link{vcr.neural.newdata}}
}

\examples{
vcrout <- vcr.da.train(iris[, 1:4], iris[, 5])
silplot(vcrout)
# For more examples, we refer to the vignettes:
\dontrun{
vignette("Discriminant_analysis_examples")
vignette("K_nearest_neighbors_examples")
vignette("Support_vector_machine_examples")
vignette("Rpart_examples")
vignette("Forest_examples")
vignette("Neural_net_examples")
}
}

\name{data_bookReviews}
\alias{data_bookReviews}
\docType{data}
\title{
Amazon book reviews data
}
\description{
This is a subset of the data used in the paper, which
was assembled by Prettenhofer and Stein (2010). It contains 1000 reviews
of books on Amazon, of which 500 were selected from the original training data
and 500 from the test data.

The full dataset has been used for a variety of things, including
classification using svm. The subset was chosen small enough to keep the computation
time low, while still containing the examples in the paper.
}

\usage{data("data_bookReviews")}

\format{
  A data frame with 1000 observations on the following 2 variables.
  \describe{
    \item{\code{review}}{the review in text format (character)}
    \item{\code{sentiment}}{factor indicating the sentiment of the review: negative (1) or positive (2)}
  }
}


\source{
Prettenhofer, P., Stein, B. (2010). Cross-language text classification using
structural correspondence learning. \emph{Proceedings of the 48th annual
meeting of the association for computational linguistics},  1118-1127.
}

\examples{
data(data_bookReviews)
# Example review:
data_bookReviews[5, 1]

# The data are used in:
\dontrun{
vignette("Support_vector_machine_examples")
}
}

\name{vcr.knn.train}
\alias{vcr.knn.train}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{
Carry out a k-nearest neighbor classification on training data, and prepare to visualize its results.
}
\description{
Carries out a k-nearest neighbor classification on the training data. Various additional output is produced for the purpose of constructing graphical displays such as the \code{\link{classmap}}.
}
\usage{
vcr.knn.train(X, y, k)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{X}{This can be a rectangular matrix or data frame of (already standardized) measurements, or a dist object obtained from \code{\link[stats:dist]{stats::dist}} or \code{\link[cluster:daisy]{cluster::daisy}}. Missing values are not allowed.
}
\item{y}{factor with the given (observed) class labels. There need to be non-missing \code{y} in order to be able to train the classifier.}
\item{k}{the number of nearest neighbors used. It can be selected by running cross-validation using a different package.}
}

\value{
A list with components: \cr

\item{yint}{number of the given class of each case. Can contain \code{NA}'s.}
\item{y}{given class label of each case. Can contain \code{NA}'s.}
\item{levels}{levels of \code{y}}
\item{predint}{predicted class number of each case. Always exists.}
\item{pred}{predicted label of each case.}
\item{altint}{number of the alternative class. Among the classes different from the given class, it is the one with the highest posterior probability. Is \code{NA} for cases whose \code{y} is missing.}
\item{altlab}{label of the alternative class. Is \code{NA} for cases whose \code{y} is missing.}
\item{PAC}{probability of the alternative class. Is \code{NA} for cases whose \code{y} is missing.}
\item{figparams}{parameters used to compute \code{fig}.}
\item{fig}{distance of each case \eqn{i} from each class \eqn{g}. Always exists.}
\item{farness}{farness of each case from its given class. Is \code{NA} for cases whose \code{y} is missing.}
\item{ofarness}{for each case \eqn{i}, its lowest  \code{fig[i,g]} to any class \eqn{g}. Always exists.}
\item{k}{the requested number of nearest neighbors, from the arguments. Will also be used for  classifying new data.}
\item{ktrues}{for each case this contains the actual number of elements in its neighborhood. This can be higher than \code{k} due to ties.}
\item{counts}{a matrix with 3 columns, each row representing a case. For the neighborhood of each case it says how many members it has from the given class, the predicted class, and the alternative class. The first and third entry is \code{NA} for cases whose \code{y} is missing.}
\item{X}{If the argument \code{X} was a data frame or matrix of coordinates, \code{as.matrix(X)} is returned here. This is useful for classifying new data.}
}

\references{
Raymaekers J., Rousseeuw P.J., Hubert M. (2021). Class maps for visualizing classification results. \emph{Technometrics}, 64(2), 151–165. \doi{10.1080/00401706.2021.1927849}
}

\author{
 Raymaekers J., Rousseeuw P.J.
}

\seealso{
\code{\link{vcr.knn.newdata}}, \code{\link{classmap}}, \code{\link{silplot}}, \code{\link{stackedplot}}
}

\examples{
vcrout <- vcr.knn.train(iris[, 1:4], iris[, 5], k = 5)
confmat.vcr(vcrout)
stackedplot(vcrout)
classmap(vcrout, "versicolor", classCols = 2:4)
# The cases misclassified as virginica are shown in blue.

# For more examples, we refer to the vignette:
\dontrun{
vignette("K_nearest_neighbors_examples")
}
}

\name{vcr.da.newdata}
\alias{vcr.da.newdata}

\title{
Carry out discriminant analysis on new data, and prepare to visualize its results.
}
\description{
Predicts class labels for new data by discriminant analysis, using the output of \code{\link{vcr.da.train}} on the training data. For new data cases whose label in \code{yintnew} is non-missing, additional output is produced for constructing graphical displays such as the \code{\link{classmap}}.
}
\usage{
vcr.da.newdata(Xnew, ynew=NULL, vcr.da.train.out)
}

\arguments{
  \item{Xnew}{data matrix of the new data, with the same number of columns as in the training data. Missing values are not allowed.
}
\item{ynew}{factor with class membership of each new case. Can be \code{NA} for some or all cases. If \code{NULL}, is assumed to be \code{NA} everywhere.}
\item{vcr.da.train.out}{output of \code{\link{vcr.da.train}} on the training data.}
}

\value{
A list with components: \cr

\item{yintnew}{number of the given class of each case. Can contain \code{NA}'s.}
\item{ynew }{given class label of each case. Can contain \code{NA}'s.}
\item{levels }{levels of the response, from \code{vcr.da.train.out}.}
\item{predint}{predicted class number of each case. Always exists.}
\item{pred}{predicted label of each case.}
\item{altint}{number of the alternative class. Among the classes different from the given class, it is the one with the highest posterior probability. Is \code{NA} for cases whose \code{ynew} is missing.}
\item{altlab}{label of the alternative class. Is \code{NA} for cases whose \code{ynew} is missing.}
\item{PAC}{probability of the alternative class. Is \code{NA} for cases whose \code{ynew} is missing.}
\item{fig}{distance of each case \eqn{i} to each class \eqn{g}. Always exists.}
\item{farness}{farness of each case \eqn{i} from its given class. Is \code{NA} for cases whose \code{ynew} is missing.}
\item{ofarness}{For each case \eqn{i}, its lowest \code{fig[i,g]} to any class \code{g}. Always exists.}
\item{classMS}{list with center and covariance matrix of each class, from \code{vcr.da.train.out}.}
\item{lCurrent}{log of mixture density of each case in its given class. Is \code{NA} for cases with missing \code{ynew}.}
\item{lPred}{log of mixture density of each case in its predicted class. Always exists.}
\item{lAlt}{log of mixture density of each case in its alternative class. Is \code{NA} for cases with missing \code{ynew}.}
}
\references{
Raymaekers J., Rousseeuw P.J., Hubert M. (2021). Class maps for visualizing classification results. \emph{Technometrics}, 64(2), 151–165. \doi{10.1080/00401706.2021.1927849}
}

\author{
 Raymaekers J., Rousseeuw P.J.
}

\seealso{
\code{\link{vcr.da.train}}, \code{\link{classmap}}, \code{\link{silplot}}, \code{\link{stackedplot}}
}

\examples{

vcr.train <- vcr.da.train(iris[, 1:4], iris[, 5])
inds <- c(51:150) # a subset, containing only 2 classes
iris2 <- iris[inds, ] # fake "new" data
iris2[c(1:10, 51:60), 5] <- NA
vcr.test <- vcr.da.newdata(iris2[, 1:4], iris2[, 5], vcr.train)
vcr.test$PAC[1:25] # between 0 and 1. Is NA where the response is.
plot(vcr.test$PAC, vcr.train$PAC[inds]); abline(0, 1) # match
plot(vcr.test$farness, vcr.train$farness[inds]); abline(0, 1) # match
confmat.vcr(vcr.train) # for comparison
confmat.vcr(vcr.test)
stackedplot(vcr.train) # for comparison
stackedplot(vcr.test)
classmap(vcr.train, "versicolor", classCols = 2:4) # for comparison
classmap(vcr.test, "versicolor", classCols = 2:4) # has fewer points

# For more examples, we refer to the vignette:
\dontrun{
vignette("Discriminant_analysis_examples")
}
}

\name{qresplot}
\alias{qresplot}

\title{
Draw a quasi residual plot of PAC versus a data feature
}
\description{
Draw a quasi residual plot to visualize classification results.
The vertical axis of the quasi residual plot shows each case's probability of alternative class (PAC). The horizontal axis
shows the feature given as the second argument in the function call.
}
\usage{
qresplot(PAC, feat, xlab = NULL, xlim = NULL,
         main = NULL, identify = FALSE, gray = TRUE,
         opacity = 1, squareplot = FALSE, plotLoess = FALSE,
         plotErrorBars = FALSE, plotQuantiles = FALSE,
         grid = NULL, probs = c(0.5, 0.75),
         cols = NULL, fac = 1, cex = 1,
         cex.main = 1.2, cex.lab = 1,
         cex.axis = 1, pch = 19)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
\item{PAC}{vector with the PAC values of a classification,
typically the \code{$PAC} in the return of a call to a function \code{vcr.*.*}}
\item{feat}{the PAC will be plotted versus this data
feature. Note that feat does not have to be
one of the explanatory variables of the model.
It can be another variable, a combination of
variables (like a sum or a principal component
score), the row number of the cases if they
were recorded succesively, etc.}
\item{xlab}{label for the horizontal axis, i.e. the name
of variable feat.}
\item{xlim}{limits for the horizontal axis. If \code{NULL}, the
range of feat is used.}
\item{main}{title for the plot.}
\item{identify }{if \code{TRUE}, left-click on a point to get its number, then ESC to exit.}
\item{gray}{logical, if \code{TRUE} (the default) the plot region
where \eqn{PAC < 0.5} gets a light gray background.
Points in this region were classified into
their given class, and the points above this
region were misclassified.}
\item{opacity}{determines opacity of plotted dots.
Value between 0 and 1, where 0 is transparent
and 1 is opaque.}
\item{squareplot}{if \code{TRUE}, the horizontal and vertical axis will
get the same length.}
\item{plotLoess}{if \code{TRUE}, a standard loess curve is fitted and
superimposed on the plot. May not work well if
feat is discrete with few values.
At most one of the options \code{plotLoess},
\code{plotErrorbars}, or \code{plotQuantiles} can be selected.}
\item{plotErrorBars}{if \code{TRUE}, the average PAC and its standard
error are computed on the intervals of a grid
(see option grid). Then a red curve connecting
the averages is plotted, as well as two blue
curves corresponding to the average plus or
minus one standard error. At most one of the options \code{plotLoess},
\code{plotErrorbars}, or \code{plotQuantiles} can be selected.}
\item{plotQuantiles}{if \code{TRUE}, one or more quantiles of the PAC
are computed on the intervals of a grid
(see option grid). The quantiles correspond
the probabilities in option probs.
Then the curves connecting the quantiles
are plotted. At most one of the options \code{plotLoess},
\code{plotErrorbars}, or \code{plotQuantiles} can be selected.}
\item{grid}{only used when \code{plotErrorBars} or \code{plotQuantiles}
are selected. This is a vector with increasing
feat values, forming the grid. If \code{NULL}, the
grid consists of the minimum and the maximum
of feat, with 9 equispaced points between them.}
\item{probs}{only used when \code{plotQuantiles} is selected. This
is a vector with probabilities determining the
quantiles. If \code{NULL}, defaults to \code{c(0.5, 0.75)}.}
\item{cols}{only used when plotquantiles is selected.
A vector with the colors of the quantile curves.
If \code{NULL} the cols are taken as 2, 3, ...}
\item{fac}{only used when \code{plotLoess}, \code{plotErrorBars} or
\code{plotQuantiles} are selected. A real number to
multiply the resulting curves. A value \code{fac > 1}
can be useful to better visualize the curves
when they would be too close to zero.
By default (\code{fac = 1}) this is not done.}
\item{cex}{passed on to \code{\link[graphics]{plot}}.}
\item{cex.main}{same, for title.}
\item{cex.lab}{same, for labels on horizontal and vertical axes.}
\item{cex.axis}{same, for axes.}
\item{pch}{plot character for the points, defaults to 19.}

}


\value{

\item{coordinates}{a matrix with 2 columns containing the
coordinates of the plotted points. This makes it
easier to add text next to interesting points.
If \code{identify = TRUE}, the attribute ids of coordinates
contains the row numbers of the identified points
in the matrix coordinates.

}
}
\references{
Raymaekers J., Rousseeuw P.J.(2021). Silhouettes and quasi residual plots for neural nets and tree-based classifiers. \emph{Journal of Computational and Graphical Statistics}, 31(4), 1332–1343. \doi{10.1080/10618600.2022.2050249}
}

\author{
 Raymaekers J., Rousseeuw P.J.
}


\examples{
library(rpart)
data("data_titanic")
traindata <- data_titanic[which(data_titanic$dataType == "train"), -13]
set.seed(123) # rpart is not deterministic
rpart.out <- rpart(y ~ Pclass + Sex + SibSp +
                    Parch + Fare + Embarked,
                  data = traindata, method = 'class', model = TRUE)
mytype <- list(nominal = c("Name", "Sex", "Ticket", "Cabin", "Embarked"), ordratio = c("Pclass"))
x_train <- traindata[, -12]
y_train <- traindata[,  12]
vcrtrain <- vcr.rpart.train(x_train, y_train, rpart.out, mytype)
# Quasi residual plot versus age, for males only:
PAC <- vcrtrain$PAC[which(x_train$Sex == "male")]
feat <- x_train$Age[which(x_train$Sex == "male")]
qresplot(PAC, feat, xlab = "Age (years)", opacity = 0.5,
         main = "quasi residual plot for male passengers",
         plotLoess = TRUE)
text(x = 14, y = 0.60, "loess curve", col = "red", cex = 1)
}

\name{vcr.svm.train}
\alias{vcr.svm.train}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{
Prepare for visualization of a support vector machine classification on training data.
}
\description{
Produces output for the purpose of constructing graphical displays such as the \code{\link{classmap}}. The user first needs to run a support vector machine classification on the data by \code{\link[e1071:svm]{e1071::svm}}, with the option \code{probability = TRUE}. This classification can be with two or more classes. The output of \code{\link[e1071:svm]{e1071::svm}} is then an argument to \code{\link{vcr.svm.train}}. As \code{\link[e1071:svm]{e1071::svm}} does not output the data itself, it needs to be given as well, in the arguments \code{X} and \code{y}.
}
\usage{
vcr.svm.train(X, y, svfit, ortho = FALSE)
}

\arguments{

\item{X}{matrix of data coordinates, as used in \code{\link[e1071:svm]{e1071::svm}}. Missing values are not allowed.}
\item{y}{factor with the given (observed) class labels. It is crucial that X and y are exactly the same as in the call to \code{\link[e1071:svm]{e1071::svm}}.}
\item{svfit}{an object returned by \code{\link[e1071:svm]{e1071::svm}}, called with exactly the same \code{X} and \code{y} as above.}
\item{ortho}{If \code{TRUE}, will compute farness in the orthogonal complement of the vector beta given by \code{\link[e1071:svm]{e1071::svm}}. Is only possible for 2 classes, else there would be several beta vectors.}
}

\value{
A list with components: \cr
\item{yint}{number of the given class of each case. Can contain \code{NA}'s.}
\item{y}{given class label of each case. Can contain \code{NA}'s.}
\item{levels}{levels of the response \code{y}.}
\item{predint}{predicted class number of each case. Always exists.}
\item{pred}{predicted label of each case.}
\item{altint}{number of the alternative class. Among the classes different from the given class, it is the one with the highest posterior probability. Is \code{NA} for cases whose \code{y} is missing.}
\item{altlab}{label of the alternative class. Is \code{NA} for cases whose \code{y} is missing.}
\item{PAC}{probability of the alternative class. Is \code{NA} for cases whose \code{y} is missing.}
\item{figparams}{parameters used in \code{fig}, can be used for new data.}
\item{fig}{distance of each case \eqn{i} from each class \eqn{g}. Always exists.}
\item{farness}{farness of each case from its given class. Is \code{NA} for cases whose \code{y} is missing.}
\item{ofarness}{for each case \eqn{i}, its lowest \code{fig[i,g]} to any class \code{g}. Always exists.}
\item{svfit}{as it was input, will be useful for new data.}
\item{X}{the matrix of data coordinates from the arguments. This is useful for classifying new data.}
}

\references{
Raymaekers J., Rousseeuw P.J., Hubert M. (2021). Class maps for visualizing classification results. \emph{Technometrics}, 64(2), 151–165. \doi{10.1080/00401706.2021.1927849}
}

\author{
 Raymaekers J., Rousseeuw P.J.
}

\seealso{
\code{\link{vcr.knn.newdata}}, \code{\link{classmap}}, \code{\link{silplot}}, \code{\link{stackedplot}}, \code{\link[e1071:svm]{e1071::svm}}
}

\examples{
library(e1071)
set.seed(1); X <- matrix(rnorm(200 * 2), ncol = 2)
X[1:100, ] <- X[1:100, ] + 2
X[101:150, ] <- X[101:150, ] - 2
y <- as.factor(c(rep("blue", 150), rep("red", 50)))
cols <- c("deepskyblue3", "red")
plot(X, col = cols[as.numeric(y)], pch = 19)
# We now fit an SVM with radial basis kernel to the data:
set.seed(1) # to make the result of svm() reproducible.
svmfit <- svm(y~., data = data.frame(X = X, y = y),
scale = FALSE, kernel = "radial", cost = 10,
gamma = 1, probability = TRUE)
plot(svmfit$decision.values, col = cols[as.numeric(y)]); abline(h = 0)
# so the decision values separate the classes reasonably well.
plot(svmfit, data = data.frame(X = X, y = y), X.2~X.1, col = cols)
# The boundary is far from linear (but in feature space it is).
vcr.train <- vcr.svm.train(X, y, svfit = svmfit)
confmat.vcr(vcr.train)
stackedplot(vcr.train, classCols = cols)
classmap(vcr.train, "blue", classCols = cols)
classmap(vcr.train, "red", classCols = cols)

# For more examples, we refer to the vignette:
\dontrun{
vignette("Support_vector_machine_examples")
}
}

\name{vcr.da.train}
\alias{vcr.da.train}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{
Carry out discriminant analysis on training data, and prepare to visualize its results.
}
\description{
Custom DA function which prepares for graphical displays such as the \code{\link{classmap}}. The disciminant analysis itself is carried out by the maximum a posteriori rule, which maximizes the density of the mixture.
}
\usage{
vcr.da.train(X, y, rule = "QDA", estmethod = "meancov")
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{X}{a numerical matrix containing the predictors in its columns. Missing values are not allowed.
}
\item{y}{a factor with the given class labels.}
\item{rule}{either "\code{QDA}" for quadratic discriminant analysis or "\code{LDA}"  for linear discriminant analysis.}
\item{estmethod}{function for location and covariance estimation.
Should return a list with the center \code{$m} and the covariance matrix \code{$S}. The default is \code{"meancov"} (classical mean and covariance matrix), and the option \code{"DetMCD"} (based on \code{\link[robustbase:covMcd]{robustbase::covMcd}}) is also provided.}
}


\value{
A list with components: \cr
  \item{yint}{number of the given class of each case. Can contain \code{NA}'s.}
  \item{y}{given class label of each case. Can contain \code{NA}'s.}
  \item{levels}{levels of \code{y}}
  \item{predint}{predicted class number of each case. For each case this is the class with the highest posterior probability. Always exists.}
  \item{pred}{predicted label of each case.}
  \item{altint}{number of the alternative class. Among the classes different from the given class, it is the one with the highest posterior probability. Is \code{NA} for cases whose \code{y} is missing.}
  \item{altlab}{label of the alternative class. Is \code{NA} for cases whose \code{y} is missing.}
  \item{PAC}{probability of the alternative class. Is \code{NA} for cases whose \code{y} is missing.}
  \item{figparams}{parameters for computing \code{fig}, can be used for new data.}
  \item{fig}{distance of each case \eqn{i} from each class \eqn{g}. Always exists.}
  \item{farness}{farness of each case from its given class. Is \code{NA} for cases whose \code{y} is missing.}
  \item{ofarness}{for each case \code{i}, its lowest  \code{fig[i,g]} to any class \code{g}. Always exists.}
  \item{classMS}{list with center and covariance matrix of each class}
  \item{lCurrent}{log of mixture density of each case in its given class. Is \code{NA} for cases with missing \code{y}.
}
  \item{lPred}{log of mixture density of each case in its predicted class. Always exists.
}
  \item{lAlt}{log of mixture density of each case in its alternative class. Is \code{NA} for cases with missing \code{y}.}
}
\references{
Raymaekers J., Rousseeuw P.J., Hubert M. (2021). Class maps for visualizing classification results. \emph{Technometrics}, 64(2), 151–165. \doi{10.1080/00401706.2021.1927849}
}

\author{
 Raymaekers J., Rousseeuw P.J.
}

\seealso{
\code{\link{vcr.da.newdata}}, \code{\link{classmap}}, \code{\link{silplot}}, \code{\link{stackedplot}}
}

\examples{

data("data_floralbuds")
X <- data_floralbuds[, 1:6]; y <- data_floralbuds[, 7]
vcrout <- vcr.da.train(X, y, rule = "QDA")
# For linear discriminant analysis, put rule = "LDA".
confmat.vcr(vcrout) # There are a few outliers
cols <- c("saddlebrown", "orange", "olivedrab4", "royalblue3")
stackedplot(vcrout, classCols = cols)
classmap(vcrout, "bud", classCols = cols)

# For more examples, we refer to the vignette:
\dontrun{
vignette("Discriminant_analysis_examples")
}
}

\name{data_latenc}
\alias{data_latenc}
\docType{data}
\title{
Latenc data
}
\description{
Latency times in seconds of exploratory behavior in fish during three experimental rounds.

}

\usage{data("data_latenc")}

\format{
  A data frame with 40 observations on the following variables.
  \describe{
  \item{round 1}{Latency times for round 1 of the experiment.}
  \item{round 2}{Latency times for round 1 of the experiment.}
  \item{round 3}{Latency times for round 1 of the experiment.}
}
}


\source{
Kerr, Nicky R., and Travis Ingram. "Personality does not predict individual niche variation in a freshwater fish." Behavioral Ecology 32.1 (2021): 159-167.
}

\examples{
data(data_latenc)
str(data_latenc)

# The data are used in:
\dontrun{
vignette("bixplot_examples")
}
}

\name{makeFV}
\alias{makeFV}
\title{
Constructs feature vectors from a kernel matrix.
}
\description{
Constructs feature vectors from a kernel matrix.
}
\usage{
makeFV(kmat, transfmat = NULL, precS = 1e-12)
}

\arguments{

\item{kmat}{a kernel matrix. If \code{transfmat} is \code{NULL}, we are dealing with training data and then \code{kmat} must be a square kernel matrix (of size \eqn{n} by \eqn{n} when there are \eqn{n} cases). Such a PSD matrix kmat can e.g. be produced by \code{\link{makeKernel}} or by \code{\link[kernlab:kernelMatrix]{kernlab::kernelMatrix}}. If on the other hand \code{transfmat} is not \code{NULL}, we are dealing with a test set. See details for the precise working.}
\item{transfmat}{transformation matrix. If not \code{NULL}, it is the value \code{transfmat} of \code{\link{makeFV}} on training data. It has to be a square matrix, with as many rows as there were training data.}
\item{precS}{if not \code{NULL}, eigenvalues of \code{kmat} below \code{precS} will be set equal to precS.}
}

\details{
If \code{transfmat} is non-\code{NULL}, we are dealing with a test set.
Denote the number of cases in the test set by \eqn{m \geq 1}. Each row of \code{kmat} of the test set then must contain the kernel values of a new case with all cases in the training set. Therefore the kernel matrix kmat must have dimensions \eqn{m} by \eqn{n}. The matrix \code{kmat} can e.g. be produced by \code{\link{makeKernel}}. It can also be obtained by running \code{\link[kernlab:kernelMatrix]{kernlab::kernelMatrix}} on the union of the training set and the test set, yielding an \eqn{(n+m)} by \eqn{(n+m)} matrix, from which one then takes the \eqn{[(n+1):m , 1:n]} submatrix.
}


\value{
A list with components: \cr
\item{Xf}{When makeKV is applied to the training set, \code{Xf} has coordinates of \eqn{n} points (vectors), the plain inner products of which equal the kernel matrix of the training set. That is, \code{kmat} = \code{Xf} \code{Xf}'. The \code{Xf} are expressed in an orthogonal basis in which the variance of the coordinates is decreasing, which is useful when plotting the first few coordinates. When \code{\link{makeFV}} is applied to a test set, \code{Xf} are coordinates of the feature vectors of the test set in the same space as those of the training set, and then \code{kmat} = \code{Xf} \%*\% \code{Xf of training data}.
  }
\item{transfmat}{square matrix for transforming kmat to \code{Xf}.}
}

\references{
Raymaekers J., Rousseeuw P.J., Hubert M. (2021). Class maps for visualizing classification results. \emph{Technometrics}, 64(2), 151–165. \doi{10.1080/00401706.2021.1927849}
}

\author{
 Raymaekers J., Rousseeuw P.J., Hubert, M.
}

\seealso{
\code{\link{makeKernel}}
}

\examples{
library(e1071)
set.seed(1); X <- matrix(rnorm(200 * 2), ncol = 2)
X[1:100, ] <- X[1:100, ] + 2
X[101:150, ] <- X[101:150, ] - 2
y <- as.factor(c(rep("blue", 150), rep("red", 50)))
cols <- c("deepskyblue3", "red")
plot(X, col = cols[as.numeric(y)], pch = 19)
# We now fit an SVM with radial basis kernel to the data:
svmfit <- svm(y~.,  data = data.frame(X = X, y = y),  scale = FALSE,
             kernel = "radial", cost = 10, gamma = 1, probability = TRUE)
Kxx <- makeKernel(X, svfit = svmfit)
outFV <- makeFV(Kxx)
Xf <- outFV$Xf # The data matrix in this feature space.
dim(Xf) # The feature vectors are high dimensional.
# The inner products of Xf match the kernel matrix:
max(abs(as.vector(Kxx - crossprod(t(Xf), t(Xf))))) # 3.005374e-13 # tiny, OK
range(rowSums(Xf^2)) # all points in Xf lie on the unit sphere.
pairs(Xf[, 1:5], col = cols[as.numeric(y)])
# In some of these we see spherical effects, e.g.
plot(Xf[, 1], Xf[, 5], col = cols[as.numeric(y)], pch = 19)
# The data look more separable here than in the original
# two-dimensional space.

# For more examples, we refer to the vignette:
\dontrun{
vignette("Support_vector_machine_examples")
}
}

\name{vcr.neural.train}
\alias{vcr.neural.train}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{
Prepare for visualization of a neural network classification on training data.
}
\description{
Produces output for the purpose of constructing graphical displays such as the \code{\link{classmap}}. The user first needs train a neural network. The representation of the data in a given layer (e.g. the final layer before applying the softmax function) then serves as the argument \code{X} to  \code{\link{vcr.neural.train}}.
}

\usage{
vcr.neural.train(X, y, probs, estmethod = meancov)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{X}{the coordinates of the \code{n} objects of the training
                data, in the layer chosen by the user. Missing
                values are not allowed.
}
\item{y}{factor with the given class labels of the objects.
                Make sure that the levels are in the same order as
                used in the neural net, i.e. the columns of its
                binary "once-hot-encoded" response vectors.}
\item{probs}{posterior probabilities obtained by the neural
                net, e.g. in keras. For each case (row of \code{X}),
                the classes have probabilities that add up to 1.
                Each row of the matrix probs contains these
                probabilities. The columns of probs must be in
                the same order as the levels of \code{y}.}
\item{estmethod}{function for location and covariance estimation.
                Should return a list with \code{$m} and \code{$S}.
                Can be \code{meancov}
                (classical mean and covariance matrix) or \code{DetMCD}.
                If one or more classes have a singular covariance
                matrix, the function automatically switches to
                the PCA-based farness used in \code{\link{vcr.svm.train}}.}

}


\value{
A list with components: \cr
 \item{X}{the coordinates of the \code{n} objects of the training
                data, in the layer chosen by the user.
}
  \item{yint}{number of the given class of each case. Can contain \code{NA}'s.}
  \item{y}{given class label of each case. Can contain \code{NA}'s.}
  \item{levels}{levels of \code{y}}
  \item{predint}{predicted class number of each case. For each case this is the class with the highest posterior probability. Always exists.}
  \item{pred}{predicted label of each case.}
  \item{altint}{number of the alternative class. Among the classes different from the given class, it is the one with the highest posterior probability. Is \code{NA} for cases whose \code{y} is missing.}
  \item{altlab}{label of the alternative class. Is \code{NA} for cases whose \code{y} is missing.}
  \item{ncolX}{number of columns in \code{X}.  Keep??}
   \item{PAC}{probability of the alternative class. Is \code{NA} for cases whose \code{y} is missing.}
  \item{computeMD}{Whether or not the farness is computed using the Mahalanobis distance.}
  \item{classMS}{list with center and covariance matrix of each class}
   \item{PCAfits}{if not \code{NULL}, PCA fits to each class, estimated from the training data but also useful for new data.}
  \item{figparams}{parameters for computing \code{fig}, can be used for new data.}
  \item{fig}{distance of each case \eqn{i} from each class \eqn{g}. Always exists.}
  \item{farness}{farness of each case from its given class. Is \code{NA} for cases whose \code{y} is missing.}
  \item{ofarness}{for each case \code{i}, its lowest  \code{fig[i,g]} to any class \code{g}. Always exists.}
}
\references{
Raymaekers J., Rousseeuw P.J.(2021). Silhouettes and quasi residual plots for neural nets and tree-based classifiers. \emph{Journal of Computational and Graphical Statistics}, 31(4), 1332–1343. \doi{10.1080/10618600.2022.2050249}
}

\author{
 Raymaekers J., Rousseeuw P.J.
}

\seealso{
\code{\link{vcr.neural.newdata}}, \code{\link{classmap}}, \code{\link{silplot}}, \code{\link{stackedplot}}
}

\examples{
# For examples, we refer to the vignette:
\dontrun{
vignette("Neural_net_examples")
}
}
